// Import TableMaster
var TableMaster = require('./tables');

// Import Card
var Card = require('./card/card');

// Import deck module
var deck = require('./card/deck');

// Constructor
var Table = function(tableId, host)
{
	this.id = tableId;
	this.host = host;
	this.isLocked = false;

	// Sessions = tables + players
	this._sessions = {};
	this._tables = {};
	this._players = {};

	// For allocating player IDs
	this._playerIdAllocator = 0;

	// ID -> Card mapping of cards
	this._cards = {};
}

// Create cards
Table.prototype.reset = function()
{
	// Clear the cards
	this._cards = {};

	// Create Card objects based on the sprites (52 cards)
	var sprites = deck.sprites;
	for(var num in sprites)
	{
		// Create the given card at the given position, with ref to table
		var card = new Card(num, sprites[num], this);

		// Calculate position of the card (TODO: offset? center?)
		var pos = num % 8;
		var row = Math.floor(num / 8);
		card.x = pos * 1.25; // 1.25% between cards
		card.y = row * 20.0; // 20.0% between rows
		//console.log(pos + ' * 0.1 = ' + card.x);

		// All cards start with the same Z-index
		card.z = 0;

		// All cards start with no rotation
		card.rot = 0;

		// All cards start facing down (not flipped)
		card.flip = false;

		// Add to the table
		this._cards[card.id] = card;
	}

	// Done!
	console.log(this + ': spawned ' + sprites.length + ' cards!');

	// Broadcast all cards
	this.broadcast('cards', this.getCardList());
}

// Broadcasting events
Table.prototype.broadcast = function(event, data, exclude)
{
	// Log for debug
	console.log(this + ' --> "' + event + "' " + JSON.stringify(data));

	// Send to all sessions
	for(var i in this._sessions)
	{
		var session = this._sessions[i];
		if(session != exclude)
		{
			session.send(event, data);
		}
	}
}

// Get list with basic player info
Table.prototype.getPlayerList = function()
{
	var players = [];
	for(var i in this._players)
	{
		var session = this._players[i];
		players.push(session.player);
	}

	return players;
}

// Get list with all cards on the table
Table.prototype.getCardList = function()
{
	var cards = [];
	for(var i in this._cards)
	{
		var card = this._cards[i];
		cards.push
		({
			'id': card.id,
			'sprite': card.sprite,
			'x': card.x,
			'y': card.y,
			'z': card.z,
			'rot': card.rot,
			'flip': card.flip
		});
	}

	return cards;
}

// Adds a new, unidentified session
Table.prototype.addSession = function(session)
{
	// Already identified, or already in table?
	if(session.nick || session.id in this._sessions)
	{
		return;
	}

	// Add to all sessions
	session.table = this;
	this._sessions[session.id] = session;

	// Is this a table?
	if(session.type == 'table')
	{
		// Add to tables!
		this._tables[session.id] = session;

		// Table has no host yet?
		if(!this._host)
		{
			this._host = session;
		}
	}

	// Welcome!
	console.log(session + ' joined #' + this.id);
}

// Removes a session, optionally destroys the table
Table.prototype.removeSession = function(session)
{
	// Session in table?
	if(session.id in this._sessions)
	{
		// Remove from session map
		delete this._sessions[session.id];

		// Identified as player?
		if(session.player)
		{
			delete this._players[session.player.id];
			this.broadcast('removePlayer', { id: session.player.id });
		}

		// Ciao!
		console.log(session + ' left #' + this.id);

		// Destroy table?
		if(this._sessions.length == 0)
		{
			TableMaster.removeTable(this);
			console.log(this + ' is empty now and should be destroyed');
		}
	}
}

// Check nickname
Table.prototype.isValidNickname = function(nick)
{
	// Check length
	var result = {};
	if(nick.length < 2)
	{
		result.error = 'tooShort';
	}
	else if(nick.length > 32)
	{
		result.error = 'tooLong';
	}

	// Player with nick exists? (case insensitive)
	else
	{
		var nickLowerCase = nick.toLowerCase();
		var inUse = false;
		for(var i in this._players)
		{
			var session = this._players[i];
			if(session.player.nick.toLowerCase() == nickLowerCase)
			{
				inUse = true;
				break;
			}
		}

		// Evaluate result
		if(inUse)
		{
			result.error = 'inUse';
		}
		else
		{
			result.ok = true;
		}
	}

	// The result either has an 'error' or 'ok'
	return result;
}

// Add and announce a player
Table.prototype.addPlayer = function(session, info)
{
	// Already a player?
	if(session.player)
	{
		return false;
	}
	else
	{
		// Set player info and allocate ID
		session.player = info;
		session.player.id = ++this._playerIdAllocator;

		// Add to player list and announce to all table sessions
		this._players[session.player.id] = session;
		this.broadcast('addPlayer', session.player);

		// Yay!
		return true;
	}
}

// String representation
Table.prototype.toString = function()
{
	return 'Table #' + this.id;
}

// Expose class
module.exports = Table;