package org.openhabbo;

import org.openhabbo.access.AccessControl;
import org.openhabbo.catalog.Catalog;
import org.openhabbo.core.Logger;
import org.openhabbo.core.PropertiesBox;
import org.openhabbo.core.ShutdownHook;
import org.openhabbo.items.ItemManager;
import org.openhabbo.modules.ModuleManager;
import org.openhabbo.net.NetManager;
import org.openhabbo.remote.RemoteServer;
import org.openhabbo.rooms.RoomManager;
import org.openhabbo.rooms.RoomNavigator;
import org.openhabbo.sessions.SessionManager;
import org.openhabbo.storage.SQLDatabase;
import org.openhabbo.users.UserManager;

/**
 * A new Habbo Hotel (www.habbo.com) game server emulator written by Nils 'Nillus' W.
 */
public final class OpenHABBO
{
	private boolean alive = false;
	private long startTime = 0;
	private String propsFile = null;
	private PropertiesBox props = null;
	private SQLDatabase db = null;
	private ModuleManager modMgr = null;

	public OpenHABBO(String propsFile)
	{
		this.alive = false;
		this.startTime = 0;
		this.propsFile = propsFile;
		this.props = new PropertiesBox();
		this.db = null;
		this.modMgr = new ModuleManager();

		System.out.println("1111111111111111111111111111111111111111111");
		System.out.println("1100000000000000000000000000000000000000011");
		System.out.println("1101111         OpenHABBO           1111011");
		System.out.println("1101111111111111111111111111111111111111011");
		System.out.println("1101111  COPYRIGHT (C) 2011 Nils W. 1111011");
		System.out.println("1101111111111111111111111111111111111111011");
		System.out.println("1101111      Parts written by:      1111011");
		System.out.println("1101111     Roy 'Meth0d' de Jong    1111011");
		System.out.println("1101111     Joe 'Joeh' Hegarty      1111011");
		System.out.println("1101111111111111111111111111111111111111011");
		System.out.println("1101111     www.openhabbo.org       1111011");
		System.out.println("1100000000000000000000000000000000000000011");
		System.out.println("1111111111111111111111111111111111111111111");
		System.out.println();
	}

	public boolean reloadProperties()
	{
		boolean succes = this.props.load(this.propsFile);
		if(succes)
		{
			Logger.getInstance().info("OpenHABBO", String.format("Loaded %d properties.", this.props.size()));
		}
		else
		{
			Logger.getInstance().warning("OpenHABBO", "Failed loading system properties.");
		}
		return succes;
	}

	public boolean start()
	{
		// Log startup time
		long beginTime = System.currentTimeMillis();
		Logger.getInstance().info("OpenHABBO", "Starting...");

		// Load system properties
		if(!this.reloadProperties())
		{
			return false;
		}

		// Prepare SQLDatabase
		this.db = new SQLDatabase(this.props);
		if(this.db.test())
		{
			Logger.getInstance().info("SQLDatabase", "Connection successful.");
			if(!this.db.getQueryQueue().open())
			{
				return false;
			}
			else
			{
				this.db.getConnMgr().startMonitor();
			}
		}
		else
		{
			Logger.getInstance().warning("SQLDatabase", "Connection failed.");
			return false;
		}

		// Load modules/plugins
		this.modMgr.reloadModules();

		// Load accesscontrol
		AccessControl.createInstance();
		AccessControl.getInstance().reloadUserRoles();

		// Load user manager
		UserManager.createInstance();

		// Load session manager
		int expectedSessions = this.props.getInt("sessions.expected", 100);
		SessionManager.createInstance(expectedSessions);

		// Load item definitions
		ItemManager.createInstance();
		ItemManager.getInstance().reloadDefinitions();

		// Create RoomManager, spawn initial workers and reload model data
		int numRoomWorkers = this.props.getInt("rooms.workers", Runtime.getRuntime().availableProcessors());
		RoomManager.createInstance(numRoomWorkers);
		RoomManager.getInstance().reloadModels();

		// Load Navigator, cache stuff, etc
		int navigatorRefreshTime = this.props.getInt("rooms.navigator.refreshtime", 5);
		RoomNavigator.createInstance(navigatorRefreshTime);
		RoomNavigator.getInstance().reload();

		// Load Catalog
		Catalog.createInstance();
		Catalog.getInstance().reload();

		// Setup network manager for game clients
		int portNumber = this.props.getInt("net.portnumber", 1337);
		int numNetWorkers = this.props.getInt("net.workers", Runtime.getRuntime().availableProcessors());
		NetManager.createInstance(portNumber, numNetWorkers);
		NetManager.getInstance().reloadBlacklist();
		if(!NetManager.getInstance().getListener().start())
		{
			return false;
		}
		
		// Setup remoteserver for RCON'ish issueing of commands
		int remotePortNumber = OpenHABBO.getInstance().getProperties().getInt("remote.portnumber", 1338);
		RemoteServer.createInstance(remotePortNumber);
		RemoteServer.getInstance().addAllowedIp("0:0:0:0:0:0:0:1%0");
		RemoteServer.getInstance().addAllowedIp("127.0.0.1");
		if(!RemoteServer.getInstance().start())
		{
			return false;
		}

		// Start module pulsing
		this.modMgr.start();

		// Attach shutdown hook
		Runtime.getRuntime().addShutdownHook(new ShutdownHook(this));

		// Startup okay!
		this.alive = true;
		this.startTime = System.currentTimeMillis();
		Logger.getInstance().info("OpenHABBO", String.format("Started in %d ms.", (this.startTime - beginTime)));
		return true;
	}

	public void stop(String reason)
	{
		// Only start if not already shutdown'd yet
		if(!this.alive)
		{
			return;
		}
		Logger.getInstance().info("OpenHABBO", String.format("Entering shutdown sequence... (reason: %s)", reason));

		// Stop the remote listener
		RemoteServer.getInstance().stop();
		
		// Stop the game listener
		NetManager.getInstance().getListener().stop();

		// Shutdown game subsystems
		SessionManager.getInstance().shutdown();
		RoomManager.getInstance().shutdown();

		// Unload modules/plugins
		this.modMgr.unloadModules();
		this.modMgr.stop();

		// Stop net workers, no more reading and sending from now on
		NetManager.getInstance().stopWorkers();

		// Close SQLDatabase
		this.db.getQueryQueue().close();
		this.db.getConnMgr().closeAll();
		this.db.getConnMgr().stopMonitor();

		// Sequence complete!
		this.alive = false;
		Logger.getInstance().info("OpenHABBO", "Shutdown sequence completed.");
	}

	public String versionString()
	{
		return "OpenHABBO v0.8.4-dev 'Skylark'";
	}

	public boolean alive()
	{
		return this.alive;
	}

	public long getStartTime()
	{
		return this.startTime;
	}

	public PropertiesBox getProperties()
	{
		return this.props;
	}

	public SQLDatabase getDb()
	{
		return this.db;
	}

	public ModuleManager getModMgr()
	{
		return this.modMgr;
	}

	/* Singleton */
	private static OpenHABBO instance;

	public final static OpenHABBO getInstance()
	{
		return instance;
	}

	public static OpenHABBO createInstance(String propsFile)
	{
		OpenHABBO.instance = new OpenHABBO(propsFile);
		Logger.getInstance().info("OpenHABBO", OpenHABBO.instance.versionString());
		return OpenHABBO.instance;
	}
}
