package org.openhabbo;

import mod_remote.RemoteServer;

import org.openhabbo.access.AccessControl;
import org.openhabbo.catalog.Catalog;
import org.openhabbo.core.Logger;
import org.openhabbo.core.PropertiesBox;
import org.openhabbo.core.ShutdownHook;
import org.openhabbo.items.ItemManager;
import org.openhabbo.modules.ModuleManager;
import org.openhabbo.net.NetManager;
import org.openhabbo.rooms.RoomManager;
import org.openhabbo.rooms.RoomNavigator;
import org.openhabbo.sessions.SessionManager;
import org.openhabbo.storage.SQLDatabase;
import org.openhabbo.support.SupportCenter;
import org.openhabbo.users.UserManager;
import org.openhabbo.util.StartupCleanup;

/**
 * A new Habbo Hotel (www.habbo.com) game server emulator written by Nils 'Nillus' W.
 */
public final class OpenHABBOphat
{
	private boolean alive;
	private long startTime;
	private String propsFile;
	private PropertiesBox props;
	private SQLDatabase db;
	private ModuleManager modMgr;
	private RemoteServer remoteServer;
	private NetManager netMgr;
	private SessionManager sessionMgr;

	private AccessControl accessCtrl;
	private UserManager userMgr;
	private ItemManager itemMgr;
	private RoomManager roomMgr;
	private RoomNavigator navigator;
	private Catalog catalog;
	private SupportCenter supportCenter;

	public OpenHABBOphat(PropertiesBox props)
	{
		this.props = new PropertiesBox();

		System.out.println("1111111111111111111111111111111111111111111");
		System.out.println("1100000000000000000000000000000000000000011");
		System.out.println("1101111         OpenHABBO           1111011");
		System.out.println("1101111111111111111111111111111111111111011");
		System.out.println("1101111  COPYRIGHT (C) 2011 Nils W. 1111011");
		System.out.println("1101111111111111111111111111111111111111011");
		System.out.println("1101111      Parts written by:      1111011");
		System.out.println("1101111     Roy 'Meth0d' de Jong    1111011");
		System.out.println("1101111     Joe 'Joeh' Hegarty      1111011");
		System.out.println("1101111111111111111111111111111111111111011");
		System.out.println("1101111     www.openhabbo.org       1111011");
		System.out.println("1100000000000000000000000000000000000000011");
		System.out.println("1111111111111111111111111111111111111111111");
		System.out.println();
	}

	public boolean reloadProperties()
	{
		boolean succes = this.props.load(this.propsFile);
		if(succes)
		{
			Logger.getInstance().info("OpenHABBO", String.format("Loaded %d properties.", this.props.size()));
		}
		else
		{
			Logger.getInstance().warning("OpenHABBO", "Failed loading system properties.");
		}
		return succes;
	}

	public boolean start()
	{
		// Log startup time
		long beginTime = System.currentTimeMillis();
		Logger.getInstance().info("OpenHABBO", "Starting...");

		// Load system properties
		if(!this.reloadProperties())
		{
			return false;
		}

		// Prepare SQLDatabase
		this.db = new SQLDatabase(this.props);
		if(this.db.test())
		{
			Logger.getInstance().info("SQLDatabase", "Connection successful.");
			if(!this.db.getQueryQueue().open())
			{
				return false;
			}
			else
			{
				this.db.getConnMgr().startMonitor();
			}
		}
		else
		{
			Logger.getInstance().warning("SQLDatabase", "Connection failed.");
			return false;
		}

		// Load modules/plugins
		this.modMgr = new ModuleManager();
		this.modMgr.reloadModules();

		// Create SessionManager
		int expectedSessions = this.props.getInt("sessions.expected", 100);
		this.sessionMgr = new SessionManager(expectedSessions);

		// Create AccessControl
		this.accessCtrl = new AccessControl();
		this.accessCtrl.reloadUserRoles();

		// Create user manager
		this.userMgr = new UserManager();

		// Create ItemManager + load definitions
		this.itemMgr = new ItemManager();
		this.itemMgr.reloadDefinitions();

		// Create RoomManager, spawn initial workers and reload model data
		int expectedRoomInstances = this.props.getInt("rooms.instances.expected", 50);
		int initialRoomWorkers = (this.props.get("rooms.workers.initial", "auto").equals("auto")) ? ((expectedRoomInstances / 2) / RoomManager.MAX_LOAD_PER_WORKER) : this.props.getInt("rooms.workers.initial", 0);
		this.roomMgr = new RoomManager(expectedRoomInstances, initialRoomWorkers);
		this.roomMgr.reloadModels();

		// Load Navigator, cache stuff, etc
		this.navigator = new RoomNavigator(OpenHABBO.getInstance().getProperties().getInt("rooms.navigator.refreshtime", 5));
		this.navigator.reload();

		// Load Catalog
		this.catalog = new Catalog();
		this.catalog.reload();

		// Start SupportCenter
		this.supportCenter = new SupportCenter();
		this.supportCenter.start();

		// Setup network manager for game clients
		int portNumber = this.props.getInt("net.portnumber", 1337);
		int initialNetWorkers = (this.props.get("net.workers.initial", "auto").equals("auto")) ? ((expectedSessions / 2) / NetManager.MAX_LOAD_PER_WORKER) : this.props.getInt("net.workers.initial", 0);
		this.netMgr = new NetManager(portNumber, initialNetWorkers);
		this.netMgr.reloadBlacklist();
		if(!this.netMgr.getListener().start())
		{
			return false;
		}

		// Create & start remoteserver for RCON'ish issueing of commands
		this.remoteServer = new RemoteServer(this.props.getInt("remote.portnumber", 1338));
		this.remoteServer.addAllowedIp("0:0:0:0:0:0:0:1%0");
		this.remoteServer.addAllowedIp("127.0.0.1");
		this.remoteServer.start();

		// Attach shutdown hook
		Runtime.getRuntime().addShutdownHook(new ShutdownHook(this));

		// Clean some database things
		StartupCleanup.cleanup();

		// Startup okay!
		this.alive = true;
		this.startTime = System.currentTimeMillis();
		Logger.getInstance().info("OpenHABBO", String.format("Started in %d ms.", (this.startTime - beginTime)));
		return true;
	}

	public void stop(String reason)
	{
		// Only start if not already shutdown'd yet
		if(!this.alive)
		{
			return;
		}
		Logger.getInstance().info("OpenHABBO", String.format("Initiating shutdown sequence... (reason: %s)", reason));

		// Stop the remote command server
		this.remoteServer.stop();

		// Stop the game listener
		this.netMgr.getListener().stop();

		// Shutdown game subsystems
		this.sessionMgr.onShutdown();
		this.roomMgr.onShutdown();
		this.supportCenter.stop();

		// Stop net workers, no more reading and sending from now on
		this.netMgr.stopWorkers();

		// Unload modules/plugins
		this.modMgr.unloadModules();

		// Close SQLDatabase
		this.db.getQueryQueue().close();
		this.db.getConnMgr().closeAll();
		this.db.getConnMgr().stopMonitor();

		// Sequence complete!
		this.alive = false;
		Logger.getInstance().info("OpenHABBO", "Shutdown sequence completed.");
	}

	public String versionString()
	{
		return "OpenHABBO v0.8.4-dev 'Skylark'";
	}

	public boolean alive()
	{
		return this.alive;
	}

	public long getStartTime()
	{
		return this.startTime;
	}

	public PropertiesBox getProperties()
	{
		return this.props;
	}

	public SQLDatabase getDb()
	{
		return this.db;
	}

	public ModuleManager getModMgr()
	{
		return this.modMgr;
	}

	/*
	public RemoteServer getRemoteServer()
	{
		return this.remoteServer;
	}

	public NetManager getNetMgr()
	{
		return this.netMgr;
	}

	public SessionManager getSessionMgr()
	{
		return this.sessionMgr;
	}

	public AccessControl getAccessCtrl()
	{
		return this.accessCtrl;
	}

	public UserManager getUserMgr()
	{
		return this.userMgr;
	}

	public ItemManager getItemMgr()
	{
		return this.itemMgr;
	}

	public RoomManager getRoomMgr()
	{
		return this.roomMgr;
	}

	public RoomNavigator getNavigator()
	{
		if(this.navigator.requiresRefresh())
		{
			this.navigator.refresh();
		}
		return this.navigator;
	}

	public Catalog getCatalog()
	{
		return this.catalog;
	}

	public SupportCenter getSupportCenter()
	{
		return this.supportCenter;
	}*/

	/* Singleton */
	private static OpenHABBO instance;

	public static OpenHABBO getInstance()
	{
		return instance;
	}

	public static OpenHABBO createInstance(String propsFile)
	{
		OpenHABBO instance = new OpenHABBO(propsFile);
		Logger.getInstance().info("OpenHABBO", instance.versionString());
		OpenHABBO.instance = instance;
		return instance;
	}
}
