package org.openhabbo.chatcommands;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.openhabbo.access.UserRole;
import org.openhabbo.core.Logger;
import org.openhabbo.sessions.Session;

public final class ChatCmdRegistry
{
	private final static Map<String, ChatCommandHandler> entries = new HashMap<String, ChatCommandHandler>();

	private ChatCmdRegistry()
	{
		
	}
	
	public final static void register(String cmd, ChatCommandHandler handler)
	{
		entries.put(cmd, handler);
	}

	public final static void unregister(String name)
	{
		entries.remove(name);
	}

	public final static int count()
	{
		return entries.size();
	}
	
	public static Map<String, ChatCommandHandler> getAvailableChatCommands(UserRole role)
	{
		Map<String, ChatCommandHandler> available = new HashMap<String, ChatCommandHandler>();
		for(Entry<String, ChatCommandHandler> cmd : entries.entrySet())
		{
			String right = cmd.getValue().getRequiredRight();
			if(right == null || role.hasRight(right))
			{
				available.put(cmd.getKey(), cmd.getValue());
			}
		}
		return available;
	}

	public static boolean parseAndHandleCommand(Session session, String text)
	{
		// Actually a chat command?
		if(text.indexOf(ChatCommandHandler.COMMAND_HEAD) != 0)
		{
			// Doesn't start with the COMMAND_HEAD character
			return false;
		}

		// Attempt to separate command and body
		String[] segments = text.substring(1).split(" ", 2);

		// Handler registered?
		ChatCommandHandler handler = entries.get(segments[0]);
		if(handler == null)
		{
			// Doesn't exist...
			return false;
		}

		// User has the required right?
		String right = handler.getRequiredRight();
		if(right != null && !session.getUser().role.hasRight(right))
		{
			return false;
		}

		// Invoke handler and catch exceptions
		String body = (segments.length == 2 ? segments[1] : null);
		try
		{
			handler.handle(session, body);
		}
		catch(Exception ex)
		{
			Logger.getInstance().error("ChatCmdRegistry", String.format("Error processing chatcommand '%s'. (body=%s)", segments[0], body), ex);
		}

		// Handled
		return true;
	}
}
