package org.openhabbo.items;

import java.util.List;

import org.openhabbo.OpenHABBO;
import org.openhabbo.rooms.instances.RoomAvatar;
import org.openhabbo.rooms.instances.RoomInstance;
import org.openhabbo.rooms.instances.RoomMap;
import org.openhabbo.rooms.instances.RoomTile;
import org.openhabbo.sessions.EServerMessage;
import org.openhabbo.sessions.Session;
import org.openhabbo.sessions.messages.SerializableObject;
import org.openhabbo.sessions.messages.ServerMessage;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLHelper;
import org.openhabbo.users.User;

public abstract class Item implements SerializableObject
{
	public final static String ERROR_CUSTOMDATA = Character.toString((char) 1);

	public int id;
	public ItemDefinition definition;
	public int ownerId;
	public int roomId;
	public int x;
	public int y;
	public float z;
	public byte rotation;
	public String wallPosition;
	public String customData;
	public int teleporterId;

	public RoomInstance room;

	private int pulseTimer;
	private boolean requiresDbUpdate;

	public boolean inRoom()
	{
		return(this.roomId > 0);
	}

	public boolean is(char flag)
	{
		return this.definition.is(flag);
	}

	public void requireDbUpdate()
	{
		requiresDbUpdate = true;
	}

	public void dbUpdated()
	{
		requiresDbUpdate = false;
	}

	public boolean requiresDbUpdate()
	{
		return requiresDbUpdate;
	}

	public boolean requiresPulse()
	{
		return(this.pulseTimer > 0);
	}

	protected boolean hasTimeout()
	{
		return this.requiresPulse();
	}

	protected void setTimeout(int pulses)
	{
		this.pulseTimer = pulses;
	}

	protected void clearTimeout()
	{
		this.pulseTimer = 0;
	}

	public void pulse()
	{
		// Pulse!
		this.pulseTimer--;
		this.onPulse();

		// End of timeout?
		if(this.pulseTimer <= 0)
		{
			this.clearTimeout();
			this.onTimeout();
		}
	}

	public void setPersistentData(String data)
	{
		this.customData = data;
		if(this.inRoom())
		{
			this.refreshInRoom();
			this.requireDbUpdate();
		}
		else
		{
			this.update();
		}
	}

	public void setTemporaryData(String data)
	{
		String backup = this.customData;
		this.customData = data;
		this.refreshInRoom();
		this.customData = backup;
	}

	public void refreshInRoom()
	{
		if(this.inRoom())
		{
			ServerMessage msg = new ServerMessage();
			if(this.is(EItemBehavior.FLOORITEM))
			{
				msg.init(EServerMessage.ObjectDataUpdate);
				msg.appendStr(this.id);
				msg.appendStr(this.customData);
			}
			else if(this.is(EItemBehavior.WALLITEM))
			{
				msg.init(EServerMessage.ItemUpdate);
				msg.appendObject(this);
			}
			this.room.broadcast(msg);
		}
	}

	public RoomTile getTile()
	{
		return new RoomTile(this.x, this.y);
	}

	public RoomTile getTileInFront()
	{
		int x = this.x;
		int y = this.y;
		if(this.rotation == 0)
		{
			x--;
		}
		else if(this.rotation == 2)
		{
			x++;
		}
		else if(this.rotation == 4)
		{
			y++;
		}
		else if(this.rotation == 6)
		{
			y--;
		}
		return new RoomTile(x, y);
	}

	public RoomTile getTileBehind()
	{
		int x = this.x;
		int y = this.y;
		if(this.rotation == 0)
		{
			y++;
		}
		else if(this.rotation == 2)
		{
			x--;
		}
		else if(this.rotation == 4)
		{
			y--;
		}
		else if(this.rotation == 6)
		{
			x++;
		}
		return new RoomTile(x, y);
	}

	public void update()
	{
		if(this.id > 0)
		{
			if(this.is(EItemBehavior.FLOORITEM))
			{
				OpenHABBO.getInstance().getDb().execute("UPDATE items SET owner_id=" + this.ownerId + ",room_id=" + this.roomId + ",x=" + this.x + ",y=" + this.y + ",z=" + this.z + ",rotation=" + this.rotation + ",customdata='" + SQLHelper.escape(this.customData) + "' WHERE id=" + this.id);
			}
			else if(this.is(EItemBehavior.WALLITEM))
			{
				OpenHABBO.getInstance().getDb().execute("UPDATE items SET owner_id=" + this.ownerId + ",room_id=" + this.roomId + ",wallposition = '" + SQLHelper.escape(this.wallPosition) + "',customdata='" + SQLHelper.escape(this.customData) + "' WHERE id=" + this.id);
			}
			this.dbUpdated();
		}
	}

	public void delete()
	{
		if(this.id > 0)
		{
			OpenHABBO.getInstance().getDb().execute("DELETE FROM items WHERE id=" + this.id);
		}
	}

	public String generatePurchaseCustomData(User user, String rawData)
	{
		// rawData is discarded, impl classes should return proper-formatted data, customdata is only allowed on a few item types, and this is how we deal with 'em
		return null;
	}

	public void onPurchase(List<Item> shipping)
	{
		// Create this item and add to shipping
		ItemManager.getInstance().storeItem(this);
		shipping.add(this);

		// Optionally overridden by impl class
	}

	public boolean onPlace(Session session)
	{
		// Optionally overridden by impl class
		return true;
	}

	public boolean onPickup(Session session)
	{
		// Optionally overridden by impl class
		return true;
	}

	public void onDelete()
	{
		// Optionally overridden by impl class
	}

	public void onTrigger(Session session, int instruction)
	{
		// Optionally overridden by impl class
	}

	protected void onPulse()
	{
		// Optionally overridden by impl class
	}

	protected void onTimeout()
	{
		// Optionally overridden by impl class
	}
	
	public void onGenerateFloorMapTile(RoomMap map, RoomTile tile)
	{
		if(this.is(EItemBehavior.CAN_STAND_ON_TOP))
		{
			map.height[tile.x][tile.y] = (this.z + this.definition.height);
		}
		else if(this.is(EItemBehavior.CAN_SIT_ON_TOP))
		{
			map.state[tile.x][tile.y] = RoomMap.INTERACTIVE;
		}
		else if(this.is(EItemBehavior.CAN_LAY_ON_TOP))
		{
			map.state[tile.x][tile.y] = RoomMap.INTERACTIVE;
			if(this.rotation == 0 || this.rotation == 4)
			{
				map.redirect[tile.x][tile.y] = new RoomTile(tile.x, this.y);
			}
			else
			{
				map.redirect[tile.x][tile.y] = new RoomTile(this.x, tile.y);
			}
		}
		else
		{
			if(map.state[tile.x][tile.y] != RoomMap.INTERACTIVE)
			{
				map.state[tile.x][tile.y] = RoomMap.CLOSED;
			}
		}
	}
	
	public boolean onAvatarOnTop(RoomAvatar avatar)
	{
		if(this.is(EItemBehavior.CAN_SIT_ON_TOP))
		{
			avatar.z = this.z;
			avatar.headRotation = avatar.bodyRotation = this.rotation;
			avatar.addStatus("sit", Float.toString(this.definition.height));
			return true;
		}
		else if(this.is(EItemBehavior.CAN_LAY_ON_TOP))
		{
			avatar.z = this.z;
			avatar.headRotation = avatar.bodyRotation = this.rotation;
			avatar.addStatus("lay", Float.toString(this.definition.height) + " null");
			return true;
		}
		
		return false;
	}

	@Override
	public void serialize(ServerMessage msg)
	{
		if(this.is(EItemBehavior.FLOORITEM))
		{
			msg.appendInt(this.id);
			msg.appendInt(this.definition.spriteId);
			msg.appendInt(this.x);
			msg.appendInt(this.y);
			msg.appendInt(this.rotation);
			msg.appendStr(Float.toString(this.z));
			msg.appendInt(0);
			msg.appendStr(this.customData);
			msg.appendInt(-1);
		}
		else if(this.is(EItemBehavior.WALLITEM))
		{
			msg.appendStr(this.id);
			msg.appendInt(this.definition.spriteId);
			msg.appendStr(this.wallPosition);
			if(this.customData == null || this.customData.isEmpty() || this.customData.indexOf(' ') == -1)
			{
				msg.appendStr(this.customData);
			}
			else
			{
				msg.appendStr(this.customData.split(" ")[0]);
			}
		}
	}

	public static Item parse(DatabaseResult row)
	{
		// Find definition
		ItemDefinition def = ItemManager.getInstance().getDefinition(row.getInt("definition_id"));
		if(def != null)
		{
			// Create instance
			Item item = def.newInstance();
			item.id = row.getInt("id");
			item.ownerId = row.getInt("owner_id");
			item.roomId = row.getInt("room_id");
			item.x = row.getInt("x");
			item.y = row.getInt("y");
			item.z = row.getFloat("z");
			item.rotation = row.getByte("rotation");
			item.wallPosition = row.getString("wallposition");
			item.customData = row.getString("customdata");
			if(item.parseExtraData(row))
			{
				return item;
			}
		}

		// Bad item
		return null;
	}

	protected boolean parseExtraData(DatabaseResult row)
	{
		// This method can be overridden by impl classes
		return true;
	}
}
