package org.openhabbo.modules;

import java.util.HashMap;
import java.util.Map;
import org.openhabbo.OpenHABBO;
import org.openhabbo.core.Logger;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLHelper;

public abstract class Module
{
	private boolean loaded;

	private long pulseInterval;
	private long lastPulseTime;

	private final Map<String, String> config = new HashMap<String, String>(0);
	private final ModulePayload payload = new ModulePayload();
	
	public boolean loaded()
	{
		return this.loaded;
	}

	public final boolean load()
	{
		// Already loaded?
		if(this.loaded)
		{
			return true;
		}

		// Inject payload, attempt to load
		this.payload.inject();
		this.loaded = this.onLoad();
		if(!this.loaded)
		{
			this.payload.eject();
		}
		return this.loaded;
	}

	public final void unload()
	{
		if(this.loaded)
		{
			this.loaded = false;
			if(this.hasPulseEnabled())
			{
				this.onPulse();
				this.disablePulseTimer();
			}
			this.payload.eject();
			this.onUnload();
		}
	}
	
	public final ModulePayload getAutoLoader()
	{
		return this.payload;
	}

	protected boolean onLoad()
	{
		// Optionally overriden
		return true;
	}

	protected void onUnload()
	{
		// Optionally overriden
	}

	protected void onPulse()
	{
		// Optionally overriden
	}

	protected void enablePulseTimer(int pulseSeconds)
	{
		this.pulseInterval = (pulseSeconds * 1000);
	}

	protected void disablePulseTimer()
	{
		this.pulseInterval = 0;
	}

	public boolean hasPulseEnabled()
	{
		return(this.pulseInterval > 0);
	}

	public final void pulse()
	{
		long nowTime = System.currentTimeMillis();
		if((nowTime - this.lastPulseTime) > this.pulseInterval)
		{
			try
			{
				this.onPulse();
			}
			catch(Exception ex)
			{
				Logger.getInstance().error(this, "Error on pulse.", ex);
			}
			finally
			{
				this.lastPulseTime = nowTime;
			}
		}
	}

	public String getStatusString()
	{
		// Optionally overriden to provide statistics etc
		return null;
	}

	public final void reloadConfig()
	{
		// Clear current data
		this.config.clear();

		// Load fresh data from database
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT `key`,`value` FROM system_moduleconfig WHERE module='" + this.getMetaInfo().name() + "'");
		while(result.row())
		{
			this.config.put(result.getString("key"), result.getString("value"));
		}
		result.release();
	}

	public final String getConfigValue(String key)
	{
		return this.config.get(key);
	}

	public final void setConfigValue(String key, String value)
	{
		// Get rid of old value
		this.deleteConfigValue(key);

		// Insert new value
		this.config.put(key, value);
		OpenHABBO.getInstance().getDb().execute("DELETE FROM system_moduleconfig WHERE module='" + this.getMetaInfo().name() + "' AND `key`='" + SQLHelper.escape(key) + "'");
		OpenHABBO.getInstance().getDb().execute("INSERT INTO system_moduleconfig(module,`key`,`value`) VALUES ('" + this.getMetaInfo().name() + "','" + SQLHelper.escape(key) + "','" + SQLHelper.escape(value) + "')");
	}

	public final void deleteConfigValue(String key)
	{
		if(this.config.containsKey(key))
		{
			this.config.remove(key);
			OpenHABBO.getInstance().getDb().execute("DELETE FROM system_moduleconfig WHERE module='" + this.getMetaInfo().name() + "' AND `key`='" + SQLHelper.escape(key) + "'");
		}
	}
	
	public final ModuleInfo getMetaInfo()
	{
		return this.getClass().getAnnotation(ModuleInfo.class);
	}
	
	@Override
	public final String toString()
	{
		return this.getMetaInfo().name() + "/" + this.getMetaInfo().version();
	}
}
