package org.openhabbo.modules;

import java.util.HashMap;
import java.util.Map;

import org.openhabbo.OpenHABBO;
import org.openhabbo.core.Logger;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLHelper;

public abstract class Module
{
	public final String name;
	public final Version version;
	public final String description;
	public final String author;
	public final String url;

	private long pulseInterval;
	private long lastPulseTime;
	private final Map<String, Version> requirements = new HashMap<String, Version>(0);
	private final Map<String, String> config = new HashMap<String, String>(0);

	public Module(Version version, String description, String author, String url)
	{
		this.name = this.getClass().getSimpleName();
		this.version = version;
		this.description = description;
		this.author = author;
		this.url = url;
	}

	public final void require(String module)
	{
		this.require(module, null);
	}

	public final void require(String module, Version version)
	{
		this.requirements.put(module, version);
	}

	public final Map<String, Version> getRequirements()
	{
		return this.requirements;
	}

	public abstract boolean onLoad();

	public abstract void onUnload();

	protected void onPulse()
	{
		// Optionally overriden
	}
	
	public void enablePulseTimer(int pulseSeconds)
	{
		this.pulseInterval = (pulseSeconds * 1000);
	}
	
	public void disablePulseTimer()
	{
		this.pulseInterval = 0;
	}
	
	public boolean hasPulseEnabled()
	{
		return (this.pulseInterval > 0);
	}
	
	public final void pulse()
	{
		long nowTime = System.currentTimeMillis();
		if((nowTime - this.lastPulseTime) > this.pulseInterval)
		{
			try
			{
				this.onPulse();
			}
			catch(Exception ex)
			{
				Logger.getInstance().error(this, "Error on pulse.", ex);
			}
			finally
			{
				this.lastPulseTime = nowTime;
			}
		}
	}

	public String getStatusString()
	{
		// Optionally overriden to provide statistics etc
		return null;
	}

	public final void reloadConfig()
	{
		// Clear current data
		this.config.clear();

		// Load fresh data from database
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT module,`key`,`value` FROM system_moduleconfig");
		while(result.row())
		{
			this.config.put(result.getString("key"), result.getString("value"));
		}
		result.release();
	}

	public final String getConfigValue(String key)
	{
		return this.config.get(key);
	}

	public final void setConfigValue(String key, String value)
	{
		// Get rid of old value
		this.deleteConfigValue(key);

		// Insert new value
		this.config.put(key, value);
		OpenHABBO.getInstance().getDb().execute("DELETE FROM system_moduleconfig WHERE module='" + this.name + "' AND `key`='" + SQLHelper.escape(key) + "'");
		OpenHABBO.getInstance().getDb().execute("INSERT INTO system_moduleconfig(module,`key`,`value`) VALUES ('" + this.name + "','" + SQLHelper.escape(key) + "','" + SQLHelper.escape(value) + "')");
	}

	public final void deleteConfigValue(String key)
	{
		if(this.config.containsKey(key))
		{
			this.config.remove(key);
			OpenHABBO.getInstance().getDb().execute("DELETE FROM system_moduleconfig WHERE module='" + this.name + "' AND `key`='" + SQLHelper.escape(key) + "'");
		}
	}

	public final String toString()
	{
		return this.getClass().getSimpleName() + "/" + this.version.toString();
	}
}
