package org.openhabbo.net;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;

import org.openhabbo.core.Logger;
import org.openhabbo.sessions.messages.ClientMessage;
import org.openhabbo.util.Base64Encoding;
import org.openhabbo.util.CrossdomainPolicy;

public final class NetConnection
{
	private final static int READBUFFER_SIZE = 1024;
	private final static int SENDBUFFER_SIZE = (READBUFFER_SIZE * 8);

	public interface ConnectionHandler
	{
		public void onNetNewMessage(ClientMessage msg);

		public void onNetClose(String reason);
	}

	private SocketChannel channel;
	private SelectionKey key;
	private boolean closed;
	private NetWorker worker;
	private NetConnection.ConnectionHandler handler;
	private int msgFullLength;
	private ByteBuffer readBuffer;
	private ByteBuffer sendBuffer;

	public NetConnection(SocketChannel channel)
	{
		this.channel = channel;
	}

	public void closeSocket()
	{
		if(!this.closed)
		{
			this.closed = true;
			this.readBuffer = null;
			this.sendBuffer = null;
			if(this.worker != null)
			{
				this.worker.unregister(this);
			}
			try
			{
				this.channel.close();
			}
			catch(IOException ex)
			{
				Logger.getInstance().error("NetConnection", "Failed to close socket.", ex);
			}
		}
	}

	public void close(String reason)
	{
		this.closeSocket();
		if(this.handler != null)
		{
			this.handler.onNetClose(reason);
		}
	}

	public boolean assignWorker(NetWorker worker, Selector selector)
	{
		try
		{
			this.channel.configureBlocking(false);
			this.key = this.channel.register(selector, SelectionKey.OP_READ | SelectionKey.OP_WRITE, this);
			this.readBuffer = ByteBuffer.allocateDirect(NetConnection.READBUFFER_SIZE);
			this.sendBuffer = ByteBuffer.allocateDirect(NetConnection.SENDBUFFER_SIZE);
		}
		catch(IOException ex)
		{
			return false;
		}
		this.worker = worker;
		return true;
	}

	public void send(byte[] data)
	{
		try
		{
			if(this.key.isValid())
			{
				// Lock while writing
				synchronized(this.sendBuffer)
				{
					this.sendBuffer.put(data);
					this.key.interestOps(SelectionKey.OP_READ | SelectionKey.OP_WRITE);
				}
			}
		}
		catch(Exception ex)
		{
			Logger.getInstance().error(this.handler, String.format("Error enqueing %d bytes of data for sending, packet too large?", data.length), ex);
		}
	}

	public void onWrite() throws IOException
	{
		// Lock while reading
		synchronized(this.sendBuffer)
		{
			this.key.interestOps(SelectionKey.OP_READ);
			this.sendBuffer.flip();
			this.channel.write(this.sendBuffer);
			if(this.sendBuffer.hasRemaining())
			{
				this.sendBuffer.flip();
				this.key.interestOps(SelectionKey.OP_READ | SelectionKey.OP_WRITE);
			}
			else
			{
				this.sendBuffer.clear();
			}
		}
	}

	public boolean onRead() throws IOException
	{
		// Read all available data into the buffer
		int read = this.channel.read(this.readBuffer);
		if(read <= 0)
		{
			return false;
		}

		// Flip to readable mode
		this.readBuffer.flip();

		// Process the entire buffer
		while(this.readBuffer.hasRemaining())
		{
			// Waiting for new message?
			if(this.msgFullLength == 0)
			{
				byte[] header = new byte[3];
				this.readBuffer.get(header);
				if(header[0] == 64)
				{
					this.msgFullLength = Base64Encoding.decodeInt(header);
					if(this.msgFullLength <= 0 || this.msgFullLength > NetConnection.READBUFFER_SIZE)
					{
						return false;
					}
				}
				else
				{
					Logger.getInstance().debug(this.handler, "requesting policy file");
					this.send(CrossdomainPolicy.getXml().getBytes());
					return true;
				}
			}

			// Message complete?
			if(this.readBuffer.remaining() >= this.msgFullLength)
			{
				// Copy all msgdata into a new array
				byte[] bytes = new byte[this.msgFullLength];
				this.readBuffer.get(bytes);
				this.msgFullLength = 0;

				// Message received!
				ClientMessage msg = new ClientMessage(bytes, bytes.length);
				this.handler.onNetNewMessage(msg);
			}
		}

		// Prepare the buffer for the next data
		if(this.msgFullLength == 0)
		{
			this.readBuffer.clear();
		}
		else
		{
			this.readBuffer.flip();
		}
		return true;
	}

	public String getIpAddress()
	{
		return this.channel.socket().getInetAddress().getHostAddress();
	}

	public boolean closed()
	{
		return this.closed;
	}

	public NetConnection.ConnectionHandler getHandler()
	{
		return this.handler;
	}

	public void setHandler(NetConnection.ConnectionHandler handler)
	{
		this.handler = handler;
	}
}
