package org.openhabbo.rooms;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.openhabbo.OpenHABBO;
import org.openhabbo.core.Logger;
import org.openhabbo.rooms.instances.RoomInstance;
import org.openhabbo.rooms.instances.RoomWorker;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLHelper;

public final class RoomManager
{
	//private Map<Integer, Room> cache;
	private Map<String, RoomModel> models;
	private Map<Integer, RoomInstance> instances;
	private RoomWorker[] workers;

	private RoomManager(int numWorkers)
	{
		//this.cache = new ConcurrentHashMap<Integer, Room>();
		this.models = new ConcurrentHashMap<String, RoomModel>();
		this.instances = new ConcurrentHashMap<Integer, RoomInstance>();
		this.workers = new RoomWorker[numWorkers];
		for(int index = 0; index < numWorkers; index++)
		{
			this.workers[index] = new RoomWorker(index, this);
			this.workers[index].start();
			Logger.getInstance().info("RoomMgr", String.format("Spawned RoomWorker #%d.", index));
		}
	}

	public void shutdown()
	{
		// Shutdown workers
		for(int index = 0; index < this.workers.length; index++)
		{
			this.workers[index].stop();
			this.workers[index] = null;
		}
		Logger.getInstance().info("RoomMgr", "Stopped all workers.");

		// Shutdown roominstances?
		if(!this.instances.isEmpty())
		{
			Logger.getInstance().info("RoomMgr", "Unloading all room instances...");
			for(RoomInstance instance : this.instances.values())
			{
				instance.unload();
			}
			this.instances.clear();
			Logger.getInstance().info("SessionMgr", "All room instances unloaded.");
		}
	}

	public void reloadModels()
	{
		this.models.clear();

		// Load all model data and parse them
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT * FROM roommodels");
		while(result.row())
		{
			RoomModel model = RoomModel.parse(result);
			model.parseMap();
			this.models.put(model.name, model);
		}
		result.release();

		Logger.getInstance().info("RoomMgr", String.format("Reloaded %d room models.", this.models()));
	}

	public RoomModel getModel(String name)
	{
		return this.models.get(name);
	}

	public int models()
	{
		return this.models.size();
	}
	
	public RoomInstance getRoomInstance(int roomId, boolean allowCreate)
	{
		// Look for existing instance
		RoomInstance instance = this.instances.get(roomId);
		if(instance != null)
		{
			return instance;
		}

		// Allowed to create a new instance?
		if(allowCreate)
		{
			return this.createRoomInstance(roomId);
		}
		else
		{
			return null;
		}
	}

	private RoomWorker getLowestLoadedWorker()
	{
		int lowest = 0, bestIndex = -1;
		for(int index = 0; index < this.workers.length; index++)
		{
			int load = this.workers[index].load();
			if(bestIndex == -1 || load < lowest)
			{
				lowest = load;
				bestIndex = index;
			}
		}
		return this.workers[bestIndex];
	}

	private RoomInstance createRoomInstance(int roomId)
	{
		// Get room info
		Room info = this.getRoomInfo(roomId);
		if(info == null)
		{
			return null;
		}

		// Get room model
		RoomModel model = this.getModel(info.modelName);
		if(model == null)
		{
			Logger.getInstance().warning("RoomMgr", String.format("Room #%d refers to the missing room model \"%s\". Could not create room instance.", info.id, info.modelName));
			return null;
		}

		// Create instance and load data like furniture
		RoomInstance instance = new RoomInstance(info, model);
		try
		{
			instance.load();
		}
		catch(Exception ex)
		{
			Logger.getInstance().error("RoomMgr", String.format("Error loading instance of room #%d.", roomId), ex);
			return null;
		}

		// Assign to worker
		this.getLowestLoadedWorker().assign(instance);

		// Add to instance map
		this.instances.put(instance.getInfo().id, instance);

		// Instance created!
		Logger.getInstance().info("RoomMgr", String.format("Created instance for room #%d.", info.id));
		return instance;
	}

	public void unloadRoomInstance(int roomId)
	{
		// Attempt to remove from the instance map
		RoomInstance instance = this.instances.remove(roomId);

		// Instance removed? Process further unloading
		if(instance != null)
		{
			// Un-assign from worker
			this.workers[instance.getWorkerIndex()].unassign(roomId);

			// Unload/save data
			instance.unload();

			// Unloaded successfully!
			Logger.getInstance().info("RoomMgr", String.format("Unloaded instance of room #%d.", roomId));
		}
	}

	public Room getRoomInfo(int roomId)
	{
		// Lookup cache
		/*Room room = this.cache.get(roomId);
		if(room != null)
		{
			return room;
		}*/
		
		// Lookup room instances
		RoomInstance instance = this.getRoomInstance(roomId, false);
		if(instance != null)
		{
			return instance.getInfo();
		}

		// Lookup database
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r LEFT JOIN users u ON u.id = r.owner_id WHERE r.id =" + roomId);
		Room room = result.row() ? Room.parse(result) : null;
		result.release();
		return room;
	}
	
	public List<Room> getRoomsForUser(int userId)
	{
		ArrayList<Room> rooms = new ArrayList<Room>(5);

		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE r.owner_id=" + userId);
		while(result.row())
		{
			rooms.add(Room.parse(result));
		}
		result.release();

		return rooms;
	}

	public int createGuestRoom(int userId, String name, String modelName)
	{
		return OpenHABBO.getInstance().getDb().executeReturnKey("INSERT INTO rooms(owner_id,name,model_name) VALUES (" + userId + ",'" + SQLHelper.escape(name) + "','" + SQLHelper.escape(modelName) + "')");
	}

	public void deleteGuestRoom(int roomId)
	{
		OpenHABBO.getInstance().getDb().execute("DELETE FROM rooms WHERE id=" + roomId);
		OpenHABBO.getInstance().getDb().execute("DELETE FROM items WHERE room_id=" + roomId);
		this.unloadRoomInstance(roomId);
	}

	public Map<Integer, String> getRightsForRoom(int roomId)
	{
		// Initialize an empty map with init cap of 3
		Map<Integer, String> users = new HashMap<Integer, String>(3);

		// Query the user IDs + usernames
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT user_id,username FROM room_rights JOIN users ON id=user_id WHERE room_id=" + roomId);
		while(result.row())
		{
			users.put(result.getInt("user_id"), result.getString("username"));
		}
		result.release();

		// Return the users
		return users;
	}

	public void addRoomRights(int roomId, int userId)
	{
		OpenHABBO.getInstance().getDb().execute("INSERT INTO room_rights(room_id,user_id) VALUES(" + roomId + "," + userId + ")");
	}

	public void removeRoomRights(int roomId, int userId)
	{
		OpenHABBO.getInstance().getDb().execute("DELETE FROM room_rights WHERE room_id=" + roomId + " AND user_id=" + userId);
	}

	public int instances()
	{
		return this.instances.size();
	}

	/* Begin singleton */
	private static RoomManager instance;

	public final static RoomManager getInstance()
	{
		return instance;
	}

	public static void createInstance(int numWorkers)
	{
		instance = new RoomManager(numWorkers);
	}
	/* End singleton */

}
