package org.openhabbo.rooms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.openhabbo.OpenHABBO;
import org.openhabbo.core.Logger;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLHelper;

public class RoomNavigator
{
	private long lastRefresh;
	private int maxAge;

	private Object lock = new Object();
	private Map<Integer, RoomCategory> categories = new HashMap<Integer, RoomCategory>();
	private List<NavigatorFrontpageItem> frontpage = new ArrayList<NavigatorFrontpageItem>();
	private List<Room> busiestGuestRooms = new ArrayList<Room>();
	private List<Room> topRatedGuestRooms = new ArrayList<Room>();

	private RoomNavigator(int maxAgeSeconds)
	{
		this.maxAge = (maxAgeSeconds * 1000);
	}

	public void reload()
	{
		synchronized(this.lock)
		{
			// Load categories
			this.categories.clear();
			DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT * FROM navigator_categories ORDER BY ordernum ASC");
			while(result.row())
			{
				if(result.getBoolean("enabled"))
				{
					RoomCategory category = RoomCategory.parse(result);
					this.categories.put(category.id, category);
				}
			}

			// Load frontpage items
			this.frontpage.clear();
			result.query("SELECT id,name,bannertype,type,inner_id,image,image_src,enabled FROM navigator_frontpage ORDER BY ordernum ASC");
			while(result.row())
			{
				if(result.getBoolean("enabled"))
				{
					NavigatorFrontpageItem fpi = new NavigatorFrontpageItem();
					fpi.id = result.getInt("id");
					fpi.name = result.getString("name");
					fpi.bannerType = result.getInt("bannertype");
					fpi.image = result.getString("image");
					fpi.imageIsExternal = result.getString("image_src").equals("external");

					// Get inner item
					String innerType = result.getString("type");
					int innerId = result.getInt("inner_id");
					if(innerType.equals("room"))
					{
						fpi.room = RoomManager.getInstance().getRoomInfo(innerId);
					}
					else
					{
						fpi.category = this.getCategory(innerId);
					}

					// Inner item exists? (don't enable invalid frontpage items)
					if(fpi.room != null || fpi.category != null)
					{
						this.frontpage.add(fpi);
					}
					else
					{
						Logger.getInstance().warning("RoomNavigator", String.format("Frontpage item #%d (\"%s\") refers to the missing %s #%d. Item ignored.", fpi.id, fpi.name, innerType, innerId));
					}
				}
			}
			result.release();

			// Yay!
			Logger.getInstance().info("RoomNavigator", String.format("Reloaded %d guestroom categories and %d frontpage items.", this.categories(), this.frontpageItems()));
		}
	}

	public boolean requiresRefresh()
	{
		return((System.currentTimeMillis() - this.lastRefresh) > this.maxAge);
	}

	public void refresh()
	{
		synchronized(this.lock)
		{
			// Reload category rooms
			for(RoomCategory category : this.categories.values())
			{
				category.rooms.clear();
				DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE category_id=" + category.id + " ORDER BY users_now DESC,rating DESC LIMIT 40");
				while(result.row())
				{
					Room room = Room.parse(result);
					category.rooms.put(room.id, room);
				}
				result.release();
			}

			// Refresh toprated rooms
			this.topRatedGuestRooms.clear();
			DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE r.owner_id > 0 ORDER BY rating DESC LIMIT 40");
			while(result.row())
			{
				this.topRatedGuestRooms.add(Room.parse(result));
			}
			result.release();

			// Refresh frontpage items
			for(NavigatorFrontpageItem fpi : this.frontpage)
			{
				if(fpi.room != null)
				{
					Room newInfo = RoomManager.getInstance().getRoomInfo(fpi.room.id);
					if(newInfo != null)
					{
						fpi.room = newInfo;
					}
				}
			}

			// This is for debug only
			Logger.getInstance().debug("RoomNavigator", "Refreshed successfully.");
		}
	}

	public RoomCategory getCategory(int categoryId)
	{
		return this.categories.get(categoryId);
	}

	public int categories()
	{
		return this.categories.size();
	}

	public Collection<RoomCategory> getCategories()
	{
		return this.categories.values();
	}

	public int frontpageItems()
	{
		return this.frontpage.size();
	}

	public List<NavigatorFrontpageItem> getFrontpageItems()
	{
		return this.frontpage;
	}

	public List<Room> getBusiestGuestRooms()
	{
		return this.busiestGuestRooms;
	}

	public List<Room> getTopRatedRooms()
	{
		return this.topRatedGuestRooms;
	}

	public List<Room> getGuestRoomSearchResult(String criteria, boolean byTags)
	{
		ArrayList<Room> rooms = new ArrayList<Room>();

		criteria = SQLHelper.escape(criteria);
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id = r.owner_id WHERE u.username='" + criteria + "' OR r.name LIKE '" + criteria + "%' OR r.tags LIKE '%" + criteria + "%' LIMIT 30;");
		while(result.row())
		{
			rooms.add(Room.parse(result));
		}
		result.release();

		return rooms;
	}

	/* Begin singleton */
	private static RoomNavigator instance;

	public final static RoomNavigator getInstance()
	{
		if(instance.requiresRefresh())
		{
			instance.refresh();
		}
		return instance;
	}

	public static void createInstance(int maxAgeSeconds)
	{
		instance = new RoomNavigator(maxAgeSeconds);
	}
	/* End singleton */
}
