package org.openhabbo.rooms.instances;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

import org.openhabbo.core.DynamicObject;
import org.openhabbo.events.Event;
import org.openhabbo.items.Item;
import org.openhabbo.rooms.pathfinding.PathfinderNode;
import org.openhabbo.rooms.pathfinding.RotationCalculator;
import org.openhabbo.sessions.EServerMessage;
import org.openhabbo.sessions.messages.SerializableObject;
import org.openhabbo.sessions.messages.ServerMessage;
import org.openhabbo.util.ChatUtil;

public abstract class RoomAvatar extends DynamicObject implements SerializableObject
{
	public final static Event onPulse = new Event("RoomAvatar.OnPulse");
	public final static Event onUnidle = new Event("RoomAvatar.OnUnidle");

	// Room
	public int id;
	private RoomInstance room;
	private boolean invalidated;

	// Position
	public int x;
	public int y;
	public float z;
	public byte headRotation;
	public byte bodyRotation;
	public Item currentItem;
	
	// Statuses
	public final HashMap<String, String> statuses;
	public boolean requiresUpdate;

	public int idleTime;
	public int danceId;
	public int carryItemId;
	public int carryItemTimer;

	// Movement
	public int goalX;
	public int goalY;
	public int nextX;
	public int nextY;
	public float nextZ;
	public byte movementTries;
	private boolean moveLock;
	public boolean overrideNextTile;
	public List<PathfinderNode> path;
	public boolean startNewPath;

	public RoomAvatar()
	{
		this.statuses = new HashMap<String, String>();
		this.path = Collections.emptyList();
		this.clearMovement();
	}

	public RoomInstance getRoom()
	{
		return this.room;
	}

	public void setRoom(RoomInstance room)
	{
		this.room = room;
	}

	public void invalidate()
	{
		this.invalidated = true;
	}

	public final boolean invalidated()
	{
		return this.invalidated;
	}

	public void pulse()
	{
		if(RoomAvatar.onPulse.fire(this))
		{
			this.idleTime++;
		}

		// Deal with carry item
		if(this.carryItemId > 0)
		{
			this.carryItemTimer--;
			if(this.carryItemTimer <= 0)
			{
				this.stopCarryItem();
			}
		}
	}

	public void unidle()
	{
		if(RoomAvatar.onUnidle.fire(this))
		{
			this.idleTime = 0;
		}
	}

	public boolean at(RoomTile tile)
	{
		return this.at(tile.x, tile.y);
	}

	public boolean at(int x, int y)
	{
		return(this.x == x && this.y == y);
	}

	public boolean touches(RoomTile tile)
	{
		return this.touches(tile.x, tile.y);
	}

	public boolean touches(int x, int y)
	{
		return this.room.getMap()	.tilesTouch(this.x, this.y, x, y);
	}
	
	public void updatePosition(int x, int y, float z)
	{
		// Step off current item
		if(this.currentItem != null)
		{
			this.currentItem.onAvatarStepOff(this);
		}
		
		// Update the position
		this.x = x;
		this.y = y;
		this.z = z;
		this.requiresUpdate = true;
		
		// Step on new item
		for(Item item : this.room.getFloorItemsOnTile(this.x, this.y))
		{
			if(item.onAvatarStepOn(this))
			{
				this.currentItem = item;
				break;
			}
		}
	}

	public void requestNewTrip(RoomTile goal)
	{
		this.startMovingTo(goal.x, goal.y);
	}

	public void startMovingTo(int goalX, int goalY)
	{
		this.unidle();
		this.path.clear();
		this.goalX = goalX;
		this.goalY = goalY;
		this.nextX = this.nextY = -1;
		this.requiresUpdate = startNewPath = true;
	}

	public final boolean isMoving()
	{
		return(this.goalX != -1);
	}

	public void clearMovement()
	{
		// Clear path?
		if(!this.path.isEmpty())
		{
			this.path.clear();
		}

		// Clear movement variables
		this.movementTries = 0;
		this.goalX = this.goalY = this.nextX = this.nextY = -1;
	}

	public boolean canMove()
	{
		return !this.moveLock;
	}

	public void lockMovement()
	{
		this.moveLock = true;
		this.clearMovement();
	}

	public void unlockMovement()
	{
		this.moveLock = false;
	}

	public void lookTo(int tileX, int tileY, boolean lockBody)
	{
		if(this.nextX != -1 || (tileX == this.x && tileY == this.y) || this.statuses.containsKey("lay"))
		{
			return;
		}

		byte rotation = RotationCalculator.calculateDirection(this.x, this.y, tileX, tileY);
		int diff = this.bodyRotation - rotation;
		this.headRotation = this.bodyRotation;

		if(lockBody || this.statuses.containsKey("sit"))
		{
			if(this.bodyRotation == 2 || this.bodyRotation == 4)
			{
				if(diff > 0)
				{
					this.headRotation = (byte) (this.bodyRotation - 1);
				}
				else if(diff < 0)
				{
					this.headRotation = (byte) (this.bodyRotation + 1);
				}
			}
			else if(this.bodyRotation == 0 || this.bodyRotation == 6)
			{
				if(diff > 0)
				{
					this.headRotation = (byte) (this.bodyRotation - 1);
				}
				else if(diff < 0)
				{
					this.headRotation = (byte) (this.bodyRotation + 1);
				}
			}
		}
		else if(diff <= -2 || diff >= 2)
		{
			this.headRotation = rotation;
			this.bodyRotation = rotation;
		}
		else
		{
			this.headRotation = rotation;
		}
		this.unidle();
		this.requiresUpdate = true;
	}

	public void stopCarryItem()
	{
		this.carryItem(0);
	}

	public void carryItem(int itemId)
	{
		// Set the ID + timer
		this.carryItemId = itemId;
		if(itemId > 0)
		{
			this.carryItemTimer = 240;
		}
		else
		{
			this.carryItemTimer = 0;
		}

		// Notify the roominstance
		ServerMessage msg = new ServerMessage(EServerMessage.CarryObject);
		msg.appendInt(this.id);
		msg.appendInt(this.carryItemId);
		this.room.broadcast(msg);
	}

	public void chat(String text, boolean isShout)
	{
		// Broadcast chat to room
		ServerMessage msg = new ServerMessage(isShout ? EServerMessage.Shout : EServerMessage.Chat);
		msg.appendInt(this.id);
		msg.appendStr(text);
		msg.appendInt(ChatUtil.getChatGesture(text));
		this.room.broadcast(msg);

		// Unidle
		this.unidle();
		
		// Interact with all other room avatars
		for(RoomAvatar avatar : this.room.getAllAvatars())
		{
			if(avatar.id != this.id)
			{
				avatar.onChatFromAvatar(this, text, isShout);
			}
		}
	}

	protected abstract void onChatFromAvatar(RoomAvatar from, String text, boolean isShout);

	public final void addStatus(String key)
	{
		this.addStatus(key, null);
	}

	public final void addStatus(String key, String value)
	{
		this.statuses.put(key, value);
		this.requiresUpdate = true;
	}

	public final void removeStatus(String key)
	{
		this.statuses.remove(key);
		this.requiresUpdate = true;
	}
	
	public final void appendStatusString(ServerMessage msg)
	{
		msg.appendInt(this.id);
		msg.appendInt(this.x);
		msg.appendInt(this.y);
		msg.appendStr(this.z);
		msg.appendInt(this.headRotation);
		msg.appendInt(this.bodyRotation);
		msg.append('/');
		for(Entry<String, String> status : this.statuses.entrySet())
		{
			msg.append(status.getKey());
			if(status.getValue() != null)
			{
				msg.append(' ');
				msg.append(status.getValue());
			}
			msg.append('/');
		}
		msg.appendStr('/');
	}

	public abstract String getAvatarKey();
}
