package org.openhabbo.rooms.instances;

import java.util.ArrayList;
import java.util.List;

import org.openhabbo.core.Logger;
import org.openhabbo.items.Item;
import org.openhabbo.rooms.RoomModel;

public final class RoomMap
{
	public final static byte OPEN = (1 << 0);
	public final static byte CLOSED = (1 << 1);
	public final static byte INTERACTIVE = (1 << 2);
	
	private RoomModel model;
	
	public byte[][] state;
	public float[][] height;
	public boolean[][] avatars;
	public RoomTile[][] redirect;

	public RoomMap(RoomModel model)
	{
		this.model = model;
		this.state = new byte[model.getMaxX()][model.getMaxY()];
		this.height = new float[model.getMaxX()][model.getMaxY()];
		this.avatars = new boolean[model.getMaxX()][model.getMaxY()];
		this.redirect = new RoomTile[model.getMaxX()][model.getMaxY()];
	}

	public boolean tileExists(RoomTile tile)
	{
		return this.tileExists(tile.x, tile.y);
	}

	public boolean tileExists(int x, int y)
	{
		return(x >= 0 && x < this.model.getMaxX() && y >= 0 && y < this.model.getMaxY());
	}

	public boolean tileIsPassable(int x, int y)
	{
		return this.tileIsPassable(x, y, false);
	}
	
	public boolean tileIsPassable(RoomTile tile, boolean lastStep)
	{
		return this.tileIsPassable(tile.x, tile.y, lastStep);
	}

	public boolean tileIsPassable(int x, int y, boolean lastStep)
	{
		// Tile exists?
		if(!this.tileExists(x, y))
		{
			return false;
		}

		// Avatar here?
		if(this.avatars[x][y])
		{
			return false;
		}

		// Blocked?
		if(this.state[x][y] == RoomMap.CLOSED)
		{
			return false;
		}

		// Closed item tiles are only Passable with last step of path
		if(!lastStep && (this.state[x][y] == (RoomMap.CLOSED | RoomMap.INTERACTIVE)))
		{
			return false;
		}

		// Passable!
		return true;
	}

	public List<RoomTile> getAffectedTiles(Item obj)
	{
		return this.getAffectedTiles(obj.definition.length, obj.definition.width, obj.x, obj.y, obj.rotation);
	}

	public List<RoomTile> getAffectedTiles(byte length, byte width, int x, int y, byte rotation)
	{
		List<RoomTile> tiles = new ArrayList<RoomTile>();

		// Account rotation
		if(length != width)
		{
			if(rotation == 0 || rotation == 4)
			{
				byte swap = length;
				length = width;
				width = swap;
			}
		}
		
		// Determine tiles (within map) and add them
		for(int iX = x; iX < x + length; iX++)
		{
			for(int iY = y; iY < y + width; iY++)
			{
				if(this.tileExists(iX, iY))
				{
					tiles.add(new RoomTile(iX, iY));
				}
			}
		}

		return tiles;
	}

	public boolean tilesTouch(int x, int y, int x2, int y2)
	{
		return(Math.abs(x - x2) <= 1 && Math.abs(y - y2) <= 1);
	}
	
	public void reset()
	{
		// Get default floormap from room model
		char[][] modelMap = this.model.getMap();
		
		// Create default floor map
		for(int y = 0; y < this.model.getMaxY(); y++)
		{
			for(int x = 0; x < this.model.getMaxX(); x++)
			{
				// Grab tile
				char tile = modelMap[x][y];

				// Process tile
				this.redirect[x][y] = null;
				if(Character.isDigit(tile))
				{
					this.state[x][y] = RoomMap.OPEN;
					this.height[x][y] = Integer.parseInt(Character.toString(tile));
				}
				else
				{
					this.state[x][y] = RoomMap.CLOSED;
					this.height[x][y] = 0.0f;
				}
			}
		}
	}
	
	public void applyFloorItem(Item item)
	{
		// Process all tiles this item currently resides on
		for(RoomTile tile : this.getAffectedTiles(item))
		{
			// Out of range?
			if(!this.tileExists(tile))
			{
				Logger.getInstance().warning("RoomMap", String.format("Floor item #%d at tile [%d,%d] was residing on tiles outside the boundaries of the map. Floor item not applied to maps.", item.id, item.x, item.y));
				return;
			}
			
			// Increase heightmap
			if (item.z >= this.height[tile.x][tile.y])
			{
				this.height[tile.x][tile.y] = item.z;
			}
			
			// Top item on the tile? Tile state is determined by top item on the tile!
			if(item.z >= this.height[tile.x][tile.y])
			{
				item.onGenerateFloorMapTile(this, tile);
			}
		}
	}
	
	public RoomModel getModel()
	{
		return this.model;
	}
}
