package org.openhabbo.rooms.instances;

import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;

import org.openhabbo.core.Logger;
import org.openhabbo.rooms.RoomManager;

public final class RoomWorker implements Runnable
{
	private int index;
	private RoomManager mgr;

	private Thread thread;
	private boolean running;
	private List<Integer> toRemove;
	private Map<Integer, RoomInstance> instances;

	public RoomWorker(int index, RoomManager mgr)
	{
		this.index = index;
		this.mgr = mgr;
		this.toRemove = new CopyOnWriteArrayList<Integer>();
		this.instances = new ConcurrentHashMap<Integer, RoomInstance>(RoomManager.MAX_LOAD_PER_WORKER);
	}

	public void start()
	{
		if(!this.running)
		{
			this.running = true;
			this.thread = new Thread(this, this.toString());
			this.thread.start();
		}
	}

	public void stop()
	{
		this.running = false;
		this.thread.interrupt();
		this.instances.clear();
	}

	public void add(RoomInstance instance)
	{
		this.instances.put(instance.getInfo().id, instance);
	}

	public void remove(int roomId)
	{
		this.toRemove.add(roomId);
	}

	@Override
	public void run()
	{
		while(this.running)
		{
			// Measure start time, pulse all instances, measure end time
			long startTime = System.nanoTime();
			for(RoomInstance instance : this.instances.values())
			{
				try
				{
					instance.pulse();
					if(instance.requiresUnload())
					{
						this.mgr.unloadRoomInstance(instance.getInfo().id);
					}
				}
				catch(Exception ex)
				{
					Logger.getInstance().error(this, String.format("Error pulsing room instance #%d.", instance.getInfo().id), ex);
				}
			}
			if(this.toRemove.size() > 0)
			{
				for(Integer roomId : this.toRemove)
				{
					this.instances.remove(roomId);
				}
				this.toRemove.clear();
			}
			long endTime = System.nanoTime();

			// Now sleep the essential amount of ms
			double sleepTime = (endTime - startTime);
			sleepTime /= 1000000;
			sleepTime = (RoomInstance.PULSE_INTERVAL - sleepTime);
			if(sleepTime > 0.0)
			{
				try
				{
					Thread.sleep((long) sleepTime);
				}
				catch(InterruptedException ex)
				{
					
				}
			}
		}
	}

	public int index()
	{
		return this.index;
	}

	public RoomManager getMgr()
	{
		return this.mgr;
	}

	public int load()
	{
		return this.instances.size();
	}

	public boolean running()
	{
		return this.running;
	}

	@Override
	public String toString()
	{
		return "RoomWorker #" + this.index;
	}
}
