package org.openhabbo.rooms.pathfinding;

/*
 * Code: Project Thor Pathfinding
 Author: Joe Hegarty

 Copyright (c) 2009 Joe Hegarty

 Permission is hereby granted, free of charge, to the Blunk Server Team to obtain a copy
 of the above software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute copies of the Software,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

import java.util.ArrayList;
import java.util.List;

/**
 * A quick messy Java port of a pathfinder written by Joe 'Joeh' Hegarty for Thor server.
 * 
 * @author Joe Hegarty (ported by Nillus)
 */
public final class JoehPathfinder
{
	private static final int HEURISTIC_ESTIMATE = 2;
	private static final int MAX_CYCLE_LIMIT = 2500;

	private final byte[][] map;
	private final boolean[][] userMap;
	private float[][] heightMap;
	private final int maxX;
	private final int maxY;
	private NodeQueue listOpen;
	private List<PathfinderNode> listClosed;
	private boolean pathFound = false;

	private final float maxAscend;
	private final float maxDescend;

	public JoehPathfinder(byte[][] map, boolean[][] userMap, float[][] heightMap, float maxAscend, float maxDescend)
	{
		this.map = map;
		this.userMap = userMap;
		this.heightMap = heightMap;
		this.maxAscend = maxAscend;
		this.maxDescend = maxDescend;

		// -1 because of zero based arrays
		this.maxX = userMap.length - 1;
		this.maxY = userMap[0].length - 1;

		this.listOpen = new NodeQueue();
		this.listClosed = new ArrayList<PathfinderNode>();
	}

	public List<PathfinderNode> findPath(int x, int y, int goalX, int goalY)
	{
		byte[][] Direction = new byte[][] { { 0, -1 }, { 1, 0 }, { 0, 1 }, { -1, 0 } };

		int openListCount = 0;
		int closedListCount = 0;

		PathfinderNode parent = new PathfinderNode();
		parent.x = x;
		parent.y = y;
		parent.parentX = x;
		parent.parentY = y;
		parent.h = HEURISTIC_ESTIMATE;
		parent.f = HEURISTIC_ESTIMATE;
		parent.g = 0;

		listOpen.push(parent);
		openListCount++;

		while(openListCount > 0)
		{
			openListCount--;
			parent = listOpen.pop();
			if(parent.x == goalX && parent.y == goalY)
			{
				pathFound = true;
				listClosed.add(parent);
				closedListCount++;
				break;
			}

			if(closedListCount > MAX_CYCLE_LIMIT) break;

			for(int i = 0; i < 4; i++)
			{
				PathfinderNode newNode = new PathfinderNode();
				newNode.x = (short) (parent.x + Direction[i][0]);
				newNode.y = (short) (parent.y + Direction[i][1]);

				if(newNode.x < 0 || newNode.y < 0 || newNode.x > this.maxX || newNode.y > this.maxY || map[newNode.x][newNode.y] != 0 || this.userMap[newNode.x][newNode.y])
				{
					continue;
				}

				int newG = parent.g + (this.map[newNode.x][newNode.y]) + 1; // Cost
				if(newG == parent.g) // Same cost as the parent, why taking this route?
				continue;

				float hParent = this.heightMap[parent.x][parent.y];
				float hNew = this.heightMap[newNode.x][newNode.y];
				if(hParent - maxDescend > hNew || hParent + maxAscend < hNew) continue;

				int listIndex = -1;
				for(int j = 0; j < openListCount; j++)
				{
					PathfinderNode curNode = listOpen.get(j);
					if(curNode.x == newNode.x && curNode.y == newNode.y)
					{
						listIndex = j;
						break;
					}
				}
				if(listIndex != -1 && listOpen.get(listIndex).g <= newG) // Better node already in open list
				continue;

				listIndex = -1;
				for(int j = 0; j < closedListCount; j++)
				{
					PathfinderNode curNode = listClosed.get(j);
					if(curNode.x == newNode.x && curNode.y == newNode.y)
					{
						listIndex = j;
						break;
					}
				}
				if(listIndex != -1 && listClosed.get(listIndex).g <= newG) // Better node already in closed list
				continue;

				newNode.parentX = parent.x;
				newNode.parentY = parent.y;
				newNode.g = newG;

				// Calculate heuristic
				int xD1 = parent.x - goalX;
				int xD2 = x - goalX;
				int yD1 = parent.y - goalY;
				int yD2 = y - goalY;

				newNode.h = HEURISTIC_ESTIMATE * (Math.max(Math.abs(newNode.x - goalX), Math.abs(newNode.y - goalY)));
				newNode.h = (int) (newNode.h + Math.abs(xD1 * yD2 - xD2 * yD1) * 0.001);
				newNode.f = newNode.g + newNode.h;
				listOpen.push(newNode);
				openListCount++;
			}
			listClosed.add(parent);
			closedListCount++;
		}

		if(this.pathFound)
		{
			PathfinderNode topNode = listClosed.get(listClosed.size() - 1);
			for(int j = closedListCount - 1; j >= 0; j--)
			{
				PathfinderNode pNode = listClosed.get(j);
				if(topNode.parentX == pNode.x && topNode.parentY == pNode.y || j == closedListCount - 1)
				{
					topNode = pNode;
				}
				else
				{
					listClosed.remove(j);
				}
			}

			return listClosed;
		}

		return new ArrayList<PathfinderNode>(0);
	}
}
