package org.openhabbo.rooms.pathfinding;

/*
 * Code: Project Thor Pathfinding
 Author: Joe Hegarty

 Copyright (c) 2009 Joe Hegarty

 Permission is hereby granted, free of charge, to the Blunk Server Team to obtain a copy
 of the above software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute copies of the Software,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

import java.util.ArrayList;
import java.util.List;

import org.openhabbo.rooms.instances.RoomMap;

/**
 * A quick messy Java port of a pathfinder written by Joe 'Joeh' Hegarty for Thor server.
 * 
 * @author Joe Hegarty (ported by Nillus)
 */
public final class JoehPathfinderStaticImpl
{
	private static final int HEURISTIC_ESTIMATE = 2;
	private static final int MAX_CYCLE_LIMIT = 2500;
	private static final byte[][] DIRECTION = new byte[][] { { 0, -1 }, { 1, 0 }, { 0, 1 }, { -1, 0 } };

	public static List<PathfinderNode> findPath(int x, int y, int goalX, int goalY, byte[][] map, boolean[][] userMap, float[][] heightMap, float maxAscend, float maxDescend)
	{
		int openListCount = 0;
		int closedListCount = 0;

		int maxX = userMap.length - 1;
		int maxY = userMap[0].length - 1;

		PathfinderNode parent = new PathfinderNode();
		parent.x = x;
		parent.y = y;
		parent.parentX = x;
		parent.parentY = y;
		parent.h = HEURISTIC_ESTIMATE;
		parent.f = HEURISTIC_ESTIMATE;
		parent.g = 0;

		// Create closedlist
		List<PathfinderNode> listClosed = new ArrayList<PathfinderNode>();

		// Create openlist
		NodeQueue listOpen = new NodeQueue();
		listOpen.push(parent);
		openListCount++;

		boolean pathFound = false;
		while(openListCount > 0)
		{
			openListCount--;
			parent = listOpen.pop();
			if(parent.x == goalX && parent.y == goalY)
			{
				pathFound = true;
				listClosed.add(parent);
				closedListCount++;
				break;
			}

			if(closedListCount > MAX_CYCLE_LIMIT) break;

			for(int i = 0; i < 4; i++)
			{
				PathfinderNode newNode = new PathfinderNode();
				newNode.x = (short) (parent.x + DIRECTION[i][0]);
				newNode.y = (short) (parent.y + DIRECTION[i][1]);

				// Tile out of range?
				if(newNode.x < 0 || newNode.y < 0 || newNode.x > maxX || newNode.y > maxY)
				{
					continue;
				}

				// Avatar here?
				if(userMap[newNode.x][newNode.y])
				{
					continue;
				}

				// Tile not open?
				if(!((map[newNode.x][newNode.y] & RoomMap.OPEN) == RoomMap.OPEN))
				{
					continue;
				}

				// Perform height check
				float hParent = heightMap[parent.x][parent.y];
				float hNew = heightMap[newNode.x][newNode.y];
				if(hParent - maxDescend > hNew || hParent + maxAscend < hNew)
				{
					continue;
				}

				// Calculate node cost
				int newG = parent.g + 1;

				// Check if node already in list
				int listIndex = -1;
				for(int j = 0; j < openListCount; j++)
				{
					PathfinderNode curNode = listOpen.get(j);
					if(curNode.x == newNode.x && curNode.y == newNode.y)
					{
						listIndex = j;
						break;
					}
				}
				if(listIndex != -1 && listOpen.get(listIndex).g <= newG) // Better node already in open list
				continue;

				listIndex = -1;
				for(int j = 0; j < closedListCount; j++)
				{
					PathfinderNode curNode = listClosed.get(j);
					if(curNode.x == newNode.x && curNode.y == newNode.y)
					{
						listIndex = j;
						break;
					}
				}
				if(listIndex != -1 && listClosed.get(listIndex).g <= newG) // Better node already in closed list
				continue;

				newNode.parentX = parent.x;
				newNode.parentY = parent.y;
				newNode.g = newG;

				// Calculate heuristic
				int xD1 = parent.x - goalX;
				int xD2 = x - goalX;
				int yD1 = parent.y - goalY;
				int yD2 = y - goalY;

				newNode.h = HEURISTIC_ESTIMATE * (Math.max(Math.abs(newNode.x - goalX), Math.abs(newNode.y - goalY)));
				newNode.h = (int) (newNode.h + Math.abs(xD1 * yD2 - xD2 * yD1) * 0.001);
				newNode.f = newNode.g + newNode.h;
				listOpen.push(newNode);
				openListCount++;
			}
			listClosed.add(parent);
			closedListCount++;
		}

		if(pathFound)
		{
			PathfinderNode topNode = listClosed.get(listClosed.size() - 1);
			for(int j = closedListCount - 1; j >= 0; j--)
			{
				PathfinderNode pNode = listClosed.get(j);
				if(topNode.parentX == pNode.x && topNode.parentY == pNode.y || j == closedListCount - 1)
				{
					topNode = pNode;
				}
				else
				{
					listClosed.remove(j);
				}
			}

			return listClosed;
		}

		return new ArrayList<PathfinderNode>(0);
	}
}
