package org.openhabbo.sessions;


import org.openhabbo.core.DynamicObject;
import org.openhabbo.core.Logger;
import org.openhabbo.events.Event;
import org.openhabbo.items.ItemInventory;

import org.openhabbo.net.NetConnection;
import org.openhabbo.rooms.Room;
import org.openhabbo.rooms.RoomManager;
import org.openhabbo.rooms.instances.RoomInstance;
import org.openhabbo.rooms.instances.RoomUser;
import org.openhabbo.sessions.messages.ClientMessage;
import org.openhabbo.sessions.messages.ServerMessage;
import org.openhabbo.users.User;
import org.openhabbo.users.UserManager;
import org.openhabbo.users.achievements.AchievementManager;
import org.openhabbo.users.subscriptions.SubscriptionManager;

public final class Session extends DynamicObject implements NetConnection.ConnectionHandler
{
	// Events
	public final static Event onAuthenticate = new Event("Session.OnAuthenticate");
	public final static Event onLogin = new Event("Session.OnLogin");
	public final static Event onLogout = new Event("Session.OnLogout");
	public final static Event onRoomEnter = new Event("Session.OnRoomEnter");
	public final static Event onRoomExit = new Event("Session.OnRoomExit");
	public final static Event onChat = new Event("Session.OnChat");

	// Core
	public final long id;
	private NetConnection conn;

	// Message handling
	private long lastMsgTime;
	private ServerMessage msg;
	private ServerMessage responseObject;

	// Session variables
	private User user;
	private int accessLogId;
	private ItemInventory itemInventory;
	private RoomInstance roomInstance;
	private RoomUser roomUser;

	public int waitingForDoorbell;
	public int authenticatedTeleporterId;

	public Session(long sessionId, NetConnection conn)
	{
		this.id = sessionId;
		this.conn = conn;
		this.msg = new ServerMessage();
		this.responseObject = new ServerMessage();
	}

	public void onEnd()
	{
		if(!this.conn.closed())
		{
			this.conn.closeSocket();
		}
		if(this.authenticated())
		{
			this.onLogout();
			SessionManager.getInstance().removeAuthedSession(this);
		}
	}

	public void authenticate(String ticket)
	{
		// User exists for this auth ticket?
		User user = UserManager.getInstance().getUserByAuthTicket(ticket);
		if(user == null)
		{
			SessionManager.getInstance().endSession(this.id, "invalid auth ticket");
			return;
		}

		// Run auth checks
		if(!Session.onAuthenticate.fire(this, user))
		{
			return;
		}

		// Load subscriptions, rebuild rights list
		user.subscriptions = SubscriptionManager.getInstance().getSubscriptionsForUser(user.id);
		user.rebuildRightsList();

		// Insufficient permissions?
		if(user.role == null || !user.rights.contains("fuse_login"))
		{
			SessionManager.getInstance().endSession(this.id, String.format("user #%d is not allowed to login because of insufficient role privileges", user.id));
			return;
		}

		// Add to session list
		this.user = user;
		SessionManager.getInstance().addAuthedSession(this);

		// Send user right list
		this.msg.init(EServerMessage.UserRights);
		this.msg.appendInt(this.user.specialRightsLevel);
		this.msg.appendStrList(this.user.rights);
		this.send(this.msg);

		// Fuck knows
		this.msg.init(290);
		this.msg.appendBool(true);
		this.msg.appendBool(false);
		this.send(this.msg);
		
		// Enable info feed
		this.msg.init(517);
		this.msg.appendBool(true);
		this.send(this.msg);
		
		// Login OK!
		this.msg.init(EServerMessage.AuthenticationOK);
		this.send(this.msg);

		// Run other logic
		this.onLogin();
	}

	private void onLogin()
	{
		// Reload achievements
		this.user.achievements = AchievementManager.getInstance().getAchievementsForUser(this.user.id);

		// Reload item/pets inventory
		this.itemInventory = new ItemInventory(this);
		this.itemInventory.reload();

		// Update last online time
		user.updateLastOnline(true);

		// Execute other on-login tasks
		Session.onLogin.fire(this);
	}

	private void onLogout()
	{
		// Run hooks
		Session.onLogout.fire(this);

		// Stop trading, leave the room, etc etc
		this.leaveRoomSilent();

		// Clear inventories to speed up garbagecollection
		this.itemInventory.clear();

		// Update last online time
		user.updateLastOnline(false);

		// Save misc things
		user.saveMisc();
	}

	public void tryGetRoomAccess(int roomId, String password)
	{
		// Leave any room instance the user might be in now
		this.leaveRoomSilent();

		// Get/create room instance
		RoomInstance room = RoomManager.getInstance().getRoomInstance(roomId, true);
		if(room == null)
		{
			return;
		}

		// The room owner can ALWAYS enter the room, never has to check for user limit, bans, state, etc
		boolean skipAccessChecks = (this.getUser().id == room.getInfo().ownerId) || this.getUser().rights.contains("fuse_enter_any_room");
		if(!skipAccessChecks)
		{
			int errorCode = 0;
			
			// Room is full?
			if(room.getInfo().usersNow >= room.getInfo().usersMax)
			{
				// Can't override user limit?
				if(!this.getUser().rights.contains("fuse_enter_full_rooms"))
				{
					errorCode = 1;
				}
			}

			// User is banned?
			else if(room.userIsBanned(this.getUser().id))
			{
				errorCode = 4;
			}
			
			// Any RoomEntryErrors?
			if(errorCode != 0)
			{
				this.msg.init(EServerMessage.RoomEntryError);
				this.msg.appendInt(errorCode);
				this.send(this.msg);
				this.kickFromRoom();
				return;
			}

			// Password room?
			if(room.getInfo().state == Room.STATE_PASSWORD && !room.getInfo().password.equals(password))
			{
				this.sendGenericError(-100002);
				this.kickFromRoom();
				return;
			}

			// Doorbell room?
			else if(room.getInfo().state == Room.STATE_DOORBELL)
			{
				// Ring the doorbell, await response
				if(room.ringDoorbell(this))
				{
					// Cool, let's wait
					this.waitingForDoorbell = room.getInfo().id;

					// Waiting...
					this.msg.init(EServerMessage.Doorbell);
					this.msg.appendStr("");
					this.send(this.msg);
				}
				else
				{
					// NOBODY HEARD YOU, FOREVER ALONE
					this.msg.init(EServerMessage.FlatAccessDenied);
					this.send(this.msg);
					this.kickFromRoom();
				}
				return;
			}
		}

		// Send room stats tracker
		this.msg.init(EServerMessage.RoomStatsTracker);
		this.msg.append("/client/");
		if(room.getInfo().isPublicRoom())
		{
			this.msg.append("public/");
			this.msg.append(room.getInfo().modelName);
			this.msg.append("/0");
		}
		else
		{
			this.msg.append("private/");
			this.msg.append(Integer.toString(room.getInfo().id));
			this.msg.append("/id");
		}
		this.send(this.msg);

		// Forward to room instance
		this.goToRoom(roomId);
	}

	public void goToRoom(int roomId)
	{
		// Get the running instance, it should be there by now...
		RoomInstance instance = RoomManager.getInstance().getRoomInstance(roomId, false);
		if(instance == null)
		{
			return;
		}

		// Send room model
		this.msg.init(EServerMessage.RoomReady);
		this.msg.appendStr(instance.getModel().name);
		this.msg.appendInt(instance.getInfo().id);
		this.send(this.msg);

		// This is the current room instance now
		this.roomInstance = instance;
	}

	public boolean enterRoom()
	{
		this.roomUser = this.roomInstance.addSession(this);
		if(this.roomUser != null)
		{
			if(Session.onRoomEnter.fire(this))
			{
				return true;
			}
			else
			{
				// Silent kick (a hook specified that roomentry should not continue)
				this.kickFromRoom();
			}
		}
		else
		{
			// Hacky entry?
			this.kickFromRoom("Invalid room entry!");
		}

		// Do not enter room
		return false;
	}

	private boolean leaveRoomSilent()
	{
		// Reset room ID
		this.waitingForDoorbell = 0;

		// Leave room instance
		if(this.roomInstance != null)
		{
			if(Session.onRoomExit.fire(this))
			{
				this.roomInstance.removeSession(this);
				this.roomInstance = null;
				this.roomUser = null;
			}
			else
			{
				// Did not leave room
				return false;
			}
		}

		// Left room!
		return true;
	}

	public void kickFromRoom()
	{
		this.kickFromRoom(null);
	}

	public void kickFromRoom(String text)
	{
		// SHE GAVE ME THE PLEASE-FACE
		if(!this.leaveRoomSilent())
		{
			// Room exit cancelled
			return;
		}

		// I GAVE HER THE LEAVE-FACE
		this.msg.init(EServerMessage.CloseConnection);
		this.send(this.msg);

		// BODY LIKE A BOTTLE, PRETTY LITTLE MODEL MODEL MODEEL
		if(text != null)
		{
			this.sendNotif(text);
		}
	}

	@Override
	public void onNetNewMessage(ClientMessage msg)
	{
		// Fucking heartbeat! ("D{")
		if(msg.msgId == 315) return;

		// Log that this session received a message!
		this.lastMsgTime = System.currentTimeMillis();
		Logger.getInstance().debug(this, "<-- " + msg.toString());

		// Message contains protocol-EOF byte? This is _never_ acceptable because of exploitable purposes
		if(msg.contains(ServerMessage.MSG_EOF))
		{
			SessionManager.getInstance().endSession(this.id, "MSG_EOF detected in incoming message", true);
			return;
		}

		// These are the only responding handlers for SendSessionParams and SsoLogin
		if(!this.authenticated() && (msg.msgId != 206 && msg.msgId != 415))
		{
			Logger.getInstance().warning(this, String.format("msg #%d (\"%s\") is an invalid pre-auth msg.", msg.msgId, msg.header()));
		}
		else
		{
			// Lookup message handler and execute handling logic
			SessionMessageHandler handler = SessionMsgRegistry.get().lookup(msg.msgId);
			if(handler != null)
			{
				System.out.println(handler.getClass().getSimpleName());
				try
				{
					handler.handle(this, msg, this.responseObject);
				}
				catch(Exception ex)
				{
					Logger.getInstance().error(this, String.format("Error handling msg #%d (\"%s\").", msg.msgId, msg.header()), ex);
				}
				this.responseObject.reset();
			}
			else
			{
				Logger.getInstance().warning(this, String.format("SessionMsgRegistry does not have an entry for msg #%d. (\"%s\")", msg.msgId, msg.header()));
			}
		}
	}

	@Override
	public void onNetClose(String reason)
	{
		SessionManager.getInstance().endSession(this.id, reason);
	}

	public void send(ServerMessage msg)
	{
		Logger.getInstance().debug(this, "--> " + msg.toString());
		this.conn.send(msg.getResult());
	}

	public void sendGenericError(int code)
	{
		this.msg.init(EServerMessage.GenericError);
		this.msg.appendInt(code);
		this.send(this.msg);
	}

	public void sendNotif(String text)
	{
		this.sendNotif(text, null);
	}

	public void sendNotif(String text, String url)
	{
		this.msg.init(EServerMessage.Notif);
		this.msg.appendStr(text);
		this.msg.appendStr(url);
		this.send(this.msg);
	}

	public void sendSystemNotif(String text)
	{
		this.msg.init(810);
		this.msg.appendBool(true);
		this.msg.appendStr(text);
		this.send(this.msg);
	}

	public void rebuildAndSendUserRights()
	{
		if(this.user != null)
		{
			// Aggegrate the list
			user.rebuildRightsList();

			// Send the rights list (role, subscriptions)
			this.msg.init(EServerMessage.UserRights);
			this.msg.appendStrList(user.rights);
			this.send(this.msg);
		}
	}

	public void sendCreditBalance()
	{
		this.msg.init(EServerMessage.CreditBalance);
		this.msg.appendStr(this.user.credits + ".0");
		this.send(this.msg);
	}

	public void sendPixelBalance()
	{
		this.sendPixelsAddedNotif(0);
	}

	public void sendPixelsAddedNotif(int added)
	{
		this.msg.init(EServerMessage.PixelsReceivedNotification);
		this.msg.appendInt(this.user.pixels);
		this.msg.appendInt(added);
		this.send(this.msg);
	}

	public void sendRoomPrivileges()
	{
		if(this.roomUser != null)
		{
			if(this.roomUser.hasRoomRights)
			{
				this.msg.init(EServerMessage.YouAreController);
				this.send(this.msg);
			}
			if(this.roomUser.hasRoomOwner)
			{
				this.msg.init(EServerMessage.YouAreOwner);
				this.send(this.msg);
			}
		}
	}

	public final NetConnection getConnection()
	{
		return this.conn;
	}

	public final long getLastMsgTime()
	{
		return this.lastMsgTime;
	}

	public final boolean authenticated()
	{
		return(this.user != null);
	}

	public final User getUser()
	{
		return this.user;
	}

	public final int getAccessLogId()
	{
		return this.accessLogId;
	}

	public void setAccessLogId(int accessLogId)
	{
		if(this.accessLogId == 0)
		{
			this.accessLogId = accessLogId;
		}
	}

	public final ItemInventory getItemInventory()
	{
		return this.itemInventory;
	}

	public final boolean inRoom()
	{
		return(this.roomInstance != null);
	}

	public final RoomInstance getRoomInstance()
	{
		return this.roomInstance;
	}

	public final RoomUser getRoomUser()
	{
		return this.roomUser;
	}

	@Override
	public String toString()
	{
		return "Session #" + this.id;
	}
}
