package org.openhabbo.storage;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.openhabbo.core.Logger;

public final class DatabaseConnectionManager implements Runnable
{
	private final static long MAX_IDLE_SECONDS = 120;

	private DatabaseEndpoint database;
	private String driver;
	private int maxConnections;

	private boolean monitorAlive;
	private Thread monitorThread;

	private int connectionCounter;
	private List<DatabaseConnection> connections;

	public DatabaseConnectionManager(DatabaseEndpoint database, String driver, int maxConnections)
	{
		this.database = database;
		this.driver = driver;
		this.maxConnections = maxConnections;
		this.connections = new ArrayList<DatabaseConnection>(maxConnections);
	}

	public boolean test()
	{
		// Register connection driver package
		String driverClassName = this.driver;
		try
		{
			Driver driver = (Driver) Class.forName(driverClassName).newInstance();
			DriverManager.registerDriver(driver);
		}
		catch(Exception ex)
		{
			Logger.getInstance().error("DatabaseConnectionMgr", String.format("Can't find driver '%s', please copy the driver library to the local Java extensions directory (lib/ext).", driverClassName), null);
			return false;
		}

		// Create test connection
		DatabaseConnection test = this.getConnection();
		if(test != null)
		{
			this.releaseConnection(test);
			return true;
		}
		else
		{
			return false;
		}
	}

	public void startMonitor()
	{
		if(!this.monitorAlive)
		{
			// Create monitor thread on a low priority
			this.monitorThread = new Thread(this, "DatabaseConnectionManager monitor");
			this.monitorThread.setPriority(Thread.MIN_PRIORITY);

			// Start the monitor thread
			this.monitorAlive = true;
			this.monitorThread.start();

			Logger.getInstance().info("DatabaseConnectionMgr", "Started connection monitor.");
		}
	}

	public void stopMonitor()
	{
		if(this.monitorAlive)
		{
			// Stop the monitor thread
			this.monitorAlive = false;
			this.monitorThread.interrupt();

			Logger.getInstance().info("DatabaseConnectionMgr", "Stopped connection monitor.");
		}
	}

	public void closeAll()
	{
		synchronized(this.connections)
		{
			// Close all open connections
			for(DatabaseConnection conn : this.connections)
			{
				try
				{
					conn.getInnerConnection().close();
				}
				catch(Exception ex)
				{
					ex.printStackTrace();
				}
			}
			this.connections.clear();

			// Woo woo!
			Logger.getInstance().info("DatabaseConnectionMgr", "Closed all connections.");
		}
	}

	public DatabaseConnection getConnection()
	{
		synchronized(this.connections)
		{
			// Any available connections?
			Iterator<DatabaseConnection> iterator = this.connections.iterator();
			while(iterator.hasNext())
			{
				DatabaseConnection conn = iterator.next();
				if(conn.isAvailable())
				{
					try
					{
						// If it's not closed, then this is the connection to return!
						if(!conn.getInnerConnection().isClosed())
						{
							conn.available(false);
							return conn;
						}
						else
						{
							iterator.remove();
							Logger.getInstance().warning("DatabaseConnectionMgr", String.format("Connection #%d was dead but still in the available connections list, removed.", conn.id));
						}
					}
					catch(SQLException ex)
					{
						// /care
					}
				}
			}

			// No open available connections. Allowed to create a new connection?
			if(this.maxConnections <= 0 || this.connections.size() < this.maxConnections)
			{
				try
				{
					Connection innerConn = DriverManager.getConnection(this.database.url, this.database.username, this.database.password);
					DatabaseConnection conn = new DatabaseConnection(this.connectionCounter++, innerConn, this);
					conn.available(true);
					this.connections.add(conn);
					Logger.getInstance().debug("DatabaseConnectionMgr", "Created new connection.");

					// Try again
					return this.getConnection();
				}
				catch(Exception ex)
				{
					Logger.getInstance().error("DatabaseConnectionMgr", "Could not create a new connection! JDBC reported: " + ex.getMessage(), null);

					return null;
				}
			}

			// Can't create more connections at the moment? Wait for a returning connection, but no longer than x seconds
			try
			{
				this.connections.wait(5 * 1000);
			}
			catch(InterruptedException ex)
			{
				// Waitlock triggered! Yay, a connection returned!
				return this.getConnection();
			}

			// This is real bad, please have mercy as there's nothing we can do right now, database is dead
			Logger.getInstance().warning("DatabaseConnectionMgr", "No available connections, not allowed to create a new connection and waiting for returning connections timed out.");
			return null;
		}
	}

	public void releaseConnection(DatabaseConnection conn)
	{
		// Actually valid connection passed in?
		if(conn == null) return;

		// Update the connection and notify an arbitrary thread that is waiting that are waiting
		synchronized(this.connections)
		{
			conn.available(true);
			this.connections.notify();
		}
	}

	@Override
	public void run()
	{
		// Hardcoded cleanup interval
		final long cleanupInterval = 30 * 1000;

		// Run all the time
		while(this.monitorAlive)
		{
			try
			{
				Thread.sleep(cleanupInterval);
			}
			catch(InterruptedException ex)
			{
				return;
			}

			// Close + remove available connections that have been idle for too long
			int cleanedUpCount = 0;
			synchronized(this.connections)
			{
				Iterator<DatabaseConnection> iterator = this.connections.iterator();
				while(iterator.hasNext())
				{
					DatabaseConnection conn = iterator.next();
					if(conn.isAvailable())
					{
						if(conn.getIdleSeconds() > DatabaseConnectionManager.MAX_IDLE_SECONDS)
						{
							this.close(conn);
							iterator.remove();
							cleanedUpCount++;
						}
					}
				}
			}

			// Any connections closed + removed? Write debug log
			if(cleanedUpCount > 0)
			{
				Logger.getInstance().debug("DatabaseConnectionMgr", String.format("Cleaned up %d connection(s).", cleanedUpCount));
			}
		}
	}

	private void close(DatabaseConnection conn)
	{
		try
		{
			conn.getInnerConnection().close();
		}
		catch(SQLException e)
		{
			Logger.getInstance().warning("DatabaseConnectionMgr", String.format("Error closing database connection #%d.", conn.id));
		}
	}

	public int count()
	{
		return this.connections.size();
	}
}
