package mod_achievement;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.openhabbo.OpenHABBO;
import org.openhabbo.core.Logger;
import org.openhabbo.sessions.EServerMessage;
import org.openhabbo.sessions.Session;
import org.openhabbo.sessions.SessionManager;
import org.openhabbo.sessions.messages.ServerMessage;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLDatabase;
import org.openhabbo.users.User;
import org.openhabbo.users.UserManager;

public final class AchievementManager
{
	private Map<Integer, Achievement> achievements;

	private AchievementManager()
	{
		this.achievements = Collections.emptyMap();
	}

	public void reload()
	{
		// Create new collection
		Map<Integer, Achievement> tmpAchievements = new HashMap<Integer, Achievement>();

		// Load from database
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT * FROM achievements");
		while(result.row())
		{
			Achievement ach = new Achievement();
			ach.id = result.getInt("id");
			ach.maxLevel = result.getInt("max_level");
			ach.badge = result.getString("badge");
			ach.badgeDynamicLevel = result.getBoolean("badge_dynamiclevel");
			ach.pixelsBase = result.getInt("pixels_base");
			ach.pixelsMultiplier = result.getFloat("pixels_multiplier");
			tmpAchievements.put(ach.id, ach);
		}
		result.release();

		// Set instance field
		this.achievements = Collections.unmodifiableMap(tmpAchievements);

		// Done!
		Logger.getInstance().info("AchievementMgr", String.format("Reloaded %d achievements.", this.achievements()));
	}

	public Achievement getAchievement(int achievementId)
	{
		return this.achievements.get(achievementId);
	}

	public Collection<Achievement> getAllAchievements()
	{
		return this.achievements.values();
	}

	public Map<Integer, Integer> getAchievementsForUser(int userId)
	{
		Map<Integer, Integer> achs = new HashMap<Integer, Integer>();

		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT achievement_id,level FROM user_achievements WHERE user_id=" + userId);
		while(result.row())
		{
			achs.put(result.getInt("achievement_id"), result.getInt("level"));
		}
		result.release();

		return achs;
	}

	public void unlockAchievementForUser(int userId, int achievementId)
	{
		this.unlockAchievementForUser(userId, achievementId, Achievement.START_LEVEL);
	}

	public void unlockAchievementForUser(int userId, int achievementId, int level)
	{
		// User and achievement exists?
		User user = UserManager.getInstance().getUserById(userId, true);
		Achievement ach = this.getAchievement(achievementId);
		if(user == null || ach == null)
		{
			return;
		}

		// Valid level?
		if(level > ach.maxLevel)
		{
			level = ach.maxLevel;
		}

		// Calculate pixel reward
		int pixels = ach.getPixels(level);

		// Shorthand variable
		SQLDatabase db = OpenHABBO.getInstance().getDb();

		// Delete old achievement level (if any)
		db.execute("DELETE FROM user_achievements WHERE user_id=" + userId + " AND achievement_id=" + achievementId);

		// New achievement level
		db.execute("INSERT INTO user_achievements(user_id,achievement_id,level) VALUES (" + userId + "," + achievementId + "," + level + ")");

		// Give pixels
		user.pixels += pixels;
		user.saveCreditsAndPixels();

		// User online?
		Session session = SessionManager.getInstance().getSessionByUserId(userId);
		if(session != null)
		{
			// Update achievement list
			session.getUser().achievements.put(achievementId, level);

			// Notify of achievement gain
			ServerMessage msg = new ServerMessage(EServerMessage.AchievementNotification);
			msg.appendInt(achievementId);
			msg.appendInt(level);
			msg.appendStr(ach.getBadge(level));
			if(level == Achievement.START_LEVEL)
			{
				// No previous badge
				msg.appendStr("");
			}
			else
			{
				// Previous badge
				msg.appendStr(ach.getBadge(level - 1));
			}
			session.send(msg);

			// Notify of pixel gain
			session.sendPixelsAddedNotif(pixels);
		}
	}

	public int achievements()
	{
		return this.achievements.size();
	}

	/* Begin singleton */
	private final static AchievementManager instance = new AchievementManager();

	public final static AchievementManager getInstance()
	{
		return instance;
	}
	/* End singleton */
}
