package org.openhabbo;

import org.openhabbo.access.AccessControl;
import org.openhabbo.catalog.Catalog;
import org.openhabbo.chatcommands.ChatCommandHandlerManager;
import org.openhabbo.core.Logger;
import org.openhabbo.core.PropertiesBox;
import org.openhabbo.core.ShutdownHook;
import org.openhabbo.items.ItemManager;
import org.openhabbo.messenger.InstantMessengerService;
import org.openhabbo.modules.ModuleHost;
import org.openhabbo.modules.ModuleManager;
import org.openhabbo.net.NetManager;
import org.openhabbo.remote.RemoteServer;
import org.openhabbo.rooms.RoomManager;
import org.openhabbo.rooms.RoomNavigator;
import org.openhabbo.sessions.SessionManager;
import org.openhabbo.storage.SQLDatabase;
import org.openhabbo.support.SupportCenter;
import org.openhabbo.users.UserManager;
import org.openhabbo.users.achievements.AchievementManager;
import org.openhabbo.util.StartupCleanup;

/**
 * A new Habbo Hotel (www.habbo.com) game server emulator written by Nils 'Nillus' W.
 */
public class OpenHABBO
{
	private boolean m_alive;
	private long m_startTime;
	private String m_propsFile;
	private Logger m_logger;
	private PropertiesBox m_props;
	private SQLDatabase m_db;
	private ModuleManager m_modMgr;
	private RemoteServer m_remoteServer;
	private NetManager m_netMgr;
	private SessionManager m_sessionMgr;

	private AccessControl m_accessCtrl;
	private UserManager m_userMgr;
	private AchievementManager m_achievementMgr;
	private InstantMessengerService m_imSvc;
	private ItemManager m_itemMgr;
	private RoomManager m_roomMgr;
	private RoomNavigator m_navigator;
	private Catalog m_catalog;
	private SupportCenter m_supportCenter;
	private ChatCommandHandlerManager m_chatCmdHandlerMgr;

	public OpenHABBO(String propsFile)
	{
		m_propsFile = propsFile;
		m_logger = new Logger("openhabbo");
		m_props = new PropertiesBox();

		System.out.println("1111111111111111111111111111111111111111111");
		System.out.println("1100000000000000000000000000000000000000011");
		System.out.println("1101111         OpenHABBO           1111011");
		System.out.println("1101111111111111111111111111111111111111011");
		System.out.println("1101111  COPYRIGHT (C) 2011 Nils W. 1111011");
		System.out.println("1101111111111111111111111111111111111111011");
		System.out.println("1101111      Parts written by:      1111011");
		System.out.println("1101111     Roy 'Meth0d' de Jong    1111011");
		System.out.println("1101111     Joe 'Joeh' Hegarty      1111011");
		System.out.println("1101111111111111111111111111111111111111011");
		System.out.println("1101111     www.openhabbo.org       1111011");
		System.out.println("1100000000000000000000000000000000000000011");
		System.out.println("1111111111111111111111111111111111111111111");
		System.out.println();
	}

	public boolean reloadProperties()
	{
		boolean succes = m_props.load(m_propsFile);
		if(succes)
		{
			m_logger.info("OpenHABBO", String.format("Loaded %d properties.", m_props.size()));
		}
		else
		{
			m_logger.warning("OpenHABBO", "Failed loading system properties.");
		}
		return succes;
	}

	public boolean start()
	{
		// Log startup time
		long beginTime = System.currentTimeMillis();
		m_logger.info("OpenHABBO", "Starting...");

		// Load system properties
		if(!this.reloadProperties())
		{
			return false;
		}

		// Prepare SQLDatabase
		m_db = new SQLDatabase(m_props, m_logger);
		if(m_db.test())
		{
			m_logger.info("SQLDatabase", "Connection successful.");
			if(!m_db.getQueryQueue().open())
			{
				return false;
			}
			else
			{
				m_db.getConnMgr().startMonitor();
			}
		}
		else
		{
			m_logger.warning("SQLDatabase", "Connection failed.");
			return false;
		}

		// Create SessionManager
		int expectedSessions = m_props.getInt("sessions.expected", 100);
		m_sessionMgr = new SessionManager(expectedSessions);

		// Create AccessControl
		m_accessCtrl = new AccessControl();
		m_accessCtrl.reloadUserRoles();

		// Create user manager
		m_userMgr = new UserManager();

		// Create AchievementManager
		m_achievementMgr = new AchievementManager();
		m_achievementMgr.reload();

		// Create InstantMessengerService
		m_imSvc = new InstantMessengerService(500);

		// Create ItemManager + load definitions
		m_itemMgr = new ItemManager();
		m_itemMgr.reloadDefinitions();

		// Create RoomManager, spawn initial workers and reload model data
		int expectedRoomInstances = m_props.getInt("rooms.instances.expected", 50);
		int initialRoomWorkers = (m_props.get("rooms.workers.initial", "auto").equals("auto")) ? ((expectedRoomInstances / 2) / RoomManager.MAX_LOAD_PER_WORKER) : m_props.getInt("rooms.workers.initial", 0);
		m_roomMgr = new RoomManager(expectedRoomInstances, initialRoomWorkers);
		m_roomMgr.reloadModels();

		// Load Navigator, cache stuff, etc
		m_navigator = new RoomNavigator(OpenHABBO.getInstance().getProperties().getInt("rooms.navigator.refreshtime", 5));
		m_navigator.reload();

		// Load Catalog
		m_catalog = new Catalog();
		m_catalog.reload();

		// Start SupportCenter
		m_supportCenter = new SupportCenter();
		m_supportCenter.start();

		// Load ChatCommandHandlerManager
		m_chatCmdHandlerMgr = new ChatCommandHandlerManager();
		m_chatCmdHandlerMgr.registerStandardCommands();

		// Load modules/plugins
		m_modMgr = new ModuleManager();
		m_modMgr.loadModules();

		// Setup network manager for game clients
		int portNumber = m_props.getInt("net.portnumber", 1337);
		int initialNetWorkers = (m_props.get("net.workers.initial", "auto").equals("auto")) ? ((expectedSessions / 2) / NetManager.MAX_LOAD_PER_WORKER) : m_props.getInt("net.workers.initial", 0);
		m_netMgr = new NetManager(portNumber, initialNetWorkers);
		m_netMgr.reloadBlacklist();
		if(!m_netMgr.getListener().start())
		{
			return false;
		}

		// Create & start remoteserver for RCON'ish issueing of commands
		m_remoteServer = new RemoteServer(m_props.getInt("remote.portnumber", 1338));
		m_remoteServer.addAllowedIp("0:0:0:0:0:0:0:1%0");
		m_remoteServer.addAllowedIp("127.0.0.1");
		m_remoteServer.getCmdMgr().registerStandardCommands();
		m_remoteServer.start();

		// Attach shutdown hook
		Runtime.getRuntime().addShutdownHook(new ShutdownHook(this));

		// Clean some database things
		StartupCleanup.cleanup();

		// Startup okay!
		m_alive = true;
		m_startTime = System.currentTimeMillis();
		m_logger.info("OpenHABBO", String.format("Started in %d ms.", (m_startTime - beginTime)));
		return true;
	}

	public void stop(String reason)
	{
		// Only start if not already shutdown'd yet
		if(!m_alive)
		{
			return;
		}
		m_logger.info("OpenHABBO", String.format("Initiating shutdown sequence... (reason: %s)", reason));

		// Stop the remote command server
		m_remoteServer.stop();

		// Stop the game listener
		m_netMgr.getListener().stop();

		// Shutdown game subsystems
		m_sessionMgr.onShutdown();
		m_roomMgr.onShutdown();
		m_imSvc.onShutdown();
		m_supportCenter.stop();

		// Stop net workers, no more reading and sending from now on
		m_netMgr.stopWorkers();

		// Unload modules/plugins
		m_modMgr.unloadModules();

		// Close SQLDatabase
		m_db.getQueryQueue().close();
		m_db.getConnMgr().closeAll();
		m_db.getConnMgr().stopMonitor();

		// Sequence complete!
		m_alive = false;
		m_logger.info("OpenHABBO", "Shutdown sequence completed.");
	}

	public final String versionString()
	{
		return "OpenHABBO v0.0.1-dev 'Skylark'";
	}

	public boolean alive()
	{
		return m_alive;
	}

	public long getStartTime()
	{
		return m_startTime;
	}

	public Logger getLogger()
	{
		return m_logger;
	}

	public PropertiesBox getProperties()
	{
		return m_props;
	}

	public SQLDatabase getDb()
	{
		return m_db;
	}

	public ModuleManager getModMgr()
	{
		return m_modMgr;
	}

	public RemoteServer getRemoteServer()
	{
		return m_remoteServer;
	}

	public NetManager getNetMgr()
	{
		return m_netMgr;
	}

	public SessionManager getSessionMgr()
	{
		return m_sessionMgr;
	}

	public AccessControl getAccessCtrl()
	{
		return m_accessCtrl;
	}

	public UserManager getUserMgr()
	{
		return m_userMgr;
	}

	public AchievementManager getAchievementMgr()
	{
		return m_achievementMgr;
	}

	public InstantMessengerService getImSvc()
	{
		return m_imSvc;
	}

	public ItemManager getItemMgr()
	{
		return m_itemMgr;
	}

	public RoomManager getRoomMgr()
	{
		return m_roomMgr;
	}

	public RoomNavigator getNavigator()
	{
		if(m_navigator.requiresRefresh())
		{
			m_navigator.refresh();
		}
		return m_navigator;
	}

	public Catalog getCatalog()
	{
		return m_catalog;
	}

	public SupportCenter getSupportCenter()
	{
		return m_supportCenter;
	}

	public ChatCommandHandlerManager getChatCmdMgr()
	{
		return m_chatCmdHandlerMgr;
	}

	/* Singleton */
	private static OpenHABBO m_instance;

	public static OpenHABBO getInstance()
	{
		return m_instance;
	}

	private static ModuleHost m_modHost;

	public static ModuleHost getModHost()
	{
		return m_modHost;
	}

	public static OpenHABBO createInstance(String propsFile)
	{
		OpenHABBO instance = new OpenHABBO(propsFile);
		instance.getLogger().info("OpenHABBO", instance.versionString());
		m_instance = instance;
		return instance;
	}
}
