package org.openhabbo.chatcommands;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.openhabbo.OpenHABBO;
import org.openhabbo.access.UserRole;
import org.openhabbo.sessions.Session;
import org.openhabbo.util.PackageClassFinder;

public class ChatCommandHandlerManager
{
	private Map<String, ChatCommandHandler> m_handlers;

	public ChatCommandHandlerManager()
	{
		m_handlers = new HashMap<String, ChatCommandHandler>();
	}

	public void registerHandler(String cmd, ChatCommandHandler handler)
	{
		m_handlers.put(cmd, handler);
		OpenHABBO.getInstance().getLogger().info("ChatCmdMgr", String.format("Registered chatcommand '%s%s'.", ChatCommandHandler.COMMAND_HEAD, cmd));
	}

	public void registerHandler(ChatCommandHandler handler)
	{
		this.registerHandler(handler.getCommand(), handler);
	}

	public void unregisterHandler(String cmd)
	{
		m_handlers.remove(cmd);
	}

	public List<ChatCommandHandler> getAvailableChatCommands(UserRole role)
	{
		List<ChatCommandHandler> available = new ArrayList<ChatCommandHandler>();
		for(ChatCommandHandler cmd : m_handlers.values())
		{
			String right = cmd.getRequiredRight();
			if(right == null || role.hasRight(cmd.getRequiredRight()))
			{
				available.add(cmd);
			}
		}
		return available;
	}

	public boolean parseAndHandleCommand(Session session, String text)
	{
		// Actually a chat command?
		if(text.indexOf(ChatCommandHandler.COMMAND_HEAD) != 0)
		{
			// Doesn't start with the COMMAND_HEAD character
			return false;
		}

		// Attempt to separate command and body
		String[] segments = text.substring(1).split(" ", 2);

		// Handler registered?
		ChatCommandHandler handler = m_handlers.get(segments[0]);
		if(handler == null)
		{
			// Doesn't exist...
			return false;
		}

		// User has the required right?
		String right = handler.getRequiredRight();
		if(right != null && !session.getUser().role.hasRight(right))
		{
			return false;
		}

		// Invoke handler and catch exceptions
		String body = (segments.length == 2 ? segments[1] : null);
		try
		{
			handler.handle(session, body);
		}
		catch(Exception ex)
		{
			OpenHABBO.getInstance().getLogger().error("ChatCmdMgr", String.format("Error processing chatcommand '%s'. (body=%s)", segments[0], body), ex);
		}

		// Handled
		return true;
	}

	public void registerStandardCommands()
	{
		for(Class<ChatCommandHandler> cmdClass : PackageClassFinder.getClasses("org.openhabbo.chatcommands.std"))
		{
			try
			{
				this.registerHandler(cmdClass.newInstance());
			}
			catch(Exception ex)
			{
				ex.printStackTrace();
			}
		}
	}
}
