package org.openhabbo.net;

import java.util.ArrayList;
import java.util.List;

import org.openhabbo.OpenHABBO;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLHelper;

/*
 * This is how the networking layers work at accepting new connections:
 * NetListener is a ServerSocketChannel listening for connections, once it receives a connection request, it will check if the remote host is blacklisted. If so, connection is dropped. If not blacklisted, it will forward the NetConnection to the NetManager. (NetManager.onAcceptedByListener)
 * NetManager will find the most suitable NetWorker for the newly accepted NetConnection, or spawn a new NetWorker if needed. It will then enqueue the NetConnection with the NetWorker so it's interests can be registered on the NetWorker thread, rather than the NetListener thread.
 * NetWorker will wakeup from it's blocking select() call and attempt to register any pending NetConnections, getting it's selector interested in events from the NetConnection's channel. If the connection has been closed in the meantime, it will be disposed and registration fails. If registration succeeds, it will notify the NetManager that the NetConnection has been accepted by the NetWorker. (NetManager.onAcceptedByProcessor)
 * NetManager will now ask the SessionManager to start a new Session for the newly accepted NetConnection.
 */
public class NetManager
{
	public final static int MAX_LOAD_PER_WORKER = 50;

	private NetListener m_listener;
	private List<String> m_blacklist;
	private List<NetWorker> m_workers;

	public NetManager(int port, int initialWorkers)
	{
		m_listener = new NetListener(port, this);
		m_workers = new ArrayList<NetWorker>(initialWorkers);
		OpenHABBO.getInstance().getLogger().info("NetMgr", String.format("Spawning %d initial workers...", initialWorkers));
		for(int x = 0; x < initialWorkers; x++)
		{
			this.spawnNewWorker();
		}
		m_blacklist = new ArrayList<String>();
	}

	public void stopWorkers()
	{
		for(NetWorker worker : m_workers)
		{
			worker.stop();
		}
		m_workers.clear();
		OpenHABBO.getInstance().getLogger().info("NetMgr", "Stopped all workers.");
	}

	private int spawnNewWorker()
	{
		int index = m_workers.size();
		m_workers.add(new NetWorker(index, this));
		m_workers.get(index).start();
		OpenHABBO.getInstance().getLogger().info("NetMgr", String.format("Spawned NetWorker #%d.", index));
		return index;
	}

	private NetWorker getLowestLoadedWorker()
	{
		int lowest = 999, bestIndex = -1;
		for(int index = 0; index < m_workers.size(); index++)
		{
			int load = m_workers.get(index).load();
			if((load < NetManager.MAX_LOAD_PER_WORKER) && load < lowest)
			{
				lowest = load;
				bestIndex = index;
			}
		}
		if(bestIndex == -1)
		{
			bestIndex = this.spawnNewWorker();
		}
		return m_workers.get(bestIndex);
	}

	public void onAcceptedByListener(NetConnection conn)
	{
		// Check for blacklist status
		if(this.ipIsBlacklisted(conn.getIpAddress()))
		{
			OpenHABBO.getInstance().getLogger().warning("NetMgr", String.format("Denying connection request from blacklisted IP %s.", conn.getIpAddress()));
			conn.closeSocket();
			return;
		}

		// Enqueue it for the best worker
		NetWorker worker = this.getLowestLoadedWorker();
		worker.enqueueNewConnection(conn);
	}

	public void onAcceptedByWorker(NetConnection conn)
	{
		OpenHABBO.getInstance().getSessionMgr().startSession(conn);
	}

	public boolean ipIsBlacklisted(String ip)
	{
		return m_blacklist.contains(ip);
	}

	public void reloadBlacklist()
	{
		m_blacklist.clear();

		// Reload all the blacklisted IPs
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT ip FROM server_ipblacklist");
		while(result.row())
		{
			m_blacklist.add(result.getString("ip"));
		}
		result.release();

		OpenHABBO.getInstance().getLogger().info("NetMgr", String.format("Reloaded %d blacklisted IP(s).", m_blacklist.size()));
	}

	public void blacklistIp(String ip, String info)
	{
		// Already blacklisted?
		if(this.ipIsBlacklisted(ip))
		{
			return;
		}

		// Add to internal blacklist
		m_blacklist.add(ip);

		// Insert new IP to blacklist table
		OpenHABBO.getInstance().getDb().execute("INSERT INTO server_ipblacklist(ip,timestamp,info) VALUES ('" + SQLHelper.escape(ip) + "','" + SQLHelper.getDateTimeStr() + "','" + SQLHelper.escape(info) + "')");

		// Log it
		OpenHABBO.getInstance().getLogger().info("NetMgr", String.format("Blacklisted IP address %s (\"%s\").", ip, info));
	}

	public NetListener getListener()
	{
		return m_listener;
	}
}
