package org.openhabbo.remote;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.List;

import org.openhabbo.OpenHABBO;

public class RemoteServer implements Runnable
{
	public final static int PACKET_SIZE = 1024;

	private int m_port;
	private DatagramSocket m_socket;

	private boolean m_alive;
	private Thread m_thread;

	private List<String> m_allowedIps;
	private RemoteCommandHandlerManager m_cmdHandlerMgr;

	public RemoteServer(int port)
	{
		m_port = port;
		m_allowedIps = new ArrayList<String>();
		m_cmdHandlerMgr = new RemoteCommandHandlerManager();
	}

	public boolean start()
	{
		// Not already started?
		if(m_alive)
		{
			return false;
		}
		else
		{
			// Bind socket
			try
			{
				m_socket = new DatagramSocket(m_port);
			}
			catch(SocketException ex)
			{
				OpenHABBO.getInstance().getLogger().error("RemoteServer", String.format("Error binding to UDP/%d.", m_port), ex);
				return false;
			}

			// Create thread
			m_thread = new Thread(this, "RemoteServer (UDP/" + m_port + ")");

			// Start thread
			m_alive = true;
			m_thread.start();

			// Started!
			OpenHABBO.getInstance().getLogger().info("RemoteServer", String.format("Listening for remote commands on UDP port %d.", m_port));
			return true;
		}
	}

	public void stop()
	{
		if(m_alive)
		{
			// Close socket
			m_alive = false;
			m_thread.interrupt();

			// Close the socket
			m_socket.close();
		}
	}

	public void addAllowedIp(String ip)
	{
		synchronized(m_allowedIps)
		{
			m_allowedIps.add(ip);
		}
	}

	public void removeAllowedIp(String ip)
	{
		synchronized(m_allowedIps)
		{
			m_allowedIps.remove(ip);
		}
	}

	@Override
	public void run()
	{
		// This packet is re-used all the time
		DatagramPacket packet = new DatagramPacket(new byte[RemoteServer.PACKET_SIZE], RemoteServer.PACKET_SIZE);

		// While the server is running...
		while(m_alive)
		{
			// Receive the next packet
			try
			{
				m_socket.receive(packet);
			}
			catch(IOException ex)
			{
				// OpenHABBO.getInstance().getLogger().error("RemoteServer", "Error reading new UDP packet.", ex);
				continue;
			}

			// Sender is allowed?
			String senderIp = packet.getAddress().getHostAddress();
			if(m_allowedIps.contains(senderIp))
			{
				// Get data as a string and parse "COMMAND arg1 arg2 arg3 etc"
				String data = new String(packet.getData(), 0, packet.getLength());
				String parts[] = data.split(" ", 2);

				// Dispatch command to command handler
				m_cmdHandlerMgr.handleCommand(parts[0], (parts.length == 2) ? parts[1] : null);
			}
			else
			{
				// Lol hax
				OpenHABBO.getInstance().getLogger().warning("RemoteServer", String.format("IP '%s' is not allowed to issue remote commands. Command ignored.", senderIp));
			}
		}
	}

	public int getPort()
	{
		return m_port;
	}

	public RemoteCommandHandlerManager getCmdMgr()
	{
		return m_cmdHandlerMgr;
	}
}
