package org.openhabbo.rooms.instances;

import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.openhabbo.OpenHABBO;
import org.openhabbo.items.EItemBehaviour;
import org.openhabbo.items.Item;
import org.openhabbo.modules.ModuleResource;
import org.openhabbo.net.messages.ServerMessage;
import org.openhabbo.rooms.Room;
import org.openhabbo.rooms.RoomModel;
import org.openhabbo.rooms.pathfinding.JoehPathfinderStaticImpl;
import org.openhabbo.rooms.pathfinding.PathfinderNode;
import org.openhabbo.rooms.pathfinding.RotationCalculator;
import org.openhabbo.sessions.EServerMessage;
import org.openhabbo.sessions.Session;
import org.openhabbo.storage.DatabaseConnection;
import org.openhabbo.util.ChatUtil;

public final class RoomInstance extends ModuleResource
{
	/**
	 * Affecting this value will speed up movement of room users, bot AI, etc. The value is in milliseconds.
	 */
	public final static int PULSE_INTERVAL = 470;

	private Object m_lock;
	private Room m_info;
	private RoomModel m_model;
	private RoomMap m_map;

	private int m_workerIndex;
	private int m_idleTimer;
	private boolean m_unloading;

	private Map<Integer, Item> m_items;

	private Map<Integer, String> m_rights;
	private Map<Integer, Long> m_bans = null;

	private int m_virtualIdCounter;
	private Map<Integer, RoomUser> m_users;
	private Map<String, RoomUser> m_usersByUsername = new HashMap<String, RoomUser>();

	public RoomInstance(Room info, RoomModel model)
	{
		// Core room info
		m_lock = new Object();
		m_info = info;
		m_model = model;
		m_map = new RoomMap(model.getMaxX(), model.getMaxY());

		// Worker
		m_workerIndex = -1;
		m_idleTimer = 0;
		m_unloading = false;

		// Content
		m_items = new HashMap<Integer, Item>();
		m_rights = null;
		m_bans = null;

		// Users
		m_virtualIdCounter = 0;
		m_users = new HashMap<Integer, RoomUser>();
		m_usersByUsername = new HashMap<String, RoomUser>();
	}

	public Room getInfo()
	{
		return m_info;
	}

	public RoomModel getModel()
	{
		return m_model;
	}

	public RoomMap getMap()
	{
		return m_map;
	}

	public int getWorkerIndex()
	{
		return m_workerIndex;
	}

	public void setWorkerIndex(int index)
	{
		m_workerIndex = index;
	}

	public boolean requiresUnload()
	{
		final int maxIdleSeconds = 10;
		return(m_idleTimer > ((maxIdleSeconds * 1000) / RoomInstance.PULSE_INTERVAL));
	}

	@Override
	public String toString()
	{
		return "RoomInstance #" + m_info.id;
	}

	public boolean onLoad()
	{
		synchronized(m_lock)
		{
			// Reload bans & rights
			m_bans = new HashMap<Integer, Long>();
			m_rights = OpenHABBO.getInstance().getRoomMgr().getRightsForRoom(m_info.id);

			// Reload items
			m_items.clear();
			for(Item item : OpenHABBO.getInstance().getItemMgr().getItemsForRoom(m_info.id))
			{
				m_items.put(item.id, item);
			}

			// Generate map
			this.generateFloorMap();
		}

		return true;
	}

	public void onUnload()
	{
		synchronized(m_lock)
		{
			// Flag that this room instance is unloading
			m_unloading = true;

			// Remove all users
			for(RoomUser user : m_users.values())
			{
				if(user.getSession() != null)
				{
					user.getSession().kickFromRoom(null);
				}
			}
			m_users.clear();
			m_usersByUsername.clear();
			m_rights.clear();
			m_bans.clear();
			m_info.usersNow = 0;
			m_info.saveUserCount();

			// Save & clear items
			this.saveChangedItems();
			m_items.clear();
		}
	}

	private void saveChangedItems()
	{
		// Get dedicated database connection
		DatabaseConnection conn = OpenHABBO.getInstance().getDb().getConnMgr().getConnection();
		if(conn == null)
		{
			return;
		}

		// Prepare statement for updating items, batch everything and execute
		try
		{
			PreparedStatement query = conn.getInnerConnection().prepareStatement("UPDATE items SET x=?,y=?,z=?,rotation=?,wallposition=?,customdata=? WHERE id=?");
			for(Item item : m_items.values())
			{
				if(item.requiresDbUpdate())
				{
					query.setInt(1, item.x);
					query.setInt(2, item.y);
					query.setFloat(3, item.z);
					query.setByte(4, item.rotation);
					query.setString(5, item.wallPosition);
					query.setString(6, item.customData);
					query.setInt(7, item.id);
					query.addBatch();
					item.dbUpdated();
				}
			}
			query.executeBatch();
			query.close();
		}
		catch(SQLException ex)
		{
			OpenHABBO.getInstance().getLogger().error(this, "Error saving changed items.", ex);
		}

		// Release the dedicated database connection
		OpenHABBO.getInstance().getDb().getConnMgr().releaseConnection(conn);
	}

	public RoomUser getRoomUserByVirtualId(int virtualId)
	{
		synchronized(m_lock)
		{
			for(RoomUser user : m_users.values())
			{
				if(user.virtualId == virtualId)
				{
					return user;
				}
			}
		}
		return null;
	}

	public RoomUser getRoomUserByUserId(int userId)
	{
		return m_users.get(userId);
	}

	public RoomUser getRoomUserByUsername(String username)
	{
		return m_usersByUsername.get(username.toLowerCase());
	}

	public RoomUser getUserOnTile(int x, int y)
	{
		synchronized(m_lock)
		{
			for(RoomUser user : m_users.values())
			{
				if(user.x == x && user.y == y)
				{
					return user;
				}
			}
		}
		return null;
	}

	public void refreshUserOnTile(int x, int y)
	{
		RoomUser user = this.getUserOnTile(x, y);
		if(user != null)
		{
			// Remove old statuses and calculate new
			user.removeInteractiveStatuses();
			this.applyTileToUser(user);

			// Immediate refresh (no waiting on roominstance pulse)
			ServerMessage msg = new ServerMessage(EServerMessage.UserUpdate);
			msg.appendInt(1);
			user.appendStatusString(msg);
			this.broadcast(msg);
		}
	}

	public Collection<RoomUser> getRoomUsers()
	{
		return m_users.values();
	}

	public RoomUser addSession(Session session)
	{
		// If room instance is unloading, then don't support this operation anymore!
		if(m_unloading)
		{
			return null;
		}

		// Create roomuser with unique virtual ID
		RoomUser user = new RoomUser(m_virtualIdCounter++);
		user.setSession(session);
		user.setUserInfo(session.getUser());
		user.effectId = session.getEffectInventory().getCurrentEffect();

		// Position user at spawn position
		if(session.authenticatedTeleporterId == 0)
		{
			user.x = m_model.doorX;
			user.y = m_model.doorY;
			user.z = m_model.doorZ;
			user.headRotation = user.bodyRotation = m_model.doorRotation;
		}
		else
		{
			// Teleporter exists?
			Item tele = this.getItem(session.authenticatedTeleporterId);
			if(tele == null)
			{
				return null;
			}
			else
			{
				// Spawn in teleporter
				user.x = user.x;
				user.y = user.y;
				user.z = user.z;
				session.authenticatedTeleporterId = 0;

				// Broadcast activity
				// this.getInteractor().broadcastTeleporterActivity(obj.ID, obj.definition.sprite, user.getUserObject().name, false);
			}
		}

		// Determine privileges if in guestroom
		if(this.getInfo().isGuestRoom())
		{
			boolean isAnyRoomOwner = user.getUser().role.hasRight("is_any_room_owner");
			user.hasRoomOwner = (isAnyRoomOwner || (user.getUser().id == this.getInfo().ownerId));
			user.hasRoomRights = (user.hasRoomOwner || m_rights.containsKey(user.getUser().id));
			if(user.hasRoomRights)
			{
				user.statuses.put("flatctrl", user.hasRoomOwner ? "useradmin" : null);
			}
		}

		// Add to user collection
		synchronized(m_lock)
		{
			m_users.put(user.getUser().id, user);
			m_usersByUsername.put(user.getUser().username.toLowerCase(), user);
		}

		// Announce to other users
		ServerMessage msg = new ServerMessage(EServerMessage.Users);
		msg.appendInt(1); // 1 item
		msg.appendObject(user);
		this.broadcast(msg);

		// Update current user count
		synchronized(m_info)
		{
			m_info.usersNow++;
			m_info.saveUserCount();
		}

		// Done!
		OpenHABBO.getInstance().getLogger().debug(this, String.format("User #%d (%s) entered.", user.getUser().id, user.getUser().username));
		return user;
	}

	public void removeSession(Session session)
	{
		// If room instance is unloading, then don't support this operation anymore!
		if(m_unloading)
		{
			return;
		}

		// Remove user from collections
		RoomUser user;
		synchronized(m_lock)
		{
			user = m_users.remove(session.getUser().id);
			m_usersByUsername.remove(session.getUser().username.toLowerCase());
		}

		// Actually user removed?
		if(user != null)
		{
			// Invalidate user for any references floating around
			user.invalidate();

			// Unblock map
			m_map.users[user.x][user.y] = false;

			// Broadcast leaving
			ServerMessage msg = new ServerMessage(EServerMessage.UserRemove);
			msg.appendStr(Integer.toString(user.virtualId));
			this.broadcast(msg);

			// Update current user count
			synchronized(m_info)
			{
				m_info.usersNow--;
				m_info.saveUserCount();
			}

			// Exited successfully!
			OpenHABBO.getInstance().getLogger().debug(this, String.format("User #%d (%s) exited.", user.getUser().id, user.getUser().username));
		}
	}

	public void moderationKick(int issuerId, String message)
	{
		// Maintain collection with users to kick
		List<RoomUser> toKick = new ArrayList<RoomUser>();

		// Don't kick the issuer
		synchronized(m_lock)
		{
			for(RoomUser user : m_usersByUsername.values())
			{
				if(user.getSession() != null && user.getUser().id != issuerId)
				{
					toKick.add(user);
				}
			}
		}

		// Kick the targets
		for(RoomUser user : toKick)
		{
			user.getSession().kickFromRoom(message);
		}
	}

	public void broadcast(ServerMessage msg)
	{
		this.broadcast(msg.getResult());
	}

	public void broadcast(byte[] data)
	{
		OpenHABBO.getInstance().getLogger().debug(this, "--> " + new String(data));
		synchronized(m_lock)
		{
			for(RoomUser user : m_users.values())
			{
				if(user.getSession() != null)
				{
					user.getSession().getConnection().send(data);
				}
			}
		}
	}

	public void addRoomBan(int userId, int durationSeconds)
	{
		long expireTime = System.currentTimeMillis() + (durationSeconds * 1000);
		synchronized(m_lock)
		{
			m_bans.put(userId, expireTime);
		}
	}

	public void removeRoomBan(int userId)
	{
		synchronized(m_lock)
		{
			m_bans.remove(userId);
		}
	}

	public boolean userIsBanned(int userId)
	{
		Long expireTime = m_bans.get(userId);
		if(expireTime != null)
		{
			if(System.currentTimeMillis() < expireTime)
			{
				return true;
			}
			else
			{
				this.removeRoomBan(userId);
			}
		}
		return false;
	}

	public boolean addRights(int userId, String username)
	{
		if(m_rights.containsKey(userId))
		{
			return false;
		}
		else
		{
			synchronized(m_rights)
			{
				m_rights.put(userId, username);
			}
			OpenHABBO.getInstance().getRoomMgr().addRoomRights(m_info.id, userId);
			return true;
		}
	}

	public boolean removeRights(int userId)
	{
		if(!m_rights.containsKey(userId))
		{
			return false;
		}
		else
		{
			synchronized(m_rights)
			{
				m_rights.remove(userId);
			}
			OpenHABBO.getInstance().getRoomMgr().removeRoomRights(m_info.id, userId);
			return true;
		}
	}

	public boolean userHasRights(int userId)
	{
		return m_rights.containsKey(userId);
	}

	public Set<Entry<Integer, String>> getUsersWithRights()
	{
		return m_rights.entrySet();
	}

	public void chat(RoomUser source, String text, boolean shout)
	{
		// Broadcast chat to room
		ServerMessage msg = new ServerMessage(shout ? EServerMessage.Shout : EServerMessage.Chat);
		msg.appendInt(source.virtualId);
		msg.appendStr(text);
		msg.appendInt(ChatUtil.getChatGesture(text));
		this.broadcast(msg);

		// Rotate all heads to the source of chat (if possible)
		for(RoomUser user : m_users.values())
		{
			if(user.virtualId != source.virtualId)
			{
				user.lookTo(source.x, source.y, true);
			}
		}

		// TODO: trigger bot/pet responses etc
	}

	private void generateFloorMap()
	{
		// Get default floormap from room model
		char[][] modelMap = m_model.getMap();

		// Determine map boundaries
		int maxX = modelMap.length;
		int maxY = modelMap[0].length;

		// Create default floor map
		for(int y = 0; y < maxY; y++)
		{
			for(int x = 0; x < maxX; x++)
			{
				// Grab tile
				char tile = modelMap[x][y];

				// Determine tile
				if(tile != 'x' && Character.isDigit(tile))
				{
					m_map.state[x][y] = RoomMap.OPEN;
					m_map.height[x][y] = Float.parseFloat(Character.toString(tile));
				}
				else
				{
					m_map.state[x][y] = RoomMap.CLOSED;
					m_map.height[x][y] = 0.0f;
				}
			}
		}

		// Apply floor items to map
		synchronized(m_lock)
		{
			for(Item item : m_items.values())
			{
				if(item.is(EItemBehaviour.FLOORITEM))
				{
					this.applyFloorItemToMaps(item, m_map.state, m_map.height, null);
				}
			}
		}

		// Restore door tile
		m_map.state[m_model.doorX][m_model.doorY] = RoomMap.OPEN;
		m_map.height[m_model.doorX][m_model.doorY] = m_model.doorZ;
	}

	private void applyFloorItemToMaps(Item obj, byte[][] tmpMap, float[][] tmpHeightmap, RoomTile[][] tmpRedirectMap)
	{
		// TRIGGERS ARE SIMPLE. K.
		if(obj.is(EItemBehaviour.TRIGGER))
		{
			tmpMap[obj.x][obj.y] = RoomMap.INTERACTIVE;
			return;
		}

		for(RoomTile tile : RoomMap.getAffectedTiles(obj, true))
		{
			// Out of range?
			if(!(tile.x >= 0 && tile.x < tmpMap.length && tile.y >= 0 && tile.y < tmpMap[0].length))
			{
				OpenHABBO.getInstance().getLogger().warning(this, String.format("Floor item #%d at tile [%d,%d] was outside the boundaries of the map. Floor item not applied to maps.", obj.id, obj.x, obj.y));
				return;
			}

			// Increase heightmap
			if(obj.z >= tmpHeightmap[tile.x][tile.y])
			{
				tmpHeightmap[tile.x][tile.y] = obj.z;
			}

			// Is this object a door?
			if(obj.is(EItemBehaviour.DOOR))
			{
				// Is the door open?
				if(obj.customData != null && obj.customData.equals("O"))
				{
					tmpMap[tile.x][tile.y] = RoomMap.OPEN;
				}
				else
				{
					tmpMap[tile.x][tile.y] = RoomMap.CLOSED;
				}

				// Doors haven't got further handling
				return;
			}

			// Is this object the top item on the tile?
			if((obj.z >= tmpHeightmap[tile.x][tile.y]))
			{
				if(obj.is(EItemBehaviour.CAN_STAND_ON_TOP))
				{
					if(obj.z > tmpHeightmap[tile.x][tile.y])
					{
						tmpMap[tile.x][tile.y] = RoomMap.OPEN;
						tmpHeightmap[tile.x][tile.y] = (obj.z + obj.definition.height);
					}
				}
				else if(obj.is(EItemBehaviour.CAN_SIT_ON_TOP))
				{
					tmpMap[tile.x][tile.y] = RoomMap.INTERACTIVE;
				}
				else if(obj.is(EItemBehaviour.CAN_LAY_ON_TOP))
				{
					tmpMap[tile.x][tile.y] = RoomMap.INTERACTIVE;

					// Redirect to top
					if(obj.rotation == 0 || obj.rotation == 4)
					{
						tmpRedirectMap[tile.x][tile.y] = new RoomTile(tile.x, obj.y);
					}
					else
					{
						tmpRedirectMap[tile.x][tile.y] = new RoomTile(obj.x, tile.y);
					}
				}
				else
				{
					// Block the tile if there is no interactive object on top
					if(tmpMap[tile.x][tile.y] != RoomMap.INTERACTIVE)
					{
						tmpMap[tile.x][tile.y] = RoomMap.CLOSED;
					}
				}
			}
		}
	}

	public Item getItem(int itemId)
	{
		return m_items.get(itemId);
	}

	public List<Item> getItems(char behaviourFlag)
	{
		List<Item> items = new ArrayList<Item>();
		synchronized(m_lock)
		{
			for(Item item : m_items.values())
			{
				if(item.is(behaviourFlag))
				{
					items.add(item);
				}
			}
		}
		return items;
	}

	public List<Item> getFloorItemsOnTile(int X, int Y)
	{
		List<Item> objs = new ArrayList<Item>(0);
		synchronized(m_lock)
		{
			for(Item obj : m_items.values())
			{
				if(obj.is(EItemBehaviour.FLOORITEM))
				{
					for(RoomTile tile : RoomMap.getAffectedTiles(obj, true))
					{
						if(tile.x == X && tile.y == Y)
						{
							objs.add(obj);
						}
					}
				}
			}
		}
		return objs;
	}

	public boolean placeFloorItem(Item item, RoomUser user, int x, int y)
	{
		return this.moveFloorItem(item, user, x, y, (byte) 0, true);
	}

	public boolean moveFloorItem(Item item, RoomUser user, int x, int y, byte rotation)
	{
		return this.moveFloorItem(item, user, x, y, rotation, false);
	}

	private boolean moveFloorItem(Item item, RoomUser user, int newX, int newY, byte newRotation, boolean isNew)
	{
		// Validate new rotation, this can really trouble client and server!
		if(newRotation != 0 && newRotation != 2 && newRotation != 4 && newRotation != 6)
		{
			return false;
		}

		// Get the tiles this object would reside on if this item was placed here
		List<RoomTile> newTiles = RoomMap.getAffectedTiles(item.definition.length, item.definition.width, newX, newY, newRotation, true);

		// Check those tiles for 'free-ness'
		for(RoomTile tile : newTiles)
		{
			if(!m_map.exists(tile.x, tile.y))
			{
				return false; // Out of map range / bad tile
			}

			// Is there a SpaceUser on this tile?
			if(this.getUserOnTile(tile.x, tile.y) != null)
			{
				// Can rotate on different tile than root?
				if(newRotation == item.rotation && (tile.x != item.x || tile.y != item.y))
				{
					return false;
				}
			}
		}

		// Calculate new height
		float newZ = m_map.height[newX][newY];

		// Trying to stack on itself? (by rotating an object that is already in the space)
		if(!isNew)
		{
			if(item.rotation == newRotation && item.x == newX && item.y == newY && item.z != newZ)
			{
				return false;
			}
		}

		// Now check all the new tiles to determine new height
		for(RoomTile tile : newTiles)
		{
			// Get the objects that are 'leaning' on this tile somehow
			List<Item> itemsOnTile = this.getFloorItemsOnTile(tile.x, tile.y);

			// Check those objects and determine top height
			for(Item tileItem : itemsOnTile)
			{
				// Is this not the object that is being moved now?
				if(tileItem.id != item.id)
				{
					// Is this the top item?
					if((tileItem.z + tileItem.definition.height) > newZ)
					{
						// Can stack on top?
						if(tileItem.is(EItemBehaviour.CAN_STACK_ON_TOP))
						{
							newZ = tileItem.z + tileItem.definition.height;
						}
						else
						{
							return false;
						}
					}
				}
			}
		}

		// Max stacking height
		final int maxItemZ = 15;
		newZ = (newZ > maxItemZ) ? maxItemZ : newZ;

		// Trigger the interactor
		if(!item.onPlace(user))
		{
			return false;
		}

		// Backup the old position
		int oldX = item.x;
		int oldY = item.y;
		byte oldRotation = item.rotation;

		// Set object on new position
		item.x = newX;
		item.y = newY;
		item.z = newZ;
		item.rotation = newRotation;

		// If item is new to room...
		if(isNew)
		{
			item.room = this;
			item.roomId = m_info.id;
			item.ownerId = m_info.ownerId;
			item.update();
			synchronized(m_lock)
			{
				m_items.put(item.id, item);
			}
		}
		else
		{
			item.requireDbUpdate();
		}

		// Broadcast placement
		ServerMessage msg = new ServerMessage(isNew ? EServerMessage.ObjectAdd : EServerMessage.ObjectUpdate);
		msg.appendObject(item);
		this.broadcast(msg);

		// Generate the new map
		this.generateFloorMap();

		// Restore RoomUsers on old RoomTiles?
		if(!isNew)
		{
			for(RoomTile tile : RoomMap.getAffectedTiles(item.definition.length, item.definition.width, oldX, oldY, oldRotation, true))
			{
				this.refreshUserOnTile(tile.x, tile.y);
			}
		}

		// Refresh RoomUsers on new RoomTiles
		for(RoomTile tile : RoomMap.getAffectedTiles(item, true))
		{
			this.refreshUserOnTile(tile.x, tile.y);
		}

		// Move OK!
		return true;
	}

	public boolean placeWallItem(Item item, RoomUser user, String position)
	{
		// TODO: check if valid pos etc

		// Trigger the interactor
		if(!item.onPlace(user))
		{
			return false;
		}

		// Update the item with new info
		item.room = this;
		item.roomId = m_info.id;
		item.ownerId = m_info.ownerId;
		item.wallPosition = position;
		item.update();
		synchronized(m_lock)
		{
			m_items.put(item.id, item);
		}

		// Broadcast update
		ServerMessage msg = new ServerMessage(EServerMessage.ItemAdd);
		msg.appendObject(item);
		this.broadcast(msg);

		// Move OK!
		return true;
	}

	public Item pickupItem(int itemId, RoomUser user)
	{
		// Find the item
		Item item = this.getItem(itemId);

		// Pre-pickup checks
		if(!item.onPickup(user))
		{
			return null;
		}

		// Remove from collection
		synchronized(m_lock)
		{
			m_items.remove(item.id);
		}

		// Notify clients of removal
		ServerMessage msg = new ServerMessage();
		if(item.is(EItemBehaviour.FLOORITEM))
		{
			msg.init(EServerMessage.ObjectRemove);
			msg.appendStr(Integer.toString(item.id));
			msg.appendStr("");
			msg.appendBool(false);
		}
		else if(item.is(EItemBehaviour.WALLITEM))
		{
			msg.init(EServerMessage.ItemRemove);
			msg.appendStr(Integer.toString(item.id));
			msg.appendBool(false);
		}
		this.broadcast(msg);

		// Update floormaps if flooritem
		if(item.is(EItemBehaviour.FLOORITEM))
		{
			// Re-generate floormap
			this.generateFloorMap();

			// Refresh users on tile
			for(RoomTile tile : RoomMap.getAffectedTiles(item, true))
			{
				this.refreshUserOnTile(tile.x, tile.y);
			}
		}

		// Clear item
		item.room = null;
		item.roomId = 0;
		item.ownerId = 0;
		item.x = 0;
		item.y = 0;
		item.z = 0;
		item.rotation = 0;
		item.wallPosition = null;

		// Picked up successfully!
		return item;
	}

	public void pulse()
	{
		// Don't allw modification of the room
		synchronized(m_lock)
		{
			// Pulse the items, process timed items, etc
			this.pulseItems();

			// Pulse the users, kick idle sessions, etc
			this.pulseUsers();

			// Handle games? Or something? LOL!
		}
	}

	private void pulseItems()
	{
		for(Item item : m_items.values())
		{
			if(item.requiresPulse())
			{
				item.pulse();
			}
		}
	}

	private void pulseUsers()
	{
		// These values are dealt with at the end of the pulse
		int updates = 0;
		boolean hasSessions = false;
		Session toKickFromDoor = null;

		// Process all users (sessions/bots/pets
		for(RoomUser user : m_users.values())
		{
			if(user.getSession() != null)
			{
				hasSessions = true;
			}

			// Any updates required?
			if(user.requiresUpdate)
			{
				updates++;
				if(user.goalX != -1)
				{
					// Destination reached?
					if(user.x == user.goalX && user.y == user.goalY)
					{
						// Trip complete!
						user.clearMovement();

						// Is SpaceUser on an interactive tile?
						if(m_map.state[user.x][user.y] == RoomMap.INTERACTIVE)
						{
							this.applyTileToUser(user);
						}
						else if(user.x == m_model.doorX && user.y == m_model.doorY)
						{
							toKickFromDoor = user.getSession();
						}
					}
					else
					{
						// Determine next tile in path
						PathfinderNode next;
						if(user.overrideNextTile)
						{
							// Override used!
							user.overrideNextTile = false;

							// Substitute the next step
							next = new PathfinderNode();
							next.X = user.goalX;
							next.Y = user.goalY;
						}
						else
						{
							// Start new path?
							if(user.startNewTrip)
							{
								user.startNewTrip = false;
								this.startUserMovement(user, user.goalX, user.goalY, user.overrideNextTile);
							}

							// Pop the next tile from the path
							next = this.getMovingUserNextTile(user);
						}

						// Next tile available?
						if(next == null)
						{
							// Retry?
							if(++user.movementTries < 3)
							{
								user.requestNewTrip(user.goalX, user.goalY);
							}
							else
							{
								user.clearMovement();
							}

							// Is RoomUser on an interactive tile?
							if(m_map.state[user.x][user.y] == RoomMap.INTERACTIVE)
							{
								this.applyTileToUser(user);
							}
						}
						else
						{
							// Moved a tile
							user.movementTries = 0;

							// Set next coordinate for the next phase
							user.nextX = next.X;
							user.nextY = next.Y;
							user.nextZ = m_map.height[next.X][next.Y];
							user.requiresUpdate = true;
						}
					}
				}
			}
		}

		// Any updates to be broadcasted?
		if(updates > 0)
		{
			ServerMessage msg = new ServerMessage(EServerMessage.UserUpdate);
			msg.appendInt(updates);
			for(RoomUser user : m_users.values())
			{
				if(user.requiresUpdate)
				{
					if(user.nextX == -1)
					{
						user.appendStatusString(msg);
						user.requiresUpdate = false;
					}
					else
					{
						// Rotate users body
						byte rotation = RotationCalculator.calculateHumanMoveDirection(user.x, user.y, user.nextX, user.nextY);

						// Apply rotation
						user.headRotation = rotation;
						user.bodyRotation = rotation;

						// Remove interactive statuses
						user.removeInteractiveStatuses();

						// Add the movement status for the status update
						user.statuses.put("mv", user.nextX + "," + user.nextY + "," + Float.toString(user.nextZ).replace(',', '.'));
						user.appendStatusString(msg);
						user.statuses.remove("mv");

						// Unblock previous tile
						m_map.users[user.x][user.y] = false;

						// Block the next tile (if walkthrough is disabled)
						user.x = user.nextX;
						user.y = user.nextY;
						user.z = user.nextZ;
						user.nextX = -1;
						m_map.users[user.x][user.y] = !m_info.allowWalkthrough;
					}
				}
			}
			this.broadcast(msg);
		}

		// Update idle timer
		if(hasSessions)
		{
			m_idleTimer = 0;
			if(toKickFromDoor != null)
			{
				toKickFromDoor.kickFromRoom(null);
			}
		}
		else
		{
			m_idleTimer++;
		}
	}

	private void startUserMovement(RoomUser usr, int goalX, int goalY, boolean overrideNext)
	{
		// Tile exists?
		if(!m_map.exists(goalX, goalY))
		{
			// WTF TILE AINT ON MAP?!
			return;
		}

		// Override?
		if(overrideNext)
		{
			usr.goalX = goalX;
			usr.goalY = goalY;
			usr.overrideNextTile = true;
		}
		else
		{
			// One move to seat etc?
			boolean interactiveTile = false;
			if(m_map.state[goalX][goalY] == RoomMap.INTERACTIVE)
			{
				if(!m_map.users[goalX][goalY])
				{
					interactiveTile = true;
					m_map.state[goalX][goalY] = RoomMap.OPEN; // Free for one tick
				}
			}

			// Find path using the beloved Joe 'Joeh' Hegarty 2D A* pathfinder
			boolean noDescLimit = m_info.isGuestRoom();
			List<PathfinderNode> path = JoehPathfinderStaticImpl.findPath(usr.x, usr.y, goalX, goalY, m_map.state, m_map.users, m_map.height, 1.5f, noDescLimit ? 100.0f : 2.0f);

			// Restore map
			if(interactiveTile)
			{
				m_map.state[goalX][goalY] = RoomMap.INTERACTIVE;
			}

			// Is there a path found?
			if(path.size() > 0)
			{
				// Remove current tile node
				path.remove(0);

				// Set this path and destination to space user
				usr.path = path;
				usr.goalX = goalX;
				usr.goalY = goalY;
			}
		}
	}

	private PathfinderNode getMovingUserNextTile(RoomUser user)
	{
		int tileAmount = user.path.size();
		if(tileAmount > 0)
		{
			PathfinderNode next = user.path.remove(0);
			if(next == null || m_map.state[next.X][next.Y] == RoomMap.CLOSED)
			{
				// Invalid tile: tile blocked now
				return null;
			}
			else if(m_map.users[next.X][next.Y])
			{
				// Invalid tile: user here
				return null;
			}
			else
			{
				final boolean CROSS_DIAGONAL = true;
				if(tileAmount > 1 && CROSS_DIAGONAL)
				{
					// Peek + test next node (already popped one, so this is like path[1])
					PathfinderNode node = user.path.get(0);
					if(m_map.walkable(node.X, node.Y))
					{
						// Can't skip more than one tile!
						if(!(Math.abs(user.x - node.X) > 1 || Math.abs(user.y - node.Y) > 1))
						{
							byte X1 = 0;
							byte X2 = 0;
							if(node.X > user.x)
							{
								X1 = -1;
								X2 = 1;
							}
							else
							{
								X1 = 1;
								X2 = -1;
							}

							// Valid shortcut?
							if(m_map.walkable(node.X + X1, node.Y) && m_map.walkable(user.x + X2, user.y))
							{
								next = user.path.remove(0);
							}
						}
					}
				}
			}

			// The next tile!
			return next;
		}
		else
		{
			return null;
		}
	}

	private void applyTileToUser(RoomUser user)
	{
		for(Item obj : this.getFloorItemsOnTile(user.x, user.y))
		{
			if(applyFloorItemToUser(user, obj))
			{
				return;
			}
		}
	}

	private boolean applyFloorItemToUser(RoomUser user, Item obj)
	{
		if(obj.is(EItemBehaviour.CAN_SIT_ON_TOP))
		{
			user.z = obj.z;
			user.headRotation = obj.rotation;
			user.bodyRotation = obj.rotation;
			user.danceId = 0;
			user.statuses.put("sit", Float.toString(obj.z + obj.definition.height));
			return true;
		}

		return false;
	}
}
