package org.openhabbo.sessions;

import java.util.List;

import org.openhabbo.OpenHABBO;
import org.openhabbo.items.ItemInventory;
import org.openhabbo.messenger.InstantMessenger;
import org.openhabbo.modules.ModuleResource;
import org.openhabbo.net.NetConnection;
import org.openhabbo.net.messages.ClientMessage;
import org.openhabbo.net.messages.ServerMessage;
import org.openhabbo.rooms.Room;
import org.openhabbo.rooms.instances.RoomInstance;
import org.openhabbo.rooms.instances.RoomUser;
import org.openhabbo.support.ModerationBan;
import org.openhabbo.users.User;
import org.openhabbo.users.effects.AvatarEffectInventory;

public class Session extends ModuleResource implements NetConnection.ConnectionHandler
{
	public final long id;
	private NetConnection m_conn;

	// Message handling
	private long m_lastMsgTime;
	private ServerMessage m_msg;
	private ServerMessage m_responseObject;

	// Session variables
	private User m_user;
	private int m_accessLogId;
	private InstantMessenger m_messenger;
	private ItemInventory m_itemInventory;
	private AvatarEffectInventory m_effectInventory;
	private RoomInstance m_roomInstance;
	private RoomUser m_roomUser;

	public int authenticatedRoomId;
	public int authenticatedTeleporterId;

	public Session(long sessionId, NetConnection conn)
	{
		this.id = sessionId;
		m_conn = conn;
		m_msg = new ServerMessage();
		m_responseObject = new ServerMessage();

		this.onCreate();
	}

	private void onCreate()
	{

	}

	public void onEnd()
	{
		if(!m_conn.closed())
		{
			m_conn.closeSocket();
		}
		if(this.authenticated())
		{
			this.onLogout();
			OpenHABBO.getInstance().getSessionMgr().removeAuthedSession(this);
		}
	}

	public void authenticate(String ticket)
	{
		// User exists for this auth ticket?
		User user = OpenHABBO.getInstance().getUserMgr().getUserByAuthTicket(ticket);
		if(user == null)
		{
			OpenHABBO.getInstance().getSessionMgr().endSession(this.id, "invalid auth ticket");
			return;
		}

		// User is banned?
		List<ModerationBan> bans = OpenHABBO.getInstance().getSupportCenter().getBansForUserId(user.id, false);
		if(bans.size() > 0)
		{
			// Send the first ban and exit auth routine
			this.sendBan(bans.get(0));
			return;
		}

		// Insufficient permissions?
		if(user.role == null || !user.role.hasRight("fuse_login"))
		{
			OpenHABBO.getInstance().getSessionMgr().endSession(this.id, String.format("user #%d is not allowed to login because of insufficient role privileges", user.id));
			return;
		}

		// Add to session list
		m_user = user;
		OpenHABBO.getInstance().getSessionMgr().addAuthedSession(this);

		// Send user right list
		this.sendUserRights();

		// Login OK!
		m_msg.init(EServerMessage.AuthenticationOK);
		this.send(m_msg);

		// Run other logic
		this.onLogin();

		// Helloooo!
		this.sendNotif("Hello and welcome back to OpenHABBO 'Skylark'!", "http://www.openhabbo.org");
	}

	private void onLogin()
	{
		OpenHABBO instance = OpenHABBO.getInstance();

		// Reload badges and sort out slots
		m_user.badges = instance.getUserMgr().getBadgesForUser(m_user.id);
		m_user.equipBadges();

		// Reload tags
		m_user.tags = instance.getUserMgr().getTagsForUser(m_user.id);

		// Reload achievements
		m_user.achievements = instance.getAchievementMgr().getAchievementsForUser(m_user.id);

		// Create empty messenger component
		m_messenger = new InstantMessenger(this);

		// Reload item/pets inventory
		m_itemInventory = new ItemInventory(m_user.id);
		m_itemInventory.reload();

		// Reload effect inventory
		m_effectInventory = new AvatarEffectInventory(m_user.id);
		m_effectInventory.reload();
	}

	private void onLogout()
	{
		// Stop trading, leave the room, etc etc
		this.leaveRoom();

		// Logout from messenger
		if(m_messenger.initialized())
		{
			m_messenger.offline();
		}

		// Clear inventories to speed up garbagecollection
		m_itemInventory.clear();
		m_effectInventory.clear();

		// Save other stuff
		m_user.saveMisc();
		if(m_user.requiresBadgeUpdate)
		{
			m_user.saveBadges();
		}
		if(m_user.badges.size() > 0)
		{
			m_user.saveBadges();
		}
	}

	public void tryRoom(int roomId, String password)
	{
		// Leave any room instance the user might be in now
		this.leaveRoom();

		// Get/create room instance
		RoomInstance room = OpenHABBO.getInstance().getRoomMgr().getRoomInstance(roomId, true);
		if(room == null)
		{
			return;
		}

		// Room is full?
		if(room.getInfo().usersNow >= room.getInfo().usersMax)
		{
			// Can't override user limit?
			if(!this.getUser().role.hasRight("fuse_enter_full_rooms"))
			{
				// "Room is full"
				m_msg.init(EServerMessage.RoomEntryError);
				m_msg.appendInt(1);
				this.send(m_msg);

				// Kick to hotel view
				m_msg.init(EServerMessage.CloseConnection);
				this.send(m_msg);

				return;
			}
		}

		// The room owner can ALWAYS enter the room, never has to check for user limit, bans, state, etc
		if(this.getUser().id != room.getInfo().ownerId)
		{
			// Room is full?
			if(room.getInfo().usersNow >= room.getInfo().usersMax)
			{
				// Can't override user limit?
				if(!this.getUser().role.hasRight("fuse_enter_full_rooms"))
				{
					// "Room is full"
					m_msg.init(EServerMessage.RoomEntryError);
					m_msg.appendInt(1);
					this.send(m_msg);

					// Kick to hotel view
					m_msg.init(EServerMessage.CloseConnection);
					this.send(m_msg);

					return;
				}
			}

			// Execute access checks for room state?
			if(!this.getUser().role.hasRight("fuse_enter_any_room"))
			{
				// Password room?
				if(room.getInfo().state == Room.STATE_PASSWORD && !room.getInfo().password.equalsIgnoreCase(password))
				{
					this.sendGenericError(-100002);
					this.kickFromRoom(null);
					return;
				}
				// Doorbell room?
				else if(room.getInfo().state == Room.STATE_DOORBELL)
				{

				}

				// User is banned?
				if(room.userIsBanned(this.getUser().id))
				{
					// "You are banned from this room" message
					m_msg.init(EServerMessage.RoomEntryError);
					m_msg.appendInt(4);
					this.send(m_msg);

					// Kick to hotel viw
					m_msg.init(EServerMessage.CloseConnection);
					this.send(m_msg);

					return;
				}
			}
		}

		// Send room stats tracker url (?)
		m_msg.init(166);
		m_msg.append("/client/");
		if(room.getInfo().isPublicRoom())
		{
			m_msg.append("public/");
			m_msg.append(room.getInfo().modelName);
			m_msg.append("/0");
		}
		else
		{
			m_msg.append("private/");
			m_msg.append(Integer.toString(room.getInfo().id));
			m_msg.append("/id");
		}
		this.send(m_msg);

		// Send room model
		m_msg.init(EServerMessage.RoomReady);
		m_msg.appendStr(room.getModel().name);
		m_msg.appendInt(room.getInfo().id);
		this.send(m_msg);

		// Forward to room instance
		this.goToRoom(roomId);
	}

	public void goToRoom(int roomId)
	{
		// Get the running instance, it should be there by now...
		RoomInstance instance = OpenHABBO.getInstance().getRoomMgr().getRoomInstance(roomId, false);
		if(instance == null)
		{
			return;
		}

		// If entering a guestroom, then send more data, client doesn't ask for it
		if(instance.getInfo().isGuestRoom())
		{
			// Send wallpaper/floor/landscape if not default data
			if(!instance.getInfo().wallpaper.equals("0.0"))
			{
				m_msg.init(EServerMessage.RoomProperty);
				m_msg.appendStr("wallpaper");
				m_msg.appendStr(instance.getInfo().wallpaper);
				this.send(m_msg);
			}
			if(!instance.getInfo().floor.equals("0.0"))
			{
				m_msg.init(EServerMessage.RoomProperty);
				m_msg.appendStr("floor");
				m_msg.appendStr(instance.getInfo().floor);
				this.send(m_msg);
			}
			if(!instance.getInfo().landscape.equals("0.0"))
			{
				m_msg.init(EServerMessage.RoomProperty);
				m_msg.appendStr("landscape");
				m_msg.appendStr(instance.getInfo().landscape);
				this.send(m_msg);
			}

			// Send rating
			m_msg.init(EServerMessage.RoomRating);
			m_msg.appendInt(instance.getInfo().rating);
			this.send(m_msg);

			// Send ongoing room event
			m_msg.init(EServerMessage.RoomEvent);
			m_msg.appendStr("-1");
			this.send(m_msg);
		}

		// This is the current room instance now
		m_roomInstance = instance;
	}

	public boolean enterRoom()
	{
		m_roomUser = m_roomInstance.addSession(this);
		if(m_roomUser != null)
		{
			return true;
		}
		else
		{
			this.kickFromRoom("Invalid room entry!");
			return false;
		}
	}

	private void leaveRoom()
	{
		// Reset room ID if not teleporting
		if(this.authenticatedTeleporterId == 0)
		{
			this.authenticatedRoomId = 0;
		}

		// Leave room instance
		if(m_roomInstance != null)
		{
			m_roomInstance.removeSession(this);
			m_roomInstance = null;
			m_roomUser = null;
		}
	}

	public void kickFromRoom(String text)
	{
		// SHE GAVE ME THE PLEASE-FACE
		this.leaveRoom();

		// I GAVE HER THE LEAVE-FACE
		m_msg.init(EServerMessage.CloseConnection);
		this.send(m_msg);

		// BODY LIKE A BOTTLE, PRETTY LITTLE MODEL MODEL MODEEL
		if(text != null)
		{
			this.sendNotif(text);
		}
	}

	@Override
	public void onNetNewMessage(ClientMessage msg)
	{
		// Fucking heartbeat! ("D{")
		if(msg.msgId == 315) return;

		// Log that this session received a message!
		m_lastMsgTime = System.currentTimeMillis();
		OpenHABBO.getInstance().getLogger().debug(this, "<-- " + msg.toString());

		// Message contains protocol-EOF byte? This is _never_ acceptable because of exploitable purposes
		if(msg.contains(ServerMessage.MSG_EOF))
		{
			OpenHABBO.getInstance().getNetMgr().blacklistIp(this.getConnection().getIpAddress(), "MSG_EOF");
			OpenHABBO.getInstance().getSessionMgr().endSession(this.id, "MSG_EOF detected in incoming message");
			return;
		}

		// These are the only responding handlers for SendSessionParams and SsoLogin
		if(!this.authenticated() && (msg.msgId != 206 && msg.msgId != 415))
		{
			OpenHABBO.getInstance().getLogger().warning(this, String.format("msg #%d (\"%s\") is an invalid pre-auth msg.", msg.msgId, msg.header()));
		}
		else
		{
			// Lookup message handler and execute handling logic
			SessionMessageHandler handler = SessionMessageHandlerRegistry.lookup(msg.msgId);
			if(handler != null)
			{
				System.out.println(handler.getClass().getSimpleName());
				try
				{
					handler.handle(this, msg, m_responseObject);
				}
				catch(Exception ex)
				{
					OpenHABBO.getInstance().getLogger().error(this, String.format("Error handling msg #%d (\"%s\").", msg.msgId, msg.header()), ex);
				}
				m_responseObject.reset();
			}
			else
			{
				OpenHABBO.getInstance().getLogger().warning(this, String.format("SessionMessageHandlerRegistry does not have an entry for msg #%d. (\"%s\")", msg.msgId, msg.header()));
			}
		}
	}

	@Override
	public void onNetClose(String reason)
	{
		OpenHABBO.getInstance().getSessionMgr().endSession(this.id, reason);
	}

	public void send(ServerMessage msg)
	{
		OpenHABBO.getInstance().getLogger().debug(this, "--> " + msg.toString());
		m_conn.send(msg.getResult());
	}

	public void sendGenericError(int code)
	{

	}

	public void sendBan(ModerationBan ban)
	{
		m_msg.init(EServerMessage.UserBanned);
		m_msg.appendStr(ban.message);
		// TODO: "Date the ban will be lifted: %s" etc
		this.send(m_msg);
	}

	public void sendNotif(String text)
	{
		this.sendNotif(text, null);
	}

	public void sendNotif(String text, String url)
	{
		m_msg.init(EServerMessage.Notif);
		m_msg.appendStr(text);
		m_msg.appendStr(url);
		this.send(m_msg);
	}

	public void sendUserRights()
	{
		if(m_user != null)
		{
			m_msg.init(EServerMessage.UserRights);
			m_msg.appendInt(m_user.role.rights.length);
			for(String right : m_user.role.rights)
			{
				m_msg.appendStr(right);
			}
			this.send(m_msg);
		}
	}

	public void sendCreditBalance()
	{
		m_msg.init(EServerMessage.CreditBalance);
		m_msg.appendStr(m_user.credits + ".0");
		this.send(m_msg);
	}

	public void sendPixelBalance()
	{
		this.sendPixelsAddedNotif(0);
	}

	public void sendPixelsAddedNotif(int added)
	{
		m_msg.init(EServerMessage.PixelsReceivedNotification);
		m_msg.appendInt(m_user.pixels);
		m_msg.appendInt(added);
		this.send(m_msg);
	}

	public void sendSubscriptionData(String subscriptionId)
	{
		m_msg.init(EServerMessage.ScrSendUserInfo);
		m_msg.appendStr(subscriptionId);
		m_msg.appendInt(0); // days left
		m_msg.appendBool(true); // show expire date?
		m_msg.appendInt(0); // months left
		this.send(m_msg);
	}

	public void sendRoomPrivileges()
	{
		if(m_roomUser != null)
		{
			if(m_roomUser.hasRoomRights)
			{
				m_msg.init(EServerMessage.YouAreController);
				this.send(m_msg);
			}
			if(m_roomUser.hasRoomOwner)
			{
				m_msg.init(EServerMessage.YouAreOwner);
				this.send(m_msg);
			}
		}
	}

	public NetConnection getConnection()
	{
		return m_conn;
	}

	public long getLastMsgTime()
	{
		return m_lastMsgTime;
	}

	public boolean authenticated()
	{
		return(m_user != null);
	}

	public User getUser()
	{
		return m_user;
	}

	public int getAccessLogId()
	{
		return m_accessLogId;
	}

	public void setAccessLogId(int accessLogId)
	{
		if(this.m_accessLogId == 0)
		{
			this.m_accessLogId = accessLogId;
		}
	}

	public InstantMessenger getInstantMessenger()
	{
		return m_messenger;
	}

	public ItemInventory getItemInventory()
	{
		return m_itemInventory;
	}

	public AvatarEffectInventory getEffectInventory()
	{
		return m_effectInventory;
	}

	public boolean inRoom()
	{
		return(m_roomInstance != null);
	}

	public RoomInstance getRoomInstance()
	{
		return m_roomInstance;
	}

	public RoomUser getRoomUser()
	{
		return m_roomUser;
	}

	@Override
	public String toString()
	{
		return "Session #" + this.id;
	}
}
