package org.openhabbo.storage;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.openhabbo.core.Logger;

public final class DatabaseConnectionManager implements Runnable
{
	private final static long MAX_IDLE_SECONDS = 120;

	private Logger m_logger;
	private DatabaseEndpoint m_database;
	private String m_driver;
	private int m_maxConnections;

	private boolean m_monitorAlive;
	private Thread m_monitorThread;

	private int m_connectionCounter;
	private List<DatabaseConnection> m_connections;

	public DatabaseConnectionManager(Logger logger, DatabaseEndpoint databaseEworkeroint, String driver, int maxConnections)
	{
		m_logger = logger;
		m_database = databaseEworkeroint;
		m_driver = driver;
		m_maxConnections = maxConnections;
		m_connections = new ArrayList<DatabaseConnection>(maxConnections);
	}

	public boolean test()
	{
		// Register connection driver package
		String driverClassName = m_driver;
		try
		{
			Driver driver = (Driver) Class.forName(driverClassName).newInstance();
			DriverManager.registerDriver(driver);
		}
		catch(Exception ex)
		{
			m_logger.error("DatabaseConnectionMgr", "Cannot find driver '" + m_driver + "'", null);
			return false;
		}

		// Create test connection
		DatabaseConnection test = this.getConnection();
		if(test != null)
		{
			this.releaseConnection(test);
			return true;
		}
		else
		{
			return false;
		}
	}

	public void startMonitor()
	{
		if(!m_monitorAlive)
		{
			// Create monitor thread on a low priority
			m_monitorThread = new Thread(this, "DatabaseConnectionManager monitor");
			m_monitorThread.setPriority(Thread.MIN_PRIORITY);

			// Start the monitor thread
			m_monitorAlive = true;
			m_monitorThread.start();

			m_logger.info("DatabaseConnectionMgr", "Started connection monitor.");
		}
	}

	public void stopMonitor()
	{
		if(m_monitorAlive)
		{
			// Stop the monitor thread
			m_monitorAlive = false;
			m_monitorThread.interrupt();

			m_logger.info("DatabaseConnectionMgr", "Stopped connection monitor.");
		}
	}

	public void closeAll()
	{
		synchronized(m_connections)
		{
			// Close all open connections
			for(DatabaseConnection conn : m_connections)
			{
				try
				{
					conn.getInnerConnection().close();
				}
				catch(Exception ex)
				{
					ex.printStackTrace();
				}
			}
			m_connections.clear();

			// Woo woo!
			m_logger.info("DatabaseConnectionMgr", "Closed all connections.");
		}
	}

	public DatabaseConnection getConnection()
	{
		synchronized(m_connections)
		{
			// Any available connections?
			Iterator<DatabaseConnection> iterator = m_connections.iterator();
			while(iterator.hasNext())
			{
				DatabaseConnection conn = iterator.next();
				if(conn.isAvailable())
				{
					try
					{
						// If it's not closed, then this is the connection to return!
						if(!conn.getInnerConnection().isClosed())
						{
							// Exception ex = new Exception();
							// System.out.println("acquired: " + Logger.convertStackTraceToString(ex));
							conn.available(false);
							return conn;
						}
						else
						{
							iterator.remove();
							m_logger.warning("DatabaseConnectionMgr", String.format("Connection #%d was dead but still in the available connections list, removed.", conn.id));
						}
					}
					catch(SQLException ex)
					{
						// /care
					}
				}
			}

			// No open available connections. Allowed to create a new connection?
			if(m_maxConnections <= 0 || m_connections.size() < m_maxConnections)
			{
				try
				{
					Connection innerConn = DriverManager.getConnection(m_database.url, m_database.username, m_database.password);
					DatabaseConnection conn = new DatabaseConnection(m_connectionCounter++, innerConn, this);
					conn.available(true);
					m_connections.add(conn);
					m_logger.debug("DatabaseConnectionMgr", String.format("Created new connection #%d.", conn.id));

					// Try again
					return this.getConnection();
				}
				catch(Exception ex)
				{
					m_logger.error("DatabaseConnectionMgr", "Could not create a new connection! JDBC reported: " + ex.getMessage(), null);

					return null;
				}
			}

			// Can't create more connections at the moment? Wait for a returning connection, but no longer than x seconds
			try
			{
				m_connections.wait(5 * 1000);
			}
			catch(InterruptedException ex)
			{
				// Waitlock triggered! Yay, a connection returned!
				return this.getConnection();
			}

			// This is real bad
			m_logger.warning("DatabaseConnectionMgr", "No available connections, not allowed to create a new connection and waiting for returning connections timed out.");
			return null;
		}
	}

	public void releaseConnection(DatabaseConnection conn)
	{
		// Actually valid connection passed in?
		if(conn == null) return;

		// Update the connection and notify an arbitrary thread that is waiting that are waiting
		synchronized(m_connections)
		{
			conn.available(true);
			// System.out.println("released");
			m_connections.notify();
		}
	}

	@Override
	public void run()
	{
		// Hardcoded cleanup interval
		final long cleanupInterval = 30 * 1000;

		// Run all the time
		while(m_monitorAlive)
		{
			try
			{
				Thread.sleep(cleanupInterval);
			}
			catch(InterruptedException ex)
			{
				return;
			}

			// Close + remove available connections that have been idle for too long
			int cleanedUpCount = 0;
			synchronized(m_connections)
			{
				Iterator<DatabaseConnection> iterator = m_connections.iterator();
				while(iterator.hasNext())
				{
					DatabaseConnection conn = iterator.next();
					if(conn.isAvailable())
					{
						if(conn.getIdleSeconds() > DatabaseConnectionManager.MAX_IDLE_SECONDS)
						{
							this.close(conn);
							iterator.remove();
							cleanedUpCount++;
						}
					}
				}
			}

			// Any connections closed + removed? Write debug log
			if(cleanedUpCount > 0)
			{
				m_logger.debug("DatabaseConnectionMgr", String.format("Cleaned up %d connection(s).", cleanedUpCount));
			}
		}
	}

	private void close(DatabaseConnection conn)
	{
		try
		{
			conn.getInnerConnection().close();
		}
		catch(SQLException e)
		{
			m_logger.warning("DatabaseConnectionMgr", String.format("Error closing database connection #%d.", conn.id));
		}
	}

	public Logger getLogger()
	{
		return m_logger;
	}

	public int count()
	{
		return m_connections.size();
	}
}
