package org.openhabbo.storage;

import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

public class DatabaseQueryQueue implements Runnable
{
	private SQLDatabase m_db;
	private DatabaseConnection m_conn;
	private Statement m_st;
	private List<String> m_queries;

	private boolean m_alive;
	private Thread m_thread;

	public DatabaseQueryQueue(SQLDatabase db)
	{
		m_db = db;
		m_queries = new ArrayList<String>(50);
	}

	public boolean open()
	{
		// Not already open?
		if(this.isOpen())
		{
			return false;
		}

		// Get database connection
		DatabaseConnection conn = m_db.getConnMgr().getConnection();
		if(conn == null)
		{
			m_db.getLogger().warning("DatabaseQueryQueue", "Could not open because no database connection was available.");
			return false;
		}

		// Attempt creating the statement that will be used to process queue's with
		try
		{
			m_st = conn.getInnerConnection().createStatement();
			m_conn = conn;
		}
		catch(SQLException ex)
		{
			m_db.getConnMgr().releaseConnection(conn);
			m_db.getLogger().error("DatabaseQueryQueue", "Error creating statement!", ex);
			return false;
		}

		// Create thread
		m_thread = new Thread(this, "DatabaseQueryQueue");
		m_thread.setPriority(Thread.MIN_PRIORITY);

		// Start thread
		m_alive = true;
		m_thread.start();

		// Open for business
		return true;
	}

	public void close()
	{
		// Actually open?
		if(!this.isOpen())
		{
			return;
		}

		// Abort the thread
		m_alive = false;
		m_thread.interrupt();

		// Flush the queue
		this.process();

		// Close the statement
		try
		{
			m_st.close();
		}
		catch(SQLException ex)
		{
			m_db.getLogger().error("DatabaseQueryQueue", "Error closing statement.", ex);
		}

		// Release the connection
		m_db.getConnMgr().releaseConnection(m_conn);
	}

	public void enqueue(String sql)
	{
		synchronized(m_queries)
		{
			m_queries.add(sql);
		}
	}

	private void process()
	{
		// Don't bother with empty queues
		if(m_queries.size() == 0)
		{
			return;
		}

		// Execute all the queries
		synchronized(m_queries)
		{
			for(String query : m_queries)
			{
				try
				{
					m_st.executeUpdate(query);
				}
				catch(SQLException ex)
				{
					m_db.getLogger().error("DatabaseQueryQueue", String.format("Error executing \"%s\", error message: %s.", query, ex.getMessage()), null);
				}
			}

			// Clear the queue
			m_queries.clear();
		}
	}

	@Override
	public void run()
	{
		while(m_alive)
		{
			// Process queue
			this.process();

			// Process the queue every second
			try
			{
				Thread.sleep(1000);
			}
			catch(InterruptedException e)
			{
				return;
			}
		}
	}

	public boolean isOpen()
	{
		return(m_st != null);
	}

	public SQLDatabase getDb()
	{
		return m_db;
	}
}
