package org.openhabbo.support;

import java.sql.Types;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.openhabbo.OpenHABBO;
import org.openhabbo.net.messages.ServerMessage;
import org.openhabbo.rooms.instances.RoomInstance;
import org.openhabbo.sessions.EServerMessage;
import org.openhabbo.sessions.Session;
import org.openhabbo.storage.DatabaseConnection;
import org.openhabbo.storage.DatabaseResult;
import org.openhabbo.storage.SQLHelper;

public class SupportCenter implements Runnable
{
	private Thread m_thread;
	private boolean m_alive;
	private List<ChatlogEntry> m_chatlogStack;

	public SupportCenter()
	{
		m_thread = new Thread(this, "SupportCenter worker");
		m_chatlogStack = new ArrayList<ChatlogEntry>();
	}

	public void start()
	{
		m_alive = true;
		m_thread.start();
		OpenHABBO.getInstance().getLogger().info("SupportCenter", "Started worker.");
	}

	public void stop()
	{
		m_alive = false;
		m_thread.interrupt();
		OpenHABBO.getInstance().getLogger().info("SupportCenter", "Stopped worker.");
	}

	public void onShutdown()
	{
		this.stop();
	}

	public void writeSupportLog(int issuerId, int targetId, String action, String message, String extraInfo)
	{
		OpenHABBO.getInstance().getDb().execute("INSERT INTO support_log(timestamp,issuer_id,action,target_id,message,extra_info) VALUES('" + SQLHelper.getDateTimeStr() + "'," + issuerId + ",'" + SQLHelper.escape(action) + "'," + targetId + ",'" + SQLHelper.escape(message) + "','" + SQLHelper.escape(extraInfo) + "')");
	}

	public boolean issueUserAlert(int issuerId, int userId, String message, String extraInfo)
	{
		Session sess = OpenHABBO.getInstance().getSessionMgr().getSessionByUserId(userId);
		if(sess == null)
		{
			return false;
		}
		else
		{
			sess.sendNotif(message);
			this.writeSupportLog(issuerId, userId, "alert", message, extraInfo);
			return true;
		}
	}

	public boolean issueUserKick(int issuerId, int userId, String message, String extraInfo)
	{
		Session sess = OpenHABBO.getInstance().getSessionMgr().getSessionByUserId(userId);
		if(sess == null)
		{
			return false;
		}
		else
		{
			sess.kickFromRoom(message);
			this.writeSupportLog(issuerId, userId, "kick", message, extraInfo);
			return true;
		}
	}

	public boolean issueRoomAlert(int issuerId, int roomId, String message, String extraInfo)
	{
		RoomInstance room = OpenHABBO.getInstance().getRoomMgr().getRoomInstance(roomId, false);
		if(room == null)
		{
			return false;
		}
		else
		{
			ServerMessage msg = new ServerMessage(EServerMessage.Notif);
			msg.appendStr(message);
			room.broadcast(msg);

			this.writeSupportLog(issuerId, roomId, "roomalert", message, extraInfo);
			return true;
		}
	}

	public boolean issueRoomKick(int issuerId, int roomId, String message, String extraInfo)
	{
		RoomInstance room = OpenHABBO.getInstance().getRoomMgr().getRoomInstance(roomId, false);
		if(room == null)
		{
			return false;
		}
		else
		{
			room.moderationKick(issuerId, message);
			this.writeSupportLog(issuerId, roomId, "roomkick", message, extraInfo);
			return true;
		}
	}

	public boolean issueUserBan(int issuerId, int userId, String message, int hours, String extraInfo)
	{
		return false;
	}

	public boolean issueIpBan(int issuerId, String ip, String message, int hours, String extraInfo)
	{
		return false;
	}

	public List<ModerationBan> getBansForIp(String ip, boolean includeExpired)
	{
		List<ModerationBan> bans = new ArrayList<ModerationBan>(0);
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT * FROM support_bans WHERE ip='" + SQLHelper.escape(ip) + "'");
		while(result.row())
		{
			ModerationBan ban = ModerationBan.parse(result);
			if(includeExpired || !ban.hasExpired())
			{
				bans.add(ban);
			}
		}
		return bans;
	}

	public List<ModerationBan> getBansForUserId(int userId, boolean includeExpired)
	{
		List<ModerationBan> bans = new ArrayList<ModerationBan>(0);
		DatabaseResult result = OpenHABBO.getInstance().getDb().query("SELECT * FROM support_bans WHERE user_id=" + userId);
		while(result.row())
		{
			ModerationBan ban = ModerationBan.parse(result);
			if(includeExpired || !ban.hasExpired())
			{
				bans.add(ban);
			}
		}
		result.release();
		return bans;
	}

	public void addChatlog(ChatlogEntry entry)
	{
		synchronized(m_chatlogStack)
		{
			m_chatlogStack.add(entry);
		}
	}

	@Override
	public void run()
	{
		while(m_alive)
		{
			try
			{
				Thread.sleep(15 * 1000);
			}
			catch(InterruptedException e)
			{
				return;
			}

			// Any chatlogs to bulk insert?
			synchronized(m_chatlogStack)
			{
				if(m_chatlogStack.size() > 0)
				{
					// Create connection
					DatabaseConnection conn = OpenHABBO.getInstance().getDb().getConnMgr().getConnection();
					if(conn != null)
					{
						try
						{
							// Create PreparedStatement and batch all the INSERT queries
							PreparedStatement query = conn.getInnerConnection().prepareStatement("INSERT INTO support_chatlog(user_id,room_id,timestamp,type,text,receiver_id) VALUES (?,?,?,?,?,?)");
							for(ChatlogEntry entry : m_chatlogStack)
							{
								query.setInt(1, entry.userId);
								query.setInt(2, entry.roomId);
								query.setTimestamp(3, entry.timestamp);
								query.setString(4, entry.type.toString());
								query.setString(5, entry.text);
								if(entry.receiverId == 0)
								{
									query.setNull(6, Types.INTEGER);
								}
								else
								{
									query.setInt(6, entry.receiverId);
								}
								query.addBatch();
							}

							// Execute the batch and close the statement
							query.executeBatch();
							query.close();
						}
						catch(SQLException ex)
						{
							OpenHABBO.getInstance().getLogger().error("SupportCenter", "Error bulk inserting chatlog stack!", ex);
						}
						finally
						{
							OpenHABBO.getInstance().getDb().getConnMgr().releaseConnection(conn);
						}

						// Done!
						m_chatlogStack.clear();
					}
				}
			}
		}
	}
}
