package org.openhabbo.util;

import java.io.File;
import java.net.URL;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

public class PackageClassFinder
{
	/**
	 * Scans all classes accessible from the context class loader which belong to the given package and subpackages.
	 * 
	 * @param packageName
	 *        The base package
	 * @return The classes
	 */
	@SuppressWarnings({ "rawtypes" })
	public static List<Class> getClasses(String packageName)
	{
		ArrayList<Class> classes = new ArrayList<Class>();
		try
		{
			ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
			String path = packageName.replace('.', '/');
			Enumeration<URL> resources = classLoader.getResources(path);
			List<File> dirs = new ArrayList<File>();
			while(resources.hasMoreElements())
			{
				URL resource = resources.nextElement();
				String fileName = resource.getFile();
				String fileNameDecoded = URLDecoder.decode(fileName, "UTF-8");
				dirs.add(new File(fileNameDecoded));
			}
			for(File directory : dirs)
			{
				classes.addAll(findClasses(directory, packageName));
			}
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
		return classes;
	}

	/**
	 * Recursive method used to find all classes in a given directory and subdirs.
	 * 
	 * @param directory
	 *        The base directory
	 * @param packageName
	 *        The package name for classes found inside the base directory
	 * @return The classes
	 * @throws ClassNotFoundException
	 */
	@SuppressWarnings("rawtypes")
	private static List<Class> findClasses(File directory, String packageName) throws ClassNotFoundException
	{
		List<Class> classes = new ArrayList<Class>();
		if(!directory.exists())
		{
			return classes;
		}
		File[] files = directory.listFiles();
		for(File file : files)
		{
			String fileName = file.getName();
			if(file.isDirectory())
			{
				classes.addAll(findClasses(file, packageName + "." + fileName));
			}
			else if(fileName.endsWith(".class") && !fileName.contains("$"))
			{
				Class _class;
				try
				{
					_class = Class.forName(packageName + '.' + fileName.substring(0, fileName.length() - 6));
				}
				catch(ExceptionInInitializerError e)
				{
					_class = Class.forName(packageName + '.' + fileName.substring(0, fileName.length() - 6), false, Thread.currentThread().getContextClassLoader());
				}
				classes.add(_class);
			}
		}
		return classes;
	}

	@SuppressWarnings({ "rawtypes", "unchecked" })
	public <T> List<T> getClassesByType(String packageName)
	{
		List<T> matches = new ArrayList<T>();
		for(Class c : getClasses(packageName))
		{
			try
			{
				T test = (T) c.newInstance();
				matches.add(test);
			}
			catch(Exception ex)
			{
				// Cast failed
			}
		}
		return matches;
	}
}
