package org.openhabbo.net;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.concurrent.Executors;

import org.jboss.netty.bootstrap.ServerBootstrap;
import org.jboss.netty.channel.Channel;
import org.jboss.netty.channel.ChannelPipeline;
import org.jboss.netty.channel.ChannelPipelineFactory;
import org.jboss.netty.channel.Channels;
import org.jboss.netty.channel.group.ChannelGroup;
import org.jboss.netty.channel.group.DefaultChannelGroup;
import org.jboss.netty.channel.socket.nio.NioServerSocketChannelFactory;
import org.openhabbo.OpenHABBO;
import org.openhabbo.core.Logger;

public final class NetServer
{
	private ServerBootstrap server;
	private Channel listener;
	private final ChannelGroup channels;
	
	public NetServer()
	{
		this.server = null;
		this.listener = null;
		this.channels = new DefaultChannelGroup();
	}

	public boolean start(int port)
	{
		// Already running?
		if(this.server != null)
		{
			return false;
		}
		
		// Create server boss and worker threadpools using defaults
		ServerBootstrap server = new ServerBootstrap(new NioServerSocketChannelFactory(Executors.newCachedThreadPool(), Executors.newCachedThreadPool()));
		
		// Setup pipeline factory
		final GroupingHandler grouper = new GroupingHandler(this.channels);
		server.setPipelineFactory(new ChannelPipelineFactory()
		{
			public ChannelPipeline getPipeline() throws Exception
			{
				ChannelPipeline pipeline = Channels.pipeline(grouper, NetHandshakeHandler.getInstance());
				return pipeline;
			}
		});

		// Bind and start listening for connections
		try
		{
			Channel listener = server.bind(new InetSocketAddress(port));
			if(listener.isOpen())
			{
				this.server = server;
				this.listener = listener;
				
				// Yay!
				Logger.getInstance().info("NetServer", String.format("Listening for game connections on TCP port %d.", port));
				return true;
			}
		}
		catch(Exception ex)
		{
			
		}
		
		// Failed whatever
		Logger.getInstance().warning("NetServer", String.format("Could not start listening on TCP port %d! Port is already taken/security error?", port));
		return false;
	}
	
	public void stop()
	{
		if(this.server != null)
		{
			// Close listener
			this.listener.close();
			this.listener = null;
			
			// Done!
			Logger.getInstance().info("NetServer", "Closed listener.");
		}
	}

	public void close()
	{
		// Close all connections
		this.channels.close();
		this.channels.clear();
		
		// Destroy bootstrapper
		this.server.releaseExternalResources();
		this.server = null;
	}
	
	public boolean allowChannel(Channel channel)
	{
		// Limit configured?
		int limit = OpenHABBO.getInstance().getProperties().getInt("net.max.per.ip", 0);
		if(limit > 0)
		{
			// Get address of joining connection
			InetAddress address = ((InetSocketAddress)channel.getRemoteAddress()).getAddress();
				
			// Check all open channels
			int count = 0;
			for(Channel someChannel : this.channels)
			{
				// Same IP?
				if(((InetSocketAddress)someChannel.getRemoteAddress()).getAddress().equals(address))
				{
					// Increment and check count
					count++;
					if(count == limit)
					{
						Logger.getInstance().info("NetServer", String.format("Denying connection request from %s, because this host already has %d open connections. (max)", address.getHostAddress(), count));
						return false;
					}
				}
			}
		}
			
		// Allow!
		return true;
	}
	
	/* Begin singleton */
	private static NetServer instance;

	public static NetServer getInstance()
	{
		return instance;
	}

	public static void createInstance()
	{
		instance = new NetServer();
	}
	/* End singleton */
}
