package com.fjordev.popbird

import scala.collection.mutable

import org.joda.time.DateTime

import akka.actor.{ActorSystem}

import Account._

/**
 * Represents one Twitter account that has successfully been authorized with an {@link Application}.
 * @param app The Twitter application that was authorized with.
 * @param token The OAuth2 token of this account, supplied by the application.
 * @param secret The OAuth2 secret of this account, supplied by the application.
 * @param joinedAt Datetime representing the moment of registration wtih the service.
 */
class Account(val app: Application, val token: String, val secret: String, val joinedAt: DateTime = DateTime.now())
{
  /**
   * The unique Twitter user ID of the account.
   */
  val userId: Long = token.substring(0, token.indexOf('-')).toLong

  /**
   * Twitter @screenName of the account, minus the '@' sign. (aka 'username')
   */
  var screenName: String = _

  /**
   * Display name of the account, such as 'John Doe' or 'ACME Inc'.
   */
  var name: String = _

  /**
   * HTTP/HTTPS URL to the largest available profile image of the account.
   */
  var profileImageURL: String = _

  /**
   * ISO-xxx language code that is used when performing a tweet search.
   */
  var searchLanguage: String = _

  /**
   * Keywords that will be used when performing a tweet search.
   */
  val keywords = mutable.Set[Keyword]()

  /**
   * Relevant users that will be used to import followers.
   */
  val relevantUsers = mutable.Set[RelevantUser]()

  /**
   * Contact email of the account.
   */
  var email: String = _

  /**
   * Indicates whether the account is currently suspended by Twitter.
   */
  var suspended: Boolean = _

  /**
   * Indicates the current service status of the account.
   */
  var status: AccountStatus.Value = AccountStatus.Active

  /**
   * Flushes the account data to the database.
   */
  def save() = println("DB saved " + this)

  def actor(implicit system: ActorSystem) = system.actorSelection(app.actor(system).path + "/" + userId)
}

object Account
{

  /**
   * Keyword belonging to an {@link Account}. An account can have many keywords and a keyword can contain whitespaces.
   * @param account The associated account.
   * @param keyword The keyword string.
   */
  case class Keyword(account: Account, keyword: String)

  /**
   * A Twitter user that is already being followed by relevant targets for an {@link Account}. Will be used to 'import' followers.
   * @param account The associated account.
   * @param userId The Twitter user ID of the target user.
   * @param screenName The latest screen name of the target user.
   */
  case class RelevantUser(account: Account, userId: Long, var screenName: String = "")

  /**
   * An enumeration of valid statuses for an {@link Account}.
   */
  object AccountStatus extends Enumeration
  {
    /**
     * The account is active and operating.
     */
    val Active = Value

    /**
     * The account has a pending payment and is not operating.
     */
    val PaymentPending = Value

    /**
     * The account is currently suspended due to abuse.
     */
    val Abuse = Value
  }

}