package net.nillus.habbo;

import java.util.Map.Entry;

import net.nillus.habbo.core.Console;
import net.nillus.habbo.core.Logger;
import net.nillus.habbo.core.Logger.ELogLevel;
import net.nillus.habbo.modules.ModuleInfo;
import net.nillus.habbo.rooms.RoomManager;
import net.nillus.habbo.sessions.SessionManager;
import net.nillus.habbo.util.Base64Encoding;
import net.nillus.habbo.util.WireEncoding;

public final class ConsoleFrontend
{
	public static void main(String[] args)
	{
		// Exit code
		int code = 0;

		// Set logger output path
		Logger.getInstance().setOutputFile("habbod.log");

		// Create + start instance
		habbod instance = habbod.createInstance("habbod.properties");
		if(!instance.start())
		{
			Logger.getInstance().info("habbod", "Startup failed.");
			code = 1;
		}
		else
		{
			while(instance.alive())
			{
				String line = Console.readLine();
				try
				{
					if(line != null)
					{
						ConsoleFrontend.executeCommand(instance, line);
					}
				}
				catch(Exception ex)
				{
					Logger.getInstance().error("ConsoleFrontend", String.format("Error executing command line '%s'.", line), ex);
				}
			}
		}

		// Exit the JVM with a helpful exit code
		System.out.println("Press a key to exit.");
		Console.readKey();
		System.exit(code);
	}

	private static void executeCommand(habbod instance, String line)
	{
		String cmd = line.split(" ", 2)[0].toLowerCase();

		// Execute the appropriate logic for each command
		if(cmd.equals("reloadprops"))
		{
			instance.reloadProperties();
		}
		else if(cmd.equals("reloadblacklist"))
		{
			//NetManager.getInstance().reloadBlacklist();
		}
		else if(cmd.equals("shutdown"))
		{
			String[] args = line.split(" ", 3);

			// Any arguments?
			if(args.length > 1)
			{
				// Shutdown notification ingame?
				int timeoutSeconds = Integer.parseInt(args[1]);
				if(args.length == 3)
				{
					SessionManager.getInstance().sendNotifToAll(String.format("habbod is shutting down!\r\r%s\r\rYou will be disconnected soon....", args[2]));
				}

				// Wait out the timeout, then enter shutdown sequence...
				Logger.getInstance().info("ConsoleFrontend", String.format("Entering shutdown sequence in %d seconds...", timeoutSeconds));
				try
				{
					Thread.sleep(timeoutSeconds * 1000);
				}
				catch(InterruptedException e)
				{

				}
			}

			// Kill the instance!
			instance.stop("ConsoleFrontend command 'shutdown' sent");
		}
		else if(cmd.equals("stats"))
		{
			long uptime = (System.currentTimeMillis() - instance.getStartTime());
			long days = (uptime / (1000 * 60 * 60 * 24));
			long hours = (uptime - days * (1000 * 60 * 60 * 24)) / (1000 * 60 * 60);
			long minutes = (uptime - days * (1000 * 60 * 60 * 24) - hours * (1000 * 60 * 60)) / (1000 * 60);
			long seconds = (uptime - days * (1000 * 60 * 60 * 24) - hours * (1000 * 60 * 60) - minutes * (1000 * 60)) / (1000);
			System.out.println(String.format("habbod uptime: %d day(s), %d hour(s), %d minute(s) and %d second(s).", days, hours, minutes, seconds));
			System.out.println(String.format("JVM memory usage: %d kilobytes.", Runtime.getRuntime().totalMemory() / 1024));
			System.out.println(String.format("Open database connections: %d.", instance.getDb().getConnMgr().count()));
			System.out.println(String.format("I have %d authenticated session(s) and I run %d room instance(s).", SessionManager.getInstance().authedSessions(), RoomManager.getInstance().getAllRoomInstances().size()));
		}
		else if(cmd.equals("debug"))
		{
			String args = line.split(" ")[1];
			Logger.getInstance().setMinImportance(args.equals("on") ? ELogLevel.Debug : ELogLevel.Info);
		}
		else if(cmd.equals("memclean"))
		{
			Runtime.getRuntime().gc();
			Logger.getInstance().info("ConsoleFrontend", "JVM GC executed.");
		}
		else if(cmd.equals("modules"))
		{
			System.out.println("Begin of module list:");
			for(Entry<ModuleInfo, String> module : instance.getModMgr().getLoadedModules().entrySet())
			{
				// TODO: print module info
			}
			System.out.println("End of module list.");
		}
		else if(cmd.equals("reloadmodules"))
		{
			instance.getModMgr().reloadModules();
		}
		else if(cmd.equals("encode"))
		{
			String[] args = line.split(" ", 3);
			if(args.length >= 3)
			{
				int i = Integer.parseInt(args[2]);
				if(args[1].equals("wire"))
				{
					System.out.println("WireEncoding.encodeInt(" + i + ") = " + new String(WireEncoding.encodeInt(i)));
				}
				else if(args[1].equals("base64"))
				{
					System.out.println("Base64Encoding.encodeInt(" + i + ") = " + new String(Base64Encoding.encodeInt(i, 2)));
				}
			}
		}
		else if(cmd.equals("decode"))
		{
			String[] args = line.split(" ", 3);
			if(args.length >= 3)
			{
				if(args[1].equals("wire"))
				{
					System.out.println("WireEncoding.decodeInt(" + args[2] + ") = " + WireEncoding.decodeInt(args[2].getBytes())[0]);
				}
				else if(args[1].equals("base64"))
				{
					System.out.println("Base64Encoding.decodeInt(" + args[2] + ") = " + Base64Encoding.decodeInt(args[2].getBytes()));
				}
			}
		}
		else if(cmd.equals("help"))
		{
			System.out.println("The following commands are available: ");
			System.out.println("- reloadprops - Reloads the openhabbo.properties file");
			System.out.println("- reloadblacklist - Reloads the IP blacklist");
			System.out.println("- shutdown [timeout] [notif] - Gracefully shutdowns the server, after waiting for an optional [timeout] and [notif]ying online users");
			System.out.println("- stats - Shows various system stats");
			System.out.println("- modules - Shows loaded modules");
			System.out.println("- memclean - Runs the JVM garbagecollector");
			System.out.println("- encode %base64|wire %integer - Base64 / Wire encodes %integer and prints the result as a string");
			System.out.println("- decode %base64|wire %string - Base64 / Wire decodes %string and prints the result as an integer");
		}
		else
		{
			Logger.getInstance().info("ConsoleFrontend", String.format("Unknown command '%s', type 'help' for a list of commands.", cmd));
		}
	}
}
