package net.nillus.habbo.rooms.instances;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

import mod_pets.RoomPet;
import net.nillus.habbo.communication.messages.outgoing.EServerMessage;
import net.nillus.habbo.communication.messages.outgoing.SerializableObject;
import net.nillus.habbo.communication.messages.outgoing.ServerMessage;
import net.nillus.habbo.core.DynamicObject;
import net.nillus.habbo.events.Event;
import net.nillus.habbo.rooms.pathfinding.PathfinderNode;
import net.nillus.habbo.rooms.pathfinding.RotationCalculator;
import net.nillus.habbo.util.ChatUtil;


public abstract class RoomAvatar extends DynamicObject implements SerializableObject
{
	public final static Event onPulse = new Event("RoomAvatar.OnPulse");
	public final static Event onUnidle = new Event("RoomAvatar.OnUnidle");

	// Room
	public int avatarId;
	private RoomInstance room;
	private boolean invalidated;

	// Position
	/*
	public int x;
	public int y;
	public float z;*/
	public byte headRotation;
	public byte bodyRotation;
	public RoomTile tile;
	
	// Statuses
	public final HashMap<String, String> statuses;
	public boolean requiresUpdate;

	public int idleTime;

	// Movement
	public RoomTile goal;
	public RoomTile next;
	public byte movementTries;
	private boolean moveLock;
	public boolean overrideNextTile;
	public List<PathfinderNode> path;
	public boolean executePathfinder;
	public byte pathfindingAllowedOn;

	public RoomAvatar()
	{
		this.statuses = new HashMap<String, String>();
		this.path = Collections.emptyList();
		this.clearMovement();
	}
	
	public RoomInstance getRoom()
	{
		return this.room;
	}

	public void setRoom(RoomInstance room)
	{
		this.room = room;
	}

	public void invalidate()
	{
		this.invalidated = true;
	}

	public final boolean invalidated()
	{
		return this.invalidated;
	}

	public void pulse()
	{
		if(RoomAvatar.onPulse.fire(this))
		{
			this.idleTime++;
		}
	}

	public void unidle()
	{
		if(RoomAvatar.onUnidle.fire(this))
		{
			this.idleTime = 0;
		}
	}

	public boolean at(RoomTile tile)
	{
		return(this.tile == tile);
	}
	
	public RoomTile getTile()
	{
		return this.tile;//this.room.getMap().getTile(this.x, this.y);
	}
	
	public void setGoal(RoomTile goal)
	{
		// Unidle the avatar
		this.unidle();
		
		// Get rid of current path
		this.clearMovement();
		
		// Set the goal
		if(goal != null)
		{
			this.goal = goal;
			this.executePathfinder = true;
			
			// Include in next instance pulse
			this.requiresUpdate = true;
		}
	}

	public final boolean hasGoal()
	{
		return(this.goal != null);
	}

	public void clearMovement()
	{
		this.goal = null;
		//this.next = null;
		this.movementTries = 0;
		this.path.clear();
	}

	public boolean canMove()
	{
		return !this.moveLock;
	}

	public void lockMovement()
	{
		this.moveLock = true;
		this.clearMovement();
	}

	public void unlockMovement()
	{
		this.moveLock = false;
	}

	public void lookTo(RoomTile tile, boolean lockBody)
	{
		// Can't rotate in this context?
		if(this.hasGoal() || tile.equals(this.getTile()) || this.statuses.containsKey("lay"))
		{
			return;
		}

		// Calculate new rotation
		byte rotation = RotationCalculator.calculateDirection(this.tile.x, this.tile.y, tile.x, tile.y);
		int diff = this.bodyRotation - rotation;
		this.headRotation = this.bodyRotation;

		// Apply new rotation
		if(lockBody || this.statuses.containsKey("sit"))
		{
			if(this.bodyRotation == 2 || this.bodyRotation == 4)
			{
				if(diff > 0)
				{
					this.headRotation = (byte) (this.bodyRotation - 1);
				}
				else if(diff < 0)
				{
					this.headRotation = (byte) (this.bodyRotation + 1);
				}
			}
			else if(this.bodyRotation == 0 || this.bodyRotation == 6)
			{
				if(diff > 0)
				{
					this.headRotation = (byte) (this.bodyRotation - 1);
				}
				else if(diff < 0)
				{
					this.headRotation = (byte) (this.bodyRotation + 1);
				}
			}
		}
		else if(diff <= -2 || diff >= 2)
		{
			this.headRotation = rotation;
			this.bodyRotation = rotation;
		}
		else
		{
			this.headRotation = rotation;
		}
		
		// Unidle and require update (rotation changed)
		this.unidle();
		this.requiresUpdate = true;
	}

	public void chat(String text, boolean isShout)
	{
		// Broadcast chat to room
		ServerMessage msg = new ServerMessage(isShout ? EServerMessage.Shout : EServerMessage.Chat);
		msg.appendInt(this.avatarId);
		msg.appendStr(text);
		msg.appendInt(ChatUtil.getChatGesture(text));
		this.room.broadcast(msg);

		// Unidle
		this.unidle();
		
		// Interact with all other room avatars
		for(RoomAvatar avatar : this.room.getAllAvatars())
		{
			if(avatar.avatarId != this.avatarId)
			{
				avatar.onChatFromAvatar(this, text, isShout);
			}
		}
	}

	protected abstract void onChatFromAvatar(RoomAvatar from, String text, boolean isShout);

	public final void addStatus(String key)
	{
		this.addStatus(key, null);
	}

	public final void addStatus(String key, String value)
	{
		this.statuses.put(key, value);
		this.requiresUpdate = true;
	}

	public final void removeStatus(String key)
	{
		this.statuses.remove(key);
		this.requiresUpdate = true;
	}
	
	public void removeNonMoveStatuses()
	{
		this.removeStatus("sit");
		this.removeStatus("lay");
	}
	
	public void refresh(boolean instant)
	{
		this.updatePosition(this.tile, instant);
	}
	
	public void updatePosition(RoomTile to, boolean instantStatusUpdate)
	{
		// Step off current tile
		RoomTile from = this.tile;
		from.stepOff(this);
		
		// Step on new tile
		to.stepOn(this);
		
		if(!instantStatusUpdate)
		{
			this.requiresUpdate = true;
		}
		else
		{
			// Send instant update
			ServerMessage msg = new ServerMessage(EServerMessage.UserUpdate);
			msg.appendInt(1);
			this.appendStatusString(msg);
			this.room.broadcast(msg);
			
			// Don't require further update
			this.requiresUpdate = false;
		}
	}
	
	public final void appendStatusString(ServerMessage msg)
	{
		msg.appendInt(this.avatarId);
		msg.appendInt(this.tile.x);
		msg.appendInt(this.tile.y);
		msg.appendStr(this.tile.height);
		msg.appendInt(this.headRotation);
		msg.appendInt(this.bodyRotation);
		for(Entry<String, String> status : this.statuses.entrySet())
		{
			msg.append(status.getKey());
			if(status.getValue() != null)
			{
				msg.append(' ');
				msg.append(status.getValue());
			}
			msg.append('/');
		}
		if(this instanceof RoomPet)
		{
			if(this.tile.is(RoomTile.WATER))
			{
				msg.append("swm");
				msg.append('/');
			}
		}
		msg.appendStr('/');
	}

	public abstract String getAvatarKey();
}
