package net.nillus.habbo.game.items;

import java.util.List;

import net.nillus.habbo.habbod;
import net.nillus.habbo.communication.EServerMessage;
import net.nillus.habbo.communication.SerializableObject;
import net.nillus.habbo.communication.ServerMessage;
import net.nillus.habbo.game.items.wired.triggers.WiredUserStepsOffItemTrigger;
import net.nillus.habbo.game.items.wired.triggers.WiredUserStepsOnItemTrigger;
import net.nillus.habbo.game.items.wired.triggers.WiredUserTogglesItemTrigger;
import net.nillus.habbo.game.rooms.instances.RoomAvatar;
import net.nillus.habbo.game.rooms.instances.RoomInstance;
import net.nillus.habbo.game.rooms.instances.RoomMap;
import net.nillus.habbo.game.rooms.instances.RoomTile;
import net.nillus.habbo.game.rooms.instances.RoomUser;
import net.nillus.habbo.game.users.User;
import net.nillus.habbo.sessions.Session;
import net.nillus.habbo.storage.DatabaseResult;
import net.nillus.habbo.storage.SQLHelper;
import net.nillus.habbo.util.FlagsUtil;


public abstract class Item implements SerializableObject
{
	public final static String ERROR_DATA = Character.toString((char) 1);

	public int id;
	public ItemDefinition definition;
	public int ownerId;
	public int roomId;
	public int x;
	public int y;
	public float z;
	public byte rotation;
	public String wallPosition;
	public String state;
	public String data;
	private boolean requiresDbUpdate;

	private RoomInstance room;
	private int pulseTimer;
	
	public final boolean is(char flag)
	{
		return this.definition.is(flag);
	}

	public final void requireDbUpdate()
	{
		requiresDbUpdate = true;
	}

	public final void dbUpdated()
	{
		requiresDbUpdate = false;
	}

	public final boolean requiresDbUpdate()
	{
		return requiresDbUpdate;
	}

	public final boolean inRoom()
	{
		return(this.room != null);
	}

	public final RoomInstance getRoom()
	{
		return this.room;
	}

	public final void setRoom(RoomInstance newRoom)
	{
		if(this.room != null || newRoom != null)
		{
			if(newRoom != null)
			{
				this.room = newRoom;
				this.onRoomAdd(newRoom);
			}
			else
			{
				this.onRoomRemove(this.room);
				this.room.removeTriggers(this.id);
				this.room = null;
			}
		}
	}

	public boolean requiresPulse()
	{
		return this.hasTimeout();
	}

	protected final boolean hasTimeout()
	{
		return(this.pulseTimer > 0);
	}

	protected final void setTimeout(int pulses)
	{
		this.pulseTimer = pulses;
	}

	protected final void clearTimeout()
	{
		this.pulseTimer = 0;
	}

	public final void pulse()
	{		
		// Pulse!
		this.onPulse();
		if(this.pulseTimer > 0)
		{
			this.pulseTimer--;
		}
		
		// End of timeout?
		if(this.pulseTimer == 0)
		{
			this.clearTimeout();
			this.onTimeout();
		}
	}

	public final void setState(String state)
	{
		this.state = state;
		if(this.inRoom())
		{
			this.refreshInRoom();
			this.requireDbUpdate();
		}
		else
		{
			this.update();
		}
	}

	public final void setTempState(String state)
	{
		String backup = this.state;
		this.state = state;
		this.refreshInRoom();
		this.state = backup;
	}

	public final void refreshInRoom()
	{
		if(this.inRoom())
		{
			ServerMessage msg = new ServerMessage();
			if(this.is(EItemFlags.FLOORITEM))
			{
				msg.init(EServerMessage.ObjectDataUpdate);
				msg.appendStr(this.id);
				msg.appendStr(this.state);
			}
			else if(this.is(EItemFlags.WALLITEM))
			{
				msg.init(EServerMessage.ItemUpdate);
				msg.appendObject(this);
			}
			this.room.broadcast(msg);
		}
	}

	public final RoomTile getTile()
	{
		return this.room.getMap().getTile(this.x, this.y);
	}
	
	public List<RoomTile> getTiles()
	{
		return this.room.getMap().getAffectedTiles(this);
	}

	public final RoomTile getTileInFront()
	{
		int x = this.x;
		int y = this.y;
		if(this.rotation == 0)
		{
			y--;
		}
		else if(this.rotation == 2)
		{
			x++;
		}
		else if(this.rotation == 4)
		{
			y++;
		}
		else if(this.rotation == 6)
		{
			x--;
		}
		return this.room.getMap().getTile(x, y);
	}

	public final RoomTile getTileBehind()
	{
		int x = this.x;
		int y = this.y;
		if(this.rotation == 0)
		{
			y++;
		}
		else if(this.rotation == 2)
		{
			x--;
		}
		else if(this.rotation == 4)
		{
			y--;
		}
		else if(this.rotation == 6)
		{
			x++;
		}
		return this.room.getMap().getTile(x, y);
	}

	public void update()
	{
		if(this.id > 0)
		{
			if(this.is(EItemFlags.FLOORITEM))
			{
				habbod.getInstance().getDb().execute("UPDATE items SET owner_id=" + this.ownerId + ",room_id=" + this.roomId + ",x=" + this.x + ",y=" + this.y + ",z=" + this.z + ",rotation=" + this.rotation + ",state='" + SQLHelper.escape(this.state) + "',data='" + SQLHelper.escape(this.data) + "' WHERE id=" + this.id);
			}
			else if(this.is(EItemFlags.WALLITEM))
			{
				habbod.getInstance().getDb().execute("UPDATE items SET owner_id=" + this.ownerId + ",room_id=" + this.roomId + ",wallposition = '" + SQLHelper.escape(this.wallPosition) + "',state='" + SQLHelper.escape(this.state) + "',data='" + SQLHelper.escape(this.data) + "' WHERE id=" + this.id);
			}
			this.dbUpdated();
		}
	}

	public void delete()
	{
		if(this.id > 0)
		{
			habbod.getInstance().getDb().execute("DELETE FROM items WHERE id=" + this.id);
		}
	}
	
	public String processPurchaseDataToState(User user, String purchaseData)
	{
		// purchaseData is discarded, impl classes should return proper-formatted data, is only allowed on a few item types, and this is how we deal with 'em
		// This is the default state for all items
		return "0";
	}

	public void onPurchase(List<Item> shipping)
	{
		// Optionally overridden by impl class

		// Create this item and add to shipping
		ItemManager.getInstance().storeItem(this);
		shipping.add(this);
	}

	protected void onParse()
	{
		// Optionally overridden by impl class
	}

	protected void onRoomAdd(RoomInstance room)
	{
		// Optionally overridden by impl class
	}

	protected void onRoomRemove(RoomInstance room)
	{
		// Optionally overriden by impl class
	}

	public boolean onPlace(Session session, RoomInstance room)
	{
		// Optionally overridden by impl class
		return true;
	}

	public boolean onPickup(Session session)
	{
		// Optionally overridden by impl class
		return true;
	}

	public void onDelete()
	{
		// Optionally overridden by impl class
	}

	protected void onPulse()
	{
		// Optionally overridden by impl class
	}

	protected void onTimeout()
	{
		// Optionally overridden by impl class
	}

	public void onGenerateFloorMapTile(RoomMap map, RoomTile tile)
	{
		tile.effectId = 0;
		if(this.is(EItemFlags.CAN_STAND_ON_TOP))
		{
			if(!FlagsUtil.isSet(tile.state, RoomTile.CLOSED))
			{
				tile.topItem = this;
				tile.height = (this.z + this.definition.height);
				tile.state = RoomTile.OPEN;
			}
		}
		else if(this.is(EItemFlags.CAN_SIT_ON_TOP))
		{
			tile.height = this.z;
			tile.topItem = this;
			tile.state = (RoomTile.CLOSED | RoomTile.LASTSTEP);
		}
		else if(this.is(EItemFlags.CAN_LAY_ON_TOP))
		{
			tile.topItem = this;
			tile.state = (RoomTile.CLOSED | RoomTile.LASTSTEP);
			if(this.rotation == 0 || this.rotation == 4)
			{
				tile.redirect = map.getTile(tile.x, this.y);
			}
			else
			{
				tile.redirect = map.getTile(this.x, tile.y);
			}
		}
		else
		{
			// No last step item here?
			if(!FlagsUtil.isSet(tile.state, RoomTile.LASTSTEP))
			{
				// Close the tile
				tile.topItem = this;
				tile.height = (this.z + this.definition.height);
				tile.state = RoomTile.CLOSED;
			}
		}
	}

	public void onAvatarStepOn(RoomAvatar avatar)
	{
		// Default item interaction types
		if(this.is(EItemFlags.CAN_SIT_ON_TOP))
		{
			avatar.headRotation = avatar.bodyRotation = this.rotation;
			avatar.addStatus("sit", Float.toString(avatar instanceof RoomUser ? this.definition.height : this.definition.height / 2));
		}
		else if(this.is(EItemFlags.CAN_LAY_ON_TOP))
		{
			avatar.headRotation = avatar.bodyRotation = this.rotation;
			avatar.addStatus("lay", Float.toString(this.definition.height) + " null");
		}
		
		// Raise triggers
		if(avatar instanceof RoomUser)
		{
			this.getRoom().executeTriggers(this.id, (RoomUser)avatar, WiredUserStepsOnItemTrigger.class);
		}
	}

	public void onAvatarStepOff(RoomAvatar avatar)
	{
		// Raise triggers
		if(avatar instanceof RoomUser)
		{
			this.getRoom().executeTriggers(this.id, (RoomUser)avatar, WiredUserStepsOffItemTrigger.class);
		}
	}

	public final void toggle(RoomUser user, int instruction)
	{
		// Toggle
		String previousState = this.state;
		this.onToggle(user, instruction);
		
		// State changed?
		if(this.state != previousState)
		{
			// Raise triggers?
			if(user != null && this.getRoom() != null)
			{
				this.getRoom().executeTriggers(this.id, user, WiredUserTogglesItemTrigger.class);
			}
		}
	}
	
	public boolean onToggle(RoomUser user, int instruction)
	{
		// Override this
		return false;
	}

	@Override
	public final void serialize(ServerMessage msg)
	{
		if(this.is(EItemFlags.FLOORITEM))
		{
			msg.appendInt(this.id);
			msg.appendInt(this.definition.displayId);
			msg.appendInt(this.x);
			msg.appendInt(this.y);
			msg.appendInt(this.rotation);
			msg.appendStr(this.z);
			msg.appendInt(0);
			msg.appendStr(this.state);
			msg.appendInt(-1);
			msg.appendBool(false);
		}
		else if(this.is(EItemFlags.WALLITEM))
		{
			msg.appendStr(this.id);
			msg.appendInt(this.definition.displayId);
			msg.appendStr(this.wallPosition);
			msg.appendStr(this.state);
			msg.appendBool(false);
		}
	}

	public final static Item parse(DatabaseResult row)
	{
		// Find definition
		ItemDefinition def = ItemManager.getInstance().getDefinition(row.getInt("definition_id"));
		if(def != null)
		{
			// Create instance
			Item item = def.newInstance();
			item.id = row.getInt("id");
			item.ownerId = row.getInt("owner_id");
			item.roomId = row.getInt("room_id");
			item.x = row.getInt("x");
			item.y = row.getInt("y");
			item.z = row.getFloat("z");
			item.rotation = row.getByte("rotation");
			item.wallPosition = row.getString("wallposition");
			item.state = row.getString("state");
			item.data = row.getString("data");
			item.onParse();
			return item;
		}

		// Bad item
		return null;
	}
}
