package net.nillus.habbo.game.rooms.instances;

import java.util.ArrayList;
import java.util.List;

import net.nillus.habbo.game.items.Item;
import net.nillus.habbo.game.rooms.RoomModel;
import net.nillus.habbo.util.FlagsUtil;

public final class RoomMap
{
	private RoomModel model;
	private RoomTile[][] tiles;

	public RoomMap(RoomModel model)
	{
		this.model = model;
		this.tiles = new RoomTile[model.getMaxX()][model.getMaxY()];
	}
	
	public RoomTile getTile(int x, int y)
	{
		return this.tileExists(x, y) ? this.tiles[x][y] : null;
	}
	
	public boolean tileExists(int x, int y)
	{
		return(x >= 0 && x < this.tiles.length && y >= 0 && y < this.tiles[0].length);
	}

	public boolean tileIsPassable(int x, int y, byte allowedStates)
	{
		return this.tileIsPassable(x, y, allowedStates, false);
	}
	
	public final boolean tileIsPassable(int x, int y, byte allowedStates, boolean lastStep)
	{
		// Get tile
		RoomTile tile = this.getTile(x, y);
		if(tile == null)
		{
			return false;
		}

		// Avatar here?
		if(tile.hasAvatar)
		{
			return false;
		}

		// Normal tile?
		if(!tile.is(RoomTile.LASTSTEP))
		{
			// Not allowed by options?
			if(!FlagsUtil.isSet(allowedStates, tile.state))
			{
				return false;
			}
		}
		else
		{
			// Tile is last step, but move is not last step?
			if(!lastStep)
			{
				return false;
			}
		}
		
		
		// Blocked?
		if(tile.state == RoomTile.CLOSED)
		{
			return false;
		}

		// Closed item tiles are only passable with last step of path
		if(!lastStep && (tile.state == (RoomTile.CLOSED | RoomTile.LASTSTEP)))
		{
			return false;
		}

		// Passable!
		return true;
	}
	
	public List<RoomTile> getAffectedTiles(Item obj)
	{
		return this.getAffectedTiles(obj.x, obj.y, obj.definition.length, obj.definition.width, obj.rotation);
	}
	
	public List<RoomTile> getAffectedTiles(int x, int y, byte length, byte width, byte rotation)
	{
		List<RoomTile> tiles = new ArrayList<RoomTile>();

		// Account rotation?
		if(length != width)
		{
			if(rotation == 0 || rotation == 4)
			{
				byte swap = length;
				length = width;
				width = swap;
			}
		}
		
		// Determine tiles (within map) and add them
		for(int iX = x; iX < x + width; iX++)
		{
			for(int iY = y; iY < y + length; iY++)
			{
				RoomTile tile = this.getTile(iX, iY);
				tiles.add(tile != null ? tile : RoomTile.OUT_OF_RANGE);
			}
		}

		return tiles;
	}
	
	public void reset()
	{
		// Get default floormap from room model
		char[][] modelMap = this.model.getMap();
		
		// Create default floor map
		for(int y = 0; y < this.model.getMaxY(); y++)
		{
			for(int x = 0; x < this.model.getMaxX(); x++)
			{
				// Create tile if it doesn't exist yet
				RoomTile tile = this.tiles[x][y];
				if(tile == null)
				{
					tile = this.tiles[x][y] = new RoomTile(x, y);
				}
				
				// Clear tile and set default state
				tile.clear(modelMap[x][y]);
			}
		}
	}
	
	public void applyFloorItem(Item item)
	{
		// Process all tiles this item currently resides on
		for(RoomTile tile : this.getAffectedTiles(item))
		{
			// Increase heightmap
			if (item.z > tile.height)
			{
				tile.height = item.z;
			}
			
			// Top item on the tile? Tile state is determined by top item on the tile!
			if(item.z >= tile.height)
			{
				item.onGenerateFloorMapTile(this, tile);
			}
		}
	}
	
	public RoomModel getModel()
	{
		return this.model;
	}
	
	public RoomTile getDoor()
	{
		return this.getTile(this.model.doorX, this.model.doorY);
	}
}
