package habbod.communication.messages.incoming.room.engine;

import habbod.communication.ClientMessage;
import habbod.communication.EServerMessage;
import habbod.communication.EServerMessageAlternate;
import habbod.communication.RequestHandler;
import habbod.communication.ServerMessage;
import habbod.communication.messages.outgoing.room.engine.RoomEventMessageComposer;
import habbod.events.Event;
import habbod.game.rooms.Room;
import habbod.game.rooms.instances.RoomAvatar;
import habbod.game.rooms.instances.RoomInstance;
import habbod.game.rooms.instances.RoomUser;
import habbod.game.rooms.items.RoomFloorItem;
import habbod.game.rooms.items.RoomWallItem;
import habbod.sessions.Session;

import java.util.Collection;
import java.util.Collections;
import java.util.List;


public class GetRoomContentDataHandler implements RequestHandler
{
	public final static Event onSendRoomContent = new Event("OnSendRoomData");
	public final static Event onSendUserInfo = new Event("OnSendUserInfo");

	@Override
	public void handle(Session session, ClientMessage request, ServerMessage response)
	{
		// Shorthand variables
		RoomInstance instance = session.getRoomInstance();
		if(instance == null)
		{
			return;
		}
		Room room = instance.getInfo();
		
		// Send remaining chat timeout seconds
		if(session.chatTimeoutExpireTime > 0)
		{
			int remainingSeconds = (int) ((session.chatTimeoutExpireTime - System.currentTimeMillis()) / 1000);
			if(remainingSeconds > 0)
			{
				response.init(EServerMessageAlternate.ROOM_MUTED);
				response.appendInt(remainingSeconds);
				session.send(response);
			}
		}
		
		// If entering a guestroom, then send more data, client doesn't ask for it
		if(room.isGuestRoom())
		{
			// Send wallpaper/floor/landscape if not default data
			if(!room.wallpaper.isEmpty())
			{
				response.init(EServerMessage.RoomProperty);
				response.appendStr("wallpaper");
				response.appendStr(room.wallpaper);
				session.send(response);
			}
			if(!room.floor.isEmpty())
			{
				response.init(EServerMessage.RoomProperty);
				response.appendStr("floor");
				response.appendStr(room.floor);
				session.send(response);
			}
			if(!room.landscape.isEmpty())
			{
				response.init(EServerMessage.RoomProperty);
				response.appendStr("landscape");
				response.appendStr(room.landscape);
				session.send(response);
			}
		}
		
		// Passive objects
		List<RoomFloorItem> passiveObjects = Collections.emptyList();
		response.init(30);
		response.appendInt(passiveObjects.size());
		for(RoomFloorItem item : passiveObjects)
		{
		//	response.appendObject(item);
		}
		session.send(response);

		// Public rooms have a different 'interface'
		if(room.isPublicRoom())
		{
			response.init(EServerMessage.RoomEntryInfo);
			response.appendBool(false);
			response.appendStr(room.modelName);
			response.appendBool(false);
			session.send(response);
		}
		else
		{
			// Hide walls
			response.init(472);
			response.appendBool(room.hideWalls);
			session.send(response);

			// Send floor item list
			Collection<RoomFloorItem> floorItems = instance.getFloorItems();
			response.init(EServerMessage.FloorItems);
			response.appendInt(floorItems.size());
			for(RoomFloorItem item : floorItems)
			{
				response.appendObject(item);
			}
			session.send(response);

			// Send floor item list
			Collection<RoomWallItem> wallItems = instance.getWallItems();
			response.init(EServerMessage.FloorItems);
			response.appendInt(wallItems.size());
			for(RoomWallItem item : wallItems)
			{
				response.appendObject(item);
			}
			session.send(response);

			// Interface info (edit room etc)
			response.init(EServerMessage.RoomEntryInfo);
			response.appendBool(true);
			response.appendInt(room.id);
			response.appendBool(room.ownerId == session.getUser().id);
			session.send(response);
		}
		
		// Send room event
		session.send(RoomEventMessageComposer.compose(instance.getEvent()));
		
		// Entering guestroom?
		if(room.isGuestRoom())
		{
			// Get 'already rated' list
			List<Integer> rated = session.get("ratedRooms");
			if(rated != null)
			{
				ServerMessage msg = new ServerMessage(EServerMessageAlternate.ROOM_RATING_INFO);
				if(rated.contains(room.id) || session.getUser().id == room.ownerId)
				{
					// Current rating
					msg.appendInt(room.rating);
				}
				else
				{
					// Still rateable!
					msg.appendInt(-1);
				}
				session.send(msg);
			}
		}
		
		// Other room content
		GetRoomContentDataHandler.onSendRoomContent.fire(session);

		// Send avatars
		Collection<RoomAvatar> avatars = instance.getAllAvatars();
		response.init(EServerMessage.Users);
		response.appendInt(avatars.size());
		for(RoomAvatar avatar : avatars)
		{
			response.appendObject(avatar);
		}
		session.send(response);

		// Try spawning room user
		if(session.enterRoom())
		{
			session.sendRoomPrivileges();
		}
		else
		{
			return;
		}

		// Send statuses of all avatars
		response.init(EServerMessage.UserUpdate);
		response.appendInt(avatars.size());
		for(RoomAvatar avatar : avatars)
		{
			avatar.appendStatusString(response);
		}
		session.send(response);

		// Send other special user statuses
		for(RoomUser user : instance.getAllUsers())
		{
			if(user.danceId > 0)
			{
				response.init(EServerMessage.Dance);
				response.appendInt(user.avatarId);
				response.appendInt(user.danceId);
				session.send(response);
			}
			if(user.carryItemId > 0 && user.carryItemTimer > 0)
			{
				response.init(EServerMessage.CarryObject);
				response.appendInt(user.avatarId);
				response.appendInt(user.carryItemTimer);
				session.send(response);
			}
			if(user.getCurrentEffect() > 0)
			{
				response.init(475);
				response.appendInt(user.avatarId);
				response.appendInt(user.getCurrentEffect());
				response.appendBool(false);
				session.send(response);
			}
			GetRoomContentDataHandler.onSendUserInfo.fire(session, user);
		}
	}
}