package net.nillus.habbo.game.moderation;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import net.nillus.habbo.habbod;
import net.nillus.habbo.communication.EServerMessage;
import net.nillus.habbo.communication.ServerMessage;
import net.nillus.habbo.communication.messages.outgoing.moderation.ModerationTicketMessageComposer;
import net.nillus.habbo.core.Logger;
import net.nillus.habbo.game.access.AccessControl;
import net.nillus.habbo.game.rooms.RoomManager;
import net.nillus.habbo.game.rooms.instances.RoomInstance;
import net.nillus.habbo.game.rooms.instances.RoomUser;
import net.nillus.habbo.game.users.User;
import net.nillus.habbo.game.users.UserManager;
import net.nillus.habbo.sessions.Session;
import net.nillus.habbo.sessions.SessionManager;
import net.nillus.habbo.storage.DatabaseResult;
import net.nillus.habbo.storage.SQLHelper;

public final class ModerationCenter
{
	private ConcurrentHashMap<Integer, ModerationTicket> tickets;
	private ModerationPresetManager presetManager;
	
	private ModerationCenter()
	{
		this.tickets = new ConcurrentHashMap<Integer, ModerationTicket>();
		this.presetManager = new ModerationPresetManager();
	}

	public void loadPendingTickets()
	{
		this.tickets.clear();
		
		// Load all open/picked tickets from database
		DatabaseResult result = habbod.getInstance().getDb().query("SELECT * FROM moderation_tickets WHERE status=" + ModerationTicket.STATUS_OPEN + " OR status=" + ModerationTicket.STATUS_PICKED);
		while(result.row())
		{
			// Process from row
			ModerationTicket ticket = new ModerationTicket();
			ticket.id = result.getInt("id");
			ticket.priority = result.getInt("priority");
			ticket.type = result.getInt("type");
			ticket.userId = result.getInt("user_id");
			ticket.userUsername = UserManager.getInstance().getUsername(ticket.userId);
			ticket.reportedUserId = result.getInt("reported_id");
			if(ticket.reportedUserId > 0)
			{
				ticket.reportedUsername = UserManager.getInstance().getUsername(ticket.userId);
			}
			ticket.pickerId = result.getInt("picker_id");
			if(ticket.pickerId > 0)
			{
				ticket.pickerUsername = UserManager.getInstance().getUsername(ticket.pickerId);
			}
			ticket.roomId = result.getInt("room_id");
			ticket.roomName = result.getString("room_name");
			ticket.message = result.getString("message");
			ticket.timestamp = result.getTime("timestamp");
			
			// Add to map
			this.tickets.put(ticket.id, ticket);
		}
		result.release();
		
		// Yay!
		Logger.getInstance().info("ModerationCenter", String.format("Loaded %d pending CFHs from database.", this.tickets.size()));
	}

	public ModerationTicket getTicket(int ticketId)
	{
		return this.tickets.get(ticketId);
	}
	
	public ModerationTicket getUnpickedTicketForUser(int userId)
	{
		for(ModerationTicket ticket : this.tickets.values())
		{
			if(ticket.userId == userId && ticket.status == ModerationTicket.STATUS_OPEN)
			{
				return ticket;
			}
		}
		return null;
	}

	public Collection<ModerationTicket> getAllTickets()
	{
		return this.tickets.values();
	}

	public void submitTicket(Session session, int type, int reportedUserId, String message)
	{
		// Create ticket
		ModerationTicket ticket = new ModerationTicket();
		ticket.priority = 1;
		ticket.type = type;
		ticket.status = ModerationTicket.STATUS_OPEN;
		ticket.userId = session.getUser().id;
		ticket.userUsername = session.getUser().username;
		if(reportedUserId > 0)
		{
			ticket.reportedUserId = reportedUserId;
			ticket.reportedUsername = UserManager.getInstance().getUsername(reportedUserId);
		}
		ticket.message = message;
		ticket.roomId = session.getRoomInstance().getInfo().id;
		ticket.roomName = session.getRoomInstance().getInfo().name;
		ticket.timestamp = System.currentTimeMillis();
		ticket.id = habbod.getInstance().getDb().executeReturnKey("INSERT INTO moderation_tickets(priority,type,status,user_id,reported_id,room_id,room_name,message,timestamp) VALUES (" + ticket.priority + "," + ticket.type + "," + ticket.status + "," + ticket.userId + "," + ticket.reportedUserId + "," + ticket.roomId + ",'" + SQLHelper.escape(ticket.roomName) + "','" + SQLHelper.escape(ticket.message) + "','" + SQLHelper.formatDate(ticket.timestamp) + "')");
		
		// Add to map
		this.tickets.put(ticket.id, ticket);
		
		// Notify helpers
		this.broadcastTicket(ticket);
	}
	
	public void broadcastTicket(ModerationTicket ticket)
	{
		for(Session session : SessionManager.getInstance().getAllSessions())
		{
			if(session.authenticated() && session.getUser().rights.contains("fuse_mod"))
			{
				session.send(ModerationTicketMessageComposer.compose(ticket));
			}
		}
	}
	
	public UserModerationHistory getUserModerationHistory(int userId)
	{
		UserModerationHistory history = new UserModerationHistory();
		history.bans = habbod.getInstance().getDb().queryInt("SELECT COUNT(id) FROM moderation_bans WHERE user_id=" + userId);
		history.cautions = habbod.getInstance().getDb().queryInt("SELECT COUNT(id) FROM system_actionlog WHERE target_id=" + userId + " AND (action='caution' OR (action='kick' AND message != ''))");
		history.cfhs = habbod.getInstance().getDb().queryInt("SELECT COUNT(id) FROM moderation_tickets WHERE user_id=" + userId);
		history.abusiveCfhs = habbod.getInstance().getDb().queryInt("SELECT COUNT(id) FROM moderation_tickets WHERE user_id=" + userId + " AND status=" + ModerationTicket.STATUS_ABUSIVE);
	
		return history;
	}

	public void issueUserAlert(int issuerId, String username, String message, boolean isCaution) throws ModerationActionError
	{
		Session sess = SessionManager.getInstance().getSessionByUsername(username);
		this.issueUserAlert(issuerId, sess, message, isCaution);
	}

	public void issueUserAlert(int issuerId, int userId, String message, boolean isCaution) throws ModerationActionError
	{
		Session sess = SessionManager.getInstance().getSessionByUserId(userId);
		this.issueUserAlert(issuerId, sess, message, isCaution);
	}

	private void issueUserAlert(int issuerId, Session session, String message, boolean isCaution) throws ModerationActionError
	{
		if(session == null)
		{
			throw new ModerationActionError("user not online");
		}
		else if(message.isEmpty())
		{
			throw new ModerationActionError("message cannot be blank");
		}
		else
		{
			session.sendNotif(message, isCaution);
			AccessControl.getInstance().logUserAction(issuerId, session.getUser().id, isCaution ? "caution" : "alert", message);
		}
	}

	public void issueUserKick(int issuerId, String username, String message) throws ModerationActionError
	{
		Session sess = SessionManager.getInstance().getSessionByUsername(username);
		this.issueUserKick(issuerId, sess, message);
	}

	public void issueUserKick(int issuerId, int userId, String message) throws ModerationActionError
	{
		Session sess = SessionManager.getInstance().getSessionByUserId(userId);
		this.issueUserKick(issuerId, sess, message);
	}

	private void issueUserKick(int issuerId, Session session, String message) throws ModerationActionError
	{
		User issuer = UserManager.getInstance().getUserById(issuerId);
		if(session == null || !session.inRoom())
		{
			throw new ModerationActionError("user not online / not in room");
		}
		else if(issuer == null || issuer.role.level <= session.getUser().role.level)
		{
			throw new ModerationActionError("insufficient permissions");
		}
		else
		{
			if(message != null && message.isEmpty())
			{
				message = null;
			}
			session.kickFromRoom(message, true);
			AccessControl.getInstance().logUserAction(issuerId, session.getUser().id, "kick", message);
		}
	}

	public void issueRoomAlert(int issuerId, int roomId, String message, boolean isCaution) throws ModerationActionError
	{
		RoomInstance room = RoomManager.getInstance().getRoomInstance(roomId, false);
		if(room == null)
		{
			throw new ModerationActionError("room not loaded");
		}
		else if(message.isEmpty())
		{
			throw new ModerationActionError("message cannot be blank");
		}
		else
		{
			ServerMessage msg = new ServerMessage(isCaution ? EServerMessage.Caution : EServerMessage.Notif);
			msg.appendStr(message);
			room.broadcast(msg);

			AccessControl.getInstance().logUserAction(issuerId, roomId, "roomalert", message);
		}
	}

	public void issueRoomKick(int issuerId, int roomId, String message) throws ModerationActionError
	{
		User issuer = UserManager.getInstance().getUserById(issuerId);
		RoomInstance room = RoomManager.getInstance().getRoomInstance(roomId, false);
		if(issuer == null || room == null)
		{
			throw new ModerationActionError("room not loaded");
		}
		else
		{
			// Fix message
			if(message != null && message.isEmpty())
			{
				message = null;
			}

			// Collect list with users to kick
			List<RoomUser> toKick = new ArrayList<RoomUser>();
			for(RoomUser user : room.getAllUsers())
			{
				if(user.getInfo().role.level < issuer.role.level)
				{
					toKick.add(user);
				}
			}

			// Kick them
			for(RoomUser user : toKick)
			{
				user.getSession().kickFromRoom(message, true);
			}

			AccessControl.getInstance().logUserAction(issuerId, roomId, "roomkick", message);
		}
	}

	public void issueUserBan(int issuerId, int userId, String message, int hours) throws ModerationActionError
	{
		// Get issuer and target
		User issuer = UserManager.getInstance().getUserById(issuerId);
		User target = UserManager.getInstance().getUserById(userId);
		
		// Perform checks
		if(issuer == null || target == null)
		{
			throw new ModerationActionError("user does not exist");
		}
		else if(target.role.level >= issuer.role.level)
		{
			throw new ModerationActionError("insufficient permissions");
		}
		else if(message.isEmpty())
		{
			throw new ModerationActionError("message cannot be blank");
		}
		else if(hours <= 0)
		{
			throw new ModerationActionError("ban duration must be larger than 0 hours");
		}
		
		// Get calendar
		Calendar cal = Calendar.getInstance();
		
		// Create ban
		ModerationBan ban = new ModerationBan();
		ban.userId = userId;
		ban.message = message;
		ban.issuedBy = issuerId;
		ban.issuedAt = cal.getTimeInMillis();
		cal.add(Calendar.HOUR, hours);
		ban.expiresAt = cal.getTimeInMillis();
		ban.id = habbod.getInstance().getDb().executeReturnKey("INSERT INTO moderation_bans(user_id,message,expires_at,issued_at,issued_by) VALUES (" + ban.userId + ",'" + SQLHelper.escape(ban.message) + "','" + SQLHelper.formatDate(ban.expiresAt) + "','" + SQLHelper.formatDate(ban.issuedAt) + "'," + ban.issuedBy + ")");
		
		// Target online?
		Session session = SessionManager.getInstance().getSessionByUserId(ban.userId);
		if(session != null)
		{
			session.send(ModerationBan.createMsg(ban));
			SessionManager.getInstance().endSession(session.id, "user has been banned");
		}
	}
	
	public List<ModerationBan> getActiveBansForUserId(int userId)
	{
		List<ModerationBan> bans = new ArrayList<ModerationBan>(0);
		DatabaseResult result = habbod.getInstance().getDb().query("SELECT * FROM moderation_bans WHERE user_id=" + userId);
		while(result.row())
		{
			ModerationBan ban = ModerationBan.parse(result);
			if(!ban.hasExpired())
			{
				bans.add(ban);
			}
		}
		result.release();
		return bans;
	}

	/*
	public void issueIpBan(int issuerId, String ip, String message, int hours)
	{
		
	}

	public List<ModerationBan> getActiveBansForIp(String ip)
	{
		List<ModerationBan> bans = new ArrayList<ModerationBan>(0);
		DatabaseResult result = habbod.getInstance().getDb().query("SELECT * FROM moderation_bans WHERE ip='" + SQLHelper.escape(ip) + "'");
		while(result.row())
		{
			ModerationBan ban = ModerationBan.parse(result);
			if(!ban.hasExpired())
			{
				bans.add(ban);
			}
		}
		return bans;
	}*/
	
	public ModerationPresetManager getPresetManager()
	{
		return this.presetManager;
	}

	private final static ModerationCenter instance = new ModerationCenter();

	public final static ModerationCenter getInstance()
	{
		return instance;
	}
}
