package net.nillus.habbo;

import net.nillus.habbo.core.Logger;
import net.nillus.habbo.core.PropertiesBox;
import net.nillus.habbo.core.ShutdownHook;
import net.nillus.habbo.core.Logger.ELogLevel;
import net.nillus.habbo.game.access.AccessControl;
import net.nillus.habbo.game.catalog.Catalog;
import net.nillus.habbo.game.items.ItemManager;
import net.nillus.habbo.game.rooms.RoomManager;
import net.nillus.habbo.game.rooms.navigator.RoomNavigator;
import net.nillus.habbo.game.users.achievements.AchievementManager;
import net.nillus.habbo.modules.ModuleManager;
import net.nillus.habbo.net.NetServer;
import net.nillus.habbo.remote.RemoteServer;
import net.nillus.habbo.sessions.SessionManager;
import net.nillus.habbo.storage.SQLDatabase;

/**
 * A new Habbo Hotel (www.habbo.com) game server emulator written by Nils 'Nillus' W.
 */
public final class habbod
{
	private boolean alive = false;
	private long startTime = 0;
	private String propsFile = null;
	private PropertiesBox props = null;
	private SQLDatabase db = null;
	private ModuleManager modMgr = null;
	
	public habbod(String propsFile)
	{
		this.alive = false;
		this.startTime = 0;
		this.propsFile = propsFile;
		this.props = new PropertiesBox();
		this.db = null;
		this.modMgr = new ModuleManager();

		System.out.println("1111111111111111111111111111111111111111111");
		System.out.println("1100000000000000000000000000000000000000011");
		System.out.println("1101111           habbod           1111011");
		System.out.println("1101111111111111111111111111111111111111011");
		System.out.println("1101111  COPYRIGHT (C) 2011 Nils W. 1111011");
		System.out.println("1101111111111111111111111111111111111111011");
		System.out.println("1101111      Parts written by:      1111011");
		System.out.println("1101111     Joe 'Joeh' Hegarty      1111011");
		System.out.println("1101111     Roy 'Meth0d' de Jong    1111011");
		System.out.println("1100000000000000000000000000000000000000011");
		System.out.println("1111111111111111111111111111111111111111111");
		System.out.println();
	}

	public boolean reloadProperties()
	{
		boolean succes = this.props.load(this.propsFile);
		if(succes)
		{
			Logger.getInstance().info("habbod", String.format("Loaded %d properties.", this.props.size()));
		}
		else
		{
			Logger.getInstance().warning("habbod", "Failed loading system properties.");
		}
		return succes;
	}

	public boolean start()
	{
		// Set log importancy
		Logger.getInstance().setMinImportance(ELogLevel.Info);
		
		// Log startup time
		long beginTime = System.currentTimeMillis();
		Logger.getInstance().info("habbod", "Starting...");

		// Load system properties
		if(!this.reloadProperties())
		{
			return false;
		}

		// Prepare SQLDatabase
		this.db = new SQLDatabase(this.props);
		if(this.db.test())
		{
			Logger.getInstance().info("SQLDatabase", "Connection successful.");
			if(!this.db.getQueryQueue().open())
			{
				return false;
			}
			else
			{
				this.db.getConnMgr().startMonitor();
			}
		}
		else
		{
			Logger.getInstance().warning("SQLDatabase", "Connection failed.");
			return false;
		}

		// Load modules/plugins
		this.modMgr.reloadModules();

		// Load accesscontrol
		AccessControl.getInstance().reloadUserRoles();
		
		// Load achievement types
		AchievementManager.getInstance().reload();

		// Load item definitions
		ItemManager.getInstance().reloadDefinitions();

		// Create RoomManager, spawn initial workers and reload model data
		int numRoomWorkers = this.props.getInt("rooms.workers", Runtime.getRuntime().availableProcessors());
		RoomManager.createInstance(numRoomWorkers);
		RoomManager.getInstance().reloadModels();

		// Load Navigator, cache stuff, etc
		int navigatorRefreshTime = this.props.getInt("rooms.navigator.refreshtime", 5);
		RoomNavigator.createInstance(navigatorRefreshTime);
		RoomNavigator.getInstance().reload();

		// Load Catalog
		Catalog.getInstance().reload();

		// Setup network manager for game clients
		int portNumber = this.props.getInt("net.portnumber", 1337);
		NetServer.createInstance();
		if(!NetServer.getInstance().start(portNumber))
		{
			return false;
		}
		
		// Setup remoteserver for RCON'ish issueing of commands
		int remotePortNumber = habbod.getInstance().getProperties().getInt("remote.portnumber", 1338);
		RemoteServer.createInstance(remotePortNumber);
		RemoteServer.getInstance().addAllowedIp("0:0:0:0:0:0:0:1");
		RemoteServer.getInstance().addAllowedIp("0:0:0:0:0:0:0:1%0");
		RemoteServer.getInstance().addAllowedIp("127.0.0.1");
		if(!RemoteServer.getInstance().start())
		{
			return false;
		}

		// Start module pulsing
		this.modMgr.start();

		// Attach shutdown hook
		Runtime.getRuntime().addShutdownHook(new ShutdownHook(this));

		// Startup okay!
		this.alive = true;
		this.startTime = System.currentTimeMillis();
		Logger.getInstance().info("habbod", String.format("Started in %d ms.", (this.startTime - beginTime)));
		return true;
	}

	public void stop(String reason)
	{
		// Only start if not already shutdown'd yet
		if(!this.alive)
		{
			return;
		}
		Logger.getInstance().info("habbod", String.format("Entering shutdown sequence... (reason: %s)", reason));

		// Stop the remote listener
		RemoteServer.getInstance().stop();
		
		// Stop the game listener
		NetServer.getInstance().stop();

		// Shutdown game subsystems
		SessionManager.getInstance().shutdown();
		RoomManager.getInstance().shutdown();
		
		// Close any leftover connections
		NetServer.getInstance().close();

		// Unload modules/plugins
		this.modMgr.unloadModules();
		this.modMgr.stop();

		// Close SQLDatabase
		this.db.getQueryQueue().close();
		this.db.getConnMgr().closeAll();
		this.db.getConnMgr().stopMonitor();

		// Sequence complete!
		this.alive = false;
		Logger.getInstance().info("habbod", "Shutdown sequence completed.");
	}

	public String versionString()
	{
		return "habbod v0.9.2-dev 'Palladium'";
	}

	public boolean alive()
	{
		return this.alive;
	}

	public long getStartTime()
	{
		return this.startTime;
	}

	public PropertiesBox getProperties()
	{
		return this.props;
	}

	public SQLDatabase getDb()
	{
		return this.db;
	}

	public ModuleManager getModMgr()
	{
		return this.modMgr;
	}
	
	/* Singleton */
	private static habbod instance;

	public final static habbod getInstance()
	{
		return instance;
	}

	public static habbod createInstance(String propsFile)
	{
		habbod.instance = new habbod(propsFile);
		Logger.getInstance().info("habbod", habbod.instance.versionString());
		return habbod.instance;
	}
}
