package habbod2;

import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import habbod2.core.Locale;
import habbod2.core.Logger;
import habbod2.core.ShutdownHook;
import habbod2.core.StatsMonitor;
import habbod2.game.GameManager;
import habbod2.net.NetServer;
import habbod2.rcon.RconServer;
import habbod2.storage.Database;
import habbod2.storage.DatabaseEndpoint;
import habbod2.util.NamedThreadFactory;
import habbod2.util.PropertiesBox;
import habbod2.util.ScheduledTask;
import habbod2.util.ScriptHandler;
import habbod2.communication.RequestRegistry;

public class System
{
	private long startTime;
	private Logger logger;
	private PropertiesBox properties;
	private Database database;
	private Locale locale;
	private RconServer rcon;
	private NetServer server;
	private GameManager gameMgr;
	private ScheduledExecutorService executor;
	private boolean isShutdown;
	
	private System()
	{
		this.properties = new PropertiesBox();
		this.logger = new Logger();
		this.locale = new Locale(this);
		this.rcon = new RconServer(this);
		this.server = new NetServer(this);
		this.gameMgr = new GameManager(this);
		this.executor = Executors.newScheduledThreadPool(1, new NamedThreadFactory("SystemScheduler #%d"));
	}
	
	public boolean reloadProperties()
	{
		return this.loadProperties(null);
	}
	
	public boolean loadProperties(String file)
	{
		boolean success = (file == null) ? this.properties.reload() : this.properties.load(file);
		if(success)
		{
			this.logger.info(this, String.format("Loaded %d configuration properties.", this.properties.size()));
		}
		else
		{
			this.logger.info(this, String.format("Could not load %s! Please validate the habbod2 installation.", file));
		}
		return success;
	}
	
	public boolean start(String propsFile)
	{
		// Record boot start time
		long bootBegin = this.getTime();
		
		// Set output file for logger
		this.logger.setDatedOutputFile();
		
		// Print version
		this.logger.info(this, String.format("Booting %s...", this.getVersionString()));
		
		// Load properties
		if(!this.loadProperties(propsFile))
		{
			return false;
		}

		// Enable debug and stacktraces?
		if(this.properties.getBool("sys.log.debug"))
		{
			this.logger.setMinImportance(Logger.ELogLevel.Debug);
		}
		this.logger.setDumpStackTraces(this.properties.getBool("sys.log.stacktraces"));
		
		// Check support for LUA scripting
		if(ScriptHandler.hasNativeLibrary)
		{
			this.logger.info("System", "Lua scripting support available. (native)");
		}
		else
		{
			this.logger.warning("System", "No Lua scripting support available, because the native library 'luajava' is not present in java.library.path.");
		}
		
		// Setup database
		PropertiesBox p = this.properties;
		this.database = new Database(this, new DatabaseEndpoint(p.get("db.user"), p.get("db.password"), p.get("db.url"), p.get("db.driver"), p.getInt("db.pool.maxsize")));
		if(!this.database.test())
		{
			return false;
		}
		
		// Setup pool monitor
		this.database.getPool().startMonitor();
		
		// Queue UPDATE/INSERT/DELETEs for performance gain?
		if(this.properties.getBool("db.updates.queue.enabled"))
		{
			this.database.getUpdateQueue().open();
		}
		
		// Reload locale
		this.locale.setLocale(this.properties.get("sys.locale", "en-US"));
		
		// Reload + start game manager
		this.gameMgr.reload();
		this.gameMgr.start();
		
		// Reload request registry
		RequestRegistry.reload();
		
		// Enable RCON?
		if(this.properties.getBool("rcon.enabled"))
		{
			// Load commands
			this.rcon.reload();
			
			// Allow local commands?
			if(this.properties.getBool("rcon.allowlocal"))
			{
				this.rcon.allowLocalCommands();
			}
			
			// Attempt to start server
			if(!this.rcon.start(this.properties.getInt("rcon.port")))
			{
				return false;
			}
		}
		
		// Enable gameserver?
		if(this.properties.getBool("net.enabled"))
		{
			// Reload blacklist
			this.server.reloadBlacklist();
			
			// Attempt to start server
			if(!this.server.start(this.properties.getInt("net.port")))
			{
				return false;
			}
		}
		
		// Start stats monitor
		int statsIntervalSeconds = this.properties.getInt("sys.stats.update.seconds", 30);
		this.scheduleTask(new StatsMonitor(this), statsIntervalSeconds * 1000);
		
		// Attach shutdown hook
		this.getVM().addShutdownHook(new ShutdownHook(this));
		
		// Record start time
		this.startTime = this.getTime();
		this.logger.info(this, String.format("Booted in %d ms, ready for connections.", this.startTime - bootBegin));
		
		// OK!
		return true;
	}

	public void stop(String message)
	{
		// Already shutdown?
		if(this.isShutdown)
		{
			return;
		}
		
		// Entering shutdown sequence
		this.isShutdown = true;
		this.logger.info(this, String.format("Shutting down. (reason: %s)", message));
		
		// Stop any running servers first, no new connections!
		this.rcon.stop();
		this.server.stop();
		
		// Shutdown game systems
		this.gameMgr.shutdown();
		
		// Forcefully close any open connections and release resources
		this.server.close();
		
		// Shutdown executor
		this.executor.shutdownNow();
		
		// Shutdown database
		this.database.getPool().stopMonitor();
		this.database.getUpdateQueue().close();
		
		// Shutdown completed
		this.logger.info(this, String.format("Shutdown completed."));
	}
	
	public void executeTask(Runnable task)
	{
		this.executor.execute(task);
	}
	
	public void executeTask(Runnable task, long delayMillis)
	{
		this.executor.schedule(task, delayMillis, TimeUnit.MILLISECONDS);
	}
	
	public ScheduledTask scheduleTask(Runnable task, long intervalMillis)
	{
		return new ScheduledTask(this.executor.scheduleAtFixedRate(task, 0, intervalMillis, TimeUnit.MILLISECONDS));
	}
	
	public final long getTime()
	{
		return java.lang.System.currentTimeMillis();
	}

	public static long currentTimeMillis()
	{
		return java.lang.System.currentTimeMillis();
	}
	
	public String getVersionString()
	{
		return "habbod2 v2.0.1-dev 'Rocketman'";
	}
	
	public long getMemoryUsage()
	{
		Runtime vm = this.getVM();
		return (vm.totalMemory() - vm.freeMemory()) / 1024;
	}
	
	public long getStartTime()
	{
		return this.startTime;
	}
	
	public Runtime getVM()
	{
		return Runtime.getRuntime();
	}
	
	public Logger getLogger()
	{
		return this.logger;
	}
	
	public PropertiesBox getProperties()
	{
		return this.properties;
	}
	
	public Database getDb()
	{
		return this.database;
	}
	
	public Locale getLocale()
	{
		return this.locale;
	}
	
	public RconServer getRCON()
	{
		return this.rcon;
	}
	
	public NetServer getServer()
	{
		return this.server;
	}
	
	public GameManager getGameMgr()
	{
		return this.gameMgr;
	}
	
	@Override
	public String toString()
	{
		return "System";
	}
	
	private static final System instance = new System();
	
	public static final System get()
	{
		return instance;
	}
}
