package habbod2.rcon;

import habbod2.System;
import habbod2.SystemComponent;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class RconServer extends SystemComponent implements Runnable
{
	private static final int MAX_PACKET_SIZE = 1024;
	
	private DatagramSocket socket;
	private List<String> allowedIps;
	private Map<String, RconCommandHandler> handlers;
	
	private boolean alive;
	private Thread thread;

	public RconServer(System system)
	{
		super(system);
		this.allowedIps = new ArrayList<String>();
		this.handlers = new HashMap<String, RconCommandHandler>();
	}

	public boolean start(int port)
	{
		// Not already started?
		if(this.alive)
		{
			return false;
		}
		else
		{
			// Bind socket
			try
			{
				this.socket = new DatagramSocket(port);
			}
			catch(SocketException ex)
			{
				this.getLogger().error(this, String.format("Error binding to UDP/%d.", port), ex);
				return false;
			}

			// Create thread
			this.thread = new Thread(this, this.toString());

			// Start thread
			this.alive = true;
			this.thread.start();

			// Started!
			this.getLogger().info(this, String.format("Listening for commands on UDP/%d.", port));
			return true;
		}
	}

	public void stop()
	{
		if(this.alive)
		{
			// Close socket
			this.alive = false;
			this.thread.interrupt();

			// Close the socket
			this.socket.close();
			
			// Closed!
			this.getLogger().info(this, "Closed listener.");
		}
	}

	public void allowLocalCommands()
	{
		try
		{
			this.addAllowedIp(InetAddress.getLocalHost().getHostAddress().toString());
		}
		catch (UnknownHostException e)
		{
			this.getLogger().warning(this, "Failed to allow local address, lookup failed?");
		}
	}
	
	public void addAllowedIp(String ip)
	{
		synchronized(this.allowedIps)
		{
			this.allowedIps.add(ip);
		}
		this.getLogger().info(this, String.format("Allowed %s.", ip));
	}

	public void removeAllowedIp(String ip)
	{
		synchronized(this.allowedIps)
		{
			this.allowedIps.remove(ip);
		}
		this.getLogger().info(this, String.format("Unallowed %s.", ip));
	}
	
	public void addHandler(String cmd, RconCommandHandler handler)
	{
		if(handler != null)
		{
			this.handlers.put(cmd, handler);
		}
	}
	
	public void removeHandler(String cmd)
	{
		this.handlers.remove(cmd);
	}

	@Override
	public void run()
	{
		// This packet is re-used all the time
		DatagramPacket packet = new DatagramPacket(new byte[RconServer.MAX_PACKET_SIZE], RconServer.MAX_PACKET_SIZE);

		// While the server is running...
		while(this.alive)
		{
			// Receive the next packet
			try
			{
				this.socket.receive(packet);
			}
			catch(IOException ex)
			{
				continue;
			}

			// Sender is allowed?
			String senderIp = packet.getAddress().getHostAddress();
			if(this.allowedIps.contains(senderIp))
			{
				// Get data as a string and parse "COMMAND arg1 arg2 arg3 etc"
				String data = new String(packet.getData(), 0, packet.getLength());
				String parts[] = data.split(" ", 2);

				// Dispatch command to command handler
				this.handleCommand(parts[0], (parts.length == 2) ? parts[1] : null);
			}
			else
			{
				// Lol hax
				this.getLogger().warning(this, String.format("IP '%s' is not allowed to issue remote commands. Command ignored.", senderIp));
			}
		}
	}

	private boolean handleCommand(String cmd, String payload)
	{
		// Handler registered?
		RconCommandHandler handler = this.handlers.get(cmd);
		if(handler == null)
		{
			this.getLogger().warning(this, String.format("Unknown command '%s'. Command ignored.", cmd));
			return false;
		}

		// Invoke handler and catch exceptions
		try
		{
			handler.handle(payload);
		}
		catch(Exception ex)
		{
			this.getLogger().error(this, String.format("Error processing command '%s'. (payload=%s)", cmd, payload), ex);
		}

		// Handled
		return true;
	}

	public int getPort()
	{
		if(this.socket == null)
		{
			return 0;
		}
		else
		{
			return this.socket.getLocalPort();
		}
	}
}
