package habbod2.storage;

import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.Statement;
import java.sql.SQLException;

import habbod2.System;
import habbod2.core.SystemStats;

public class Database extends System.Component
{
	private DatabaseEndpoint endpoint;
	private DatabaseConnectionPool pool;
	private DatabaseUpdateQueue updateQueue;
	
	public Database(System system, DatabaseEndpoint endpoint)
	{
		super(system);
		this.endpoint = endpoint;
		this.pool = new DatabaseConnectionPool(this);
		this.updateQueue = new DatabaseUpdateQueue(this);
	}
	
	public boolean test()
	{
		// Register connection driver package
		String driverClassName = this.getEndpoint().driver;
		try
		{
			Driver driver = (Driver) Class.forName(driverClassName).newInstance();
			DriverManager.registerDriver(driver);
		}
		catch(Exception ex)
		{
			this.getLogger().error(this, String.format("Can't find driver '%s', please copy the driver library to the local Java extensions directory (lib/ext).", driverClassName));
			return false;
		}

		// Create test connection
		DatabaseConnection test = this.pool.getConnection();
		if(test != null)
		{
			this.pool.releaseConnection(test);
			this.getLogger().info(this, String.format("Connected to %s.", this.endpoint));
			return true;
		}
		else
		{
			return false;
		}
	}
	
	public void execute(String sql)
	{
		SystemStats.executedQueries++;
		
		if(this.updateQueue.isOpen())
		{
			this.updateQueue.enqueue(sql);
		}
		else
		{
			try(DatabaseConnection conn = this.pool.getConnection())
			{
				try(Statement st = conn.getInnerConnection().createStatement())
				{
					st.executeUpdate(sql);
				}
				catch(SQLException ex)
				{
					this.getLogger().error(this, "Error during INSERT.", ex);
				}
			}
		}
	}

	public int executeReturnKey(String sql)
	{
		SystemStats.executedQueries++;
		
		try(DatabaseConnection conn = this.getPool().getConnection())
		{
			try(Statement st = conn.getInnerConnection().createStatement())
			{
				st.execute(sql, Statement.RETURN_GENERATED_KEYS);
				try(ResultSet keys = st.getGeneratedKeys())
				{
					if(keys.next())
					{
						return keys.getInt(1);
					}
				}
			}
		}
		catch(Exception ex)
		{
			this.getLogger().error(this, "Error during INSERT.", ex);
		}
		
		return -1;
	}

	public DatabaseResult query(String sql)
	{
		SystemStats.executedQueries++;
		
		DatabaseConnection connection = this.pool.getConnection();
		if(connection != null)
		{
			try
			{
				ResultSet result = connection.getInnerConnection().createStatement().executeQuery(sql);
				return new DatabaseResult(connection, result);
			}
			catch(SQLException ex)
			{
				this.getLogger().error(this, "Error during SELECT.", ex);
				this.pool.releaseConnection(connection);
			}
		}
		
		return DatabaseResult.INVALID_RESULT;
	}
	
	public int queryInt(String sql)
	{
		SystemStats.executedQueries++;
		
		try(DatabaseConnection conn = this.pool.getConnection())
		{
			try(Statement st = conn.getInnerConnection().createStatement())
			{
				try(ResultSet rs = st.executeQuery(sql))
				{
					if(rs.next())
					{
						return rs.getInt(1);
					}
				}
			}
		}
		catch(Exception ex)
		{
			this.getLogger().error(this, "Error during SELECT.", ex);
		}
		
		return 0;
	}
	
	public String queryStr(String sql)
	{
		SystemStats.executedQueries++;
		
		try(DatabaseConnection conn = this.pool.getConnection())
		{
			try(Statement st = conn.getInnerConnection().createStatement())
			{
				try(ResultSet rs = st.executeQuery(sql))
				{
					if(rs.next())
					{
						return rs.getString(1);
					}
				}
			}
		}
		catch(Exception ex)
		{
			this.getLogger().error(this, "Error during SELECT.", ex);
		}
		
		return null;
	}
	
	public DatabaseEndpoint getEndpoint()
	{
		return this.endpoint;
	}
	
	public DatabaseConnectionPool getPool()
	{
		return this.pool;
	}
	
	public DatabaseUpdateQueue getUpdateQueue()
	{
		return this.updateQueue;
	}
}
