package habbod2.game.messenger;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;

import habbod2.System;
import habbod2.communication.messages.outgoing.friendlist.FriendListUpdateMessageComposer;
import habbod2.communication.messages.outgoing.friendlist.NewBuddyRequestMessageComposer;
import habbod2.game.sessions.Session;
import habbod2.game.users.User;

public class InstantMessenger
{
	private Session session;
	private boolean appearOffline;

	private MessengerBuddy self;
	private Map<Integer, MessengerBuddy> buddies;
	private Map<Integer, String> requests;
	private List<MessengerBuddy> buddiesToUpdate;
	
	public InstantMessenger(Session session)
	{
		this.session = session;
		this.buddiesToUpdate = new CopyOnWriteArrayList<MessengerBuddy>();
	}

	public void reload()
	{
		// Get service
		MessengerService svc = System.get().getGameMgr().getMessengerSvc();
		
		// Load buddies + requests
		this.buddiesToUpdate.clear();
		this.buddies = svc.getBuddyListForUser(this.session.getUser().id);
		this.requests = svc.getBuddyRequestsForUser(this.session.getUser().id);
		
		// Update own buddy info
		this.self = svc.getBuddy(this.session.getUser().id, true, true);
		self.setOnline(this);
		
		// Broadcast instant update to all buddies ('online!')
		this.updateStatus(true);
	}

	public void offline()
	{
		// Update own buddy info
		this.self.refCount--;
		this.self.setOffline();
		
		// Broadcast instant update to all buddies ('offline!')
		this.updateStatus(true);

		// Clear lists
		for(MessengerBuddy buddy : this.buddies.values())
		{
			buddy.refCount--;
		}
		this.buddies.clear();
		this.requests.clear();
	}
	
	public void addBuddy(MessengerBuddy buddy)
	{
		// Update through the update system
		buddy.refCount++;
		this.setUpdateNeeded(buddy);
		this.sendPendingUpdates();
	}

	public void removeBuddy(int userId)
	{
		MessengerBuddy buddy = this.buddies.remove(userId);
		if(buddy != null)
		{
			// Removed!
			buddy.refCount--;
			
			// Notify this user
			this.session.send(FriendListUpdateMessageComposer.compose(buddy));
		}
	}
	
	public void addRequest(MessengerBuddy buddy)
	{
		// Mutate list
		this.requests.put(buddy.userId, buddy.username);

		// Notify session
		this.session.send(NewBuddyRequestMessageComposer.compose(buddy));
	}

	public void removeRequest(int fromUserId)
	{
		this.requests.remove(fromUserId);
	}

	public void updateStatus(boolean instantUpdate)
	{
		// Get service
		MessengerService svc = System.get().getGameMgr().getMessengerSvc();
		if(svc.isShutdown())
		{
			return;
		}
		
		// Sync user + buddy
		User thisUser = this.session.getUser();
		MessengerBuddy thisBuddy = svc.getBuddy(thisUser.id, true, false);
		if(thisBuddy != null && thisBuddy != this.self)
		{
			this.self.refCount--;
			this.self = thisBuddy;
			thisBuddy.refCount++;
		}
		thisBuddy.figure = thisUser.figure;
		thisBuddy.motto = thisUser.motto;
		
		// Flag this user as 'updated' in buddylist of online buddies
		for(MessengerBuddy buddy : this.buddies.values())
		{
			if(buddy.isOnline())
			{
				buddy.messenger.setUpdateNeeded(thisBuddy);
				if(instantUpdate)
				{
					buddy.messenger.sendPendingUpdates();
				}
			}
		}
	}
	
	private void setUpdateNeeded(MessengerBuddy buddy)
	{
		// Not already pending update?
		if(buddy != null && !this.buddiesToUpdate.contains(buddy))
		{
			// Not in buddylist yet?
			if(!buddies.containsKey(buddy.userId))
			{
				this.buddies.put(buddy.userId, buddy);
			}
			
			// This list will be processed next time the client asks for it, or after a forced update
			this.buddiesToUpdate.add(buddy);
		}
	}
	
	public void sendPendingUpdates()
	{
		this.session.send(FriendListUpdateMessageComposer.compose(this.buddiesToUpdate));
		this.buddiesToUpdate.clear();
	}

	public Session getSession()
	{
		return this.session;
	}
	
	public MessengerBuddy getBuddy()
	{
		return this.self;
	}
	
	public boolean appearOffline()
	{
		return this.appearOffline;
	}

	public void appearOffline(boolean status)
	{
		this.appearOffline = status;
	}

	public boolean hasBuddy(int userId)
	{
		return this.buddies.containsKey(userId);
	}

	public MessengerBuddy getBuddy(int userId)
	{
		return this.buddies.get(userId);
	}

	public Collection<MessengerBuddy> getBuddies()
	{
		return this.buddies.values();
	}

	public boolean hasRequest(int fromUserId)
	{
		return this.requests.containsKey(fromUserId);
	}
	
	public Set<Entry<Integer, String>> getRequests()
	{
		return this.requests.entrySet();
	}
}
