package habbod2.game.moderation;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import habbod2.System;
import static habbod2.core.Locale._I;
import habbod2.communication.messages.outgoing.moderation.ModerationTicketMessageComposer;
import habbod2.communication.messages.outgoing.notifications.BroadcastMessageComposer;
import habbod2.game.GameManager;
import habbod2.game.rooms.instance.RoomInstance;
import habbod2.game.rooms.instance.RoomUser;
import habbod2.game.users.User;
import habbod2.game.users.UserManager;
import habbod2.game.sessions.Session;
import habbod2.game.sessions.SessionManager;
import habbod2.storage.Database;
import habbod2.storage.DatabaseResult;
import habbod2.storage.SQLHelper;

public class ModerationCenter extends GameManager.Component
{
	private ConcurrentHashMap<Integer, ModerationTicket> tickets;
	private ModerationPresetManager presetMgr;
	
	public ModerationCenter(GameManager gameMgr)
	{
		super(gameMgr);
		this.tickets = new ConcurrentHashMap<Integer, ModerationTicket>();
		this.presetMgr = new ModerationPresetManager(this);
	}
	
	public void reload()
	{
		this.loadPendingTickets();
		this.presetMgr.reload();
	}

	public void loadPendingTickets()
	{
		this.tickets.clear();
		
		// Get usermanager
		UserManager userMgr = this.getGameMgr().getUserMgr();
		
		// Load all open/picked tickets from database
		try(DatabaseResult result = this.getDb().query("SELECT * FROM moderation_tickets WHERE status=" + ModerationTicket.STATUS_OPEN + " OR status=" + ModerationTicket.STATUS_PICKED))
		{
			while(result.row())
			{
				// Process from row
				ModerationTicket ticket = new ModerationTicket();
				ticket.id = result.getInt("id");
				ticket.priority = result.getInt("priority");
				ticket.type = result.getInt("type");
				ticket.userId = result.getInt("user_id");
				ticket.userUsername = userMgr.getUsername(ticket.userId);
				ticket.reportedUserId = result.getInt("reported_id");
				if(ticket.reportedUserId > 0)
				{
					ticket.reportedUsername = userMgr.getUsername(ticket.userId);
				}
				ticket.pickerId = result.getInt("picker_id");
				if(ticket.pickerId > 0)
				{
					ticket.pickerUsername = userMgr.getUsername(ticket.pickerId);
				}
				ticket.roomId = result.getInt("room_id");
				ticket.roomName = result.getString("room_name");
				ticket.message = result.getString("message");
				ticket.timestamp = result.getTime("timestamp");
				
				// Add to map
				this.tickets.put(ticket.id, ticket);
			}
		}
		
		// Yay!
		this.getLogger().info(this, String.format("Loaded %d pending CFHs from database.", this.tickets.size()));
	}

	public ModerationTicket getTicket(int ticketId)
	{
		return this.tickets.get(ticketId);
	}
	
	public ModerationTicket getUnpickedTicketForUser(int userId)
	{
		for(ModerationTicket ticket : this.tickets.values())
		{
			if(ticket.userId == userId && ticket.status == ModerationTicket.STATUS_OPEN)
			{
				return ticket;
			}
		}
		return null;
	}

	public Collection<ModerationTicket> getAllTickets()
	{
		return this.tickets.values();
	}

	public void submitTicket(Session session, int type, int reportedUserId, String message)
	{
		// Create ticket
		ModerationTicket ticket = new ModerationTicket();
		ticket.priority = 1;
		ticket.type = type;
		ticket.status = ModerationTicket.STATUS_OPEN;
		ticket.userId = session.getUser().id;
		ticket.userUsername = session.getUser().username;
		if(reportedUserId > 0)
		{
			ticket.reportedUserId = reportedUserId;
			ticket.reportedUsername = this.getGameMgr().getUserMgr().getUsername(reportedUserId);
		}
		ticket.message = message;
		ticket.roomId = session.getRoom().getInfo().id;
		ticket.roomName = session.getRoom().getInfo().name;
		ticket.timestamp = System.get().getTime();
		ticket.id = this.getDb().executeReturnKey("INSERT INTO moderation_tickets(priority,type,status,user_id,reported_id,room_id,room_name,message,timestamp) VALUES (" + ticket.priority + "," + ticket.type + "," + ticket.status + "," + ticket.userId + "," + ticket.reportedUserId + "," + ticket.roomId + ",'" + SQLHelper.escape(ticket.roomName) + "','" + SQLHelper.escape(ticket.message) + "','" + SQLHelper.formatDate(ticket.timestamp) + "')");
		
		// Add to map
		this.tickets.put(ticket.id, ticket);
		
		// Notify helpers
		this.broadcastTicket(ticket);
	}
	
	public void broadcastTicket(ModerationTicket ticket)
	{
		SessionManager sessionMgr = this.getGameMgr().getSessionMgr();
		for(Session session : sessionMgr.getAllSessions())
		{
			if(session.authenticated() && session.getUser().rights.contains("fuse_mod"))
			{
				session.send(ModerationTicketMessageComposer.compose(ticket));
			}
		}
	}
	
	public UserModerationHistory getUserModerationHistory(int userId)
	{
		UserModerationHistory history = new UserModerationHistory();
		Database db = this.getDb();
		history.bans = db.queryInt("SELECT COUNT(id) FROM moderation_bans WHERE user_id=" + userId);
		history.cautions = db.queryInt("SELECT COUNT(id) FROM system_actionlog WHERE target_id=" + userId + " AND (action='caution' OR (action='kick' AND message != ''))");
		history.cfhs = db.queryInt("SELECT COUNT(id) FROM moderation_tickets WHERE user_id=" + userId);
		history.abusiveCfhs = db.queryInt("SELECT COUNT(id) FROM moderation_tickets WHERE user_id=" + userId + " AND status=" + ModerationTicket.STATUS_ABUSIVE);
	
		return history;
	}

	public void issueUserAlert(int issuerId, String username, String message, boolean isCaution) throws ModerationActionError
	{
		Session sess = this.getGameMgr().getSessionMgr().getSessionByUsername(username);
		this.issueUserAlert(issuerId, sess, message, isCaution);
	}

	public void issueUserAlert(int issuerId, int userId, String message, boolean isCaution) throws ModerationActionError
	{
		Session sess = this.getGameMgr().getSessionMgr().getSessionByUserId(userId);
		this.issueUserAlert(issuerId, sess, message, isCaution);
	}

	private void issueUserAlert(int issuerId, Session session, String message, boolean isCaution) throws ModerationActionError
	{
		if(session == null)
		{
			throw new ModerationActionError(_I("user not online"));
		}
		else if(message.isEmpty())
		{
			throw new ModerationActionError(_I("message cannot be blank"));
		}
		else
		{
			session.sendNotif(message, isCaution);
			this.getGameMgr().getAccessCtrl().logUserAction(issuerId, session.getUser().id, isCaution ? "caution" : "alert", message);
		}
	}

	public void issueUserKick(int issuerId, String username, String message) throws ModerationActionError
	{
		Session sess = this.getGameMgr().getSessionMgr().getSessionByUsername(username);
		this.issueUserKick(issuerId, sess, message);
	}

	public void issueUserKick(int issuerId, int userId, String message) throws ModerationActionError
	{
		Session sess = this.getGameMgr().getSessionMgr().getSessionByUserId(userId);
		this.issueUserKick(issuerId, sess, message);
	}

	private void issueUserKick(int issuerId, Session session, String message) throws ModerationActionError
	{
		User issuer = this.getGameMgr().getUserMgr().getUserById(issuerId);
		if(session == null || !session.inRoom())
		{
			throw new ModerationActionError(_I("user not online / not in room"));
		}
		else if(issuer == null || issuer.role.level <= session.getUser().role.level)
		{
			throw new ModerationActionError(_I("insufficient permissions"));
		}
		else
		{
			if(message != null && message.isEmpty())
			{
				message = null;
			}
			session.kickFromRoom(message, true);
			this.getGameMgr().getAccessCtrl().logUserAction(issuerId, session.getUser().id, "kick", message);
		}
	}

	public void issueRoomAlert(int issuerId, int roomId, String message, boolean isCaution) throws ModerationActionError
	{
		RoomInstance room = this.getGameMgr().getRoomMgr().getRoomInstance(roomId, false);
		if(room == null)
		{
			throw new ModerationActionError(_I("room not loaded"));
		}
		else if(message.isEmpty())
		{
			throw new ModerationActionError(_I("message cannot be blank"));
		}
		else
		{
			// Broadcast message
			room.broadcast(BroadcastMessageComposer.compose(message));
			
			// Log action
			this.getGameMgr().getAccessCtrl().logUserAction(issuerId, roomId, "roomalert", message);
		}
	}

	public void issueRoomKick(int issuerId, int roomId, String message) throws ModerationActionError
	{
		User issuer = this.getGameMgr().getUserMgr().getUserById(issuerId);
		RoomInstance room = this.getGameMgr().getRoomMgr().getRoomInstance(roomId, false);
		if(issuer == null || room == null)
		{
			throw new ModerationActionError(_I("room not loaded"));
		}
		else
		{
			// Fix message
			if(message != null && message.isEmpty())
			{
				message = null;
			}

			// Collect list with users to kick
			List<RoomUser> toKick = new ArrayList<RoomUser>();
			for(RoomUser user : room.getAllUsers())
			{
				if(user.getInfo().role.level < issuer.role.level)
				{
					toKick.add(user);
				}
			}

			// Kick them
			for(RoomUser user : toKick)
			{
				user.getSession().kickFromRoom(message, true);
			}

			// Log it
			this.getGameMgr().getAccessCtrl().logUserAction(issuerId, roomId, "roomkick", message);
		}
	}

	public void issueUserBan(int issuerId, int userId, String message, int hours) throws ModerationActionError
	{
		// Get issuer and target
		UserManager userMgr = this.getGameMgr().getUserMgr();
		User issuer = userMgr.getUserById(issuerId);
		User target = userMgr.getUserById(userId);
		
		// Perform checks
		if(issuer == null || target == null)
		{
			throw new ModerationActionError(_I("user does not exist"));
		}
		else if(target.role.level >= issuer.role.level)
		{
			throw new ModerationActionError(_I("insufficient permissions"));
		}
		else if(message.isEmpty())
		{
			throw new ModerationActionError(_I("message cannot be blank"));
		}
		else if(hours <= 0)
		{
			throw new ModerationActionError(_I("ban duration must be larger than 0 hours"));
		}
		
		// Get calendar
		Calendar cal = Calendar.getInstance();
		
		// Create ban
		ModerationBan ban = new ModerationBan();
		ban.userId = userId;
		ban.message = message;
		ban.issuedBy = issuerId;
		ban.issuedAt = cal.getTimeInMillis();
		cal.add(Calendar.HOUR, hours);
		ban.expiresAt = cal.getTimeInMillis();
		ban.id = this.getDb().executeReturnKey("INSERT INTO moderation_bans(user_id,message,expires_at,issued_at,issued_by) VALUES (" + ban.userId + ",'" + SQLHelper.escape(ban.message) + "','" + SQLHelper.formatDate(ban.expiresAt) + "','" + SQLHelper.formatDate(ban.issuedAt) + "'," + ban.issuedBy + ")");
		
		// Target online?
		Session session = this.getGameMgr().getSessionMgr().getSessionByUserId(ban.userId);
		if(session != null)
		{
			session.end("user has been banned");
		}
	}
	
	public List<ModerationBan> getActiveBansForUserId(int userId)
	{
		List<ModerationBan> bans = new ArrayList<ModerationBan>(0);
		
		try(DatabaseResult result = this.getDb().query("SELECT * FROM moderation_bans WHERE user_id=" + userId))
		{
			while(result.row())
			{
				ModerationBan ban = this.parseBanFromRow(result);
				if(!ban.hasExpired())
				{
					bans.add(ban);
				}
			}
		}
		
		return bans;
	}
	
	public void issueIpBan(int issuerId, String ip, String message, int hours) throws ModerationActionError
	{
		// Get issuer and target
		User issuer = this.getGameMgr().getUserMgr().getUserById(issuerId);
		
		// Perform checks
		if(issuer == null)
		{
			throw new ModerationActionError(_I("user does not exist"));
		}
		else if(message.isEmpty())
		{
			throw new ModerationActionError(_I("message cannot be blank"));
		}
		else if(hours <= 0)
		{
			throw new ModerationActionError(_I("ban duration must be larger than 0 hours"));
		}
		
		// Get calendar
		Calendar cal = Calendar.getInstance();
		
		// Create ban
		ModerationBan ban = new ModerationBan();
		ban.ip = ip;
		ban.message = message;
		ban.issuedBy = issuerId;
		ban.issuedAt = cal.getTimeInMillis();
		cal.add(Calendar.HOUR, hours);
		ban.expiresAt = cal.getTimeInMillis();
		ban.id = this.getDb().executeReturnKey("INSERT INTO moderation_bans(ip,message,expires_at,issued_at,issued_by) VALUES (" + SQLHelper.escape(ban.ip) + ",'" + SQLHelper.escape(ban.message) + "','" + SQLHelper.formatDate(ban.expiresAt) + "','" + SQLHelper.formatDate(ban.issuedAt) + "'," + ban.issuedBy + ")");
		
		// Disconnect all users on this IP
		for(Session session : this.getGameMgr().getSessionMgr().getAllSessions())
		{
			if(session.getConnection().getIpAddress().equals(ban.ip))
			{
				session.end("IP address has been banned");
			}
		}
	}

	public List<ModerationBan> getActiveBansForIp(String ip)
	{
		List<ModerationBan> bans = new ArrayList<ModerationBan>(0);
		
		try(DatabaseResult result = this.getDb().query("SELECT * FROM moderation_bans WHERE ip='" + SQLHelper.escape(ip) + "'"))
		{
			while(result.row())
			{
				ModerationBan ban = this.parseBanFromRow(result);
				if(!ban.hasExpired())
				{
					bans.add(ban);
				}
			}
		}
		
		return bans;
	}
	
	public List<ModerationBan> getActiveBansForUserOrIp(int userId, String ip)
	{
		List<ModerationBan> bans = new ArrayList<ModerationBan>(0);
		
		try(DatabaseResult result = this.getDb().query("SELECT * FROM moderation_bans WHERE user_id=" + userId + " OR ip='" + SQLHelper.escape(ip) + "'"))
		{
			while(result.row())
			{
				ModerationBan ban = this.parseBanFromRow(result);
				if(!ban.hasExpired())
				{
					bans.add(ban);
				}
			}
		}
		
		return bans;
	}
	
	private ModerationBan parseBanFromRow(DatabaseResult row)
	{
		ModerationBan ban = new ModerationBan();
		ban.id = row.getInt("id");
		ban.userId = row.getInt("user_id");
		ban.ip = row.getString("ip");
		ban.message = row.getString("message");
		ban.expiresAt = row.getTime("expires_at");
		ban.issuedAt = row.getTime("issued_at");
		ban.issuedBy = row.getInt("issued_by");
		ban.extraInfo = row.getString("extra_info");
		return ban;
	}
	
	public ModerationPresetManager getPresetManager()
	{
		return this.presetMgr;
	}
}
