package habbod2.game.rooms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import habbod2.game.GameComponent;
import habbod2.game.GameManager;
import habbod2.game.rooms.instances.RoomInstance;
import habbod2.game.users.User;
import habbod2.storage.DatabaseResult;
import habbod2.storage.SQLHelper;
import habbod2.util.NamedThreadFactory;

public class RoomManager extends GameComponent
{
	private Map<String, RoomModel> models;
	private ConcurrentHashMap<Integer, RoomInstance> instances;
	private ScheduledExecutorService executor;

	public RoomManager(GameManager gameMgr)
	{
		super(gameMgr);
		this.models = Collections.emptyMap();
		this.instances = new ConcurrentHashMap<Integer, RoomInstance>();
		this.executor = Executors.newScheduledThreadPool(1, new NamedThreadFactory("RoomWorker #%d"));
	}

	public void start()
	{
		// Reset usercounts from old instance
		this.resetDbUserCounts();
	}
	
	public void shutdown()
	{
		// Shutdown executor immediately
		this.executor.shutdownNow();

		// Any instances to unload?
		if(!this.instances.isEmpty())
		{
			this.getLogger().info(this, "Unloading all room instances...");
			for(RoomInstance instance : this.instances.values())
			{
				instance.unload();
			}
			this.instances.clear();
			this.getLogger().info(this, "All room instances unloaded.");
		}
	}
	
	private void resetDbUserCounts()
	{
		this.getDb().execute("UPDATE rooms SET users_now=0");
	}

	public void reloadModels()
	{
		// Create new collection
		Map<String, RoomModel> tmpModels = new HashMap<String, RoomModel>();
		
		// Load all model data
		DatabaseResult result = this.getDb().query("SELECT * FROM roommodels");
		while(result.row())
		{
			RoomModel model = new RoomModel();
			model.name = result.getString("name");
			model.doorX = result.getInt("door_x");
			model.doorY = result.getInt("door_y");
			model.doorZ = result.getFloat("door_z");
			model.doorRotation = result.getByte("door_rotation");
			model.hasSwimmingPool = result.getBoolean("has_pool");
			model.clubOnly = result.getBoolean("club_only");
			if(model.init(result.getString("heightmap")))
			{
				tmpModels.put(model.name, model);
			}
			else
			{
				this.getLogger().info(this, String.format("Room model '%s' is invalid and has not been loaded.", model.name));
			}
		}
		result.release();

		// Set instance field
		this.models = tmpModels;
		
		// Done!
		this.getLogger().info(this, String.format("Loaded %d room models.", this.models.size()));
	}

	public RoomModel getModel(String name)
	{
		return this.models.get(name);
	}

	public Collection<RoomModel> getAllRoomModels()
	{
		return this.models.values();
	}
	
	public RoomInstance getRoomInstance(int roomId, boolean allowCreate)
	{
		// Look for existing instance
		RoomInstance instance = this.instances.get(roomId);
		if(instance != null)
		{
			return instance;
		}

		// Allowed to create a new instance?
		if(allowCreate)
		{
			return this.createRoomInstance(roomId);
		}
		else
		{
			return null;
		}
	}

	private RoomInstance createRoomInstance(int roomId)
	{
		// Get room info
		Room info = this.getRoomInfo(roomId);
		if(info == null)
		{
			return null;
		}

		// Get room model
		RoomModel model = this.getModel(info.modelName);
		if(model == null)
		{
			this.getLogger().warning(this, String.format("Room #%d refers to the missing room model \"%s\". Could not create room instance.", info.id, info.modelName));
			return null;
		}

		// Create instance and load data like furniture
		RoomInstance instance = new RoomInstance(info, model);
		try
		{
			instance.load();
		}
		catch(Exception ex)
		{
			this.getLogger().error(this, String.format("Error loading instance of room #%d.", roomId), ex);
			return null;
		}

		// Assign to worker
		//this.getLowestLoadedWorker().assign(instance);

		// Add to instance map
		this.instances.put(instance.getInfo().id, instance);

		// Schedule and ensure there are <x> milliseconds between every pulse
		ScheduledFuture<?> task = this.executor.scheduleAtFixedRate(instance, 0, RoomInstance.PULSE_INTERVAL, TimeUnit.MILLISECONDS);
		instance.setTask(task);
		
		// Instance created!
		this.getLogger().debug(this, String.format("Created instance for room #%d.", info.id));
		return instance;
	}

	public void unloadRoomInstance(int roomId)
	{
		// Attempt to remove from the instance map
		RoomInstance instance = this.instances.remove(roomId);

		// Instance removed? Process further unloading
		if(instance != null)
		{
			// Un-assign from worker
			//this.workers[instance.getWorkerIndex()].unassign(roomId);

			// Cancel task
			instance.getTask().cancel(true);
			
			// Unload/save data
			instance.unload();

			// Unloaded successfully!
			this.getLogger().debug(this, String.format("Unloaded instance of room #%d.", roomId));
		}
	}

	public Room getRoomInfo(int roomId)
	{
		// Lookup room instances
		RoomInstance instance = this.getRoomInstance(roomId, false);
		if(instance != null)
		{
			return instance.getInfo();
		}

		// Lookup database...
		DatabaseResult result = this.getDb().query("SELECT r.*,u.username AS owner FROM rooms r LEFT JOIN users u ON u.id=r.owner_id WHERE r.id=" + roomId);
		Room room = result.row() ? Room.parse(result) : null;
		result.release();
		
		return room;
	}
	
	public List<Room> getRoomsForUser(int userId)
	{
		ArrayList<Room> rooms = new ArrayList<Room>(5);

		DatabaseResult result = this.getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE r.owner_id=" + userId);
		while(result.row())
		{
			rooms.add(Room.parse(result));
		}
		result.release();

		return rooms;
	}
	
	public int getRoomCountForUser(int userId)
	{
		return this.getDb().queryInt("SELECT COUNT(*) FROM rooms WHERE owner_id=" + userId);
	}

	public Room createGuestRoom(User user, String name, RoomModel model)
	{
		int roomId = this.getDb().executeReturnKey("INSERT INTO rooms(owner_id,name,model_name) VALUES (" + user.id + ",'" + SQLHelper.escape(name) + "','" + SQLHelper.escape(model.name) + "')");
		if(roomId > 0)
		{
			return this.getRoomInfo(roomId);
		}
		else
		{
			return null;
		}
	}

	public void deleteGuestRoom(int roomId)
	{
		// Delete room
		this.getDb().execute("DELETE FROM rooms WHERE id=" + roomId);
		this.getDb().execute("DELETE FROM room_rights WHERE room_id=" + roomId);
		
		// Delete content
		this.getDb().execute("DELETE FROM items WHERE room_id=" + roomId);
		
		// Unload room instance (if loaded)
		this.unloadRoomInstance(roomId);
	}

	public Map<Integer, String> getRightsForRoom(int roomId)
	{
		// Initialize an empty map with init cap of 3
		Map<Integer, String> users = new HashMap<Integer, String>(3);

		// Query the user IDs + usernames
		DatabaseResult result = this.getDb().query("SELECT user_id,username FROM room_rights JOIN users ON id=user_id WHERE room_id=" + roomId);
		while(result.row())
		{
			users.put(result.getInt("user_id"), result.getString("username"));
		}
		result.release();

		// Return the users
		return users;
	}

	public void addRoomRights(int roomId, int userId)
	{
		this.getDb().execute("INSERT INTO room_rights(room_id,user_id) VALUES(" + roomId + "," + userId + ")");
	}

	public void removeRoomRights(int roomId, int userId)
	{
		this.getDb().execute("DELETE FROM room_rights WHERE room_id=" + roomId + " AND user_id=" + userId);
	}
	
	public void logRoomEntry(int userId, int roomId)
	{
		this.getDb().execute("INSERT INTO user_roomvisits(user_id,room_id,entry) VALUES (" + userId + "," + roomId + ",'" + SQLHelper.getDateTime() + "')");
	}
	
	public void logRoomExit(int userId, int roomId)
	{
		this.getDb().execute("UPDATE user_roomvisits SET `exit` = '" + SQLHelper.getDateTime() + "' WHERE user_id=" + userId + " AND room_id=" + roomId + " AND `exit` IS NULL LIMIT 1");
	}
	
	public List<Room> getGuestRoomSearchResult(String criteria, int maxResults, boolean byTagsOnly)
	{
		ArrayList<Room> rooms = new ArrayList<Room>();

		criteria = SQLHelper.escape(criteria);
		
		DatabaseResult result;
		if(byTagsOnly)
		{
			result = this.getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id = r.owner_id WHERE tag1='" + criteria + "' OR tag2='" + criteria + "' ORDER BY r.users_now DESC LIMIT " + maxResults);
		}
		else
		{
			result = this.getDb().query("(SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id = r.owner_id WHERE u.username='" + criteria + "') UNION (SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id = r.owner_id WHERE r.name LIKE '" + criteria + "%') LIMIT " + maxResults);
		}
		while(result.row())
		{
			rooms.add(Room.parse(result));
		}
		result.release();

		return rooms;
	}
	
	public List<Room> getRecentGuestRoomsForUser(int userId, int maxResults)
	{
		List<Room> rooms = new ArrayList<Room>();

		List<Integer> exclude = new ArrayList<Integer>();
		DatabaseResult result = this.getDb().query("SELECT r.*,u.username AS owner,rv.entry FROM user_roomvisits rv JOIN rooms r ON r.id = rv.room_id JOIN users u ON u.id=r.owner_id WHERE rv.user_id=" + userId + " ORDER BY rv.entry DESC LIMIT " + maxResults);
		while(result.row())
		{
			int roomId = result.getInt("id");
			if(!exclude.contains(roomId))
			{
				rooms.add(Room.parse(result));
				exclude.add(roomId);
			}
		}
		result.release();

		return rooms;
	}
	
	public List<RoomVisit> getLastRoomVisitsForUser(int userId, int maxResults)
	{
		List<RoomVisit> visits = new ArrayList<RoomVisit>(maxResults);
		
		DatabaseResult result =  this.getDb().query("SELECT r.*,u.username AS owner,rv.entry,rv.exit FROM user_roomvisits rv JOIN rooms r ON r.id = rv.room_id LEFT JOIN users u ON u.id=r.owner_id WHERE rv.user_id=" + userId + " ORDER BY rv.entry DESC LIMIT " + maxResults);
		while(result.row())
		{
			RoomVisit visit = new RoomVisit();
			visit.room = Room.parse(result);
			visit.entry = result.getTime("entry");
			visit.exit = result.getTime("exit");
			visits.add(visit);
		}
		result.release();
		
		return visits;
	}
	
	public List<Integer> getFavoriteRoomsForUser(int userId)
	{
		List<Integer> favorites = new ArrayList<Integer>(3);

		DatabaseResult result = this.getDb().query("SELECT room_id FROM user_favoriterooms WHERE user_id=" + userId);
		while(result.row())
		{
			favorites.add(result.getInt("room_id"));
		}
		result.release();

		return favorites;
	}

	public void addFavoriteRoomForUser(int userId, int roomId)
	{
		this.getDb().execute("INSERT INTO user_favoriterooms(user_id,room_id) VALUES (" + userId + "," + roomId + ")");
	}

	public void removeFavoriteRoomForUser(int userId, int roomId)
	{
		this.getDb().execute("DELETE FROM user_favoriterooms WHERE user_id=" + userId + " AND room_id=" + roomId);
	}
	
	public Collection<RoomInstance> getAllRoomInstances()
	{
		return this.instances.values();
	}

}
