package habbod2.game.rooms.instance;

import java.util.List;
import java.util.ArrayList;

import habbod2.game.items.EItemFlags;
import habbod2.game.rooms.items.RoomFloorItem;
import habbod2.game.rooms.items.floor.RollerItem;
import habbod2.net.protocol.ServerMessage;
import habbod2.communication.messages.outgoing.room.engine.SlideObjectBundleMessageComposer;

public class RollerEngine
{
	private static final int DEFAULT_INTERVAL_TIME = (RoomInstance.PULSE_INTERVAL * 4);

	private long lastTime;
	private long intervalTime;
	private RoomInstance room;

	private ServerMessage msg;
	private List<Object> moved;
	
	private int lastCount;
	
	public RollerEngine(RoomInstance room)
	{
		this.lastTime = System.currentTimeMillis();
		this.room = room;

		// Set default interval time
		this.intervalTime = RollerEngine.DEFAULT_INTERVAL_TIME;
		
		// These are re-used all the time
		this.msg = ServerMessage.create();
		this.moved = new ArrayList<Object>();
	}
	
	public void setIntervalTime(long intervalTime)
	{
		this.intervalTime = intervalTime;
	}

	public void pulse()
	{
		// Time to roll the rollers?
		if((System.currentTimeMillis() - this.lastTime) < this.intervalTime)
		{
			return;
		}
		
		// Reset count
		this.lastCount = 0;
		
		// Search through all the items in the room
		boolean rebuildMap = false;
		for(RoomFloorItem tmpItem : this.room.getFloorItems())
		{
			// Is this a roller?
			if(tmpItem instanceof RollerItem)
			{
				// Count it
				this.lastCount++;
				
				// Cast to RollerItem
				RollerItem roller = (RollerItem) tmpItem;

				// Determine next tile
				RoomTile tile = roller.getTile();
				RoomTile next = roller.getTileInFront();
				boolean nextHasRoller = false;
				if(next != null && !next.hasAvatar)
				{
					// Check the items on the next tile
					boolean canMerge = false;
					List<RoomFloorItem> onNext = next.getItems();
					for(RoomFloorItem item : onNext)
					{
						// Check for rollers
						if(item instanceof RollerItem)
						{
							nextHasRoller = true;
						}
						
						// Check if merging is allowed
						else if(!canMerge && (item.is(EItemFlags.CAN_STAND_ON_TOP) || item.is(EItemFlags.CAN_STACK_ON_TOP)))
						{
							canMerge = true;
						}
						
						// Further searching is pointless?
						if(nextHasRoller && canMerge)
						{
							break;
						}
					}
					
					// Move the items on this roller?
					if(next.isOpen() || canMerge)
					{
						for(RoomFloorItem item : tile.getItems())
						{
							// Not already moved?
							if(item != roller && !moved.contains(item))
							{
								// Save current item height
								float fromZ = item.z;
								
								// Update item, force map rebuild
								item.x = next.x;
								item.y = next.y;
								item.z = (nextHasRoller ? item.z : item.z - roller.definition.height);
								item.requireDbUpdate();
								rebuildMap = true;
								
								// Add to list so it doesn't get moved twice in the same cycle
								moved.add(item);
								
								// Broadcast update
								this.room.broadcast(SlideObjectBundleMessageComposer.compose(item, tile, fromZ, roller, this.msg));
							}
						}
					}
					
					// An avatar can roll here?
					if(tile.hasAvatar && next.isOpen())
					{
						// Move the avatars
						for(RoomAvatar avatar : room.getAllAvatars())
						{
							// Avatar here and not moving, and not moved already?
							if(avatar.tile == tile && !avatar.hasGoal() && !moved.contains(avatar))
							{
								// Update map + avatar
								tile.stepOff(avatar, false);
								next.stepOn(avatar);
								
								// Add to list so it doesn't get moved twice in the same cycle
								moved.add(avatar);
								
								// Broadcast update (<avatar> just moved from <roller tile> to current position, using <roller>)
								room.broadcast(SlideObjectBundleMessageComposer.compose(avatar, tile, roller, this.msg));
							}
						}
					}
				}
			}
		}

		// Atleast one item moved?
		if(rebuildMap)
		{
			this.room.rebuildFloorMap();
		}

		// Done!
		if(!this.moved.isEmpty())
		{
			this.moved.clear();
		}
		this.lastTime = System.currentTimeMillis();
	}
	
	public int getCount()
	{
		return this.lastCount;
	}
}