package habbod2.game.rooms.instance;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

import habbod2.net.protocol.ServerMessage;
import habbod2.communication.messages.outgoing.room.chat.ChatMessageComposer;
import habbod2.communication.messages.outgoing.room.chat.ShoutMessageComposer;
import habbod2.communication.messages.outgoing.room.engine.UserUpdateMessageComposer;
import habbod2.game.pets.RoomPet;
import habbod2.game.rooms.geometry.RoomTile;
import habbod2.game.rooms.geometry.RotationCalculator;
import habbod2.game.rooms.pathfinding.PathfinderNode;
import habbod2.game.util.ChatUtil;

public abstract class RoomAvatar
{
	// Room
	public int avatarId;
	private RoomInstance room;
	private boolean invalidated;

	// Position
	public byte headRotation;
	public byte bodyRotation;
	public RoomTile tile;
	
	// Statuses
	public HashMap<String, String> statuses;
	public boolean requiresUpdate;

	public int idleTime;

	// Movement
	public RoomTile goal;
	public RoomTile next;
	private boolean moveLock;
	public boolean overrideNextTile;
	public List<PathfinderNode> path;
	public boolean executePathfinder;
	public byte pathfindingAllowedOn;
	
	// Following
	public RoomAvatar follow;

	public RoomAvatar()
	{
		this.statuses = new HashMap<String, String>();
		this.path = Collections.emptyList();
		this.clearMovement();
	}
	
	public RoomInstance getRoom()
	{
		return this.room;
	}

	public void setRoom(RoomInstance room)
	{
		this.room = room;
	}

	public void invalidate()
	{
		this.invalidated = true;
	}

	public boolean invalidated()
	{
		return this.invalidated;
	}

	public void pulse()
	{
		this.idleTime++;
		
		// Follow avatar?
		if(this.follow != null)
		{
			// Work out tile behind
			int x = this.follow.tile.x;
			int y = this.follow.tile.y;
			if(this.follow.bodyRotation == 0)
			{
				y++;
			}
			else if(this.follow.bodyRotation == 1)
			{
				y++;
				x--;
			}
			else if(this.follow.bodyRotation == 2)
			{
				x--;
			}
			else if(this.follow.bodyRotation == 3)
			{
				x--;
				y--;
			}
			else if(this.follow.bodyRotation == 4)
			{
				y--;
			}
			else if(this.follow.bodyRotation == 5)
			{
				y--;
				x++;
			}
			else if(this.follow.bodyRotation == 6)
			{
				x++;
			}
			
			// Not already on this tile?
			RoomTile tile = this.getRoom().getMap().getTile(x, y);
			if(tile != this.tile)
			{
				this.setGoal(tile);
			}
			
			// Unfollow?
			if(this.follow.invalidated())
			{
				this.follow = null;
			}
		}
	}

	public void unidle()
	{
		this.idleTime = 0;
	}

	public boolean at(RoomTile tile)
	{
		return(this.tile == tile);
	}
	
	public RoomTile getTile()
	{
		return this.tile;
	}
	
	public void setGoal(RoomTile goal)
	{
		// Unidle the avatar
		this.unidle();
		
		// Get rid of current path
		this.clearMovement();
		
		// Set the goal
		if(goal != null)
		{
			this.goal = goal;
			this.executePathfinder = true;
			
			// Include in next instance pulse
			this.requiresUpdate = true;
		}
	}

	public boolean hasGoal()
	{
		return(this.goal != null);
	}
	
	public boolean isMoving()
	{
		return(this.next != null);
	}

	public void clearMovement()
	{
		this.goal = null;
		//this.next = null;
		this.path.clear();
	}

	public boolean canMove()
	{
		return !this.moveLock;
	}

	public void lockMovement()
	{
		this.moveLock = true;
		this.clearMovement();
	}

	public void unlockMovement()
	{
		this.moveLock = false;
	}

	public void lookTo(RoomTile tile, boolean lockBody)
	{
		// Can't rotate in this context?
		if(this.hasGoal() || tile.equals(this.getTile()) || this.statuses.containsKey("lay"))
		{
			return;
		}

		// Calculate new rotation
		byte rotation = RotationCalculator.calculateDirection(this.tile.x, this.tile.y, tile.x, tile.y);
		int diff = this.bodyRotation - rotation;
		this.headRotation = this.bodyRotation;

		// Apply new rotation
		if(lockBody || this.statuses.containsKey("sit"))
		{
			if(this.bodyRotation == 2 || this.bodyRotation == 4)
			{
				if(diff > 0)
				{
					this.headRotation = (byte) (this.bodyRotation - 1);
				}
				else if(diff < 0)
				{
					this.headRotation = (byte) (this.bodyRotation + 1);
				}
			}
			else if(this.bodyRotation == 0 || this.bodyRotation == 6)
			{
				if(diff > 0)
				{
					this.headRotation = (byte) (this.bodyRotation - 1);
				}
				else if(diff < 0)
				{
					this.headRotation = (byte) (this.bodyRotation + 1);
				}
			}
		}
		else if(diff <= -2 || diff >= 2)
		{
			this.headRotation = rotation;
			this.bodyRotation = rotation;
		}
		else
		{
			this.headRotation = rotation;
		}
		
		// Unidle and require update (rotation changed)
		this.unidle();
		this.requiresUpdate = true;
	}

	public void chat(String message, boolean isShout)
	{
		// Unidle
		this.unidle();
		
		// Build servermessage
		int gesture = ChatUtil.getChatGesture(message);
		ServerMessage msg = isShout ? ShoutMessageComposer.compose(this, message, gesture) : ChatMessageComposer.compose(this, message, gesture);
		
		// This is a RoomUser? (real user)
		int myUserId = 0;
		if(this instanceof RoomUser)
		{	
			// Receive own message
			RoomUser user = (RoomUser)this;
			user.getSession().send(msg);
			
			// Record user ID for ignore-checks later
			myUserId = user.getInfo().id;
		}
		
		// Interact with avatars
		for(RoomAvatar avatar : this.room.getAllAvatars())
		{
			// Skip self (already handled)
			if(avatar == this)
			{
				continue;
			}
			
			// This is a RoomUser?
			if(myUserId > 0)
			{
				// The avatar is a RoomUser too?
				if(avatar instanceof RoomUser)
				{
					// Has he/she ignored this RoomUser?
					RoomUser user = (RoomUser)avatar;
					if(user.getSession().hasIgnoredUser(myUserId))
					{
						// Skip!
						continue;
					}
					else
					{
						// Receive the message
						user.getSession().send(msg);
					}
				}
			}
			
			// Extra stuff
			avatar.onChatFromAvatar(this, message, isShout);
		}
	}

	protected abstract void onChatFromAvatar(RoomAvatar from, String text, boolean isShout);

	public void addStatus(String key)
	{
		this.addStatus(key, null);
	}

	public void addStatus(String key, String value)
	{
		this.statuses.put(key, value);
		this.requiresUpdate = true;
	}

	public void removeStatus(String key)
	{
		this.statuses.remove(key);
		this.requiresUpdate = true;
	}
	
	public void removeNonMoveStatuses()
	{
		this.removeStatus("sit");
		this.removeStatus("lay");
	}
	
	public void refresh(boolean instant)
	{
		this.setTile(this.tile, instant);
	}
	
	public void setTile(RoomTile to, boolean instantStatusUpdate)
	{
		// Valid tile?
		if(to == null)
		{
			return;
		}
		
		// Step off current tile
		RoomTile from = this.tile;
		from.stepOff(this);
		
		// Step on new tile
		to.stepOn(this);
		
		// Delayed update?
		if(!instantStatusUpdate)
		{
			this.requiresUpdate = true;
		}
		else
		{
			// Send instant update
			this.room.broadcast(UserUpdateMessageComposer.compose(this));
			
			// Don't require further update
			this.requiresUpdate = false;
		}
	}
	
	public void appendStatusString(ServerMessage msg)
	{
		msg.appendInt(this.avatarId);
		msg.appendInt(this.tile.x);
		msg.appendInt(this.tile.y);
		msg.appendStr(this.tile.height);
		msg.appendInt(this.headRotation);
		msg.appendInt(this.bodyRotation);
		for(Entry<String, String> status : this.statuses.entrySet())
		{
			msg.append(status.getKey());
			if(status.getValue() != null)
			{
				msg.append(' ');
				msg.append(status.getValue());
			}
			msg.append('/');
		}
		if(this instanceof RoomPet)
		{
			if(this.tile.is(RoomTile.WATER))
			{
				msg.append("swm");
				msg.append('/');
			}
		}
		msg.appendStr('/');
	}

	public abstract String getAvatarKey();
}
