package habbod2.game.rooms.instance;

import java.util.List;
import java.util.ArrayList;

import habbod2.game.rooms.items.RoomFloorItem;

public class RoomTile
{
	public static final byte OPEN = (1 << 0);
	public static final byte CLOSED = (1 << 1);
	public static final byte LASTSTEP = (1 << 2);
	public static final byte WATER = (1 << 3);
	public static final RoomTile OUT_OF_RANGE = new RoomTile(-1, -1); 
	
	public final int x;
	public final int y;
	
	public byte state;
	public float height;
	public boolean hasAvatar;
	public boolean allowWalkthrough;
	
	public RoomTile redirect;
	public int effectId;
	
	public RoomFloorItem topItem;
	private List<RoomFloorItem> items;
	
	public RoomTile(int x, int y)
	{
		this.x = x;
		this.y = y;
		this.items = new ArrayList<RoomFloorItem>(1);
	}
	
	public boolean is(byte state)
	{
		return ((this.state & state) == state);
	}
	
	public boolean touches(RoomTile tile)
	{
		return(Math.abs(this.x - tile.x) <= 1 && Math.abs(this.y - tile.y) <= 1);
	}
	
	public boolean isOpen()
	{
		// Blocked?
		if(this.hasAvatar && !this.allowWalkthrough)
		{
			return false;
		}
		
		// Not open?
		if(!this.is(RoomTile.OPEN))
		{
			return false;
		}
		
		// Open!
		return true;
	}
	
	public boolean isPassable(byte allowedStates, boolean lastStep)
	{
		// Blocked?
		if(this.hasAvatar && !this.allowWalkthrough)
		{
			return false;
		}

		// Normal tile?
		if(!this.is(RoomTile.LASTSTEP))
		{
			// Not allowed by options?
			if((allowedStates & this.state) != this.state)
			{
				return false;
			}
		}
		else
		{
			// Tile is last step, but move is not last step?
			if(!lastStep)
			{
				return false;
			}
		}
		
		// Blocked?
		if(this.state == RoomTile.CLOSED)
		{
			return false;
		}

		// Closed item tiles are only passable with last step of path
		if(!lastStep && (this.state == (RoomTile.CLOSED | RoomTile.LASTSTEP)))
		{
			return false;
		}

		// Passable!
		return true;
	}
	
	public void clear(char defaultState, boolean allowWalkthrough)
	{
		this.topItem = null;
		this.redirect = null;
		if(Character.isDigit(defaultState))
		{
			this.state = RoomTile.OPEN;
			this.height = Integer.parseInt(Character.toString(defaultState));
		}
		else
		{
			this.state = RoomTile.CLOSED;
			this.height = 0.0f;
		}
		this.effectId = 0;
		this.allowWalkthrough = allowWalkthrough;
		this.items.clear();
	}
	
	public void addResidingItem(RoomFloorItem item)
	{
		// Add item
		this.items.add(item);
		
		// Update tile height if this item is higher
		if(item.z > this.height)
		{
			this.height = item.z;
		}
		
		// Top item?
		if(item.z >= this.height)
		{
			this.effectId = 0;
			item.onGenerateMapTile(this);
		}
	}
	
	public float getTopHeight()
	{
		if(this.topItem != null)
		{
			return this.topItem.z + this.topItem.definition.height;
		}
		else
		{
			return this.height;
		}
	}
	
	public void stepOn(RoomAvatar avatar)
	{
		// Avatar on tile
		avatar.tile = this;
		this.hasAvatar = true;
		
		// Item stepping
		if(this.topItem != null)
		{
			this.topItem.onAvatarStepOn(avatar);
		}
		
		// Effect handling
		if(avatar instanceof RoomHuman)
		{
			RoomHuman human = (RoomHuman)avatar;
			human.setEffect(RoomHuman.EFFECT_ROOM, this.effectId);
		}
	}
	
	public void stepOff(RoomAvatar avatar)
	{
		this.stepOff(avatar, true);
	}
	
	public void stepOff(RoomAvatar avatar, boolean clearStatuses)
	{
		if(clearStatuses)
		{
			avatar.removeNonMoveStatuses();
		}
		
		this.hasAvatar = false;
		if(this.topItem != null)
		{
			this.topItem.onAvatarStepOff(avatar);
		}
	}
	
	public List<RoomFloorItem> getItems()
	{
		return this.items;
	}
	
	public boolean equals(RoomTile tile)
	{
		return (tile != null && tile.x == this.x && tile.y == this.y);
	}
	
	@Override
	public String toString()
	{
		return ("[" + this.x + "," + this.y + "]");
	}
}
