package habbod2.game.rooms.instance;

import java.util.Collections;
import java.util.List;
import java.util.ArrayList;

import habbod2.game.rooms.items.RoomFloorItem;

public class RoomTile
{
	public static final byte OPEN = (1 << 0);
	public static final byte CLOSED = (1 << 1);
	public static final byte LASTSTEP = (1 << 2);
	public static final byte WATER = (1 << 3);
	public static final RoomTile OUT_OF_RANGE = new RoomTile(-1, -1); 
	
	public final int x;
	public final int y;
	
	public byte state;
	public float height;
	public boolean allowWalkthrough;
	private boolean hasAvatar;
	public boolean hasPendingAvatar;
	private int avatars;
	
	public RoomTile redirect;
	public int effectId;
	
	public RoomFloorItem topItem;
	private List<RoomFloorItem> items;
	
	public RoomTile(int x, int y)
	{
		this.x = x;
		this.y = y;
	}
	
	public boolean is(byte state)
	{
		return ((this.state & state) == state);
	}
	
	public boolean touches(RoomTile.TileObject obj)
	{
		return this.touches(obj.getTile());
	}
	
	public boolean touches(RoomTile tile)
	{
		return (tile != null && Math.abs(this.x - tile.x) <= 1 && Math.abs(this.y - tile.y) <= 1);
	}
	
	public static boolean touching(RoomTile t1, RoomTile t2)
	{
		return(Math.abs(t1.x - t2.x) <= 1 && Math.abs(t1.y - t2.y) <= 1);
	}
	
	public boolean isOpen()
	{
		// Blocked?
		if(this.hasAvatar() && !this.allowWalkthrough)
		{
			return false;
		}
		
		// Not open?
		if(!this.is(RoomTile.OPEN))
		{
			return false;
		}
		
		// Open!
		return true;
	}
	

	public void clear(char defaultState, boolean allowWalkthrough)
	{
		this.topItem = null;
		this.redirect = null;
		if(Character.isDigit(defaultState))
		{
			this.state = RoomTile.OPEN;
			this.height = Integer.parseInt(Character.toString(defaultState));
		}
		else
		{
			this.state = RoomTile.CLOSED;
			this.height = 0.0f;
		}
		this.effectId = 0;
		this.allowWalkthrough = allowWalkthrough;
		if(this.items != null)
		{
			this.items.clear();
		}
	}
	
	public void addResidingItem(RoomFloorItem item)
	{
		// Add item
		if(this.items == null)
		{
			this.items = new ArrayList<RoomFloorItem>(1);
		}
		this.items.add(item);
		
		// Update tile height if this item is higher
		if(this.topItem == null || item.z > this.topItem.z)
		{
			this.height = item.z;
		}
		
		// Top item?
		if(this.topItem == null || item.z >= this.topItem.z)
		{
			item.onGenerateMapTile(this);
			if(item.definition.effectId > 0 && this.is(RoomTile.OPEN))
			{
				this.effectId = item.definition.effectId;
			}
		}
	}
	
	public float getTopHeight()
	{
		if(this.topItem != null)
		{
			return this.topItem.z + this.topItem.definition.height;
		}
		else
		{
			return this.height;
		}
	}
	
	public void stepOn(RoomAvatar avatar)
	{
		// Avatar on tile
		avatar.x = this.x;
		avatar.y = this.y;
		this.avatars++;
		this.hasAvatar = true;
		
		// Item stepping
		if(this.topItem != null)
		{
			this.topItem.onAvatarStepOn(avatar);
		}
		avatar.z = this.height;
		
		// Effect handling
		if(avatar instanceof RoomHuman)
		{
			RoomHuman human = (RoomHuman)avatar;
			human.setEffect(RoomHuman.EFFECT_ROOM, this.effectId);
		}
	}
	
	public void stepOff(RoomAvatar avatar)
	{
		this.stepOff(avatar, true);
	}
	
	public void stepOff(RoomAvatar avatar, boolean clearStatuses)
	{
		if(clearStatuses)
		{
			avatar.removeNonMoveStatuses();
		}
		
		//if(this.avatars > 0)
		{
			this.avatars--;
			this.hasAvatar = false;
			this.hasPendingAvatar = false;
			if(this.topItem != null)
			{
				this.topItem.onAvatarStepOff(avatar);
			}
		}
	}
	
	public boolean hasAvatar()
	{
		return this.hasAvatar || this.hasPendingAvatar;
		//return this.avatars > 0;
	}
	
	public List<RoomFloorItem> getItems()
	{
		if(this.items != null)
		{
			return this.items;
		}
		else
		{
			return Collections.emptyList();
		}
	}
	
	public boolean equals(RoomTile tile)
	{
		return (tile != null && tile.x == this.x && tile.y == this.y);
	}
	
	@Override
	public String toString()
	{
		return ("[" + this.x + "," + this.y + "]");
	}
	
	public static interface TileObject
	{
		public RoomTile getTile();
	}
}
