package habbod2.game.rooms.instances;

import java.util.List;
import java.util.ArrayList;

import habbod2.game.items.EItemFlags;
import habbod2.game.rooms.items.RoomFloorItem;

public class RoomTile
{
	public static final byte OPEN = (1 << 0);
	public static final byte CLOSED = (1 << 1);
	public static final byte LASTSTEP = (1 << 2);
	public static final byte WATER = (1 << 3);
	public static final RoomTile OUT_OF_RANGE = new RoomTile(-1, -1); 
	
	public final int x;
	public final int y;
	
	public byte state;
	public float height;
	public boolean hasAvatar;
	
	public RoomTile redirect;
	public int effectId;
	
	public RoomFloorItem topItem;
	private List<RoomFloorItem> items;
	
	public RoomTile(int x, int y)
	{
		this.x = x;
		this.y = y;
		this.items = new ArrayList<RoomFloorItem>(1);
	}
	
	public boolean is(byte state)
	{
		return ((this.state & state) == state);
	}
	
	public boolean touches(RoomTile tile)
	{
		return(Math.abs(this.x - tile.x) <= 1 && Math.abs(this.y - tile.y) <= 1);
	}
	
	public boolean isOpen()
	{
		// Avatar here?
		if(this.hasAvatar)
		{
			return false;
		}
		
		// Not open?
		if(!this.is(RoomTile.OPEN))
		{
			return false;
		}
		
		// Open!
		return true;
	}
	
	public boolean isPassable(byte allowedStates, boolean lastStep)
	{
		// Avatar here?
		if(this.hasAvatar)
		{
			return false;
		}

		// Normal tile?
		if(!this.is(RoomTile.LASTSTEP))
		{
			// Not allowed by options?
			if((allowedStates & this.state) != this.state)
			{
				return false;
			}
		}
		else
		{
			// Tile is last step, but move is not last step?
			if(!lastStep)
			{
				return false;
			}
		}
		
		// Blocked?
		if(this.state == RoomTile.CLOSED)
		{
			return false;
		}

		// Closed item tiles are only passable with last step of path
		if(!lastStep && (this.state == (RoomTile.CLOSED | RoomTile.LASTSTEP)))
		{
			return false;
		}

		// Passable!
		return true;
	}
	
	public void clear(char defaultState)
	{
		this.topItem = null;
		this.redirect = null;
		this.effectId = 0;
		if(Character.isDigit(defaultState))
		{
			this.state = RoomTile.OPEN;
			this.height = Integer.parseInt(Character.toString(defaultState));
		}
		else
		{
			this.state = RoomTile.CLOSED;
			this.height = 0.0f;
		}
		this.items.clear();
	}
	
	public void addResidingItem(RoomFloorItem item)
	{
		// Add item
		this.items.add(item);
		
		// Update tile height if this item is higher
		if(item.z > this.height)
		{
			this.height = item.z;
		}
		
		// Top item?
		if(item.z >= this.height)
		{
			// Can stand on top?
			if(item.is(EItemFlags.CAN_STAND_ON_TOP))
			{
				// Tile not closed yet?
				if(!this.is(RoomTile.CLOSED))
				{
					this.state = RoomTile.OPEN;
					this.topItem = item;
					this.height = item.getTopHeight();
				}
			}
			
			// Can sit on top?
			else if(item.is(EItemFlags.CAN_SIT_ON_TOP))
			{
				this.state = (RoomTile.CLOSED | RoomTile.LASTSTEP);
				this.topItem = item;
			}
			
			// Can lay on top?
			else if(item.is(EItemFlags.CAN_LAY_ON_TOP))
			{
				this.state = (RoomTile.CLOSED | RoomTile.LASTSTEP);
				this.topItem = item;
				if(item.rotation == 0 || item.rotation == 4)
				{
					this.redirect = item.getRoom().getMap().getTile(this.x, item.y);
				}
				else
				{
					this.redirect = item.getRoom().getMap().getTile(item.x, this.y);
				}
			}
			
			// Solid item!
			else
			{
				// No last step item here?
				if(!this.is(RoomTile.LASTSTEP))
				{
					// Close the tile
					this.height = item.getTopHeight();
					this.state = RoomTile.CLOSED;
				}
			}
		}
	}
	
	public float getTopHeight()
	{
		if(this.topItem != null)
		{
			return this.topItem.z + this.topItem.definition.height;
		}
		else
		{
			return this.height;
		}
	}
	
	public RoomFloorItem getTopItem()
	{
		RoomFloorItem top = null;
		for(RoomFloorItem item : this.items)
		{
			if(top == null || item.z > top.z)
			{
				top = item;
			}
		}
		
		if(top != null && top != this.topItem)
		{
			top.toString();
		}
		
		return top;
	}
	
	public void stepOn(RoomAvatar avatar)
	{
		this.hasAvatar = true;
		
		// Position
		avatar.tile = this;
		
		// Item stepping
		if(this.topItem != null)
		{
			this.topItem.onAvatarStepOn(avatar);
		}
		
		// Effect handling
		if(avatar instanceof RoomUser)
		{
			RoomUser user = (RoomUser)avatar;
			if(this.effectId != user.getCurrentEffect())
			{
				user.setEffect(this.effectId, false);
			}
		}
	}
	
	public void stepOff(RoomAvatar avatar)
	{
		avatar.removeNonMoveStatuses();
		
		this.hasAvatar = false;
		if(this.topItem != null)
		{
			this.topItem.onAvatarStepOff(avatar);
		}
	}
	
	public List<RoomFloorItem> getItems()
	{
		return this.items;
	}
	
	public boolean equals(RoomTile tile)
	{
		return (tile != null && tile.x == this.x && tile.y == this.y);
	}
	
	@Override
	public String toString()
	{
		return ("[" + this.x + "," + this.y + "]");
	}
}
