package habbod2.game.rooms.minigames;

import java.util.Collection;

import habbod2.game.rooms.instance.RoomInstance;
import habbod2.game.rooms.instance.RoomUser;
import habbod2.game.rooms.items.RoomFloorItem;
import habbod2.game.rooms.items.minigame.ScoreboardItem;
import habbod2.game.rooms.items.minigame.TimerItem;
import habbod2.game.rooms.items.wired.triggers.GameEndsTrigger;
import habbod2.game.rooms.items.wired.triggers.GameStartsTrigger;
import habbod2.game.rooms.minigames.GameTeam;
import habbod2.game.util.GameUtil;

public abstract class Game
{
	private int id;
	private RoomInstance room;
	private GameTeam[] teams;
	private int startDelay;
	private int timeLeft;
	private boolean alive;
	private boolean decreaseTime;
	
	public Game(RoomInstance room)
	{
		// Set room
		this.room = room;
		
		// Get all team IDs
		GameTeam.ETeam[] teamIds = GameTeam.ETeam.values();
		
		// Create all empty teams
		this.teams = new GameTeam[teamIds.length];
		for(int i = 0; i < teamIds.length; i++)
		{
			this.teams[i] = new GameTeam(teamIds[i], this);
		}
	}
		
	public void start()
	{
		// Not already running?
		if(this.alive)
		{
			return;
		}
		
		// Requires timer
		TimerItem timer = this.getTimer();
		if(timer == null || timer.getTimeLeft() == 0)
		{
			return;
		}
		
		// Requires at least x teams with players
		if(this.getRequiredTeamsWithPlayers() > 0)
		{
			int teamsWithPlayers = 0;
			for(GameTeam team : this.teams)
			{
				if(team.getPlayers().size() > 0)
				{
					teamsWithPlayers++;
					continue;
				}
			}
			if(teamsWithPlayers < this.getRequiredTeamsWithPlayers())
			{
				return;
			}
		}
		
		// Prepare for start of new game
		this.id = GameUtil.getRandomInt(1, Integer.MAX_VALUE);
		this.alive = true;
		this.timeLeft = timer.getTimeLeft();
		for(GameTeam team : this.teams)
		{
			team.reset();
		}
		this.onStart();
		
		// Process items
		for(RoomFloorItem item : this.getItems())
		{
			if(item instanceof ScoreboardItem)
			{
				((ScoreboardItem)item).setScore(0);
			}
			else if(item instanceof GameStartsTrigger)
			{
				((GameStartsTrigger)item).execute(null);
			}
		}
	}
	
	public void stop()
	{
		// Stoppable?
		if(!this.alive)
		{
			return;
		}
		
		// Stopped!
		this.timeLeft = 0;
	
		// Determine winning team
		GameTeam winner = this.getTeamWithBestScore();
		if(winner != null)
		{
			for(GameUserPlayer player : winner.getPlayers())
			{
				player.getUser().wave();
			}
		}
		else
		{
			for(GameTeam team : this.getTeams())
			{
				for(GameUserPlayer player : team.getPlayers())
				{
					player.getUser().wave();
				}
			}
		}
		
		// Extra events
		this.onStop(winner);
		
		// Process items
		for(RoomFloorItem item : this.getItems())
		{
			if(item instanceof TimerItem)
			{
				((TimerItem)item).setTimeLeft(0);
			}
			else if(item instanceof GameEndsTrigger)
			{
				((GameEndsTrigger)item).execute(null);
			}
		}
	}
	
	public void pulse()
	{
		// Still alive?
		if(!this.alive)
		{
			return;
		}
		
		// Decrease time?
		if(this.timeLeft > 0)
		{
			// Every odd pulse (there's two pulses in a second)
			this.decreaseTime = !this.decreaseTime;
			if(this.decreaseTime)
			{
				// Deal with start delay?
				if(this.startDelay > 0)
				{
					this.startDelay--;
					if(this.startDelay == 0)
					{
						this.onStartDelayExpired();
					}
				}
				else
				{
					// Decrement time left?
					this.setTimeLeft(this.timeLeft - 1);
					if(this.timeLeft <= 0)
					{
						this.stop();
					}
				}
			}
		}
		
		// Process other game logic
		if(this.startDelay == 0)
		{
			this.onPulse();
		}
	}
	
	public void destroy()
	{
		this.stop();
		this.setAlive(false);
		for(GameTeam team : this.teams)
		{
			for(GameUserPlayer player : team.getPlayers())
			{
				player.getUser().setPlayer(null);
				player.getUser().clearEffect(RoomUser.EFFECT_GAME);
			}
		}
	}
	
	protected abstract void onStart();
	
	protected void onStartDelayExpired() { }
	
	protected abstract void onStop(GameTeam winner);
	
	protected abstract void onPulse();
	
	public int getId()
	{
		return this.id;
	}

	public RoomInstance getRoom()
	{
		return this.room;
	}
	
	public Collection<RoomFloorItem> getItems()
	{
		// TODO: caching?
		return this.room.getFloorItems();
	}
	
	public TimerItem getTimer()
	{
		for(RoomFloorItem item : this.getItems())
		{
			if(item instanceof TimerItem)
			{
				return (TimerItem)item;
			}
		}
		
		return null;
	}
	
	public GameTeam getTeam(GameTeam.ETeam teamId)
	{
		return this.teams[teamId.ordinal()];
	}
	
	public GameTeam[] getTeams()
	{
		return this.teams;
	}
	
	public GameTeam getTeamWithBestScore()
	{
		// Find team with best score
		GameTeam best = null;
		for(GameTeam team : this.teams)
		{
			if(team.getScore() > 0 && (best == null || team.getScore() > best.getScore()))
			{
				best = team;
			}
		}
		
		// Detect draws
		if(best != null)
		{
			for(GameTeam team : this.teams)
			{
				if(team != best && team.getScore() == best.getScore())
				{
					best = null;
					break;
				}
			}
		}
		
		// Return best team (null = no best team / draw)
		return best;
	}
	
	public int getTimeLeft()
	{
		return this.timeLeft;
	}
	
	public int getRequiredTeamsWithPlayers()
	{
		return 2;
	}
	
	public boolean isAlive()
	{
		return this.alive;
	}
	
	public void setAlive(boolean alive)
	{
		this.alive = alive;
	}
	
	public boolean isRunning()
	{
		return (this.alive && this.timeLeft > 0 && this.startDelay == 0);
	}

	public void setTimeLeft(int seconds)
	{
		this.timeLeft = seconds;
		
		// Update timer
		TimerItem timer = this.getTimer();
		if(timer != null)
		{
			timer.setTimeLeft(this.timeLeft);
		}
	}
	
	public void setStartDelay(int seconds)
	{
		this.startDelay = seconds;
	}
	
	public abstract GameUserPlayer.GameValues createValuesForPlayer();
}
