package habbod2.game.rooms.minigames;

import habbod2.game.rooms.instances.RoomInstance;
import habbod2.game.rooms.items.RoomFloorItem;
import habbod2.game.rooms.items.minigame.ScoreboardItem;
import habbod2.game.rooms.items.minigame.TimerItem;
import habbod2.game.rooms.items.wired.triggers.GameEndsTrigger;
import habbod2.game.rooms.items.wired.triggers.GameStartsTrigger;
import habbod2.game.util.GameUtil;

import habbod2.game.rooms.items.banzai.*;

public class OldGame
{
	private static final int MIN_TEAMS_WITH_PLAYERS = 2;
	
	private int id;
	private RoomInstance room;
	private GameTeam[] teams;
	private boolean isRunning;
	private int timeLeft;
	private boolean pulse;
	
	// Banzai-specific
	private int startDelay;
	private int sealableTileCount;
	
	public OldGame(RoomInstance room)
	{
		// Set room
		this.room = room;
		
		// Get all team IDs
		GameTeam.ETeam[] teamIds = GameTeam.ETeam.values();
		
		// Create all empty teams
		this.teams = new GameTeam[teamIds.length];
		for(int i = 0; i < teamIds.length; i++)
		{
			this.teams[i] = new GameTeam(teamIds[i], this);
		}
	}
	
	public void destroy()
	{
		this.stop();
		for(GameTeam team : this.teams)
		{
			for(GameUserPlayer player : team.getPlayers())
			{
				player.getUser().setPlayer(null);
				player.getUser().setEffect(0, true);
			}
		}
	}
	
	public void start()
	{
		// Not already running?
		if(this.isRunning)
		{
			return;
		}
		
		// Requires timer
		TimerItem timer = this.getTimer();
		if(timer == null || timer.getTimeLeft() == 0)
		{
			return;
		}
		
		// Requires atleast x teams with players
		int teamsWithPlayers = 0;
		for(GameTeam team : this.teams)
		{
			if(team.getPlayers().size() > 0)
			{
				teamsWithPlayers++;
				continue;
			}
		}
		if(teamsWithPlayers < MIN_TEAMS_WITH_PLAYERS)
		{
			return;
		}
		
		// Reset and process items
		this.sealableTileCount = 0;
		for(RoomFloorItem item : this.room.getFloorItems())
		{
			if(item instanceof ScoreboardItem)
			{
				((ScoreboardItem)item).setScore(0);
			}
			else if(item instanceof GameStartsTrigger)
			{
				((GameStartsTrigger)item).execute(null);
			}
			else if(item instanceof TileItem)
			{
				this.sealableTileCount++;
				((TileItem)item).setState(true, null, TileItem.LEVEL_EMPTY);
			}
			else if(item instanceof ApparatusItem)
			{
				((ApparatusItem)item).enable();
				this.startDelay = 3;
			}
		}
		
		// Start new game!
		this.id = GameUtil.getRandomInt(1, Integer.MAX_VALUE);
		this.timeLeft = timer.getTimeLeft();
		this.isRunning = true;
	}
	
	public void stop()
	{
		// Actually running?
		if(!this.isRunning)
		{
			return;
		}
		
		// Stopped!
		this.isRunning = false;
		this.timeLeft = 0;
		
		// Determine winning team
		GameTeam winner = null;
		for(GameTeam team : this.teams)
		{
			if(winner == null || team.getScore() > winner.getScore())
			{
				winner = team;
			}
		}
		
		// Is there a winning team?
		if(winner != null)
		{
			// Waving is for winners!
			for(GameUserPlayer player : winner.getPlayers())
			{
				player.getUser().wave();
			}
		}
		
		// Reset items etc
		for(RoomFloorItem item : this.room.getFloorItems())
		{
			if(item instanceof TimerItem)
			{
				((TimerItem)item).setTimeLeft(0);
			}
			else if(item instanceof GameEndsTrigger)
			{
				((GameEndsTrigger)item).execute(null);
			}
			else if(item instanceof TileItem)
			{
				TileItem tile = (TileItem)item;
				if(winner != null && tile.getTeamId() == winner.getId())
				{
					tile.flash(20);
				}
				else
				{
					tile.disable(20);
				}
			}
			else if(item instanceof ApparatusItem)
			{
				ApparatusItem apparatus = (ApparatusItem)item;
				if(winner != null)
				{
					apparatus.showWinner(winner.getId());
				}
				else
				{
					apparatus.disable();
				}
			}
		}
		
		// Reset scores etc
		for(GameTeam team : this.teams)
		{
			team.reset();
		}
	}
	
	public void pulse()
	{
		// Actually running?
		if(!this.isRunning)
		{
			return;
		}
		
		// Run every odd pulse
		this.pulse = !this.pulse;
		if(this.pulse)
		{
			// Decrement time left?
			if(this.startDelay == 0 || --this.startDelay <= 0)
			{
				this.setTimeLeft(this.timeLeft - 1);
				if(this.timeLeft <= 0)
				{
					this.stop();
				}
			}
		}
	}
	
	public int getId()
	{
		return this.id;
	}

	public RoomInstance getRoom()
	{
		return this.room;
	}
	
	public TimerItem getTimer()
	{
		for(RoomFloorItem item : this.room.getFloorItems())
		{
			if(item instanceof TimerItem)
			{
				return (TimerItem)item;
			}
		}
		
		return null;
	}
	
	public GameTeam getTeam(GameTeam.ETeam teamId)
	{
		return this.teams[teamId.ordinal()];
	}
	
	public int getTimeLeft()
	{
		return this.timeLeft;
	}
	
	public boolean isRunning()
	{
		return (this.isRunning && this.startDelay == 0);
	}

	public void setTimeLeft(int seconds)
	{
		this.timeLeft = seconds;
		
		// Update timer
		TimerItem timer = this.getTimer();
		if(timer != null)
		{
			timer.setTimeLeft(this.timeLeft);
		}
	}
	
	public void sealedTile()
	{
		this.sealableTileCount--;
		if(this.sealableTileCount == 0)
		{
			this.stop();
		}
	}
}
