package habbod2.game.rooms.minigames.game;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import habbod2.System;
import habbod2.storage.Database;
import habbod2.game.rooms.instance.RoomInstance;
import habbod2.game.rooms.items.RoomFloorItem;
import habbod2.game.rooms.items.minigame.banzai.ApparatusItem;
import habbod2.game.rooms.items.minigame.banzai.GameTileItem;
import habbod2.game.rooms.minigames.Game;
import habbod2.game.rooms.minigames.GameTeam;
import habbod2.game.rooms.minigames.GameUserPlayer;
import habbod2.game.rooms.minigames.GameUserPlayer.GameValues;

public class BanzaiGame extends Game
{
	private List<GameTileItem> pending = new ArrayList<GameTileItem>();
	private int gameEndedTimer;
	
	public BanzaiGame(RoomInstance room)
	{
		super(room);
	}
	
	protected void onStart()
	{
		this.setStartDelay(3);
		
		// Enable apparatus
		for(RoomFloorItem item : this.getItems())
		{
			if(item instanceof ApparatusItem)
			{
				((ApparatusItem) item).enable();
			}
		}
	}
	
	protected void onStartDelayExpired()
	{
		// Enable tiles (lights on)
		for(RoomFloorItem item : this.getItems())
		{
			if(item instanceof GameTileItem)
			{
				((GameTileItem)item).update(true, null, GameTileItem.LEVEL_EMPTY);
			}
		}
	}

	@Override
	protected void onStop(GameTeam winner)
	{		
		// Clear game ended timer
		this.gameEndedTimer = 0;
		
		// Update game wins?
		Database db = System.get().getDb();
		if(winner != null)
		{
			for(GameUserPlayer player : winner.getPlayers())
			{
				db.execute("UPDATE users SET bb_wins=bb_wins + 1 WHERE id=" + player.getUser().getInfo().id);
			}
		}
		for(GameTeam team : this.getTeams())
		{
			for(GameUserPlayer player : team.getPlayers())
			{
				BanzaiValues values = (BanzaiValues)player.getValues();
				if(values.sealedTileCount > 0)
				{
					db.execute("UPDATE users SET bb_tiles=bb_tiles + " + values.sealedTileCount + " WHERE id=" + player.getUser().getInfo().id);
				}
			}
		}
	}
	
	@Override
	protected void onPulse()
	{
		// Only process tiles when the game is actually running
		if(this.isRunning())
		{
			// Any pending tiles?
			if(!this.pending.isEmpty())
			{
				// Go through all the tiles
				Iterator<GameTileItem> it = this.pending.iterator();
				while(it.hasNext())
				{
					// Get tile
					GameTileItem tile = it.next();
					
					// Tile is ready?
					if(!tile.updatePendingTimer())
					{
						continue;
					}
					
					// Get player that moved over this tile
					GameUserPlayer player = tile.getPendingPlayer();
					if(player == null)
					{
						continue;
					}
					
					// Get player values
					BanzaiValues values = (BanzaiValues)player.getValues();
					if(values == null)
					{
						continue;
					}
					
					// Tile is not sealed yet?
					if(tile.getLevel() < GameTileItem.LEVEL_SEALED)
					{
						// Update the tile
						if(tile.getTeamId() != player.getTeam().getId())
						{
							tile.update(true, player.getTeam().getId(), GameTileItem.LEVEL_EMPTY);
						}
						else
						{
							tile.update(true, player.getTeam().getId(), tile.getLevel() + 1);
						}
						
						// Tile is now sealed?
						if(tile.getLevel() == GameTileItem.LEVEL_SEALED)
						{
							// Scored a point!
							player.addScore(1);
							values.sealedTileCount++;
						}
					}
					
					// Processed!
					it.remove();
					tile.clearPendingPlayer();
				}
			}
			
			// Count remaining tiles and show winner 
			int sealableTiles = 0;
			GameTeam leader = this.getTeamWithBestScore();
			for(RoomFloorItem item : this.getItems())
			{
				if(item instanceof GameTileItem)
				{
					GameTileItem tile = (GameTileItem)item;
					if(tile.getLevel() < GameTileItem.LEVEL_SEALED)
					{
						sealableTiles++;
					}
				}
				else if(item instanceof ApparatusItem)
				{
					ApparatusItem apparatus = (ApparatusItem)item;
					apparatus.showLeader((leader != null) ? leader.getId() : null);
				}
			}
			
			// Game ended?
			if(sealableTiles == 0)
			{
				this.stop();
			}
		}
		
		// Game ending phase?
		boolean gameEnded = (this.gameEndedTimer == 6);
		if(this.getTimeLeft() == 0)
		{
			// Tick!
			this.gameEndedTimer++;
			
			// Get winner
			GameTeam winner = this.getTeamWithBestScore();
			if(winner != null || gameEnded)
			{
				for(RoomFloorItem item : this.getItems())
				{
					if(item instanceof GameTileItem)
					{
						GameTileItem tile = (GameTileItem)item;
						if(gameEnded)
						{
							tile.update(false, null, GameTileItem.LEVEL_EMPTY);
						}
						else
						{
							if(winner != null && tile.getTeamId() == winner.getId())
							{
								int displayLevel = (this.gameEndedTimer % 2 == 0) ? tile.getLevel() : GameTileItem.LEVEL_EMPTY;
								tile.setState(GameTileItem.generateState(true, tile.getTeamId(), displayLevel));
							}
						}
					}
					else if(gameEnded && item instanceof ApparatusItem)
					{
						((ApparatusItem)item).disable();
					}
				}
			}
			
			// Game really ended?
			if(gameEnded)
			{
				this.setAlive(false);
			}
		}
	}
	
	public void tileUpdated(GameTileItem tile)
	{
		if(!this.pending.contains(tile))
		{
			this.pending.add(tile);
		}
	}
	
	@Override
	public GameValues createValuesForPlayer()
	{
		return new BanzaiValues();
	}
	
	public class BanzaiValues implements GameUserPlayer.GameValues
	{
		public GameTileItem pendingTile;
		public GameTileItem lastTile;
		public int sealedTileCount;
	}
}
