package habbod2.game.rooms.navigator;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;

import habbod2.game.GameManager;
import habbod2.game.rooms.Room;
import habbod2.storage.DatabaseResult;

public class RoomNavigator extends GameManager.Component
{
	private long lastRefreshTime;
	private long maxAgeTime;
	private AtomicBoolean isRefreshing;
	
	private Map<Integer, RoomCategory> categories;
	private List<NavigatorFrontpageItem> frontpage;
	private List<Room> busiestGuestRooms;
	private List<Room> topRatedGuestRooms;
	private int staffPickedCategoryId;
	
	public RoomNavigator(GameManager gameMgr)
	{
		super(gameMgr);
		this.categories = Collections.emptyMap();
		this.frontpage = Collections.emptyList();
		this.busiestGuestRooms = new ArrayList<Room>();
		this.topRatedGuestRooms = new ArrayList<Room>();
		this.isRefreshing = new AtomicBoolean();
	}

	public void reload()
	{
		// Create new collections
		Map<Integer, RoomCategory> tmpCategories = new HashMap<Integer, RoomCategory>();
		List<NavigatorFrontpageItem> tmpFrontpage = new ArrayList<NavigatorFrontpageItem>();

		// Load categories from database
		try(DatabaseResult result = this.getDb().query("SELECT * FROM navigator_categories"))
		{
			while(result.row())
			{
				if(result.getBoolean("enabled"))
				{
					RoomCategory category = new RoomCategory();
					category.id = result.getInt("id");
					category.name = result.getString("name");
					category.requiredRight = result.getString("required_right");
					if(category.requiredRight == null || category.requiredRight.isEmpty())
					{
						category.requiredRight = "default";
					}
					tmpCategories.put(category.id, category);
				}
			}
	
			// Load frontpage items from database
			result.query("SELECT * FROM navigator_frontpage ORDER BY ordernum ASC");
			while(result.row())
			{
				// Enabled?
				if(result.getBoolean("enabled"))
				{
					// Parse entry
					NavigatorFrontpageItem fpi = new NavigatorFrontpageItem();
					fpi.id = result.getInt("id");
					fpi.parentId = result.getInt("parent_id");
					fpi.name = result.getString("name");
					fpi.description = result.getString("description");
					fpi.showDetails = result.getString("image_type").equals("details");
					fpi.autoExpand = result.getBoolean("auto_expand");
					fpi.imageLabel = result.getString("image_label");
					fpi.imageExternal = result.getString("image_external");
					
					// Get inner item
					String innerType = result.getString("type");
					int innerId = result.getInt("inner_id");
					if(innerType.equals("room"))
					{
						fpi.room = this.getGameMgr().getRoomMgr().getRoomInfo(innerId);
					}
					else
					{
						fpi.category = tmpCategories.get(innerId);
					}
	
					// Inner item exists? (don't enable invalid frontpage items)
					if(fpi.room != null || fpi.category != null)
					{
						tmpFrontpage.add(fpi);
					}
					else
					{
						this.getLogger().warning(this, String.format("Frontpage item #%d (\"%s\") refers to the missing %s #%d. Item ignored.", fpi.id, fpi.name, innerType, innerId));
					}
				}
			}
		}
		
		// Load staffpicked rooms
		/*
		result.query("SELECT id FROM rooms WHERE staffpick=1");
		while(result.row()))
		{
			NavigatorFrontpageItem fpi = 
		}*/

		// Set instance fields
		this.categories = Collections.unmodifiableMap(tmpCategories);
		this.frontpage = Collections.unmodifiableList(tmpFrontpage);

		// Load config
		this.maxAgeTime = this.getConfig().getInt("game.rooms.navigator.refreshtime", 5) * 1000;
		this.staffPickedCategoryId = this.getConfig().getInt("game.rooms.navigator.staffpicks.category");
		
		// Done!
		this.getLogger().info(this, String.format("Loaded %d guestroom categories and %d frontpage items.", this.categories.size(), this.frontpage.size()));
	}
	
	private void refresh()
	{
		// Refreshing!
		this.isRefreshing.set(true);
		
		// Refresh categories
		for(RoomCategory category : this.categories.values())
		{
			// Refresh rooms in this category
			category.rooms.clear();
			try(DatabaseResult result = this.getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE category_id=" + category.id + " ORDER BY users_now DESC,rating DESC LIMIT 50"))
			{
				while(result.row())
				{
					category.rooms.add(Room.parse(result));
				}
			}
		}
		
		// Refresh rooms that are on the frontpage
		Iterator<NavigatorFrontpageItem> iterator = this.frontpage.iterator();
		while(iterator.hasNext())
		{
			// Is this a room?
			NavigatorFrontpageItem fpi = iterator.next();
			if(fpi.room != null)
			{
				// Get fresh room info, remove the frontpage item if the room was deleted
				Room newInfo = this.getGameMgr().getRoomMgr().getRoomInfo(fpi.room.id);
				if(newInfo != null)
				{
					fpi.room = newInfo;
				}
				else
				{
					iterator.remove();
				}
			}
		}
		
		// Refresh busiest rooms
		this.busiestGuestRooms.clear();
		try(DatabaseResult result = this.getDb().query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id WHERE r.users_now > 0 ORDER BY users_now DESC LIMIT 50"))
		{
			while(result.row())
			{
				this.busiestGuestRooms.add(Room.parse(result));
			}
	
			// Refresh toprated rooms
			this.topRatedGuestRooms.clear();
			result.query("SELECT r.*,u.username AS owner FROM rooms r JOIN users u ON u.id=r.owner_id ORDER BY rating DESC LIMIT 50");
			while(result.row())
			{
				this.topRatedGuestRooms.add(Room.parse(result));
			}
		}

		// This is for debug only
		this.lastRefreshTime = System.currentTimeMillis();
		this.isRefreshing.set(false);
		this.getLogger().debug(this, "Refreshed successfully.");
	}
	
	private void checkAge()
	{
		if(this.isRefreshing.get() == false)
		{
			long ageTime = System.currentTimeMillis() - this.lastRefreshTime;
			if(ageTime > this.maxAgeTime)
			{
				this.refresh();
			}
		}
	}
	
	public RoomCategory getStaffPicksCategory()
	{
		if(this.staffPickedCategoryId > 0)
		{
			return this.getCategory(this.staffPickedCategoryId);
		}
		else
		{
			return null;
		}
	}

	public RoomCategory getCategory(int categoryId)
	{
		this.checkAge();
		return this.categories.get(categoryId);
	}
	
	public Collection<RoomCategory> getCategories()
	{
		this.checkAge();
		return this.categories.values();
	}

	public List<NavigatorFrontpageItem> getFrontpageItems()
	{
		this.checkAge();
		return this.frontpage;
	}

	public List<Room> getBusiestGuestRooms()
	{
		this.checkAge();
		return this.busiestGuestRooms;
	}

	public List<Room> getTopRatedGuestRooms()
	{
		this.checkAge();
		return this.topRatedGuestRooms;
	}
}
