package habbod2.game.sessions;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import habbod2.System;
import habbod2.communication.DeferredRequestHandler;
import habbod2.communication.DeferredRequestHandlerExecutor;
import habbod2.communication.EClientMessage;
import habbod2.communication.RequestHandler;
import habbod2.communication.RequestRegistry;
import habbod2.net.protocol.ClientMessage;
import habbod2.net.protocol.ServerMessage;
import habbod2.events.Event;
import habbod2.game.GameManager;
import habbod2.game.items.ItemInventory;
import habbod2.game.items.pets.PetInventory;
import habbod2.game.items.trading.Trade;
import habbod2.game.messenger.InstantMessenger;
import habbod2.game.moderation.ModerationBan;
import habbod2.game.moderation.ModerationPresetManager;
import habbod2.game.moderation.ModerationTicket;
import habbod2.game.rooms.Room;
import habbod2.game.rooms.RoomManager;
import habbod2.game.rooms.instances.RoomAvatar;
import habbod2.game.rooms.instances.RoomInstance;
import habbod2.game.rooms.instances.RoomUser;
import habbod2.game.rooms.navigator.FavoriteRoomList;
import habbod2.game.rooms.navigator.RoomEvent;
import habbod2.game.rooms.navigator.RoomEventManager;
import habbod2.game.users.User;
import habbod2.game.users.badges.BadgeInventory;
import habbod2.game.users.effects.EffectInventory;
import habbod2.net.NetConnectionHandler;

import habbod2.communication.messages.outgoing.handshake.*;
import habbod2.communication.messages.outgoing.inventory.avatareffect.*;
import habbod2.communication.messages.outgoing.inventory.purse.*;
import habbod2.communication.messages.outgoing.moderator.*;
import habbod2.communication.messages.outgoing.navigator.*;
import habbod2.communication.messages.outgoing.notifications.*;
import habbod2.communication.messages.outgoing.room.action.*;
import habbod2.communication.messages.outgoing.room.chat.*;
import habbod2.communication.messages.outgoing.room.engine.*;
import habbod2.communication.messages.outgoing.room.permissions.*;
import habbod2.communication.messages.outgoing.room.session.*;

public class Session implements NetConnectionHandler.CommunicationHandler
{
	// Events
	public static final Event onAuthenticate = new Event("Session.OnAuthenticate");
	public static final Event onLogin = new Event("Session.OnLogin");
	public static final Event onLogout = new Event("Session.OnLogout");
	public static final Event onRebuildRightsList = new Event("User.OnRebuildRightsList");
	public static final Event onRoomEnter = new Event("Session.OnRoomEnter");
	public static final Event onRoomExit = new Event("Session.OnRoomExit");
	public static final Event onChat = new Event("Session.OnChat");

	// Core
	public long id;
	private NetConnectionHandler conn;
	private SessionManager sessionMgr;

	// Message handling
	private long lastMsgTime;
	public boolean pongOk;

	// Session variables
	private User user;
	private int accessLogId;
	private Map<Integer, String> ignoredUsers;
	private InstantMessenger messenger;
	private BadgeInventory badgeInventory;
	private ItemInventory itemInventory;
	private PetInventory petInventory;
	private EffectInventory effectInventory;
	private FavoriteRoomList favoriteRooms;
	private List<Integer> ratedRooms;
	
	// Room instance related variables
	private RoomInstance roomInstance;
	private RoomUser roomUser;
	private RoomEvent roomEvent;
	private Trade trade;
	
	public int authenticatedRoomId;
	public int authenticatedTeleporterId;
	public boolean waitingForDoorbell;
	public long chatTimeoutExpireTime;

	public Session(long sessionId, NetConnectionHandler conn, SessionManager sessionMgr)
	{
		this.id = sessionId;
		this.conn = conn;
		this.sessionMgr = sessionMgr;
	}

	public void end(String reason)
	{
		this.sessionMgr.endSession(this.id, reason);
	}
	
	public void onEnd()
	{
		this.conn.close();
		if(this.authenticated())
		{
			this.onLogout();
			this.sessionMgr.removeAuthedSession(this);
		}
	}

	public void authenticate(String ticket)
	{
		// Get game manager
		GameManager gameMgr = this.sessionMgr.getGameMgr();
				
		// User exists for this auth ticket?
		User user = gameMgr.getUserMgr().getUserByAuthTicket(ticket);
		if(user == null)
		{
			this.end("invalid auth ticket");
			return;
		}

		// Check for any bans
		List<ModerationBan> bans = gameMgr.getModerationCenter().getActiveBansForUserOrIp(user.id, this.conn.getIpAddress());
		if(!bans.isEmpty())
		{
			// Send the first ban?
			
			// User is banned!
			this.end("user is banned / IP banned");
			return;
		}
		
		// Run auth checks
		if(!Session.onAuthenticate.fire(this, user))
		{
			return;
		}

		// Insufficient permissions?
		if(user.role == null || !user.role.rights.contains("fuse_login"))
		{
			this.end(String.format("user #%d is not allowed to login because of insufficient role privileges", user.id));
			return;
		}

		// Add to session list
		this.user = user;
		gameMgr.getSessionMgr().addAuthedSession(this);

		// Send user right list
		this.rebuildAndSendUserRights();

		// Enable info feed
		this.send(InfoFeedEnableMessageComposer.compose(true,  false));
		
		// Login OK!
		this.send(AuthenticationOKMessageComposer.compose());

		// Run other logic
		this.onLogin();
	}

	private void onLogin()
	{
		// Get game manager
		GameManager gameMgr = this.sessionMgr.getGameMgr();
		
		// Reload ignored users
		this.ignoredUsers = gameMgr.getUserMgr().getIgnoredUsersForUser(this.user.id);
		
		// Reload messenger
		this.messenger = new InstantMessenger(this);
		this.messenger.reload();
		
		// Reload badges
		this.badgeInventory = new BadgeInventory(this);
		this.badgeInventory.reload();
		
		// Reload item inventory
		this.itemInventory = new ItemInventory(this);
		this.itemInventory.reload();
		
		// Reload pet inventory
		this.petInventory = new PetInventory(this);
		this.petInventory.reload();
		
		// Reload tags
		this.user.tags = gameMgr.getUserMgr().getTagsForUser(this.user.id);
		
		// Reload effects
		this.effectInventory = new EffectInventory(this);
		this.effectInventory.reload();
		
		// Reload favorite rooms
		this.favoriteRooms = new FavoriteRoomList(this);
		this.favoriteRooms.reload();
		
		// Reload rated rooms
		this.ratedRooms = new ArrayList<Integer>();
		
		// Reload achievements
		this.user.achievements = gameMgr.getAchievementMgr().getAchievementsForUser(this.user.id);

		// Update last online time
		this.user.updateLastOnline(true);

		// Push navigator settings
		this.send(NavigatorSettingsMessageComposer.compose(this.user.homeRoomId));
		
		// Push favorite rooms
		this.send(FavouriteRoomsMessageComposer.compose(this.favoriteRooms.getFavorites()));
		
		// Push effects
		this.send(AvatarEffectsMessageComposer.compose(this.effectInventory.getEffects()));
		
		// User is moderator?
		if(this.user.rights.contains("fuse_mod"))
		{
			// Send preset moderation responses
			ModerationPresetManager presets = gameMgr.getModerationCenter().getPresetManager();
			this.send(ModerationPresetsMessageComposer.compose(presets.getUserMessages(), presets.getRoomMessages()));
			
			// Send open tickets
			for(ModerationTicket ticket : gameMgr.getModerationCenter().getAllTickets())
			{
				if(ticket.status == ModerationTicket.STATUS_OPEN || ticket.status == ModerationTicket.STATUS_PICKED)
				{
					this.send(ModerationTicketMessageComposer.compose(ticket));
				}
			}
		}
		
		// Execute other on-login tasks
		Session.onLogin.fire(this);
	}

	private void onLogout()
	{
		// Run hooks
		Session.onLogout.fire(this);

		// Stop trading, leave the room, etc etc
		this.leaveRoomSilent();
		
		// End the room event, host has gone offline!
		if(this.roomEvent != null)
		{
			RoomEventManager roomEventMgr = this.sessionMgr.getGameMgr().getRoomEventMgr();
			roomEventMgr.removeEvent(this.roomEvent);
		}

		// Go offline!
		this.messenger.offline();
		
		// Clear inventory to speed up garbagecollection
		this.itemInventory.clear();
		
		// Save badge inventory
		if(this.badgeInventory.requiresSave())
		{
			this.badgeInventory.save();
		}
		this.badgeInventory.clear();

		// Update last online time
		this.user.updateLastOnline(false);

		// Save misc things
		this.user.saveMisc();
	}

	public void tryGetRoomAccess(int roomId, String password)
	{
		// Leave any room instance the user might be in now
		this.leaveRoomSilent();

		// Get/create room instance
		RoomManager roomMgr = this.sessionMgr.getGameMgr().getRoomMgr();
		RoomInstance room = roomMgr.getRoomInstance(roomId, true);
		if(room == null)
		{
			return;
		}
		
		// Connection to roomserver is open!
		this.send(OpenConnectionMessageComposer.compose());
		
		// The room owner can ALWAYS enter the room, never has to check for user limit, bans, state, etc
		boolean skipAccessChecks = (this.getUser().id == room.getInfo().ownerId) || this.getUser().rights.contains("fuse_can_enter_any_room") || this.authenticatedTeleporterId > 0;
		if(!skipAccessChecks)
		{
			int errorCode = 0;
			
			// Room is full?
			if(room.getInfo().usersNow >= room.getInfo().usersMax)
			{
				// Can't override user limit?
				if(!this.getUser().rights.contains("fuse_enter_full_rooms"))
				{
					errorCode = 1;
				}
			}

			// User is banned?
			else if(room.userIsBanned(this.getUser().id))
			{
				errorCode = 4;
			}
			
			// Any errors?
			if(errorCode != 0)
			{
				this.send(CantConnectMessageComposer.compose(errorCode));
				this.kickFromRoom();
				return;
			}

			// Password room?
			if(room.getInfo().state == Room.STATE_PASSWORD && !room.getInfo().password.equals(password))
			{
				this.sendGenericError(-100002);
				this.kickFromRoom();
				return;
			}

			// Doorbell room?
			else if(room.getInfo().state == Room.STATE_DOORBELL)
			{
				// Ring the doorbell, await response
				if(room.ringDoorbell(this))
				{
					// Cool, let's wait
					this.waitingForDoorbell = true;

					// Waiting...
					this.send(DoorbellMessageComposer.compose(null));
				}
				else
				{
					this.send(FlatAccessDeniedMessageComposer.compose());
					this.kickFromRoom();
				}
				return;
			}
		}
		
		// Guestroom accessible!
		if(room.getInfo().isGuestRoom())
		{
			this.send(FlatAccessibleMessageComposer.compose());
		}

		// Forward to room instance
		this.authenticatedRoomId = roomId;
	}

	public void goToRoom(int roomId)
	{
		// Get the running instance, it should be there by now...
		RoomManager roomMgr = this.sessionMgr.getGameMgr().getRoomMgr();
		RoomInstance instance = roomMgr.getRoomInstance(roomId, false);
		if(instance == null)
		{
			return;
		}
		
		// This is the current room instance now
		this.roomInstance = instance;

		// Room is ready!
		this.send(RoomReadyMessageComposer.compose(instance));
		
		// Send maps
		this.send(HeightMapMessageComposer.compose(instance));
		this.send(FloorHeightMapMessageComposer.compose(instance));
		
		// Send remaining chat timeout seconds
		if(this.chatTimeoutExpireTime > 0)
		{
			int remainingSeconds = (int) ((this.chatTimeoutExpireTime - System.currentTimeMillis()) / 1000);
			if(remainingSeconds > 0)
			{
				this.send(FloodControlMessageComposer.compose(remainingSeconds));
			}
		}
		
		// If entering a guestroom, then send more data, client doesn't ask for it
		Room room = instance.getInfo();
		if(room.isPublicRoom())
		{
			this.send(PublicRoomObjectsMessageComposer.compose(null));
		}
		else if(room.isGuestRoom())
		{
			// Send wallpaper/floor/landscape if not default data
			if(!room.wallpaper.isEmpty())
			{
				this.send(RoomPropertyMessageComposer.compose("wallpaper", room.wallpaper));
			}
			if(!room.floor.isEmpty())
			{
				this.send(RoomPropertyMessageComposer.compose("floor", room.floor));
			}
			if(!room.landscape.isEmpty())
			{
				this.send(RoomPropertyMessageComposer.compose("landscape", room.landscape));
			}
			
			// Send room event
			if(instance.getEvent() != null)
			{
				this.send(RoomEventMessageComposer.compose(instance.getEvent()));
			}
			
			// Send room rating
			if(this.getRatedRooms().contains(room.id) || this.getUser().id == room.ownerId)
			{
				this.send(RoomRatingMessageComposer.compose(room.rating));
			}
			else
			{
				this.send(RoomRatingMessageComposer.compose(RoomRatingMessageComposer.NOT_RATED_YET));
			}
			
			// Send visualization settings
			this.send(RoomVisualizationSettingsMessageComposer.compose(instance));

			// Send items
			this.send(ObjectsMessageComposer.compose(instance.getFloorItems()));
			this.send(ItemsMessageComposer.compose(instance.getWallItems()));
		}
		
		// Send avatars
		Collection<RoomAvatar> avatars = instance.getAllAvatars();
		this.send(UsersMessageComposer.compose(avatars));

		// Try spawning room user
		if(this.enterRoom())
		{
			this.sendRoomPermissions();
		}
		else
		{
			return;
		}

		// Send statuses of all avatars
		this.send(UserUpdateMessageComposer.compose(avatars));
		
		// Send other special user statuses
		for(RoomUser user : instance.getAllUsers())
		{
			if(user.isAsleep)
			{
				this.send(SleepMessageComposer.compose(user));
			}
			if(user.danceId > 0)
			{
				this.send(DanceMessageComposer.compose(user));
			}
			if(user.isCarryingItem())
			{
				this.send(CarryObjectMessageComposer.compose(user));
			}
			if(user.getCurrentEffect() > 0)
			{
				this.send(AvatarEffectMessageComposer.compose(user));
			}
		}
	}

	public boolean enterRoom()
	{
		this.roomUser = this.roomInstance.addSession(this);
		if(this.roomUser != null)
		{
			if(Session.onRoomEnter.fire(this))
			{
				RoomManager mgr = System.get().getGameMgr().getRoomMgr();
				mgr.logRoomEntry(this.user.id, this.roomInstance.getInfo().id);
				
				this.messenger.updateStatus(false);
				
				return true;
			}
			else
			{
				// Silent kick (a hook specified that roomentry should not continue)
				this.kickFromRoom();
			}
		}
		else
		{
			// Hacky entry?
			this.kickFromRoom("Invalid room entry! Room probably was being unloaded while you were entering! Please try again.");
		}

		// Do not enter room
		return false;
	}

	private boolean leaveRoomSilent()
	{
		// Reset room ID
		if(this.authenticatedTeleporterId == 0)
		{
			this.authenticatedRoomId = 0;
		}

		// Leave room instance
		if(this.roomInstance != null)
		{
			if(Session.onRoomExit.fire(this))
			{
				RoomManager mgr = System.get().getGameMgr().getRoomMgr();
				mgr.logRoomExit(this.user.id, this.roomInstance.getInfo().id);
				
				this.roomInstance.removeSession(this);
				this.roomInstance = null;
				this.roomUser = null;
				
				// End trade
				if(this.trade != null)
				{
					trade.end(this);
				}
				this.trade = null;
				
				// Update messenger
				this.messenger.updateStatus(false);
			}
			else
			{
				// Did not leave room
				return false;
			}
		}

		// Left room!
		return true;
	}

	public void kickFromRoom()
	{
		this.kickFromRoom(null);
	}

	public void kickFromRoom(String text)
	{
		this.kickFromRoom(text, false);
	}
	
	public void kickFromRoom(String text, boolean isCaution)
	{
		// SHE GAVE ME THE PLEASE-FACE
		if(!this.leaveRoomSilent())
		{
			// Room exit cancelled
			return;
		}

		// I GAVE HER THE LEAVE-FACE
		this.send(CloseConnectionMessageComposer.compose());

		// BODY LIKE A BOTTLE, PRETTY LITTLE MODEL MODEL MODEEL
		if(text != null && !text.isEmpty())
		{
			this.sendNotif(text, isCaution);
		}
	}

	@Override
	public void onNetNewMessage(ClientMessage msg)
	{
		// Fucking heartbeat! ("D{")
		if(msg.type == 315) return;

		// Log that this session received a message!
		this.lastMsgTime = System.get().getTime();
		
		// Message contains protocol-EOF byte? This is _never_ acceptable because of exploitable purposes
		if(msg.contains(ServerMessage.MSG_EOF))
		{
			this.sessionMgr.endSession(this.id, "MSG_EOF detected in incoming message", true);
			return;
		}

		// These are the only responding handlers when not authed
		if(!this.authenticated() && (msg.type != EClientMessage.InitCrypto && msg.type != EClientMessage.SSOTicket))
		{
			this.sessionMgr.getLogger().warning(this, String.format("msg #%d (\"%s\") is an invalid pre-auth msg.", msg.type, msg.header()));
		}
		else
		{
			// Lookup message handler and execute handling logic
			String name = RequestRegistry.getRequestName(msg.type);
			RequestHandler handler = RequestRegistry.getRequestHandler(msg.type);
			if(handler != null)
			{
				this.sessionMgr.getLogger().debug(this, "<-- " + msg.toString() + " (" + name + ")");
				try
				{
					if(handler instanceof DeferredRequestHandler)
					{
						// Handle in separate thread
						DeferredRequestHandlerExecutor.submit(this, msg, (DeferredRequestHandler) handler);
					}
					else
					{
						// Handle in current thread (Netty I/O)
						handler.handle(this, msg);
					}
				}
				catch(Exception ex)
				{
					this.sessionMgr.getLogger().error(this, String.format("Error handling msg #%d (\"%s\").", msg.type, name), ex);
				}
			}
			else
			{
				//this.sessionMgr.getLogger().debug(this, "<-- " + msg.toString());
				this.sessionMgr.getLogger().debug(this, String.format("RequestRegistry misses #%d. (\"%s\" -> %s)", msg.type, msg.header(), name));
			}
		}
	}

	@Override
	public void onNetClose(String reason)
	{
		this.end(reason);
	}

	public void send(ServerMessage msg)
	{
		System.get().getLogger().debug(this, "--> " + msg.toString());
		this.conn.send(msg);
	}

	public void sendGenericError(int code)
	{
		this.send(GenericErrorMessageComposer.compose(code));
	}

	public void sendNotif(String text)
	{
		this.sendNotif(text, null, false);
	}

	public void sendNotif(String text, boolean isCaution)
	{
		this.sendNotif(text, null, isCaution);
	}
	
	public void sendNotif(String text, String url, boolean isCaution)
	{
		this.send(BroadcastMessageComposer.compose(text));
		/*this.msg.init(isCaution ? EServerMessage.Broadcast : EServerMessage.UserBanned);
		this.msg.appendStr(text);
		this.msg.appendStr(url);
		this.send(this.msg);*/
	}

	public void sendSystemNotif(String message)
	{
		this.send(MOTDNotificationMessageComposer.compose(message));
	}
	
	public void sendWhisperNotif(String message)
	{
		if(this.inRoom())
		{
			this.send(WhisperMessageComposer.compose(this.getRoomUser(), message));
		}
	}

	public void rebuildAndSendUserRights()
	{
		if(this.user != null)
		{
			// Default right list
			user.rights = new ArrayList<String>();
			user.rights.add("default");
			user.specialRightsLevel = 2;

			// Role rights?
			if(user.role != null)
			{
				user.rights.addAll(user.role.rights);
			}
			
			// User-specific rights?
			Session.onRebuildRightsList.fire(this);

			// Send the rights list
			this.send(UserRightsMessageComposer.compose(user));
		}
	}

	public void sendCreditBalance()
	{
		this.send(CreditBalanceMessageComposer.compose(this.user));
	}

	public void sendPixelBalance()
	{
		this.sendPixelsAddedNotif(0);
	}

	public void sendPixelsAddedNotif(int added)
	{
		this.send(ActivityPointNotificationMessageComposer.compose(this.user.pixels, added));
	}
	
	public void sendPixelPearlBalance()
	{
		this.send(ActivityPointsMessageComposer.compose(this.user.pixels, this.user.pearls));
	}

	public void sendRoomPermissions()
	{
		if(this.roomUser != null)
		{
			if(this.roomUser.hasRoomRights)
			{
				this.send(YouAreControllerMessageComposer.compose());
			}
			if(this.roomUser.hasRoomOwner)
			{
				this.send(YouAreOwnerMessageComposer.compose());
			}
		}
	}

	public NetConnectionHandler getConnection()
	{
		return this.conn;
	}

	public long getLastMsgTime()
	{
		return this.lastMsgTime;
	}

	public boolean authenticated()
	{
		return(this.user != null);
	}

	public User getUser()
	{
		return this.user;
	}

	public int getAccessLogId()
	{
		return this.accessLogId;
	}

	public void setAccessLogId(int accessLogId)
	{
		if(this.accessLogId == 0)
		{
			this.accessLogId = accessLogId;
		}
	}

	public Map<Integer, String> getIgnoredUsers()
	{
		return this.ignoredUsers;
	}
	
	public boolean hasIgnoredUser(int userId)
	{
		return this.ignoredUsers.containsKey(userId);
	}
	
	public InstantMessenger getMessenger()
	{
		return this.messenger;
	}
	
	public BadgeInventory getBadgeInventory()
	{
		return this.badgeInventory;
	}
	
	public ItemInventory getItemInventory()
	{
		return this.itemInventory;
	}
	
	public PetInventory getPetInventory()
	{
		return this.petInventory;
	}
	
	public EffectInventory getEffectInventory()
	{
		return this.effectInventory;
	}
	
	public FavoriteRoomList getFavoriteRooms()
	{
		return this.favoriteRooms;
	}
	
	public List<Integer> getRatedRooms()
	{
		return this.ratedRooms;
	}

	public boolean inRoom()
	{
		return(this.roomInstance != null && this.roomUser != null);
	}

	public RoomInstance getRoomInstance()
	{
		return this.roomInstance;
	}

	public RoomUser getRoomUser()
	{
		return this.roomUser;
	}
	
	public RoomEvent getRoomEvent()
	{
		return this.roomEvent;
	}
	
	public void setRoomEvent(RoomEvent event)
	{
		this.roomEvent = event;
	}
	
	public boolean isTrading()
	{
		return (this.trade != null);
	}
	
	public Trade getTrade()
	{
		return this.trade;
	}
	
	public void setTrade(Trade trade)
	{
		this.trade = trade;
	}

	@Override
	public String toString()
	{
		return "Session #" + this.id;
	}
}
