package habbod2.game.sessions;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import habbod2.net.protocol.ServerMessage;
import habbod2.game.GameComponent;
import habbod2.game.GameManager;
import habbod2.game.users.User;
import habbod2.game.util.GameUtil;
import habbod2.net.NetConnectionHandler;

import habbod2.communication.messages.outgoing.handshake.PingMessageComposer;

public class SessionManager extends GameComponent
{
	private long sessionCounter;
	private ConcurrentHashMap<Long, Session> sessions;
	private ConcurrentHashMap<Integer, Session> sessionsByUserId;
	private ConcurrentHashMap<String, Session> sessionsByUsername;

	public SessionManager(GameManager gameMgr)
	{
		super(gameMgr);
		this.sessionCounter = 0;
		this.sessions = new ConcurrentHashMap<Long, Session>();
		this.sessionsByUserId = new ConcurrentHashMap<Integer, Session>();
		this.sessionsByUsername = new ConcurrentHashMap<String, Session>();
	}

	public void shutdown()
	{
		if(!this.sessions.isEmpty())
		{
			this.getLogger().info(this, "Ending all sessions...");
			for(Session session : this.sessions.values())
			{
				session.onEnd();
			}
			this.sessions.clear();
			this.sessionsByUserId.clear();
			this.sessionsByUsername.clear();
			this.getLogger().info(this, "All sessions ended.");
		}
	}

	public void startSession(NetConnectionHandler conn)
	{
		// Create Session object, assign it a unique ID and assign it as ConnectionHandler for the network connection
		Session session = new Session(++this.sessionCounter, conn, this);
		conn.setHandler(session);
		this.sessions.put(session.id, session);

		// Reseed random number generator every 10 connections
		if(this.sessionCounter % 10 == 0)
		{
			GameUtil.reseedRandom();
		}
		
		// New session started!
		this.getLogger().info(this, String.format("Session #%d joined from %s.", session.id, session.getConnection().getIpAddress()));
	}

	public void endSession(long sessionId, String reason)
	{
		this.endSession(sessionId, reason, false);
	}
	
	public void endSession(long sessionId, String reason, boolean blacklistIp)
	{
		// Remove session from map and end session
		Session session = this.sessions.remove(sessionId);
		if(session != null)
		{
			if(blacklistIp)
			{
				this.getSystem().getServer().blacklistIp(session.getConnection().getIpAddress(), reason);
			}
			session.onEnd();
			this.getLogger().info(this, String.format("Session #%d ended. (reason: %s)", sessionId, reason));
		}
	}

	public void addAuthedSession(Session session)
	{
		// Shorthand variable
		User user = session.getUser();

		// Already an authed session for this user ID? End it, don't tolerate concurrent logins
		Session concurrent = this.sessionsByUserId.get(user.id);
		if(concurrent != null)
		{
			this.endSession(concurrent.id, "concurrent user login");
		}

		// Add the session to the session user maps
		this.sessionsByUserId.put(user.id, session);
		this.sessionsByUsername.put(user.username.toLowerCase(), session);

		// Write the accesslog
		int accessLogId = this.getGameMgr().getAccessCtrl().logUserLogin(user.id, session.getConnection().getIpAddress());
		session.setAccessLogId(accessLogId);

		// User logged in!
		this.getLogger().info(this, String.format("User #%d (%s, role: %s) logged in on session #%d.", user.id, user.username, user.role.name, session.id));
	}

	public void removeAuthedSession(Session session)
	{
		// Shorthand variable
		User user = session.getUser();

		// Remove from session-by-user map
		if(this.sessionsByUserId.remove(user.id) != null)
		{
			this.sessionsByUsername.remove(user.username.toLowerCase());
		}

		// Write the accesslog
		this.getGameMgr().getAccessCtrl().logUserLogout(session.getAccessLogId());

		// User logged out!
		this.getLogger().info(this, String.format("User #%d (%s) logged out.", user.id, user.username));
	}
	
	public void runLatencyTest()
	{
		// Check pong status of every session
		List<Session> timeout = new ArrayList<Session>();
		ServerMessage ping = PingMessageComposer.compose();
		for(Session session : this.getAllSessions())
		{
			if(session.pongOk)
			{
				session.pongOk = false;
			}
			else
			{
				timeout.add(session);
				continue;
			}
			session.send(ping);
		}
		
		// End the timed out sessions
		for(Session session : timeout)
		{
			this.endSession(session.id, "ping/pong timeout");
		}
	}
	
	public Session getSession(long sessionId)
	{
		return this.sessions.get(sessionId);
	}

	public Session getSessionByUserId(int userId)
	{
		return this.sessionsByUserId.get(userId);
	}

	public Session getSessionByUsername(String username)
	{
		return this.sessionsByUsername.get(username.toLowerCase());
	}
	
	public Collection<Session> getAllSessions()
	{
		return this.sessions.values();
	}

	public void sendNotifToAll(String text)
	{
		this.sendNotifToAll(text, null);
	}

	public void sendNotifToAll(String text, String url)
	{
		for(Session session : this.sessionsByUserId.values())
		{
			session.sendNotif(text, url, false);
		}
	}

	public int sessions()
	{
		return this.sessions.size();
	}

	public int authedSessions()
	{
		return this.sessionsByUserId.size();
	}
}
