package habbod2.game.rooms.items.floor;

import java.util.List;
import java.util.ArrayList;

import habbod2.Instance;
import habbod2.communication.messages.outgoing.sound.NowPlayingMessageComposer;
import habbod2.game.items.ItemManager;
import habbod2.game.music.Song;
import habbod2.game.music.SongDisk;
import habbod2.game.rooms.instance.RoomInstance;
import habbod2.game.rooms.instance.RoomUser;
import habbod2.game.rooms.items.RoomFloorItem;
import habbod2.util.StringUtil;

public class JukeboxItem extends RoomFloorItem
{
	private boolean isEnabled;
	private long startTime;
	private int currentPosition;
	private Song currentSong;
	private boolean refreshNowPlaying;
	
	private List<SongDisk> disks = new ArrayList<SongDisk>();
	
	private void reloadDisks()
	{
		// Get item manager
		ItemManager itemMgr = Instance.get().getGameMgr().getItemMgr();
		
		// Clear old disks
		this.disks.clear();
		
		// Process data?
		if(this.data != null && !this.data.isEmpty())
		{
			// Go through all the item IDs
			String[] entries = this.data.split(",");
			for(String entry : entries)
			{
				// Get disk, load in this order
				int itemId = StringUtil.parseInt(entry);
				SongDisk disk = SongDisk.fromItem(itemMgr.getItem(itemId));
				if(disk != null)
				{
					this.loadDisk(disk);
				}
			}
		}
	}
	
	private void saveDisks()
	{
		this.data = "";
		for(SongDisk disk : this.disks)
		{
			this.data += disk.item.id + ",";
			disk.item.update(0);
		}
		this.update();
	}
	
	public boolean loadDisk(SongDisk disk)
	{
		// Machine full?
		if(this.disks.size() >= this.getCapacity())
		{
			return false;
		}
		
		// Already in? (huh?)
		for(SongDisk tmp : this.disks)
		{
			if(tmp.item.id == disk.item.id)
			{
				return false;
			}
		}
		
		// Load disk
		this.disks.add(disk);
		this.saveDisks();
		
		// OK!
		return true;
	}
	
	public SongDisk unloadDisk(int position)
	{
		// Unload disk
		SongDisk disk = this.disks.remove(position);
		if(disk == null)
		{
			return null;
		}
		
		// Removed current playing song?
		if(position == this.currentPosition)
		{
			this.playSong();
		}

		// Return the unloaded disk
		return disk;
	}

    public void start()
    {
    	this.isEnabled = true;
    	this.currentPosition = -1;
    	this.nextSong();
    	this.setState(1);
    }
	
	public void stop()
	{
		this.isEnabled = false;
		this.currentSong = null;
		this.currentPosition = -1;
		this.refreshNowPlaying = true;
		this.setState(0);
	}
	
	public void nextSong()
	{
		this.currentPosition++;
		this.playSong();
	}
	
	private void playSong()
	{
		// End of playlist reached?
		if(this.currentPosition >= this.disks.size())
		{
			this.currentPosition = 0;
		}
		
		// Playlist empty?
		if(this.disks.isEmpty())
		{
			this.stop();
		}
		else
		{
			this.currentSong = this.disks.get(this.currentPosition).song;
			this.startTime = System.currentTimeMillis();
			this.refreshNowPlaying = true;
		}
	}
	
	public int getCapacity()
	{
		return StringUtil.parseInt(this.definition.data);
	}
	
	public List<SongDisk> getDisks()
	{
		return this.disks;
	}
	
	public int getCurrentPosition()
	{
		return this.currentPosition;
	}
	
	public Song getCurrentSong()
	{
		return this.currentSong;
	}
	
	public long getSecondsPlaying()
	{
		return (System.currentTimeMillis() - this.startTime) / 1000;
	}
	
	public long getSyncTime()
	{
		Song song = this.getCurrentSong();
		if(song == null)
		{
			return 0;
		}
		else if(this.getSecondsPlaying() >= song.length)
		{
			return song.length;
		}
		else
		{
			return this.getSecondsPlaying() * 1000;
		}
	}

	@Override
	public void onRoomAdd(RoomInstance room)
	{
		room.setJukebox(this);
		this.reloadDisks();
		if(this.state.equals("1"))
		{
			this.start();
		}
	}
	
	@Override
	public void onRoomRemove(RoomInstance room)
	{
		room.setJukebox(null);
		this.saveDisks();
	}
	
	@Override
	public boolean onToggle(RoomUser user, int operation)
	{
		if(user == null || user.hasRoomRights)
		{
			if(user == null || operation == 0)
			{
				if(this.isEnabled)
				{
					this.stop();
				}
				else
				{
					this.start();
				}
			}
			
			return true;
		}
		else
		{
			return false;
		}
	}
	
	@Override
	public boolean requiresPulse()
	{
		return true;
	}
	
	@Override
	protected void onPulse()
	{
		// No song playing, or song has ended?
		if(this.isEnabled && (this.currentSong == null || this.getSecondsPlaying() >= this.currentSong.length + 1))
		{
			// Any disks to play?
			if(this.disks.size() > 0)
			{
				this.nextSong();
			}
			else
			{
				this.stop();
			}
			this.refreshNowPlaying = true;
		}
		
		// Broadcast current song?
		if(this.refreshNowPlaying)
		{
			this.getRoom().broadcast(NowPlayingMessageComposer.compose(this));
			this.refreshNowPlaying = false;
		}
	}
}
