package habbod2.net;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.jboss.netty.bootstrap.ServerBootstrap;
import org.jboss.netty.channel.Channel;
import org.jboss.netty.channel.ChannelPipeline;
import org.jboss.netty.channel.ChannelPipelineFactory;
import org.jboss.netty.channel.Channels;
import org.jboss.netty.channel.group.ChannelGroup;
import org.jboss.netty.channel.group.DefaultChannelGroup;
import org.jboss.netty.channel.socket.nio.NioServerSocketChannelFactory;
import org.jboss.netty.util.ThreadNameDeterminer;
import org.jboss.netty.util.ThreadRenamingRunnable;

import habbod2.System;
import habbod2.storage.DatabaseResult;
import habbod2.storage.SQLHelper;
import habbod2.util.NamedThreadFactory;

public class NetServer extends System.Component
{
	private ServerBootstrap server;
	private Channel listener;
	private ChannelGroup channels;
	private List<String> blacklist;
	private ExecutorService bossExecutor;
	private ExecutorService workerExecutor;
	
	public NetServer(System system)
	{
		super(system);
		this.server = null;
		this.listener = null;
		this.channels = new DefaultChannelGroup();
		this.blacklist = new ArrayList<String>();
		
		// Disable Netty's thread renaming, create executors
		ThreadRenamingRunnable.setThreadNameDeterminer(ThreadNameDeterminer.CURRENT);
		this.bossExecutor = Executors.newCachedThreadPool(new NamedThreadFactory("NetServer/Boss #%d"));
		this.workerExecutor = Executors.newCachedThreadPool(new NamedThreadFactory("NetServer/Worker #%d"));
	}

	public boolean start(int port)
	{
		// Already running?
		if(this.server != null)
		{
			return false;
		}
		
		// Create server boss and worker threadpools using defaults
		ServerBootstrap server = new ServerBootstrap(new NioServerSocketChannelFactory(this.bossExecutor, this.workerExecutor));
		
		// Setup pipeline factory for the local session manager
		final ConnectionHandler management = new ConnectionHandler(this, this.channels);
		final HandshakeHandler handshake = new HandshakeHandler(this.getSystem().getGameMgr().getSessionMgr());
		server.setPipelineFactory(new ChannelPipelineFactory()
		{
			public ChannelPipeline getPipeline() throws Exception
			{
				return Channels.pipeline(management, handshake);
			}
		});

		// Bind and start listening for connections
		try
		{
			Channel listener = server.bind(new InetSocketAddress(port));
			if(listener.isOpen())
			{
				this.server = server;
				this.listener = listener;
				
				// Up and running
				this.getLogger().info(this, String.format("Listening for game connections on TCP/%d.", port));
				return true;
			}
		}
		catch(Exception ex)
		{
			
		}
		
		// Failed whatever
		this.getLogger().warning(this, String.format("Could not start listening on TCP port %d! Port is already taken/security error?", port));
		return false;
	}
	
	public void stop()
	{
		if(this.server != null)
		{
			// Close listener
			this.listener.close();
			this.listener = null;
			
			// Done!
			this.getLogger().debug(this, "Closed listener.");
		}
	}

	public void close()
	{
		// Close all connections
		this.channels.close();
		this.channels.clear();
		
		// Shutdown the executors to speed up shutdown
		this.bossExecutor.shutdownNow();
		this.workerExecutor.shutdownNow();
		
		// Destroy bootstrapper
		this.server = null;
	}

	public boolean ipIsBlacklisted(String ip)
	{
		return this.blacklist.contains(ip);
	}

	public void reloadBlacklist()
	{
		synchronized(this.blacklist)
		{
			this.blacklist.clear();
			
			try(DatabaseResult result = this.getDb().query("SELECT ip FROM system_ipblacklist"))
			{
				while(result.row())
				{
					this.blacklist.add(result.getString("ip"));
				}
			}
			
			this.getLogger().info(this, String.format("Loaded %d blacklisted IP(s).", this.blacklist.size()));
		}
	}

	public void blacklistIp(String ip, String info)
	{
		// Already blacklisted?
		if(this.ipIsBlacklisted(ip))
		{
			return;
		}

		// Add to internal blacklist
		synchronized(this.blacklist)
		{
			this.blacklist.add(ip);
		}

		// Insert new IP to blacklist table
		this.getDb().execute("INSERT INTO system_ipblacklist(ip,timestamp,info) VALUES ('" + SQLHelper.escape(ip) + "','" + SQLHelper.getDateTime() + "','" + SQLHelper.escape(info) + "')");

		// Log it
		this.getLogger().info(this, String.format("Blacklisted IP address %s (\"%s\").", ip, info));
	}

	public boolean allowChannel(Channel channel)
	{
		// Get address of joining connection
		InetAddress address = ((InetSocketAddress)channel.getRemoteAddress()).getAddress();
			
		// IP is blacklisted?
		if(this.ipIsBlacklisted(address.getHostAddress()))
		{
			this.getLogger().warning(this, String.format("Denying connection request from %s, because this IP is blacklisted.", address.getHostAddress()));
			return false;
		}
		
		// Concurrent connection limit configured?
		int limit = this.getConfig().getInt("net.max.per.ip", 0);
		if(limit > 0)
		{
			// Check all open channels
			int count = 0;
			for(Channel someChannel : this.channels)
			{
				// Same IP?
				if(((InetSocketAddress)someChannel.getRemoteAddress()).getAddress().equals(address))
				{
					// Increment and check count
					count++;
					if(count == limit)
					{
						this.getLogger().info(this, String.format("Denying connection request from %s, because this host already has %d open connections. (max)", address.getHostAddress(), count));
						return false;
					}
				}
			}
		}
			
		// Allow!
		return true;
	}
}
