package habbod2.game.chatcommands;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Map.Entry;

import static habbod2.core.Locale._I;
import habbod2.game.GameManager;
import habbod2.game.sessions.Session;
import habbod2.game.users.User;
import habbod2.game.util.GameError;
import habbod2.storage.DatabaseResult;

public class ChatCommandManager extends GameManager.Component
{
	private Map<String, ChatCommandHandler> handlers;
	
	public ChatCommandManager(GameManager gameMgr)
	{
		super(gameMgr);
		this.handlers = Collections.emptyMap();
	}
	
	public void reload()
	{
		// Create new map
		Map<String, ChatCommandHandler> tmpHandlers = new HashMap<String, ChatCommandHandler>();
		
		// Load new handlers (enabled only)
		try(DatabaseResult result = this.getDb().query("SELECT * FROM system_chatcommands WHERE enabled=1"))
		{
			while(result.row())
			{
				String className = result.getString("class");
				try
				{
					// Create instance of handler
					ChatCommandHandler handler = (ChatCommandHandler) Class.forName("habbod2.game.chatcommands." + className).newInstance();
					
					// Add data to handler
					handler.initialize(result.getString("command"), result.getString("description"), result.getString("required_right"), result.getString("syntax"));
					
					// Add handler to the list
					tmpHandlers.put(handler.getCommand(), handler);
				}
				catch(Exception ex)
				{
					this.getLogger().error(this, String.format("Cannot register chatcommand '%s'.", className));
				}
			}
		}
	
		// Update the list
		this.handlers = tmpHandlers;
		
		// Log result
		this.getLogger().info(this, String.format("Loaded %d entries.", tmpHandlers.size()));
	}
	
	public ChatCommandHandler getCommand(String cmd)
	{
		return this.handlers.get(cmd);
	}
	
	public List<ChatCommandHandler> getChatCommandsForUser(User user)
	{
		List<ChatCommandHandler> available = new ArrayList<ChatCommandHandler>();
		for(Entry<String, ChatCommandHandler> cmd : handlers.entrySet())
		{
			String right = cmd.getValue().getRequiredRight();
			if(right == null || user.hasRight(right))
			{
				available.add(cmd.getValue());
			}
		}
		
		return available;
	}
	
	public boolean tryHandleCommand(Session session, String message)
	{
		// Actually a chat command?
		if(message.indexOf(ChatCommandHandler.COMMAND_HEAD) != 0)
		{
			return false;
		}

		// Attempt to separate command and body
		final String[] segments = message.substring(1).split(" ", 2);
		final String cmd = segments[0];
		final String body = (segments.length == 2) ? segments[1] : "";
		
		// Handler registered?
		ChatCommandHandler handler = this.handlers.get(cmd);
		if(handler == null)
		{
			// Not recognized
			return false;
		}

		// User has the required right?
		if(handler.getRequiredRight() != null)
		{
			String right = handler.getRequiredRight();
			if(!session.getUser().hasRight(right))
			{
				return false;
			}
		}
		

		// Process body, detect valid syntax, substitute optional args
		boolean isSyntaxError = false;
		String[] args = null;
		String[] syntax = handler.getSyntax();
		if(syntax != null)
		{
			String[] givenArgs = body.isEmpty() ? null : body.split(" ", syntax.length);
			if(givenArgs != null && givenArgs.length == syntax.length)
			{
				args = givenArgs;
			}
			else
			{
				// Pad empty args
				args = new String[syntax.length];
				for(int x = 0; x < syntax.length; x++)
				{
					args[x] = (givenArgs == null || x > givenArgs.length - 1) ? "" : givenArgs[x];
					if(syntax[x].charAt(0) != '-' && args[x].isEmpty())
					{
						// Wasn't supposed to be blank...
						isSyntaxError = true;
						break;
					}
				}
			}
		}

		// Syntax error?
		if(isSyntaxError)
		{			
			session.sendNotif(_I("Command recognized, but your syntax was invalid. Action not processed. Below follows the correct syntax:\r\r%s\r\rPlease try again.", handler.getUsage()));
		}
		else
		{
			// Execute the command
			try
			{
				handler.handle(session, args);
			}
			catch(GameError err)
			{
				session.sendNotif(_I("Command recognized, but action was not processed.\r\r%s", err.generateNotif()));
			}
			catch(Exception ex)
			{
				session.sendNotif(_I("Chatcommand recognized, but a server error occurred. Server log has been written."));
				this.getLogger().error(this, String.format("Uncaught error processing chatcommand '%s'. (body=%s)", segments[0], body), ex);
			}
		}

		// Recognized
		return true;
	}
	
	public Collection<ChatCommandHandler> getAllCommands()
	{
		return this.handlers.values();
	}
	
	public int count()
	{
		return this.handlers.size();
	}
}