package habbod2.game.items;

import habbod2.Instance;
import habbod2.storage.DatabaseResult;
import habbod2.game.rooms.items.RoomItem;
import habbod2.game.rooms.items.floor.GenericFloorItem;
import habbod2.game.rooms.items.floor.SongDiskItem;
import habbod2.game.rooms.items.wall.DecorationItem;
import habbod2.game.rooms.items.wall.GenericWallItem;
import habbod2.game.util.GameUtil;

public class ItemDefinition
{
	private static final String ROOMCLASS_PKG = "habbod2.game.rooms.items.";
	
	public int id;
	public String name;
	public String description;
	
	public char type;
	public int displayId;
	public String sprite;
	public String color;
	
	private Class<? extends RoomItem> roomClass;
	public String data;
	public int states;
	
	public byte length;
	public byte width;
	public float height;
	public boolean isSolid;
	public boolean isStackPlatform;
	public boolean isWater;
	
	public String action;
	public int effectId;
	
	public boolean isGiftable;
	public boolean isTradable;
	public boolean isRecylable;
	public boolean isMarketplaceSellable;
	public boolean isListGroupable;

	public int inventoryType;
	
	public boolean is(char type)
	{
		return (this.type == type);
	}
	
	public boolean is(Class<? extends RoomItem> roomClass)
	{
		return (this.roomClass == roomClass);
	}
	
	public static ItemDefinition parse(DatabaseResult result)
	{
		ItemDefinition def = new ItemDefinition();
		def.id = result.getInt("id");
		def.name = result.getString("name");
		def.description = result.getString("description");
		def.type = Character.toLowerCase(result.getString("type").charAt(0));
		def.displayId = result.getInt("display_id");
		def.sprite = result.getString("sprite");
		def.color = result.getString("color");
		def.loadRoomClass(result.getString("class"));
		def.data = result.getString("data");
		def.states = result.getInt("states");
		def.length = result.getByte("length");
		def.width = result.getByte("width");
		def.height = result.getFloat("height");
		if(def.height == 0.0)
		{
			def.height = 0.001f;
		}
		else
		{
			def.height = GameUtil.roundHeight(def.height);
		}
		def.isSolid = result.getBoolean("solid");
		def.isStackPlatform = result.getBoolean("stack_platform");
		def.isWater = result.getBoolean("water");
		def.action = result.getString("action").intern(); // intern() for faster comparisons (it's not an enum)
		def.effectId = result.getInt("effect_id");
		def.inventoryType = ItemDefinition.getInventoryType(def);
		def.isGiftable = result.getBoolean("giftable");
		def.isTradable = result.getBoolean("tradable");
		def.isRecylable = result.getBoolean("recyclable");
		//def.isMarketplaceSellable = result.getBoolean("sellable"); // TODO: re-enable this
		def.isListGroupable = result.getBoolean("list_groupable");
		
		return def;
	}

	@SuppressWarnings("unchecked")
	private void loadRoomClass(String name)
	{
		// Set default class just in case
		if(this.is(EItemType.FLOORITEM))
		{
			this.roomClass = GenericFloorItem.class;
		}
		else if(this.is(EItemType.WALLITEM))
		{
			this.roomClass = GenericWallItem.class;
		}
		
		// Load custom class?
		if(name != null && !name.isEmpty())
		{
			try
			{
				this.roomClass = (Class<? extends RoomItem>) Class.forName(ItemDefinition.ROOMCLASS_PKG + name);
			}
			catch (NoClassDefFoundError | Exception ex)
			{
				Instance.get().getLogger().warning("ItemDefinition", String.format("Definition #%d ('%s') refers to non-existent room class '%s', using %s.", this.id, this.sprite, name, this.roomClass.getName().substring(ItemDefinition.ROOMCLASS_PKG.length())));
			}
		}
	}
	
	public InventoryItem newItem()
	{
		InventoryItem item = new InventoryItem();
		item.definition = this;
		
		return item;
	}
	
	public RoomItem createRoomItem()
	{
		RoomItem item;
		try
		{
			item = this.roomClass.newInstance();
			item.definition = this;
			return item;
		}
		catch(Exception ex)
		{
			return null;
		}
	}
	
	public Class<? extends RoomItem> getRoomClass()
	{
		return this.roomClass;
	}
	
	public static int getInventoryType(ItemDefinition def)
	{
		if(def.is(DecorationItem.class))
		{
			switch(def.sprite)
			{
				case "wallpaper": return 2;
				case "floor": return 3;
				case "landscape": return 4;
				default: return 0;
			}
		}
		else if(def.is(SongDiskItem.class))
		{
			return 8;
		}
		else
		{
			return 0;
		}
	}
	
	public interface EItemType
	{
		public static final char FLOORITEM = 's';
		public static final char WALLITEM = 'i';
		public static final char EFFECT = 'e';
	}
	
	public interface EActorAction
	{
		public static final String SEAT = "sit";
		public static final String BED = "lay";
	}
	
}
