package habbod2.game.pets;

import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;

import habbod2.storage.DatabaseResult;
import habbod2.storage.SQLHelper;
import habbod2.game.GameManager;
import habbod2.game.items.ItemDefinition;
import habbod2.game.sessions.Session;

public class PetManager extends GameManager.Component
{
	private Map<Integer, PetType> types;
	
	public PetManager(GameManager gameMgr)
	{
		super(gameMgr);
		this.types = new HashMap<Integer, PetType>();
	}
	
	public void reload()
	{
		this.types.clear();
	
		// Load pet types
		try(DatabaseResult result = this.getDb().query("SELECT * FROM pettypes WHERE enabled=1"))
		{
			while(result.row())
			{
				PetType type = new PetType();
				type.id = result.getInt("id");
				type.canSwim = result.getBoolean("can_swim");
				type.mountHeight = result.getFloat("mount_height");
				this.types.put(type.id, type);
			}
		}
		
		// Load breeds
		try(DatabaseResult result = this.getDb().query("SELECT * FROM pettype_breeds"))
		{
			while(result.row())
			{
				PetType type = this.types.get(result.getInt("type_id"));
				if(type != null)
				{
					PetBreed breed = new PetBreed();
					breed.type = type;
					breed.id = result.getInt("breed");
					breed.name = result.getString("name");
					breed.isSellable = result.getBoolean("is_sellable");
					type.breeds.add(breed);
					if(breed.isSellable)
					{
						type.sellableBreeds.add(breed);
					}
				}
			}
		}
		
		// Load commands
		Map<Integer, PetCommand> commands = new HashMap<Integer, PetCommand>();
		try(DatabaseResult result = this.getDb().query("SELECT * FROM petcommands"))
		{
			while(result.row())
			{
				PetCommand cmd = new PetCommand();
				cmd.id = result.getInt("id");
				cmd.keyword = result.getString("keyword").toLowerCase();
				cmd.action = result.getString("action").intern();
				cmd.isStatus = result.getBoolean("is_status");
				cmd.experience = result.getInt("experience");
				cmd.level = result.getInt("level");
				cmd.isOwnerOnly = result.getBoolean("is_owneronly");
				commands.put(cmd.id, cmd);
			}
		}
		
		// Map commands to types
		try(DatabaseResult result = this.getDb().query("SELECT * FROM pettype_commands"))
		{
			while(result.row())
			{
				int typeId = result.getInt("type_id");
				PetCommand command = commands.get(result.getInt("command_id"));
				
				if(command != null)
				{
					if(typeId == -1)
					{
						for(PetType type : this.types.values())
						{
							type.commands.put(command.keyword, command);
						}
					}
					else
					{
						PetType type = this.types.get(result.getInt("type_id"));
						if(type != null)
						{
							type.commands.put(command.keyword, command);
						}
					}
				}
			}
		}
		
		// Load tricks
		try(DatabaseResult result = this.getDb().query("SELECT * FROM pettype_tricks"))
		{
			while(result.row())
			{
				int typeId = result.getInt("type_id");
				String trick = result.getString("trick");
				
				if(typeId == -1)
				{
					for(PetType type : this.types.values())
					{
						type.tricks.add(trick);
					}
				}
				else
				{
					PetType type = this.types.get(result.getInt("type_id"));
					if(type != null)
					{
						type.tricks.add(trick);
					}
				}
			}
		}
		
		// Load messages
		try(DatabaseResult result = this.getDb().query("SELECT * FROM pettype_messages"))
		{
			while(result.row())
			{
				int typeId = result.getInt("type_id");
				String event = result.getString("event");
				String message = result.getString("message");
				
				if(typeId == -1)
				{
					for(PetType type : this.types.values())
					{
						type.addEventMessage(event, message);
					}
				}
				else
				{
					PetType type = this.types.get(typeId);
					if(type != null)
					{
						type.addEventMessage(event, message);
					}
				}
			}
		}
		
		// Load accessories
		try(DatabaseResult result = this.getDb().query("SELECT * FROM pettype_accessories"))
		{
			while(result.row())
			{
				int typeId = result.getInt("type_id");
				ItemDefinition definition =  this.getGameMgr().getItemMgr().getDefinition(result.getInt("definition_id"));
				if(definition != null)
				{
					if(typeId == -1)
					{
						for(PetType type : this.types.values())
						{
							type.accessories.add(definition.id);
						}
					}
					else
					{
						PetType type = this.types.get(typeId);
						if(type != null)
						{
							type.accessories.add(definition.id);
						}
					}
				}
			}
		}
		
		this.getLogger().info(this, String.format("Loaded %d pet types.", this.types.size()));
	}
	
	public PetType getPetType(int typeId)
	{
		return this.types.get(typeId);
	}

	public Pet getPetById(int petId)
	{
		try(DatabaseResult result = this.getDb().query("SELECT p.*,u.username AS owner FROM pets p JOIN users u ON u.id=p.owner_id WHERE p.id=" + petId))
		{
			if(result.row())
			{
				return this.processPetFromRow(result);
			}
			else
			{
				return null;
			}
		}
	}
	
	public void createPet(int userId, String name, PetType type, int breed, String color)
	{
		// Create pet
		int petId = this.getDb().executeReturnKey("INSERT INTO pets(name,type,breed,color,owner_id,created_at) VALUES ('" + SQLHelper.escape(name) + "'," + type.id + "," + breed + ",'" + SQLHelper.escape(color) + "'," + userId + ",'" + SQLHelper.getDateTime() + "')");
		Pet pet = this.getPetById(petId);
		if(pet != null)
		{
			// Initialize pet
			pet.experience = 0;
			pet.energy = pet.getMaxEnergy() / 2;
			pet.happiness = pet.getMaxHappiness() / 2;
			pet.save();
			
			// Owner online?
			Session session = this.getSystem().getGameMgr().getSessionMgr().getSessionByUserId(userId);
			if(session != null)
			{
				session.getPetInventory().addPet(pet);
			}
		}
	}
	
	public List<Pet> getInventoryPetsForUser(int userId)
	{
		List<Pet> pets = new ArrayList<Pet>(0);
		
		try(DatabaseResult result = this.getDb().query("SELECT p.*,u.username AS owner FROM pets p JOIN users u ON u.id=p.owner_id WHERE p.owner_id=" + userId + " AND p.room_id=0"))
		{
			while(result.row())
			{
				Pet pet = this.processPetFromRow(result);
				if(pet != null)
				{
					pets.add(pet);
				}
			}
		}
		
		return pets;
	}

	public List<Pet> getPetsForRoom(int roomId)
	{
		List<Pet> pets = new ArrayList<Pet>(0);
		
		try(DatabaseResult result = this.getDb().query("SELECT p.*,u.username AS owner FROM pets p JOIN users u ON u.id=p.owner_id WHERE p.room_id=" + roomId))
		{
			while(result.row())
			{
				Pet pet = this.processPetFromRow(result);
				if(pet != null)
				{
					pets.add(pet);
				}
			}
		}
		
		return pets;
	}
	
	private Pet processPetFromRow(DatabaseResult row)
	{
		// Valid pet type?
		Pet pet = new Pet();
		pet.type = this.types.get(row.getInt("type"));
		if(pet.type == null)
		{
			return null;
		}
		
		// Process the remaining data
		pet.id = row.getInt("id");
		pet.name = row.getString("name");
		pet.breed = row.getByte("breed");
		pet.color = row.getString("color");
		pet.experience = row.getInt("experience");
		pet.energy = row.getInt("energy");
		pet.happiness = row.getInt("happiness");
		pet.respect = row.getInt("respect");
		pet.eatTimer = row.getInt("timer_eat");
		pet.drinkTimer = row.getInt("timer_drink");
		pet.sleepTimer = row.getInt("timer_sleep");
		pet.ownerId = row.getInt("owner_id");
		pet.ownerName = row.getString("owner");
		pet.createdTime = row.getTime("created_at");
		pet.roomId = row.getInt("room_id");
		pet.x = row.getInt("x");
		pet.y = row.getInt("y");
		pet.z = row.getFloat("z");
		
		// Calculate level (based on exp)
		pet.recalculateLevel();
		
		return pet;
	}
	
	public int types()
	{
		return this.types.size();
	}
}
