package habbod2.game.rooms.instance;

import habbod2.communication.messages.EServerMessage;
import habbod2.game.rooms.geometry.Direction;  
import habbod2.game.rooms.pathfinding.PathfinderNode;
import habbod2.net.protocol.ServerMessage;

import java.util.ArrayList;
import java.util.List;

public class ActorEngine
{
	private RoomInstance room;
	
	public ActorEngine(RoomInstance room)
	{
		this.room = room;
	}
	
	public void pulse()
	{
		// Get door
		RoomTile door = this.room.getMap().getDoor();
		
		// Process all the actors
		List<RoomActor> updates = new ArrayList<RoomActor>();
		for(RoomActor actor : this.room.getAllActors())
		{
			// Perform other tasks
			actor.pulse();
			
			// Requires status update?
			if(actor.requiresUpdate)
			{
				// Update me!
				if(updates.contains(actor))
				{
					continue;
				}
				else
				{
					updates.add(actor);
					actor.requiresUpdate = false;
				}
				
				// Moving?
				if(actor.hasGoal())
				{
					// At goal?
					if(actor.at(actor.goal))
					{
						// Trip complete!
						actor.clearMovement();
						
						// In door?
						if(actor.at(door) && actor instanceof RoomUser)
						{
							((RoomUser)actor).getSession().kickFromRoom();
						}
					}
					else
					{
						// Determine next tile in path
						RoomTile next = null;
						if(actor.overrideNextTile && actor.goal != null)
						{
							// Override used!
							actor.overrideNextTile = false;

							// Substitute the next step
							next = actor.goal;
						}
						else
						{
							// Start new path?
							if(actor.calculatePath)
							{
								// Shorthand
								RoomTile goal = actor.goal;
								
								// Redirect?
								if(goal.redirect != null)
								{
									goal = goal.redirect;
									actor.goal = goal;
								}

								// Open last step tile for pathfinder if the goal is a last step tile
								byte backup = -1;
								if(goal.is(RoomTile.LASTSTEP))
								{
									// No actor here? (quick check)
									if((!goal.hasActor() || goal.allowWalkthrough))
									{
										backup = goal.state;
										goal.state = RoomTile.OPEN;
									}
								}

								// Find path using the beloved Joe 'Joeh' Hegarty 2D A* pathfinder
								List<PathfinderNode> path = actor.pathfinder.findPath(goal.x, goal.y);

								// Restore map
								if(backup != -1)
								{
									goal.state = backup;
								}
								
								// Is there a path found?
								if(!path.isEmpty())
								{
									// Remove current tile node
									path.remove(0);

									// Set this path!
									actor.path = path;
								}
								
								// Executed pathfinder!
								actor.calculatePath = false;
							}

							// Pop the next tile from the path
							if(!actor.path.isEmpty())
							{
								// Test the next tile
								PathfinderNode tmpNode = actor.path.remove(0);
								next = this.room.getMap().getTile(tmpNode.x, tmpNode.y);
								if(!this.isPassable(next, actor, true))
								{
									next = null;
								}
								else
								{
									// Walking diagonal (hacky wacky isometrics!)
									final boolean ALLOW_DIAGONAL = true;
									if(ALLOW_DIAGONAL && !actor.path.isEmpty())
									{
										// Peek + test next node (already popped one, so this is like path[1])
										tmpNode = actor.path.get(0);
										RoomTile peek = this.room.getMap().getTile(tmpNode.x, tmpNode.y);
										if(this.isPassable(peek, actor, false))
										{
											// Can't skip more than one tile!
											if(actor.touches(peek))
											{
												byte xAdd1 = 0;
												byte xAdd2 = 0;
												if(peek.x > actor.x)
												{
													xAdd1 = -1;
													xAdd2 = 1;
												}
												else
												{
													xAdd1 = 1;
													xAdd2 = -1;
												}

												// Valid shortcut?
												RoomTile left = this.room.getMap().getTile(peek.x + xAdd1, peek.y);
												RoomTile right = this.room.getMap().getTile(actor.x + xAdd2, actor.y);
												if(left != null && right != null && this.isPassable(left, actor, false) && this.isPassable(right, actor, false))
												{
													next = peek;
													actor.path.remove(0);
												}
											}
										}
									}
								}
							}
						}
						
						// Next tile available?
						if(next == null)
						{
							actor.clearMovement();
						}
						else
						{
							// Redirect?
							if(next.redirect != null)
							{
								next = next.redirect;
							}
							
							// Set next tile
							actor.next = next;
							actor.next.hasPendingActor = true;
							
							// Mounted?
							if(actor instanceof RoomHuman)
							{
								RoomHuman human = (RoomHuman)actor;
								if(human.isMounted())
								{
									human.getMount().next = next;
									human.getMount().requiresUpdate = true;
									if(!updates.contains(human.getMount()))
									{
										updates.add(human.getMount());
									}
								}
							}
						}
					}
				}
			}
		}
		
		// Any updates? This includes simple status updates, but also movement:phase2
		if(!updates.isEmpty())
		{
			ServerMessage msg = ServerMessage.create(EServerMessage.UserUpdate);
			msg.appendInt(updates.size());
			for(RoomActor actor : updates)
			{
				if(actor.next != null)
				{
					actor.setDirection(Direction.calculate(actor.x, actor.y, actor.next.x, actor.next.y));
					if(actor.isReverseWalk)
					{
						actor.setDirection(actor.bodyDirection.invert());
					}
					actor.removeNonMoveStatuses();
					actor.statuses.put("mv", actor.next.x + "," + actor.next.y + "," + (actor.next.height + actor.getExtraDisplayHeight()));
				}
				actor.appendStatusString(msg);
				if(actor.next != null)
				{
					actor.statuses.remove("mv");
					actor.setTile(actor.next, false);
					actor.next = null;
				}
			}
			this.room.broadcast(msg);
		}
	}
	
	private boolean isPassable(RoomTile tile, RoomActor actor, boolean lastStep)
	{
		// Blocked?
		if(tile.hasActor() && !tile.allowWalkthrough)
		{
			return false;
		}

		// Normal tile?
		if(!tile.is(RoomTile.LASTSTEP))
		{
			// Not allowed by options?
			if((actor.pathfindingAllowedOn & tile.state) != tile.state)
			{
				return false;
			}
		}
		else
		{
			// Tile is last step, but move is not last step?
			if(!lastStep)
			{
				return false;
			}
		}
		
		// Blocked?
		if(tile.state == RoomTile.CLOSED)
		{
			return false;
		}

		// Closed item tiles are only passable with last step of path
		if(!lastStep && (tile.state == (RoomTile.CLOSED | RoomTile.LASTSTEP)))
		{
			return false;
		}

		// Passable!
		return true;
	}
}
