package habbod2.game.rooms.items;

import java.util.List;
import java.util.ArrayList;

import habbod2.communication.messages.outgoing.room.engine.ObjectDataUpdateMessageComposer;
import habbod2.game.items.ItemDefinition.EActorAction;
import habbod2.game.pets.RoomPet;
import habbod2.game.rooms.geometry.Direction;
import habbod2.game.rooms.instance.RoomActor;
import habbod2.game.rooms.instance.RoomTile;
import habbod2.game.rooms.instance.RoomUser;
import habbod2.game.rooms.items.wired.WiredTrigger;
import habbod2.game.rooms.items.wired.triggers.*;

public abstract class RoomFloorItem extends RoomItem implements RoomTile.TileObject
{
	public int x;
	public int y;
	public float z;
	public byte rotation;

	private List<WiredTrigger> triggers;
	
	public boolean isAction(String action)
	{
		// 'action' is interned!
		return (this.definition.action == action);
	}
	
	@Override
	public String getDbPosition()
	{
		return (this.x + " " + this.y + " " + this.z + " " + this.rotation);
	}
	
	@Override
	public void setPositionFromDb(String position)
	{
		try
		{
			String[] bits = position.split(" ");
			this.x = Integer.parseInt(bits[0]);
			this.y = Integer.parseInt(bits[1]);
			this.z = Float.parseFloat(bits[2]);
			this.rotation = Byte.parseByte(bits[3]);
		}
		catch(Exception ex)
		{
			
		}
	}
	
	@Override
	public final void refreshState(boolean isTemporarily)
	{
		if(isTemporarily || this.requiresInstantRefresh())
		{
			this.getRoom().broadcast(ObjectDataUpdateMessageComposer.compose(this));
		}
		else
		{
			this.getRoom().refreshItemState(this);
		}
	}
	
	public boolean requiresInstantRefresh()
	{
		return false;
	}
	
	public float getHeight()
	{
		return this.definition.height;
	}
	
	public float getInteractionHeight(RoomActor actor)
	{
		if(actor instanceof RoomPet)
		{
			return this.getHeight() / 2;
		}
		else
		{
			return this.getHeight();
		}
	}
	
	@Override
	public RoomTile getTile()
	{
		return this.getRoom().getMap().getTile(this.x, this.y);
	}
	
	public List<RoomTile> getTiles()
	{
		return this.getRoom().getMap().getAffectedTiles(this.x, this.y, this.definition.length, this.definition.width, this.rotation);
	}
	
	public RoomTile getTileInFront()
	{
		int x = this.x;
		int y = this.y;
		if(this.rotation == 0)
		{
			y--;
		}
		else if(this.rotation == 2)
		{
			x++;
		}
		else if(this.rotation == 4)
		{
			y++;
		}
		else if(this.rotation == 6)
		{
			x--;
		}
		return this.getRoom().getMap().getTile(x, y);
	}

	public RoomTile getTileBehind()
	{
		int x = this.x;
		int y = this.y;
		if(this.rotation == 0)
		{
			y++;
		}
		else if(this.rotation == 2)
		{
			x--;
		}
		else if(this.rotation == 4)
		{
			y--;
		}
		else if(this.rotation == 6)
		{
			x++;
		}
		return this.getRoom().getMap().getTile(x, y);
	}
	
	public void onGenerateMapTile(RoomTile tile)
	{
		// Can stand on top?
		if(!this.definition.isSolid)
		{
			// Tile not closed yet?
			if(!tile.is(RoomTile.CLOSED))
			{
				tile.state = RoomTile.OPEN;
				tile.topItem = this;
				tile.height = this.z + this.getHeight();
			}
		}
		
		// Can sit on top?
		else if(this.definition.action == EActorAction.SEAT)
		{
			tile.state = (RoomTile.CLOSED | RoomTile.LASTSTEP);
			tile.topItem = this;
			tile.effectId = 0;
		}
		
		// Can lay on top?
		else if(this.definition.action == EActorAction.BED)
		{
			tile.state = (RoomTile.CLOSED | RoomTile.LASTSTEP);
			tile.topItem = this;
			if(this.rotation == 0 || this.rotation == 4)
			{
				tile.redirect = this.getRoom().getMap().getTile(tile.x, this.y);
			}
			else
			{
				tile.redirect = this.getRoom().getMap().getTile(this.x, tile.y);
			}
			tile.effectId = 0;
		}
		
		// Solid item!
		else
		{
			// No last step item here?
			if(!tile.is(RoomTile.LASTSTEP))
			{
				tile.state = RoomTile.CLOSED;
				tile.height = this.z + this.getHeight();
				tile.topItem = this;
			}
		}
	}
	
	public void onActorStepOn(RoomActor actor)
	{
		// Default item interaction types
		if(this.isAction(EActorAction.SEAT) || this.isAction(EActorAction.BED))
		{
			actor.setDirection(Direction.get(this.rotation));
			actor.addStatus(this.definition.action.toString().toLowerCase(), Float.toString(this.getInteractionHeight(actor)));
		}
		
		// Raise triggers
		if(actor instanceof RoomUser)
		{
			this.executeTriggers((RoomUser)actor, UserStepsOnItemTrigger.class);
		}
	}

	public void onActorStepOff(RoomActor actor)
	{
		// Raise triggers
		if(actor instanceof RoomUser)
		{
			this.executeTriggers((RoomUser)actor, UserStepsOffItemTrigger.class);
		}
	}
	
	@Override
	public void onStateChange(RoomUser user)
	{
		if(user != null)
		{
			this.executeTriggers(user, UserTogglesItemTrigger.class);
		}
	}
	
	public void addTrigger(WiredTrigger trigger)
	{
		if(this.triggers == null)
		{
			this.triggers = new ArrayList<WiredTrigger>(1);
		}
		this.triggers.add(trigger);
	}
	
	public void removeTrigger(WiredTrigger trigger)
	{
		if(this.triggers != null)
		{
			this.triggers.remove(trigger);
		}
	}
	
	public void executeTriggers(RoomUser user, Class<? extends WiredTrigger> type)
	{
		if(this.triggers != null)
		{
			for(WiredTrigger trigger : this.triggers)
			{
				if(trigger.getClass() == type)
				{
					trigger.execute(user);
				}
			}
		}
	}
}
