package habbod2.game.rooms.minigames;

import java.util.Collection;

import habbod2.game.rooms.instance.RoomInstance;
import habbod2.game.rooms.instance.RoomUser;
import habbod2.game.rooms.items.RoomFloorItem;
import habbod2.game.rooms.items.minigame.TimerItem;
import habbod2.game.rooms.items.wired.actions.GiveGameScoreAction;
import habbod2.game.rooms.items.wired.triggers.GameStartsTrigger;
import habbod2.game.rooms.items.wired.triggers.GameEndsTrigger;
import habbod2.game.rooms.minigames.GameTeam;
import habbod2.game.util.GameUtil;

// Banzai: tiles flash only when all tiles are taken. When time runs out, it just restarts everything.
// Freeze: when last enemy team dies, winners stay on the field and wave.
public abstract class Game
{
	private static final int GAME_SHUTDOWN_TIME = GameUtil.secondsToRoomPulses(5);
	private static final int MAX_PLAYERS_PER_TEAM = 5;
	
	private TimerItem timer;
	private GameTeam[] teams;
	
	protected int shutdownTimer;
	
	public Game(TimerItem timer)
	{
		// Set timer
		this.timer = timer;
		
		// Get all team IDs
		GameTeam.ETeam[] teamIds = GameTeam.ETeam.values();
		
		// Create all empty teams
		this.teams = new GameTeam[teamIds.length];
		for(int i = 0; i < teamIds.length; i++)
		{
			this.teams[i] = new GameTeam(teamIds[i], this);
		}
		
		// Not shutting down obviously
		this.resetShutdownTimer();
	}
		
	public final void init()
	{
		// Time on timer?
		if(this.timer.getTime() == 0)
		{
			return;
		}
		
		// Prepare for start of new game
		this.resetShutdownTimer();
		for(GameTeam team : this.teams)
		{
			team.reset();
		}
		
		// Reset items
		for(RoomFloorItem item : this.getItems())
		{
			if(item instanceof GiveGameScoreAction)
			{
				((GiveGameScoreAction)item).resetUseCount();
			}
		}
		
		// Process other stuff
		this.onInit();
	}
	
	protected void onInit() { }
	
	public final void begin()
	{
		this.onBegin();
		for(RoomFloorItem item : this.getRoom().getFloorItems())
		{
			if(item instanceof GameStartsTrigger)
			{
				((GameStartsTrigger)item).execute(null);
			}
		}
	}
	
	protected void onBegin() { }
	
	public final void end()
	{
		this.resetShutdownTimer();
		this.onEnd();
		for(RoomFloorItem item : this.getRoom().getFloorItems())
		{
			if(item instanceof GameEndsTrigger)
			{
				((GameEndsTrigger)item).execute(null);
			}
		}
	}
	
	protected void onEnd() { }
	
	public final void pulse()
	{
		if(this.isRunning() || this.isShuttingDown())
		{
			this.onPulse();
			if(this.isShuttingDown())
			{
				this.shutdownTimer--;
				if(this.shutdownTimer == 0)
				{
					this.timer.stop();
				}
			}
		}
	}
	
	protected void onPulse() { }
	
	public final void outOfTime()
	{
		this.onOutOfTime();
		this.end();
	}
	
	protected void onOutOfTime() { }
	
	public final void complete()
	{
		// Can only complete a running game
		if(!this.isRunning())
		{
			return;
		}
		
		// The winner is always the team with the best score!
		GameTeam winner = this.getBestTeam();
		
		// Let the winners wave!
		if(winner != null)
		{
			for(GameUserPlayer player : winner)
			{
				player.getUser().wave();
			}
		}
		
		// Extra events
		this.onComplete(winner);
		
		// Pause timer, start shutdown timer
		this.timer.pause();
		this.shutdownTimer = GAME_SHUTDOWN_TIME;
		
		// Shutdown now?
		if(winner == null || winner.players() == 0)
		{
			this.shutdownTimer = 1;
		}
	}
	
	protected void onComplete(GameTeam winner) { }
	
	public void onPlayerLeave(GameUserPlayer player) { }
	
	public void destroy()
	{
		this.end();
		for(GameTeam team : this.teams)
		{
			for(GameUserPlayer player : team)
			{
				player.getUser().setPlayer(null);
				player.getUser().clearEffect(RoomUser.EFFECT_GAME);
			}
		}
	}
	
	protected void onDestroy() { }
	
	public TimerItem getTimer()
	{
		return this.timer;
	}
	
	public RoomInstance getRoom()
	{
		return this.timer.getRoom();
	}
	
	public boolean isRunning()
	{
		return this.timer.isRunning();
	}
	
	public boolean isShuttingDown()
	{
		return (this.shutdownTimer != -1);
	}
	
	public void resetShutdownTimer()
	{
		this.shutdownTimer = -1;
	}
	
	public Collection<RoomFloorItem> getItems()
	{
		// TODO: caching?
		return this.getRoom().getFloorItems();
	}
	
	public GameTeam getTeam(GameTeam.ETeam teamId)
	{
		return this.teams[teamId.ordinal()];
	}
	
	public GameTeam[] getTeams()
	{
		return this.teams;
	}
	
	public GameTeam getBestTeam()
	{
		// Find team with best score
		GameTeam best = null;
		for(GameTeam team : this.teams)
		{
			if(team.players() > 0 && team.getScore() > 0 && (best == null || team.getScore() > best.getScore()))
			{
				best = team;
			}
		}
		
		// Detect draws
		if(best != null)
		{
			for(GameTeam team : this.teams)
			{
				if(team != best && team.players() > 0 && team.getScore() == best.getScore())
				{
					return null;
				}
			}
		}
		
		// Return best team (null = no best team)
		return best;
	}
	
	public int getMaxPlayersPerTeam()
	{
		return MAX_PLAYERS_PER_TEAM;
	}
	
	public abstract int getEffectOffset();
	
	public abstract GameUserPlayer.GameValues createValuesForPlayer();
	
	public String getType()
	{
		return this.getClass().getSimpleName();
	}
}
