package habbod2.game.rooms.minigames;

import habbod2.game.rooms.instance.RoomUser;
import habbod2.game.rooms.items.RoomFloorItem;
import habbod2.game.rooms.items.minigame.GateItem;
import habbod2.game.rooms.items.minigame.ScoreboardItem;
import habbod2.game.rooms.items.wired.triggers.GameScoreAchievedTrigger;

import java.util.Iterator;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

public class GameTeam implements Iterable<GameUserPlayer>
{
	public enum ETeam
	{
		Red,
		Green,
		Blue,
		Yellow
	}
	
	private ETeam id;
	private Game game;
	private List<GameUserPlayer> players;
	private int score;
	
	public GameTeam(ETeam teamId, Game game)
	{
		this.id = teamId;
		this.game = game;
		this.players = new CopyOnWriteArrayList<GameUserPlayer>();
	}
	
	public void addPlayer(RoomUser user)
	{
		// Add player
		GameUserPlayer session = new GameUserPlayer(user, this);
		user.setPlayer(session);
		user.setEffect(RoomUser.EFFECT_GAME, this.game.getEffectOffset() + this.id.ordinal());
		this.players.add(session);
		
		// Update gates
		this.updateGates();
	}

	public void removePlayer(GameUserPlayer player)
	{
		// Player in team?
		if(this.players.remove(player))
		{
			// Clear player
			player.getUser().setPlayer(null);
			player.getUser().clearEffect(RoomUser.EFFECT_GAME);
			
			// Update gates
			this.updateGates();
			
			// Player left!
			this.game.onPlayerLeave(player);
		}
	}
	
	private void updateGates()
	{
		for(RoomFloorItem item : this.game.getItems())
		{
			if(item instanceof GateItem)
			{
				GateItem gate = (GateItem)item;
				if(gate.isCompatibleWith(this.game) && gate.getTeamId() == this.id)
				{
					gate.setPlayerCount(this.players.size());
				}
			}
		}
	}
	
	public ETeam getId()
	{
		return this.id;
	}
	
	public Game getGame()
	{
		return this.game;
	}
	
	public int getScore()
	{
		return this.score;
	}
	
	public void addScore(int points, GameUserPlayer player)
	{
		this.setScore(this.score + points, player);
	}
	
	public void setScore(int score, GameUserPlayer player)
	{
		int previous = this.score;
		this.score = (score > 0) ? score : 0;
		for(RoomFloorItem item : this.getGame().getItems())
		{
			if(item instanceof ScoreboardItem)
			{
				ScoreboardItem scoreboard = (ScoreboardItem)item;
				if(scoreboard.getTeamId() == this.id)
				{
					scoreboard.setScore(this.score);
				}
			}
			else if(player != null && item instanceof GameScoreAchievedTrigger)
			{
				GameScoreAchievedTrigger trigger = (GameScoreAchievedTrigger)item;
				int target = trigger.getParam(GameScoreAchievedTrigger.PARAM_SCORE);
				if(this.score >= target && previous < target)
				{
					trigger.execute(player.getUser());
				}
			}
		}
	}
	
	public void reset()
	{
		this.setScore(0, null);
		for(GameUserPlayer player : this.players)
		{
			player.reset();
		}
	}
	
	public int players()
	{
		return this.players.size();
	}
	
	public boolean isFull()
	{
		return this.players() >= this.getGame().getMaxPlayersPerTeam();
	}

	@Override
	public Iterator<GameUserPlayer> iterator()
	{
		return this.players.iterator();
	}
	
	public int getEffect()
	{
		int effectOffset = this.game.getEffectOffset();
		if(effectOffset > 0)
		{
			return effectOffset + this.id.ordinal();
		}
		else
		{
			return 0;
		}
	}
}
