package habbod2.game.rooms.minigames.game;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import habbod2.Instance;
import habbod2.communication.messages.outgoing.room.session.GamePlayerValueMessageComposer;
import habbod2.game.rooms.geometry.Direction;
import habbod2.game.rooms.instance.RoomTile;
import habbod2.game.rooms.instance.RoomUser;
import habbod2.game.rooms.items.RoomFloorItem;
import habbod2.game.rooms.items.minigame.GateItem;
import habbod2.game.rooms.items.minigame.TimerItem;
import habbod2.game.rooms.items.minigame.freeze.GameOverTileItem;
import habbod2.game.rooms.items.minigame.freeze.GameTileItem;
import habbod2.game.rooms.items.minigame.freeze.PowerUpBoxItem;
import habbod2.game.rooms.minigames.Game;
import habbod2.game.rooms.minigames.GameTeam;
import habbod2.game.rooms.minigames.GameUserPlayer;
import habbod2.game.rooms.minigames.GameUserPlayer.GameValues;
import habbod2.game.util.GameUtil;
import habbod2.storage.Database;

public class FreezeGame extends Game
{
	private static final int EFFECT_OFFSET = 40;
	private static final int EFFECT_INVINCIBLE_OFFSET = 49;
	private static final int EFFECT_FROZEN = 12;
	
	private static final int START_SCORE_PER_PLAYER = 40;
	private static final int START_LIVES = 3;
	private static final int MAX_LIVES = 3;
	
	private static final int START_BALLS = 1;
	
	private static final int START_RANGE = 2;
	private static final int MEGA_RANGE = 999;
	
	private static final int BALL_DURATION = 4;
	
	private static final Direction[] EXPLODE_NORMAL = new Direction[] { null, Direction.North, Direction.East, Direction.South, Direction.West };
	private static final Direction[] EXPLODE_DIAGONAL = new Direction[] { null, Direction.NorthEast, Direction.SouthEast, Direction.SouthWest, Direction.NorthWest };
	private static final Direction[] EXPLODE_MEGA = new Direction[] { null, Direction.North, Direction.NorthEast, Direction.East, Direction.SouthEast, Direction.South, Direction.SouthWest, Direction.West, Direction.NorthWest };
	
	private List<Ball> balls;
	
	public FreezeGame(TimerItem timer)
	{
		super(timer);
		this.balls = new ArrayList<Ball>();
	}
	
	@Override
	protected void onInit()
	{
		// Kick players that are not on a game tile
		for(GameTeam team : this.getTeams())
		{
			for(GameUserPlayer player : team)
			{
				if(!(player.getUser().getTile().topItem instanceof GameTileItem))
				{
					this.kickPlayer(player);
				}
			}
		}
		
		// Close gates and reset boxes
		for(RoomFloorItem item : this.getItems())
		{
			if(item instanceof GateItem)
			{
				GateItem gate = (GateItem)item;
				if(gate.isCompatibleWith(this))
				{
					gate.setClosed(true);
				}
			}
			else if(item instanceof PowerUpBoxItem)
			{
				if(!item.getTile().hasActor())
				{
					((PowerUpBoxItem)item).setPowerUp(null);
				}
			}
		}
		
		// Set scores and balls, lives, etc...
		for(GameTeam team : this.getTeams())
		{
			team.setScore(team.players() * START_SCORE_PER_PLAYER, null);
			for(GameUserPlayer player : team)
			{
				FreezeValues values = player.getValues();
				values.lives = START_LIVES;
				values.balls = START_BALLS;
				values.range = START_RANGE;
				this.getRoom().broadcast(GamePlayerValueMessageComposer.compose(player.getUser(), values.lives));
			}
		}
	}
	
	protected void onBegin()
	{
		// Power on game over tiles
		for(RoomFloorItem item : this.getItems())
		{
			if(item instanceof GameOverTileItem)
			{
				item.setState(1);
			}
		}
	}
	
	protected void onEnd()
	{
		// Dispose balls
		if(!this.balls.isEmpty())
		{
			for(Ball ball : this.balls)
			{
				ball.source.setTempState("");
			}
			this.balls.clear();
		}
		
		// Re-open gates, dispose powerups, and power-off game over tiles
		for(RoomFloorItem item : this.getItems())
		{
			if(item instanceof GateItem)
			{
				GateItem gate = (GateItem)item;
				if(gate.isCompatibleWith(this))
				{
					gate.setClosed(false);
				}
			}
			else if(item instanceof PowerUpBoxItem)
			{
				PowerUpBoxItem box = (PowerUpBoxItem)item;
				if(box.isCrushed())
				{
					box.setPowerUp(EPowerUp.None);
				}
			}
			else if(item instanceof GameOverTileItem)
			{
				item.setState(0);
			}
		}
		
		// Unfreeze players
		Database db = Instance.get().getDb();
		GameTeam winner = this.getBestTeam();
		for(GameTeam team : this.getTeams())
		{
			for(GameUserPlayer player : team)
			{
				// Reset timers
				FreezeValues values = player.getValues();
				values.frozenTimer = (values.frozenTimer > 0) ? 1 : 0;
				values.invincibleTimer = 0;
				
				// Update stats
				if(values.hits > 0)
				{
					db.execute("UPDATE users SET fz_kills=fz_kills+" + values.hits + " WHERE id=" + player.getUser().getInfo().id);
				}
				if(team == winner)
				{
					db.execute("UPDATE users SET fz_wins=fz_wins+1 WHERE id=" + player.getUser().getInfo().id);
				}
			}
		}
		this.onPulse();
	}

	@Override
	protected void onComplete(GameTeam winner)
	{	
		// Kick losers
		if(winner != null)
		{
			for(GameTeam team : this.getTeams())
			{
				if(team != winner)
				{
					for(GameUserPlayer player : team)
					{
						this.kickPlayer(player);
					}
				}
			}
		}
	}

	@Override
	protected void onPulse()
	{
		// Process players
		for(GameTeam team : this.getTeams())
		{
			for(GameUserPlayer player : team)
			{
				FreezeValues values = player.getValues();
				
				// Expire frozen status?
				if(values.frozenTimer > 0)
				{
					values.frozenTimer--;
					if(values.frozenTimer == 0)
					{
						// No longer frozen
						player.getUser().unlockMovement();
						
						// Out of lives?
						if(values.lives <= 0)
						{
							this.kickPlayer(player);
							continue;
						}
					}
				}
				
				// Expire invincible status?
				if(values.invincibleTimer > 0)
				{
					values.invincibleTimer--;
				}
				
				// Update effect
				if(values.frozenTimer == 0)
				{
					if(values.invincibleTimer == 0)
					{
						player.getUser().setEffect(RoomUser.EFFECT_GAME, EFFECT_OFFSET + player.getTeam().getId().ordinal());
					}
					else
					{
						player.getUser().setEffect(RoomUser.EFFECT_GAME, EFFECT_INVINCIBLE_OFFSET + player.getTeam().getId().ordinal());
					}
				}
			}
		}
		
		// Process flying balls!
		Iterator<Ball> balls = this.balls.iterator();
		while(balls.hasNext())
		{
			// Ball needs spawning?
			Ball ball = balls.next();
			if(ball.timer == BALL_DURATION)
			{
				ball.source.setTempState(ball.range == 999 ? 6000 : (ball.range <= 4 ? ball.range : 4) * 1000);
			}
			
			// Update timer
			if(ball.timer > 0)
			{
				ball.timer--;
			}
			else
			{
				// Explode in correct directions
				nextDir: for(Direction dir : ball.isMega() ? EXPLODE_MEGA : (ball.isDiagonal ? EXPLODE_DIAGONAL : EXPLODE_NORMAL))
				{
					for(int i = 1; i <= (dir == null ? 1 : ball.range); i++)
					{
						RoomTile tile = (dir == null) ? ball.source.getTile() : this.getRoom().getMap().getTile(ball.source.x + (i * dir.modX), ball.source.y + (i * dir.modY));
						if(tile != null)
						{
							// Poof!
							boolean hasTile = false;
							for(RoomFloorItem item : tile.getItems())
							{
								if(item instanceof GameTileItem)
								{
									hasTile = true;
									item.setTempState(11000 + ((i > 10 ? 9 : i - 1) * 100));
								}
							}
							
							// Stop exploding into this direction when there's no tile
							if(!hasTile)
							{
								continue nextDir;
							}
							
							// Break boxes
							for(RoomFloorItem item : tile.getItems())
							{
								if(item instanceof PowerUpBoxItem)
								{
									PowerUpBoxItem box = (PowerUpBoxItem)item;
									if(!box.isCrushed())
									{
										box.setPowerUp(EPowerUp.getRandom());
										continue nextDir;
									}
								}
							}
							
							// Hit players
							if(tile.hasActor())
							{
								for(RoomUser user : this.getRoom().getAllUsers())
								{
									if(user.at(tile) && user.isPlayer() && user.getPlayer().getGame() instanceof FreezeGame)
									{
										FreezeValues values = user.getPlayer().getValues();
										if(!values.isInvincible())
										{
											// FREEZE!
											user.clearMovement(); user.lockMovement();
											values.frozenTimer = GameUtil.secondsToRoomPulses(5);
											values.invincibleTimer = GameUtil.secondsToRoomPulses(10);
											user.setEffect(RoomUser.EFFECT_GAME, EFFECT_FROZEN);
											
											// Gametile in the room?
											GameOverTileItem exit = this.getExit();
											if(exit != null)
											{
												// Lose a life
												values.lives--;
												this.getRoom().broadcast(GamePlayerValueMessageComposer.compose(user, values.lives));
												
												// Lose score
												user.getPlayer().addScore(-10);
												
												// Award 'hit'-point
												if(ball.player != null && user.getPlayer() != ball.player)
												{
													((FreezeValues)ball.player.getValues()).hits++;
												}
											}
										}
										continue nextDir;
									}
								}
							}
							
							// Something else?
							if(!tile.isOpen())
							{
								continue nextDir;
							}
						}
					}
				}
				
				// EOL for this ball... :(
				balls.remove();
				if(ball.player != null)
				{
					((FreezeValues)ball.player.getValues()).balls++;
				}
			}
		}
	}
	
	/*
	public void broadcastLives(GameUserPlayer player)
	{
		FreezeValues values = player.getValues();
		String lives = "";
		for(int i = 1; i <= MAX_LIVES; i++)
		{
			lives += (values.lives >= i) ? "ƒ": "|";
		}
		this.getRoom().broadcast(WhisperMessageComposer.compose(player.getUser(), lives));
	}*/
	
	public void launchBall(GameTileItem source, GameUserPlayer player)
	{
		// Has player?
		FreezeValues values = null;
		if(player != null)
		{
			values = player.getValues();
			if(values.balls > 0)
			{
				values.balls--;
			}
			else
			{
				return;
			}
		}
		
		// Create ball
		Ball ball = new Ball();
		ball.source = source;
		if(player == null)
		{
			if(GameUtil.getRandomBool(0.10))
			{
				ball.range = MEGA_RANGE;
			}
			else
			{
				ball.range = GameUtil.getRandomInt(1, 3);
				ball.isDiagonal = GameUtil.getRandomBool(0.5);
			}
		}
		else
		{
			ball.player = player;
			ball.range = values.range;
			if(values.nextMega)
			{
				ball.range = MEGA_RANGE;
				values.nextMega = false;
			}
			else if(values.nextDiagonal)
			{
				ball.isDiagonal = true;
				values.nextDiagonal = false;
			}
		}
		
		// Spawn ball
		ball.timer = BALL_DURATION;
		this.balls.add(ball);
	}
	
	public void applyPowerUp(GameUserPlayer player, EPowerUp powerUp)
	{
		FreezeValues values = player.getValues();
		switch(powerUp)
		{
			case ExtraRange:
				values.range++;
				break;
				
			case ExtraBall:
				values.balls++;
				break;
			
			case DiagonalExplosion:
				values.nextDiagonal = true;
				break;
				
			case MegaExplosion:
				values.nextMega = true;
				break;
			
			case Life:
				if(values.lives < MAX_LIVES)
				{
					values.lives++;
				}
				this.getRoom().broadcast(GamePlayerValueMessageComposer.compose(player.getUser(), values.lives));
				break;
			
			case Shield:
				values.invincibleTimer = GameUtil.secondsToRoomPulses(10);
				break;
		}
	}
	
	private void kickPlayer(GameUserPlayer player)
	{
		// Unlock movement (just in case) and clear the effect
		RoomUser user = player.getUser();
		user.unlockMovement();
		user.clearEffect(RoomUser.EFFECT_GAME);
		
		// Warp to tile
		GameOverTileItem exit = this.getExit();
		if(exit != null && !player.getUser().at(exit))
		{
			player.getUser().warp(exit.getTile());
		}
		
		// Cleanup
		player.getUser().setPlayer(null);
	}
	
	@Override
	public void onPlayerLeave(GameUserPlayer player)
	{
		// Don't do anything when game is not running
		if(!this.isRunning())
		{
			return;
		}
		
		// Count teams with playesr
		int teams = 0;
		for(GameTeam team : this.getTeams())
		{
			if(team.players() > 0)
			{
				teams++;
			}
		}
		
		// Only one team standing?
		if(teams == 1)
		{
			this.complete();
		}
	} 
	
	
	public GameOverTileItem getExit()
	{
		for(RoomFloorItem item : this.getItems())
		{
			if(item instanceof GameOverTileItem)
			{
				return (GameOverTileItem)item;
			}
		}
		
		return null;
	}
	
	@Override
	public int getEffectOffset()
	{
		return FreezeGame.EFFECT_OFFSET;
	}
	
	@Override
	public GameValues createValuesForPlayer()
	{
		return new FreezeValues();
	}

	public class FreezeValues implements GameUserPlayer.GameValues
	{
		public int lives;
		public int balls;
		public int range;
		public boolean nextDiagonal;
		public boolean nextMega;
		
		public int frozenTimer;
		public int invincibleTimer;
		
		public int hits;
		
		public boolean isFrozen()
		{
			return (this.frozenTimer > 0);
		}
		
		public boolean isInvincible()
		{
			return (this.invincibleTimer > 0);
		}
	}
	
	public enum EPowerUp
	{
		None,
		ExtraRange,
		ExtraBall,
		DiagonalExplosion,
		MegaExplosion,
		Life,
		Shield;
		
		public static EPowerUp getRandom()
		{
			return GameUtil.getRandomElement(EPowerUp.values());
		}
	}
	
	private class Ball
	{
		public int range;
		public boolean isDiagonal;
		public GameTileItem source;
		public GameUserPlayer player;
		
		public int timer;
		
		public boolean isMega()
		{
			return (this.range == MEGA_RANGE);
		}
	}
}
