package habbod2.game.rooms.pathfinding;

/*
 * Code: Project Thor Pathfinding
 Author: Joe Hegarty

 Copyright (c) 2009 Joe Hegarty

 Permission is hereby granted, free of charge, to the Blunk habbod Team to obtain a copy
 of the above software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute copies of the Software,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 */

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import habbod2.game.rooms.instance.RoomActor;
import habbod2.game.rooms.instance.RoomMap;
import habbod2.game.rooms.instance.RoomTile;

public final class Pathfinder
{
	private static final int HEURISTIC_ESTIMATE = 2;
	private static final int MAX_CYCLE_LIMIT = 2500;
	private static final byte[][] DIRECTION = new byte[][] { { 0, -1 }, { 1, 0 }, { 0, 1 }, { -1, 0 } };
	
	private static final float MAX_ASCEND = 1.5f;
	private static final float MAX_DESCEND = 2;
	
	private final RoomActor actor;
	private final RoomMap map;
	
	private final NodeQueue openList;
	private List<PathfinderNode> closedList;
	
	public Pathfinder(RoomActor actor)
	{
		this.actor = actor;
		this.map = actor.getRoom().getMap();
		this.openList = new NodeQueue();
		this.createClosedList();
	}
	
	private void createClosedList()
	{
		if(this.closedList == null)
		{
			this.closedList = new ArrayList<PathfinderNode>();
		}
	}
	
	public List<PathfinderNode> findPath(int goalX, int goalY)
	{
		// Start the trip!
		PathfinderNode tmpNode = new PathfinderNode();
		tmpNode.x = this.actor.x;
		tmpNode.y = this.actor.y;
		tmpNode.parentX = tmpNode.x;
		tmpNode.parentY = tmpNode.y;
		tmpNode.h = HEURISTIC_ESTIMATE;
		tmpNode.f = HEURISTIC_ESTIMATE;
		tmpNode.g = 0;
		this.openList.push(tmpNode);

		// Create new closed list when needed
		this.createClosedList();
		
		// Run the pathfinding algorithm
		boolean pathFound = false;
		while(this.openList.size() > 0)
		{
			tmpNode = this.openList.pop();
			if(tmpNode.x == goalX && tmpNode.y == goalY)
			{
				pathFound = true;
				this.closedList.add(tmpNode);
				break;
			}

			// Been running for too long?
			if(this.closedList.size() > MAX_CYCLE_LIMIT) break;

			// Search in all 4 directions
			for(int i = 0; i < 4; i++)
			{
				PathfinderNode newNode = new PathfinderNode();
				newNode.x = (short) (tmpNode.x + DIRECTION[i][0]);
				newNode.y = (short) (tmpNode.y + DIRECTION[i][1]);

				// Tile exists?
				RoomTile tile = map.getTile(newNode.x, newNode.y);
				if(tile == null)
				{
					continue;
				}

				// Blocked by actor?
				if(tile.hasActor() && !tile.allowWalkthrough)
				{
					continue;
				}
				
				// No pathfinding allowed here?
				if((actor.pathfindingAllowedOn & tile.state) != tile.state)
				{
					continue;
				}

				// Perform height check
				RoomTile parentTile = map.getTileUnchecked(tmpNode.x, tmpNode.y);
				if(parentTile.height - Pathfinder.MAX_DESCEND > tile.height || parentTile.height + Pathfinder.MAX_ASCEND < tile.height)
				{
					continue;
				}

				// Calculate node cost (static)
				int newG = tmpNode.g + 1;

				// Check if node already in OPEN list
				int listIndex = -1;
				for(int j = 0; j < this.openList.size(); j++)
				{
					PathfinderNode curNode = this.openList.get(j);
					if(curNode.x == newNode.x && curNode.y == newNode.y)
					{
						listIndex = j;
						break;
					}
				}
				if(listIndex != -1 && this.openList.get(listIndex).g <= newG) // Better node already in open list
				continue;

				// Check if node already in CLOSED list
				listIndex = -1;
				for(int j = 0; j < this.closedList.size(); j++)
				{
					PathfinderNode curNode = this.closedList.get(j);
					if(curNode.x == newNode.x && curNode.y == newNode.y)
					{
						listIndex = j;
						break;
					}
				}
				if(listIndex != -1 && this.closedList.get(listIndex).g <= newG) // Better node already in closed list
				continue;

				// Update new node
				newNode.parentX = tmpNode.x;
				newNode.parentY = tmpNode.y;
				newNode.g = newG;

				// Calculate heuristic
				int xD1 = tmpNode.x - goalX;
				int xD2 = this.actor.x - goalX;
				int yD1 = tmpNode.y - goalY;
				int yD2 = this.actor.y - goalY;
				newNode.h = HEURISTIC_ESTIMATE * (Math.max(Math.abs(newNode.x - goalX), Math.abs(newNode.y - goalY)));
				newNode.h = (int) (newNode.h + Math.abs(xD1 * yD2 - xD2 * yD1) * 0.001);
				newNode.f = newNode.g + newNode.h;
				this.openList.push(newNode);
			}
			this.closedList.add(tmpNode);
		}

		// Path found?
		List<PathfinderNode> result;
		if(pathFound)
		{
			// Trace back...
			PathfinderNode topNode = this.closedList.get(this.closedList.size() - 1);
			for(int j = this.closedList.size() - 1; j >= 0; j--)
			{
				PathfinderNode node = this.closedList.get(j);
				if(topNode.parentX == node.x && topNode.parentY == node.y || j == this.closedList.size() - 1)
				{
					topNode = node;
				}
				else
				{
					this.closedList.remove(j);
				}
			}
			
			// Result is the closed list (remove it)
			result = this.closedList;
			this.closedList = null;
		}
		else
		{
			// Clear closed list
			this.closedList.clear();
			
			// Empty result
			result = Collections.emptyList();
		}

		// Dispose open list
		this.openList.clear();
		
		// No path
		return result;
	}
}
