package habbod2.game.sessions.local;

import java.util.ArrayList;
import java.util.List;

import org.jboss.netty.channel.ChannelHandlerContext;
import org.jboss.netty.channel.ChannelStateEvent;
import org.jboss.netty.channel.ExceptionEvent;
import org.jboss.netty.channel.MessageEvent;

import habbod2.Instance;
import habbod2.communication.messages.outgoing.room.chat.UserTypingMessageComposer;
import habbod2.core.Configuration;
import habbod2.game.GameManager;
import habbod2.game.rooms.Room;
import habbod2.game.rooms.instance.RoomActor;
import habbod2.game.rooms.instance.RoomTile;
import habbod2.game.rooms.instance.RoomUser;
import habbod2.game.rooms.items.RoomFloorItem;
import habbod2.game.rooms.navigator.NavigatorFrontpageItem;
import habbod2.game.rooms.navigator.RoomNavigator;
import habbod2.game.sessions.Session;
import habbod2.game.util.GameUtil;
import habbod2.game.util.WordFilter;
import habbod2.net.NetConnection;
import habbod2.net.protocol.ServerMessage;
import habbod2.storage.DatabaseResult;

public class LocalSession extends Session
{
	private long logoutTime;
	private List<String> messages;
	private int messageIndex;
	private String pendingMessage;
	private long pendingMessageTime;
	private RoomFloorItem lastItem;
	private boolean reToggleItem;
	
	public LocalSession(long sessionId)
	{
		super(sessionId, new LocalNetConnection());
		this.messages = new ArrayList<String>();
	}

	public void refreshMessages()
	{
		// Clear current messages
		this.messages.clear();
		this.messageIndex = 0;

		// Select a random user ID from chatlogs
		int userId = Instance.get().getDb().queryInt("SELECT user_id FROM chatlog WHERE RAND()<=(100/(SELECT COUNT(*) FROM chatlog)) ORDER BY RAND() LIMIT 1");

		// Get the first 150 messages
		try(DatabaseResult result = Instance.get().getDb().query("SELECT message FROM chatlog WHERE user_id=" + userId + " LIMIT 150"))
		{
			while(result.row())
			{
				this.messages.add(result.getString("message"));
			}
		}
	}

	public void pulse()
	{
		// Update ping
		this.pongOk = true;

		// Get config values
		Configuration config = Instance.get().getConfig();
		
		// Go offline?
		if(this.logoutTime == 0)
		{
			this.logoutTime = System.currentTimeMillis() + (GameUtil.getRandomInt(1, config.getInt("sessions.local.maxlifetime.minutes", 15)) * 60 * 1000);
		}
		if(System.currentTimeMillis() > this.logoutTime)
		{
			this.end("ACT_LOGOUT");
			return;
		}

		// Kick?
		if(GameUtil.getRandomBool(config.getFloat("sessions.local.chance.leaveroom", 0.010f))) // 1%
		{
			this.lastItem = null;
			this.kickFromRoom();
		}

		// Find room?
		if(!this.inRoom())
		{
			// Get navigator
			GameManager gameMgr = Instance.get().getGameMgr();
			RoomNavigator navi = gameMgr.getNavigator();

			// Get random room
			Room room;
			if(GameUtil.getRandomBool(config.getFloat("misc.sessions.local.chance.ownroom", 0.5f)))
			{
				room = GameUtil.getRandomElement(navi.getRoomsForUser(this.getUser().id));
			}
			else
			{
				List<Room> rooms = new ArrayList<Room>(navi.getBusiestGuestRooms());
				for(NavigatorFrontpageItem fpi : navi.getFrontpageItems())
				{
					if(fpi.room != null)
					{
						rooms.add(fpi.room);
					}
				}
				room = this.selectRandomRoomWeighted(rooms, config.getFloat("misc.sessions.local.fpi.weight", 3.0f));
				
				/*
				 * OR: room =
				 * GameUtil.getRandomElement(GameUtil.getRandomElement
				 * (navi.getCategories()).rooms);
				 */
			}

			// I wanna go there!
			if(room != null)
			{
				this.tryGetRoomAccess(room.id, "");
			}
		}
		else
		{
			// Get user
			RoomUser user = this.getRoomUser();

			// Process pending message
			if(this.pendingMessage != null && System.currentTimeMillis() >= this.pendingMessageTime)
			{
				user.getRoom().broadcast(UserTypingMessageComposer.compose(user, false));
				user.chat(WordFilter.getInstance().filter(this.pendingMessage), false);
				this.pendingMessage = null;
			}

			// Chat
			else if(this.pendingMessage == null && GameUtil.getRandomBool(config.getFloat("sessions.local.chance.chat", 0.05f))) // 5%
			{
				// Reset index?
				if(this.messageIndex >= this.messages.size())
				{
					this.refreshMessages();
				}

				// Select next message
				if(this.messages.size() > 0)
				{
					String message = this.messages.get(this.messageIndex++);

					// Needs 'typing...' status?
					int words = message.split(" ").length;
					this.pendingMessage = message;
					this.pendingMessageTime = System.currentTimeMillis() + (words * 1000);
					this.getRoom().broadcast(UserTypingMessageComposer.compose(this.getRoomUser(), true));
				}
			}
			
			// Standing in front of item?
			else if(!user.hasGoal())
			{
				// Last item invalidated?
				if(this.lastItem != null && this.lastItem.getRoom() == null)
				{
					this.lastItem = null;
				}
				
				// Rotate prior to retoggling?
				if(this.reToggleItem && this.lastItem != null)
				{
					user.lookTo(this.lastItem.getTile(), false);
				}
				
				// Toggle!
				RoomTile front = user.getTileInFront();
				if(front != null)
				{
					if(front.topItem != null && (front.topItem != this.lastItem || this.reToggleItem))
					{
						front.topItem.toggle(user, 0);
						this.lastItem = front.topItem;
					}
				}
				
				// Update retoggle status
				this.reToggleItem = user.hasGoal();
			}

			// Move
			boolean inDoor = user.touches(user.getRoom().getMap().getDoor());
			if(!user.hasGoal() && !this.reToggleItem && (inDoor || GameUtil.getRandomBool(config.getFloat("sessions.local.chance.move", 0.05f)))) // 5%
			{
				int radius = config.getInt("sessions.local.move.radius", 10);
				user.setGoal(user.getRoom().getMap().getTile(user.x + GameUtil.getRandomInt(-radius, radius), user.y + GameUtil.getRandomInt(-radius, radius)));
				/*if(inDoor || GameUtil.getRandomBool(0.75))
				{
					user.setGoal(user.getRoom().getMap().getTile(user.x + GameUtil.getRandomInt(-10, 10), user.y + GameUtil.getRandomInt(-10, 10)));
				}
				else
				{
					user.setGoal(GameUtil.getRandomElement(user.getRoom().getAllUsers()).getTileInFront());
				}*/
			}

			// Dance
			else if(GameUtil.getRandomBool(config.getFloat("sessions.local.chance.dance", 0.005f))) // 0.5%
			{
				if(user.danceId == 0)
				{
					user.dance(GameUtil.getRandomInt(0, 3));
				}
				else
				{
					user.dance(0);
				}
			}

			// Lookto
			if(GameUtil.getRandomBool(config.getFloat("sessions.local.chance.lookto", 0.05f))) // 5%
			{
				RoomActor target = GameUtil.getRandomElement(this.getRoom().getAllActors());
				if(target != null)
				{
					user.lookTo(target.getTile(), false);
				}
			}
			
			// Respect
			else if(GameUtil.getRandomBool(config.getFloat("sessions.local.chance.respect", 0.005f))) // 0.5%
			{
				RoomUser target = GameUtil.getRandomElement(this.getRoom().getAllUsers());
				if(target != null && target != user)
				{
					target.receiveRespectFrom(user);
				}
			}
		}
	}

	private Room selectRandomRoomWeighted(List<Room> rooms, float fpExtraWeight)
	{
		double completeWeight = 0.0;
		for(Room room : rooms)
		{
			completeWeight += room.usersNow;
			if(room.usersNow == 0 && room.isFrontpageItem) completeWeight += fpExtraWeight;
		}
		double r = Math.random() * completeWeight;
		double countWeight = 0.0;
		for(Room room : rooms)
		{
			countWeight += room.usersNow;
			if(room.usersNow == 0 && room.isFrontpageItem) countWeight += fpExtraWeight;
			if(countWeight >= r)
			{
				return room;
			}
		}
		return null;
	}

	public static class LocalNetConnection extends NetConnection
	{
		public static final String IP_ADDRESS = "LOCAL";
		
		public LocalNetConnection()
		{
			super(null);
		}

		@Override
		public boolean isOpen()
		{
			return true;
		}

		@Override
		public void close()
		{

		}

		private void requestClose(String reason)
		{
			this.close();
		}

		@Override
		public void disconnectRequested(ChannelHandlerContext ctx, ChannelStateEvent e)
		{
			this.requestClose("disconnected by user");
		}

		@Override
		public void send(ServerMessage msg)
		{

		}

		@Override
		public void messageReceived(ChannelHandlerContext ctx, MessageEvent e)
		{

		}

		@Override
		public void exceptionCaught(ChannelHandlerContext ctx, ExceptionEvent e)
		{
			this.requestClose("disconnected by user");
		}

		@Override
		public NetConnection.CommunicationHandler getHandler()
		{
			return null;
		}

		@Override
		public void setHandler(NetConnection.CommunicationHandler handler)
		{

		}

		@Override
		public String getIpAddress()
		{
			return LocalNetConnection.IP_ADDRESS;
		}
	}
}
