package prinzip.accounts.actions;

import java.util.concurrent.TimeUnit;

import org.apache.log4j.Logger;

import prinzip.accounts.Account;
import prinzip.accounts.AccountProblem;
import prinzip.apps.AppAuthorizationException;
import prinzip.apps.AppAuthorizationException.AuthorizationError;
import prinzip.engine.ClockworkAction;
import prinzip.engine.ClockworkEngine;
import prinzip.util.TwitterErrorCodes;

import twitter4j.TwitterException;
import static twitter4j.internal.http.HttpResponseCode.SERVICE_UNAVAILABLE;
import static twitter4j.internal.http.HttpResponseCode.GATEWAY_TIMEOUT;

public abstract class AccountAction extends ClockworkAction
{
	private static final Logger logger = Logger.getLogger(AccountAction.class);
	
	private Account account;
	
	public AccountAction(Account account)
	{
		this.account = account;
	}
	
	protected Account getAccount()
	{
		return this.account;
	}
	
	protected ClockworkEngine getEngine()
	{
		return this.account.getServer().getEngine();
	}
	
	@Override
	protected void onException(Exception e)
	{
		// Twitter exceptions?
		if(e instanceof TwitterException)
		{
			this.onTwitterException((TwitterException)e);
		}
		
		// Something else! (probably programming error)
		else
		{
			super.onException(e);
		}
	}
	
	private void onTwitterException(TwitterException tex)
	{
		// Could not connect to server?
		if(tex.isCausedByNetworkIssue() && !tex.getMessage().contains("HTTP"))
		{
			logger.warn(String.format("%s encountered a network issue, retrying later...", this));
			this.rescheduleAfter(30, TimeUnit.SECONDS);
		}
		
		// Authorization failed?
		else if(tex instanceof AppAuthorizationException)
		{
			this.onAppAuthorizationException((AppAuthorizationException)tex);
		}
		
		// Hit a few rate limits on the way to heaven?
		else if(tex.exceededRateLimitation())
		{
			// Determine retry after time
			int retryAfter = tex.getRetryAfter();
			if(retryAfter == -1)
			{
				retryAfter = tex.getRateLimitStatus().getSecondsUntilReset();
			}
			
			logger.warn(String.format("%s hit API rate limits, retrying after %d seconds ...", this, retryAfter), tex);
			this.rescheduleAfter(retryAfter, TimeUnit.SECONDS);
		}
		
		// Didn't even reach Twitter API?
		else if(tex.getStatusCode() == SERVICE_UNAVAILABLE || tex.getStatusCode() == GATEWAY_TIMEOUT)
		{
			if(logger.isDebugEnabled()) logger.debug("Twitter service is currently unavailable or reported a gateway timeout, rescheduling...");
			this.rescheduleAfter(30, TimeUnit.SECONDS);
		}
		
		// Twitter API error
		else
		{
			this.onTwitterApiException(tex);
		}
	}
	
	private void onAppAuthorizationException(AppAuthorizationException aex)
	{
		// Deal with it swiftly
		if(aex.getError() == AuthorizationError.BAD_ACCOUNT)
		{
			this.account.reportProblem(AccountProblem.INVALID_CREDENTIALS);
		}
		else if(aex.getError() == AuthorizationError.BAD_APP)
		{
			this.selectNewAppAndRetry();
		}
	}
	
	private void onTwitterApiException(TwitterException tex)
	{
		// Handle common API errors
		switch(tex.getErrorCode())
		{
			// Token is revoked or something, attempt to reauthorize
			case TwitterErrorCodes.BAD_TOKEN:
			{
				logger.warn(String.format("@%s has an invalid or expired token", this.account.username));
				this.selectNewAppAndRetry();
				break;
			}
			
			// Account is banned for whatever reason, end of the line
			case TwitterErrorCodes.SUSPENDED:
			{
				this.account.reportProblem(AccountProblem.SUSPENDED);
				break;
			}
			
			// Twitter is up to some proper funky shit
			case TwitterErrorCodes.OVER_CAPACITY:
			{
				if(logger.isDebugEnabled()) logger.debug("Twitter API is currently out of capacity, rescheduling...");
				this.rescheduleAfter(30, TimeUnit.SECONDS);
				break;
			}
			
			// Now this is messed up bro
			case TwitterErrorCodes.OAUTH_TIMESTAMP_OUT_OF_RANGE:
			{
				logger.error("OAuth timestamp: system clock is out of range?!");
				this.rescheduleAfter(30, TimeUnit.SECONDS);
				break;
			}
			
			// This is not handled here, just skip it
			default:
			{
				try
				{
					this.onOtherTwitterApiException(tex);
				}
				finally
				{
					this.onCompletion();
				}
			}
		}
	}
	
	protected abstract void onOtherTwitterApiException(TwitterException tex);
	
	protected void onUnhandledTwitterApiException(TwitterException tex)
	{
		logger.error(String.format("@%s encountered a completely unhandled Twitter error!", this.account.username), tex);
	}

	private void selectNewAppAndRetry()
	{
		try
		{
			this.account.selectNewApp();
			this.rescheduleAfter(10, TimeUnit.SECONDS);
		}
		catch(TwitterException authorizeEx)
		{
			this.onException(authorizeEx);
		}
	}
	
	@Override
	public String toString()
	{
		return String.format("@%s -> %s", this.account.username, this.getDescription());
	}
}
