package prinzip.accounts.actions;

import java.util.List;
import java.util.concurrent.TimeUnit;

import twitter4j.IDs;
import twitter4j.Query;
import twitter4j.QueryResult;
import twitter4j.Status;
import twitter4j.TwitterException;

import org.apache.log4j.Logger;

import com.google.common.base.Joiner;

import prinzip.accounts.Account;

public class FindTargetsAction extends AccountAction
{
	private static final Logger logger = Logger.getLogger(FindTargetsAction.class);
	
	private boolean wasFruitful;
	
	public FindTargetsAction(Account account)
	{
		super(account);
	}

	@Override
	protected void execute() throws TwitterException
	{
		// Determine amount of users to add
		Account account = this.getAccount();
		int toAdd = (Account.TARGETCACHE_SIZE - account.toFollow.size());
		
		// Needs users at all?
		if(toAdd > 0)
		{
			// Determine a strategy for maxing out the target cache
			int totalAdded = 0;
			if(account.importUsers.isEmpty())
			{
				totalAdded = this.findTargetsBySearch(toAdd);
			}
			else if(account.keywords.isEmpty())
			{
				totalAdded = this.findTargetsByImport(toAdd);
			}
			else
			{
				// Get 50% through search (most recent data), rest is imports
				int bySearch = this.findTargetsBySearch(toAdd / 2);
				int byImport = this.findTargetsByImport(toAdd - bySearch);
				
				// Still no full target cache?
				totalAdded = byImport + bySearch;
				if(totalAdded < toAdd)
				{
					int bySecondSearch = this.findTargetsBySearch(toAdd - totalAdded);
					bySearch += bySecondSearch;
					totalAdded += bySecondSearch;
				}
			}
			
			// Was a good search? 
			this.wasFruitful = (totalAdded > 0);
		}
	}
	
	@Override
	protected void onCompletion()
	{
		Account account = this.getAccount();
		if(this.wasFruitful)
		{
			account.save();
			account.nextFollowAction(null);
		}
		else
		{
			logger.info("search was less than fruitful, will retry in 5 minutes");
			this.rescheduleAfter(5, TimeUnit.MINUTES);
		}
	}
	
	/**
	 * @see https://dev.twitter.com/docs/using-search for query syntax
	 */
	private int findTargetsBySearch(int maxAmount) throws TwitterException
	{
		// Validate search terms
		Account account = this.getAccount();
		List<String> keywords = account.nextSearchKeywords();
		if(keywords.isEmpty())
		{
			logger.warn(String.format("@%s has no search keywords", account.username));
			return 0;
		}
		
		// Bake search query and request up to a maximum of 100 recent tweets
		Query query = new Query(Joiner.on(" OR ").join(keywords)).count(100);
		if(!account.searchLanguage.isEmpty())
		{
			query.setLang(account.searchLanguage);
		}
		QueryResult result = account.getTwitter().search(query);
		
		// Run through results
		int added = 0;
		for(Status tweet : result.getTweets())
		{
			if(added < maxAmount)
			{
				if(account.addFollowTarget(tweet.getUser().getId()))
				{
					added++;
				}
			}
			else
			{
				break;
			}
		}
		
		// Log the result
		if(logger.isDebugEnabled()) logger.debug(String.format("@%s searched \"%s\" (%d results in %.3f seconds, %d targets added)", account.username, query.getQuery(), result.getTweets().size(), result.getCompletedIn(), added));
		return added;
	}
	
	private int findTargetsByImport(int maxAmount) throws TwitterException
	{
		long srcUserId = this.getAccount().nextImportUser();
		if(srcUserId > 0)
		{
			return this.findTargetsByImport(srcUserId, maxAmount, 0);
		}
		else
		{
			return 0;
		}
	}
	
	private int findTargetsByImport(long srcUserId, int maxAmount, int usersTried) throws TwitterException
	{ 
		Account account = this.getAccount();
		
		// Query and evaluate + add as many targets as needed
		int received = 0, added = 0;
		if(srcUserId > 0)
		{
			try
			{
				// Query this account till satisfied or after trying all its followers
				long cursor = -1;
				IDs followers;
				do
				{
					// Get the next batch
					followers = account.getTwitter().getFollowersIDs(srcUserId, cursor);
					if(followers.getIDs().length == 0)
					{
						// Stop trying this account for now, it has no followers
						break;
					}
					
					// Process batch
					received += followers.getIDs().length;
					for(long followerId : followers.getIDs())
					{
						if(added < maxAmount)
						{
							if(account.addFollowTarget(followerId))
							{
								added++;
							}
						}
						else
						{
							break;
						}
					}
					cursor = followers.getNextCursor();
				}
				while(added < maxAmount && followers.hasNext());
			}
			catch(TwitterException e)
			{
				// Twitter API limit reached?
				if(e.exceededRateLimitation())
				{
					maxAmount = 0; // beat it you impatient fuck
					if(logger.isDebugEnabled()) logger.debug(String.format("@%s's import from #%d finished early due to API rate limits", account.username, srcUserId));
				}
				
				// User not found?
				else if(e.resourceNotFound())
				{
					account.importUsers.remove(srcUserId);
					logger.warn(String.format("@%s's import user #%d does not exist, removed", account.username, srcUserId));
				}
				
				// Something else...
				else
				{
					throw e;
				}
			}
		}
		
		// Done with this user!
		usersTried++;
		if(logger.isDebugEnabled()) logger.debug(String.format("@%s filtered %d follower IDs from #%d (%d targets added)", account.username, received, srcUserId, added));
		
		// Needs more from import?
		int need = (maxAmount - added);
		if(need > 0)
		{
			// Not all accounts tried yet?
			if(usersTried < account.importUsers.size())
			{
				// Next account available?
				long nextUserId = account.nextImportUser();
				if(nextUserId > 0 && nextUserId != srcUserId)
				{
					if(logger.isDebugEnabled()) logger.debug(String.format("@%s will continue importing, next source: #%d", account.username, nextUserId));
					return this.findTargetsByImport(nextUserId, need, usersTried);
				}
			}
		}
		
		// Stop importing
		return added;
	}

	@Override
	protected void onOtherTwitterApiException(TwitterException tex)
	{
		// Nothing to do here (for now)
		super.onUnhandledTwitterApiException(tex);
	}
	
	@Override
	protected String getDescription()
	{
		return "find_targets";
	}
}
